/*******************************************************************************
*                                                                              *
*    LICENSED MATERIALS - PROPERTY OF IBM                                      *
*                                                                              *
*    "RESTRICTED MATERIALS OF IBM"                                             *
*                                                                              *
*    5765-F84                                                                  *
*    5765-G18                                                                  *
*                                                                              *
*    (C) COPYRIGHT IBM CORP.  2004.  All rights reserved.                      *
*                                                                              *
*    US Government Users Restricted Rights - Use, duplication or disclosure    *
*    restricted by GSA ADP Schedule Contract with IBM Corp.                    *
*                                                                              *
********************************************************************************
*                                                                              *
*    Program name - <Cblacs.h> header file                                     *
*    DESCRIPTIVE-NAME = Parallel ESSL Cblacs header file.                      *
*                                                                              *
*    FUNCTION =                                                                *
*               This file must be included in any C or C++ file                *
*               containing Cblacs calls.                                       *
*                                                                              *
*      CHANGE-ACTIVITY = None                                                  *
*******************************************************************************/
 
#include <string.h>
 
#ifndef _Cblacs_H_
#define _Cblacs_H_
 
/* Initialization */
 
void Cblacs_pinfo(int* mypnum, int* nprocs);
void Cblacs_setup(int* mypnum, int* nprocs);
void Cblacs_get(int context, int request, int* value);
void Cblacs_set(int comm, int what, int* val);
void Cblacs_gridinit(int* context, char* order, int np_row, int np_col);
void Cblacs_gridmap(int* comm, int*  usrmap, int ldu, int nprow, int npcol);
 
/* Destruction */
 
void Cblacs_freebuff(int context, int wait);
void Cblacs_gridexit(int context);
void Cblacs_abort(int comm, int ercode);
void Cblacs_exit(int cntinue);
 
/* Sending */
 
void Ccgesd2d(int comm, int m, int n, float*  a, int lda, int rdest, int cdest);
void Cdgesd2d(int comm, int m, int n, double* a, int lda, int rdest, int cdest);
void Cigesd2d(int comm, int m, int n, int*    a, int lda, int rdest, int cdest);
void Csgesd2d(int comm, int m, int n, float*  a, int lda, int rdest, int cdest);
void Czgesd2d(int comm, int m, int n, double* a, int lda, int rdest, int cdest);
 
void Ccgebs2d(int comm, char* scope, char* top, int m, int n, float*  a, int lda);
void Cdgebs2d(int comm, char* scope, char* top, int m, int n, double* a, int lda);
void Cigebs2d(int comm, char* scope, char* top, int m, int n, int*    a, int lda);
void Csgebs2d(int comm, char* scope, char* top, int m, int n, float*  a, int lda);
void Czgebs2d(int comm, char* scope, char* top, int m, int n, double* a, int lda);
 
void Cctrsd2d(int comm, char* uplo, char* diag, int m, int n, float*  a, int lda, int rdest, int cdest);
void Cdtrsd2d(int comm, char* uplo, char* diag, int m, int n, double* a, int lda, int rdest, int cdest);
void Citrsd2d(int comm, char* uplo, char* diag, int m, int n, int*    a, int lda, int rdest, int cdest);
void Cstrsd2d(int comm, char* uplo, char* diag, int m, int n, float*  a, int lda, int rdest, int cdest);
void Cztrsd2d(int comm, char* uplo, char* diag, int m, int n, double* a, int lda, int rdest, int cdest);
 
void Cctrbs2d(int comm, char* scope, char* top, char* uplo, char* diag, int m, int n, float*  a, int lda);
void Cdtrbs2d(int comm, char* scope, char* top, char* uplo, char* diag, int m, int n, double* a, int lda);
void Citrbs2d(int comm, char* scope, char* top, char* uplo, char* diag, int m, int n, int*    a, int lda);
void Cstrbs2d(int comm, char* scope, char* top, char* uplo, char* diag, int m, int n, float*  a, int lda);
void Cztrbs2d(int comm, char* scope, char* top, char* uplo, char* diag, int m, int n, double* a, int lda);
 
/* Receiving */
 
void Ccgerv2d(int comm, int m, int n, float*  a, int lda, int rsrc, int csrc);
void Cdgerv2d(int comm, int m, int n, double* a, int lda, int rsrc, int csrc);
void Cigerv2d(int comm, int m, int n, int*    a, int lda, int rsrc, int csrc);
void Csgerv2d(int comm, int m, int n, float*  a, int lda, int rsrc, int csrc);
void Czgerv2d(int comm, int m, int n, double* a, int lda, int rsrc, int csrc);
 
void Ccgebr2d(int comm, char* scope, char* top, int m, int n, float*  a, int lda, int rsrc, int csrc);
void Cdgebr2d(int comm, char* scope, char* top, int m, int n, double* a, int lda, int rsrc, int csrc);
void Cigebr2d(int comm, char* scope, char* top, int m, int n, int*    a, int lda, int rsrc, int csrc);
void Csgebr2d(int comm, char* scope, char* top, int m, int n, float*  a, int lda, int rsrc, int csrc);
void Czgebr2d(int comm, char* scope, char* top, int m, int n, double* a, int lda, int rsrc, int csrc);
 
void Cctrrv2d(int comm, char* uplo, char* diag, int m, int n, float*  a, int lda, int rsrc, int csrc);
void Cdtrrv2d(int comm, char* uplo, char* diag, int m, int n, double* a, int lda, int rsrc, int csrc);
void Citrrv2d(int comm, char* uplo, char* diag, int m, int n, int*    a, int lda, int rsrc, int csrc);
void Cstrrv2d(int comm, char* uplo, char* diag, int m, int n, float*  a, int lda, int rsrc, int csrc);
void Cztrrv2d(int comm, char* uplo, char* diag, int m, int n, double* a, int lda, int rsrc, int csrc);
 
void Cctrbr2d(int comm, char* scope, char* top, char* uplo, char* diag, int m, int n, float*  a, int lda, int rsrc, int csrc);
void Cdtrbr2d(int comm, char* scope, char* top, char* uplo, char* diag, int m, int n, double* a, int lda, int rsrc, int csrc);
void Citrbr2d(int comm, char* scope, char* top, char* uplo, char* diag, int m, int n, int*    a, int lda, int rsrc, int csrc);
void Cstrbr2d(int comm, char* scope, char* top, char* uplo, char* diag, int m, int n, float*  a, int lda, int rsrc, int csrc);
void Cztrbr2d(int comm, char* scope, char* top, char* uplo, char* diag, int m, int n, double* a, int lda, int rsrc, int csrc);
 
/* Combine Operations */
 
void Ccgamx2d(int comm, char* scope, char* top, int m, int n, float*  a, int lda, int* ra, int* ca, int ldia, int rdest, int cdest);
void Cdgamx2d(int comm, char* scope, char* top, int m, int n, double* a, int lda, int* ra, int* ca, int ldia, int rdest, int cdest);
void Cigamx2d(int comm, char* scope, char* top, int m, int n, int*    a, int lda, int* ra, int* ca, int ldia, int rdest, int cdest);
void Csgamx2d(int comm, char* scope, char* top, int m, int n, float*  a, int lda, int* ra, int* ca, int ldia, int rdest, int cdest);
void Czgamx2d(int comm, char* scope, char* top, int m, int n, double* a, int lda, int* ra, int* ca, int ldia, int rdest, int cdest);
 
void Ccgamn2d(int comm, char* scope, char* top, int m, int n, float*  a, int lda, int* ra, int* ca, int ldia, int rdest, int cdest);
void Cdgamn2d(int comm, char* scope, char* top, int m, int n, double* a, int lda, int* ra, int* ca, int ldia, int rdest, int cdest);
void Cigamn2d(int comm, char* scope, char* top, int m, int n, int*    a, int lda, int* ra, int* ca, int ldia, int rdest, int cdest);
void Csgamn2d(int comm, char* scope, char* top, int m, int n, float*  a, int lda, int* ra, int* ca, int ldia, int rdest, int cdest);
void Czgamn2d(int comm, char* scope, char* top, int m, int n, double* a, int lda, int* ra, int* ca, int ldia, int rdest, int cdest);
 
void Ccgsum2d(int comm, char* scope, char* top, int m, int n, float*  a, int lda, int rdest, int cdest);
void Cdgsum2d(int comm, char* scope, char* top, int m, int n, double* a, int lda, int rdest, int cdest);
void Cigsum2d(int comm, char* scope, char* top, int m, int n, int*    a, int lda, int rdest, int cdest);
void Csgsum2d(int comm, char* scope, char* top, int m, int n, float*  a, int lda, int rdest, int cdest);
void Czgsum2d(int comm, char* scope, char* top, int m, int n, double* a, int lda, int rdest, int cdest);
 
/* Informational and Miscellaneous */
 
void Cblacs_gridinfo(int context, int* np_row, int* np_col, int* my_row, int* my_col);
int  Cblacs_pnum(int comm, int prow, int pcol);
void Cblacs_pcoord(int comm, int pnum, int* prow, int* pcol);
void Cblacs_barrier(int comm, char* scope);
 
/* Not part of standard Cblacs interface but exists in Parallel ESSL */
 
void Cblacs_gidinfo(int comm, int* agid, int* rgid, int* cgid);
 
#endif
