/*******************************************************************************
*                                                                              *
*    LICENSED MATERIALS - PROPERTY OF IBM                                      *
*                                                                              *
*    "RESTRICTED MATERIALS OF IBM"                                             *
*                                                                              *
*    5799-FQC                                                                  *
*    5765-422                                                                  *
*    5765-C41                                                                  *
*    5765-F84                                                                  *
*    5765-G18                                                                  *
*                                                                              *
*    (C) COPYRIGHT IBM CORP. 1995, 2007.  All rights reserved.                 *
*                                                                              *
*    US Government Users Restricted Rights - Use, duplication or disclosure    *
*    restricted by GSA ADP Schedule Contract with IBM Corp.                    *
*                                                                              *
********************************************************************************
*                                                                              *
*  Program name - <pessl.h> header file                                        *
*  Descriptive name - Parallel ESSL - C/C++ language header file               *
*                                                                              *
*  Function : This file must be included in any C/C++ file                     *
*             containing Parallel ESSL calls in order for the                  *
*             Parallel ESSL calls to work as documented in the                 *
*             Parallel ESSL Guide and Reference.                               *
*                                                                              *
*  Change activity -                                                           *
*               Parallel ESSL Version 1.1.0.0                 10/1995          *
*               Parallel ESSL Version 1.2.0.0                 05/1996          *
*               Parallel ESSL Version 1.2.1.0                 09/1996          *
*               Parallel ESSL Version 2.1.0.0                 12/1997          *
*               Parallel ESSL Version 2.1.1.0                 11/1998          *
*               Parallel ESSL Version 2.1.2.0                 10/1999          *
*               Parallel ESSL Version 2.2.0.0                 07/2000          *
*               Parallel ESSL Version 2.3.0.0                 12/2001          *
*               Parallel ESSL Version 3.1.0.0                 09/2003          *
*               Parallel ESSL Version 3.3.0.0                 10/2006          *
*               Parallel ESSL Version 3.3.0.1                 08/2007          *
*******************************************************************************/
 
#ifndef __cplusplus
  /* Begin C Path */
 
  #ifndef __pessl
   #define __pessl 1
 
   /* Definition of complex data types */
 
   #ifndef  _CMPLX
    #define  _CMPLX   1
    #ifndef  _REIM
     #define _REIM   1
    #endif
    typedef union { struct { float  _re, _im; } _data; double _align; } cmplx;
   #endif
 
   #ifndef  _DCMPLX
    #define  _DCMPLX   1
    #ifndef  _REIM
     #define _REIM   1
    #endif
    typedef union { struct { double _re, _im; } _data; double _align; } dcmplx;
   #endif
 
   #ifdef  _REIM
    #define RE(x)   ((x)._data._re)
    #define IM(x)   ((x)._data._im)
   #endif
 
   /* Complex arguments defined differently for C and C++ */
 
   #define _ESVPCM  cmplx
   #define _ESVPCOM dcmplx
 
   /* Scalar output arguments defined differently for C and C++ */
 
   #define _ESVPI   int *
   #define _ESVPD   double *
 
 
  #endif         /* End __pessl */
 
  /* End C Path */
 
#else
  /* Begin C++ Path */
 
  #ifndef _pessl
   #define _pessl 1
 
   /* Complex arguments defined differently for C and C++ */
 
   #ifndef __linux
   #if !defined(_COMPLEX_) && !defined(_ESV_COMPLEX_)
     #include <complex.h>
     #include <iostream.h>
     #define _ESVPCM cmplx
     #define _ESVPCOM complex
   #else
     #include <complex>
     using namespace std;
     #define _ESVPCM complex<float>
     #define _ESVPCOM complex<double>
   #endif
   #else
     #include <complex>
     using namespace std;
     #define _ESVPCM complex<float>
     #define _ESVPCOM complex<double>
   #endif
 
   /* Scalar output arguments defined differently for C and C++ */
 
   #ifndef _ESVCPTR
     #define _ESVPD   double &
     #define _ESVPI   int &
   #else
     #define _ESVPD   double *
     #define _ESVPI   int *
   #endif
 
   /* Definition of complex data types */
 
   #ifndef __linux
   #ifdef COMPLEXH
     #ifndef  _CMPLX
       #define  _CMPLX  1
       class cmplx
         {
          private:
            float _re,_im;
          public:
             cmplx() { _re = 0.0; _im = 0.0; }
             cmplx(float r, float i = 0.0) { _re = r; _im = i; }
             friend inline float sreal(const cmplx& a) { return a._re; }
             friend inline float simag(const cmplx& a) { return a._im; }
         };
     #endif
   #endif
   #endif
 
  #endif       /* End _pessl */
 
  /* End C++ Path */
#endif
 
#ifdef __cplusplus
  extern "C" {
#endif
 
/*  Level 2 Parallel BLAS  */
 
#define pdgemv   esvpdgemv
#define pzgemv   esvpzgemv
 
#define pdsymv   esvpdsymv
#define pzhemv   esvpzhemv
 
#define pdger    esvpdger
#define pzgeru   esvpzgeru
#define pzgerc   esvpzgerc
 
#define pdsyr    esvpdsyr
#define pzher    esvpzher
 
#define pdsyr2   esvpdsyr2
#define pzher2   esvpzher2
 
#define pdtrmv   esvpdtrmv
#define pztrmv   esvpztrmv
 
#define pdtrsv   esvpdtrsv
#define pztrsv   esvpztrsv
 
/*  Level 3 Parallel BLAS  */
 
#define pdgemm   esvpdgemm
#define pzgemm   esvpzgemm
 
#define pdsymm   esvpdsymm
#define pzsymm   esvpzsymm
#define pzhemm   esvpzhemm
 
#define pdtrmm   esvpdtrmm
#define pztrmm   esvpztrmm
 
#define pdtrsm   esvpdtrsm
#define pztrsm   esvpztrsm
 
#define pdsyrk   esvpdsyrk
#define pzsyrk   esvpzsyrk
#define pzherk   esvpzherk
 
#define pdsyr2k  esvpdsyr2k
#define pzsyr2k  esvpzsyr2k
#define pzher2k  esvpzher2k
 
#define pdtran   esvpdtran
#define pztranc  esvpztranc
#define pztranu  esvpztranu
 
/*  Dense Linear Algebraic Equations  */
 
#define pdgesv   esvpdgesv
#define pzgesv   esvpzgesv
 
#define pdgetrf  esvpdgetrf
#define pzgetrf  esvpzgetrf
 
#define pdgetrs  esvpdgetrs
#define pzgetrs  esvpzgetrs
 
#define pdgetri  esvpdgetri
#define pzgetri  esvpzgetri
 
#define pdgecon  esvpdgecon
#define pzgecon  esvpzgecon
 
#define pdgeqrf  esvpdgeqrf
#define pzgeqrf  esvpzgeqrf
 
#define pdgels   esvpdgels
#define pzgels   esvpzgels
 
#define pdposv   esvpdposv
#define pzposv   esvpzposv
 
#define pdpotrf  esvpdpotrf
#define pzpotrf  esvpzpotrf
 
#define pdpotrs  esvpdpotrs
#define pzpotrs  esvpzpotrs
 
#define pdpotri  esvpdpotri
#define pzpotri  esvpzpotri
 
#define pdpocon  esvpdpocon
#define pzpocon  esvpzpocon
 
/*  Banded Linear Algebraic Equations  */
 
#define pdpbsv   esvpdpbsv
#define pdpbtrf  esvpdpbtrf
#define pdpbtrs  esvpdpbtrs
 
#define pdgtsv   esvpdgtsv
#define pdgttrf  esvpdgttrf
#define pdgttrs  esvpdgttrs
 
#define pddtsv   esvpddtsv
#define pddttrf  esvpddttrf
#define pddttrs  esvpddttrs
 
#define pdptsv   esvpdptsv
#define pdpttrf  esvpdpttrf
#define pdpttrs  esvpdpttrs
 
/*  Sparse Linear Algebraic Equations  */
 
#define padinit  esvpadinit
#define pdspinit esvpdspinit
#define pdspins  esvpdspins
#define pdgeins  esvpdgeins
#define pdspasb  esvpdspasb
#define pdgeasb  esvpdgeasb
#define pdspgpr  esvpdspgpr
#define pdspgis  esvpdspgis
 
/*  Eigensystem Analysis Subroutines  */
 
#define pdsyevx  esvpdsyevx
#define pzheevx  esvpzheevx
#define pdsyevd  esvpdsyevd
 
#define pdsygvx  esvpdsygvx
#define pzhegvx  esvpzhegvx
 
#define pdsytrd  esvpdsytrd
#define pzhetrd  esvpzhetrd
#define pdsyntrd esvpdsyntrd
 
#define pdsygst  esvpdsygst
#define pzhegst  esvpzhegst
#define pdsyngst esvpdsyngst
 
#define pdgehrd  esvpdgehrd
 
#define pdgebrd  esvpdgebrd
#define pzgebrd  esvpzgebrd
 
#define pdgesvd  esvpdgesvd
#define pzgesvd  esvpzgesvd
 
/*  Fourier Transforms  */
 
#define pscft2   esvpscft2
#define pdcft2   esvpdcft2
 
#define psrcft2  esvpsrcft2
#define pdrcft2  esvpdrcft2
 
#define pscrft2  esvpscrft2
#define pdcrft2  esvpdcrft2
 
#define pscft3   esvpscft3
#define pdcft3   esvpdcft3
 
#define psrcft3  esvpsrcft3
#define pdrcft3  esvpdrcft3
 
#define pscrft3  esvpscrft3
#define pdcrft3  esvpdcrft3
 
/*  Random Number Generation Subroutines  */
 
#define pdurng   esvpdurng
 
/*  Utility  Subroutines  */
 
#define descinit esvdescinit
#define descset  esvdescset
 
#define iceil    esviceil
#define ilcm     esvilcm
 
#define indxg2l  esvindxg2l
#define indxg2p  esvindxg2p
#define indxl2g  esvindxl2g
 
#define infog1l  esvinfog1l
#define infog2l  esvinfog2l
 
#define pdlange  esvpdlange
#define pzlange  esvpzlange
 
#define pdlansy  esvpdlansy
#define pzlansy  esvpzlansy
#define pzlanhe  esvpzlanhe
 
#define numroc   esvnumroc
 
/*  Level 2 Parallel BLAS  */
 
void esvpdgemv(const char *, int, int, double, const void *, int, int,
               const int *, const void *, int, int, const int *, int,
               double, void *, int, int, const int *, int);
 
void esvpzgemv(const char *, int, int,_ESVPCOM, const void *, int, int,
               const int *, const void *, int, int, const int *, int,
               _ESVPCOM, void *, int, int, const int *, int);
 
void esvpdsymv(const char *, int, double, const void *, int, int,
               const int *, const void *, int, int, const int *, int,
               double, void *, int, int, const int *, int);
 
void esvpzhemv(const char *, int,_ESVPCOM, const void *, int, int,
               const int *, const void *, int, int, const int *, int,
               _ESVPCOM, void *, int, int, const int *, int);
 
void esvpdger(int, int, double, const void *, int, int, const int *, int,
              const void *, int, int, const int *, int, void *, int, int,
              const int *);
 
void esvpzgeru(int, int, _ESVPCOM, const void *, int, int, const int *,
               int, const void *, int, int, const int *, int, void *, int,
               int, const int *);
 
void esvpzgerc(int, int, _ESVPCOM, const void *, int, int, const int *,
               int, const void *, int, int, const int *, int, void *, int,
               int, const int *);
 
void esvpdsyr(const char *, int, double, const void *, int, int,
              const int *, int, void *, int, int, const int *);
 
void esvpzher(const char *, int, const double, const void *, int, int,
              const int *, int, void *, int, int, const int *);
 
void esvpdsyr2(const char *, int,  double, const void *, int, int,
               const int *, int, const void *, int, int, const int *, int,
               void *, int, int, const int *);
 
void esvpzher2(const char *, int, _ESVPCOM, const void *, int, int,
               const int *, int, const void *, int, int, const int *, int,
               void *, int, int, const int *);
 
void esvpdtrmv(const char *, const char *, const char *, int,
               const void *, int, int, const int *, void *, int, int,
               const int *, int);
 
void esvpztrmv(const char *, const char *, const  char *, int,
               const void *, int, int, const int *, void *, int, int,
               const int *, int);
 
void esvpdtrsv(const char *, const char *, const char *, int,
               const void *, int, int, const int *, void *, int, int,
               const int *, int);
 
void esvpztrsv(const char *, const char *, const char *, int,
               const void *, int, int, const int *, void *, int, int,
               const int *, int);
 
/*  Level 3 Parallel BLAS  */
 
void esvpdgemm(const char *, const char *, int, int, int, double,
               const void *, int, int, const int *, const void *, int,
               int, const int *, double, void * , int, int, const int *);
 
void esvpzgemm(const char *, const char *, int, int, int, _ESVPCOM,
               const void *, int, int, const int *, const void *, int,
               int, const int *, _ESVPCOM, void * , int, int, const int *);
 
void esvpdsymm(const char *, const char *, int, int, double,
               const void *, int, int, const int *, const void *, int, int,
               const int *, double, void * , int, int, const int *);
 
void esvpzsymm(const char *, const char *, int, int, _ESVPCOM,
               const void *, int, int, const int *, const void *, int,
               int, const int *,_ESVPCOM, void * , int, int, const int *);
 
void esvpzhemm(const char *, const char *, int, int, _ESVPCOM,
               const void *, int, int, const int *, const void *, int, int,
               const int *, _ESVPCOM, void * , int, int, const int *);
 
void esvpdtrmm(const char *, const char *, const char *, const char *,
               int, int, double, const void *, int, int, const int *,
               void *, int, int, const int *);
 
void esvpztrmm(const char *, const char *, const char *, const char *,
               int, int, _ESVPCOM, const void *, int, int, const int *,
               void *, int, int, const int *);
 
void esvpdtrsm(const char *, const char *, const char *, const char *,
               int, int, double, const void *, int, int, const int *,
               void *, int, int, const  int *);
 
void esvpztrsm(const char *, const char *, const char *, const char *,
               int, int, _ESVPCOM, const void *, int, int, const int *,
               void *, int, int, const int *);
 
void esvpdsyrk(const char *, const char *, int, int, double, const void *,
               int, int , const int *, double, void *, int, int, const int *);
 
void esvpzsyrk(const char *, const char *, int, int, _ESVPCOM, const void *,
               int, int , const int *,_ESVPCOM, void *, int, int, const int *);
 
void esvpzherk(const char *, const char *, int, int, double,const  void *,
               int, int , const int *, double, void *, int, int, const int *);
 
void esvpdsyr2k(const char *, const char *, int, int, double, const void *,
                int, int, const int *, const void *, int, int, const int *,
                double, void *, int, int, const int *);
 
void esvpzsyr2k(const char *, const char *, int, int, _ESVPCOM, const void *,
                int, int, const int *, const void *, int, int, const int *,
                _ESVPCOM, void *, int, int, const int *);
 
void esvpzher2k(const char *, const char *, int, int, _ESVPCOM, const void *,
                int, int, const int *, const void *, int, int, const int *,
                double, void *, int, int, const int *);
 
void esvpdtran(int, int, double, const void *, int, int, const int *,
               double, void *, int, int, const int *);
 
void esvpztranc(int, int,_ESVPCOM, const void *, int, int, const int *,
                _ESVPCOM, void *, int, int, const int *);
 
void esvpztranu(int, int, _ESVPCOM, const  void *, int, int, const int *,
                _ESVPCOM, void *, int, int, const int *);
 
/*  Dense Linear Algebraic Equations  */
 
void   esvpdgesv(int, int, void *, int, int, const int *, int *, void *,
                 int, int, const int *, _ESVPI);
 
void   esvpzgesv(int, int, void *, int, int, const int *, int *, void *,
                 int, int, const int *, _ESVPI);
 
void   esvpdgetrf(int, int, void *, int, int, const int *, int *, _ESVPI);
 
void   esvpzgetrf(int, int, void *, int, int, const int *, int *, _ESVPI);
 
void   esvpdgetrs(const char *, int, int, const void *, int, int, const int *,
                  const int *, void *, int, int, const int *, _ESVPI);
 
void   esvpzgetrs(const char *, int, int, const void *, int, int, const int *,
                  const int *, void *, int, int, const int *, _ESVPI);
 
void   esvpdgetri(int, void *, int, int, const int *, const int *, double *,
                  int, int *, int, _ESVPI);
 
void   esvpzgetri(int, void *, int, int, const int *, const int *, _ESVPCOM *,
                  int, int  *, int, _ESVPI);
 
void   esvpdgecon(const char *, int, const void *, int, int, const int *,
                  double, _ESVPD, double *, int, int *, int, _ESVPI);
 
void   esvpzgecon(const char *, int, const void *, int, int, const int *,
                  double, _ESVPD, _ESVPCOM *, int, double *, int, _ESVPI);
 
void   esvpdgeqrf(int, int, void *, int, int, const int *, void *, double *,
                  int, _ESVPI);
 
void   esvpzgeqrf(int, int, void *, int, int, const int *, void *, void *,
                  int, _ESVPI);
 
void   esvpdgels(const char *, int, int, int, void *, int, int, const int *,
                 void *, int, int, const int *, double *, int, _ESVPI);
 
void   esvpzgels(const char *, int, int, int, void *, int, int, const int *,
                 void *, int, int, const int *, _ESVPCOM *, int, _ESVPI);
 
void   esvpdposv(const char *, int, int,  void *, int, int, const int *,
                 void *, int, int, const int *, _ESVPI);
 
void   esvpzposv(const char *, int, int, void *, int, int, const int *,
                 void *, int, int, const int *, _ESVPI);
 
void   esvpdpotrf(const char *, int, void *, int, int, const int *, _ESVPI);
 
void   esvpzpotrf(const char *, int, void *, int, int, const int *, _ESVPI);
 
void   esvpdpotri(const char *, int, void *, int, int, const int *, _ESVPI);
 
void   esvpzpotri(const char *, int, void *, int, int, const int *, _ESVPI);
 
void   esvpdpocon(const char *, int, void *, int, int, const int *, double,
                  _ESVPD, double *, int, int *, int, _ESVPI);
 
void   esvpzpocon(const char *, int, void *, int, int, const int *, double,
                  _ESVPD, _ESVPCOM *, int, double *, int, _ESVPI);
 
void   esvpdpotrs(const char *, int, int, const void *, int, int,
                  const int *, void *, int, int, const int *, _ESVPI);
 
void   esvpzpotrs(const char *, int, int, const void *, int, int,
                  const int *, void *, int, int, const int *, _ESVPI);
 
/*  Banded Linear Algebraic Equations  */
 
void esvpdpbsv(const char *, int, int, int, void *, int, const int *,
               void *, int, const  int *, double *, int, _ESVPI);
 
void esvpdpbtrf(const char *, int, int, void *, int, const int *,void *,
                int, double *, int, _ESVPI);
 
void esvpdpbtrs(const char *, int, int, int, const void *, int, const int *,
                void *, int, const int *, const void *, int, double *,
                int, _ESVPI);
 
void esvpdgtsv(int, int, void *, void *, void *, int, const int *, void *,
               int,const int *, double *, int, _ESVPI);
 
void esvpdgttrf(int, void *, void *, void *, void *, int, const int *,
                int *, void *, int, double *, int, _ESVPI);
 
void esvpdgttrs(const char *, int, int, const void *, const void *,
                const void *, const void *, int, const int *, const int *,
                void *, int, const int *, const void *, int, double *,
                int, _ESVPI);
 
void esvpddtsv(int, int, void *, void *, void *, int, const int *, void *,
               int,const int *, double *, int, _ESVPI);
 
void esvpddttrf(int, void *, void *, void *, int, const int *, void *, int,
                double *, int, _ESVPI);
 
void esvpddttrs(const char *, int, int, const void *, const void *,
                const void *, int, const  int *, void *, int, const int *,
                const void *, int, double *, int, _ESVPI);
 
void esvpdptsv(int, int, void *, void *, int, const int *, void *, int,
               const int *, double *, int, _ESVPI);
 
void esvpdpttrf(int, void *, void *, int, const int *, void *, int, double *,
                int, _ESVPI);
 
void esvpdpttrs(int, int, const void *, const void *, int, const int *,
                void *, int, const int *, const void *, int, double *, int,
                _ESVPI);
 
/*  Sparse Linear Equations    */
 
void esvpadinit(int, void(*)(int *,int *,int *,int *,int *), int *, int);
 
void esvpdspinit(void *, int *, int *, int *, int *);
 
void esvpdspins(void *, int *, int *, int *, int *, int, int, const double *,
                const int *, const int *, const int *);
 
void esvpdgeins(int, void *, int, int, int, int, int, const void *, int,
                const int *);
 
void esvpdspasb(void *, int *, int *, int *, int *, const char *,
                const char *, int, _ESVPI);
 
void esvpdgeasb(int, void *, int, const int *);
 
void esvpdspgpr(int, const void *, const int *, const int *, const int *,
                void *, int, const int *, _ESVPI);
 
void esvpdspgis(const void *, const int *, const int *, const int *, int,
                const void *, int, void *, int, const void *, const int *,
                int *, void *, _ESVPI);
 
/*  Eigensystem Analysis Subroutines  */
 
void esvpdsyevx(const char *, const char *, const char *, int, void *, int,
                int, const int * , double, double, int, int, double,
                _ESVPI, _ESVPI, double *, double, void *, int, int,
                const int *,double *, int,int *, int, int *, int *,
                double *, _ESVPI);
 
void esvpzheevx(const char *, const char *, const char *, int, void *, int,
                int, const int *, double, double, int, int, double, _ESVPI,
                _ESVPI, double *, double, void *, int, int, const int *,
                _ESVPCOM *, int, double *, int, int *, int, int *, int *,
                double *, _ESVPI);
 
void esvpdsyevd(const char *, const char *, int, void *, int, int,
                const int *, double *, void *, int, int, const int *,
                double *, int, int *, int, _ESVPI);
 
void esvpdsygvx(int, const char *, const char *, const char *, int, void *,
                int, int, const int *, void *, int, int, const int *,
                double, double, int, int,  double, _ESVPI, _ESVPI, double *,
                double, void *, int, int, const int *, double *, int, int *,
                int, int *, int *, double *, _ESVPI);
 
void esvpzhegvx(int, const char *, const char *, const char *, int,
                void *, int, int, const int *, void *, int, int, const int *,
                double, double, int, int,  double, _ESVPI, _ESVPI, double *,
                double, void *, int, int, const int *, _ESVPCOM *, int,
                double *, int, int *, int, int *, int *, double *, _ESVPI);
 
void esvpdsytrd(const char *, int, void *, int, int, const int *,
                double *, double *, double *, double *, int, _ESVPI);
 
void esvpzhetrd(const char *, int, void *, int, int,const int *,
               double *,double *, _ESVPCOM *,_ESVPCOM *, int, _ESVPI);
 
void esvpdsyntrd(const char *, int, void *, int, int, const int *,
                double *, double *, double *, double *, int, _ESVPI);
 
void esvpdsygst(int, const char *, int, void *, int, int, const int *,
                const void *, int, int, const int *, _ESVPD, _ESVPI);
 
void esvpzhegst(int, const char *, int, void *, int, int, const int *,
                const void *, int, int, const int *, _ESVPD, _ESVPI);
 
void esvpdsyngst(int, const char *, int, void *, int, int, const int *,
                const void *, int, int, const int *, double, double*, int, _ESVPI);
 
void esvpdgehrd(int, int, int, void *, int, int, const int *, double *,
                double *, int, _ESVPI);
 
void esvpdgebrd(int, int, void *, int, int, const int *, double *, double *,
                double *, double *, double *, int, _ESVPI);
 
void esvpzgebrd(int, int, void *, int, int, const int *, double *, double *,
                _ESVPCOM *, _ESVPCOM *, _ESVPCOM *, int, _ESVPI);
 
void esvpdgesvd(const char *, const char*, int, int, void *, int, int,
                const int *, double *, void *, int, int, const int *,
                void *, int, int, const int *, double *, int, _ESVPI);
 
void esvpzgesvd(const char *, const char*, int, int, void *, int, int,
                const int *, double *, void *, int, int, const int *,
                void *, int, int, const int *, _ESVPCOM *, int, double *,
                int, _ESVPI);
 
/*  Fourier Transforms  */
 
void esvpscft2(void *, void *, int, int, int,  float, int, const int *);
 
void esvpdcft2(void *, void *, int, int, int,  double, int, const int *);
 
void esvpsrcft2(void *, void *, int, int, int,  float, int, const int *);
 
void esvpdrcft2(void *, void *, int, int, int,  double, int, const int *);
 
void esvpscrft2(void *, void *, int, int, int,  float, int, const int *);
 
void esvpdcrft2(void *, void *, int, int, int,  double, int, const int *);
 
void esvpscft3(void *, void *, int, int, int, int,  float, int,const int *);
 
void esvpdcft3(void *, void *, int, int, int, int, double, int,const int *);
 
void esvpsrcft3(void *,void *, int, int, int, int, float, int,const int *);
 
void esvpdrcft3(void *,void *, int, int, int, int, double, int,const int *);
 
void esvpscrft3(void *,void *, int, int, int, int, float, int,const int *);
 
void esvpdcrft3(void *,void *, int, int, int, int,  double, int,const int *);
 
/*  Random Number Generation Subroutines  */
 
void esvpdurng(_ESVPD, int, int, double *, int, int);
 
/*  Utility Subroutines  */
 
int    ipessl(void);
 
void   esvdescinit (int *,int,int,int,int,int,int,int,int,_ESVPI);
 
void   esvdescset  (int *,int,int,int,int,int,int,int,int);
 
int    esviceil    (int, int);
 
int    esvilcm     (int, int);
 
int    esvindxg2l  (int,int,int,int,int);
 
int    esvindxg2p  (int,int,int,int,int);
 
int    esvindxl2g  (int, int, int, int, int);
 
void   esvinfog1l  (int, int, int, int, int, _ESVPI, _ESVPI);
 
void   esvinfog2l  (int,int,const int *, int, int, int, int, _ESVPI, _ESVPI,
                    _ESVPI, _ESVPI);
 
double esvpdlange(const char *, int, int, const void *, int, int,
                  const int *, double *);
 
double esvpzlange(const char *, int, int,const void *, int, int,
                  const int *, double *);
 
double esvpdlansy(const char *, const char *, int, void*, int, int,
                  const int *, double *);
 
double esvpzlansy(const char *, const char *, int, void*, int, int,
                  const int *, double *);
 
double esvpzlanhe(const char *, const char *, int, void*, int, int,
                  const int *, double *);
 
int    esvnumroc(int, int, int, int, int);
 
#ifdef __cplusplus
}
#endif
