/*******************************************************************************
*                                                                              *
*    LICENSED MATERIALS - PROPERTY OF IBM                                      *
*                                                                              *
*    "RESTRICTED MATERIALS OF IBM"                                             *
*                                                                              *
*    5799-FQC                                                                  *
*    5765-422                                                                  *
*    5765-C41                                                                  *
*    5765-F84                                                                  *
*    5765-G18                                                                  *
*                                                                              *
*    (C) COPYRIGHT IBM CORP. 1995, 2007.  All rights reserved.                 *
*                                                                              *
*    US Government Users Restricted Rights - Use, duplication or disclosure    *
*    restricted by GSA ADP Schedule Contract with IBM Corp.                    *
*                                                                              *
********************************************************************************
*                                                                              *
*     MODULE NAME = esvcpivp.C                                                 *
*                                                                              *
*     MODULE-DESCRIPTION :                                                     *
*                                                                              *
*   FUNCTION = Test example from Parallel ESSL library routines                *
*                                                                              *
*       This program runs one example for each Parallel ESSL library           *
*       user callable routine and checks the results as expected.              *
*       If results match as expected, it returns return code of zero,          *
*       else it prints which routine gives wrong answers and returns           *
*       return code of 8.                                                      *
*                                                                              *
*     NOTES =                                                                  *
*                                                                              *
*     PROCESSOR = C++ Compiler                                                 *
*                                                                              *
*     Revision:                                                                *
*               Parallel ESSL Version 1.1.0.0                 10/1995          *
*               Parallel ESSL Version 1.2.0.0                 05/1996          *
*               Parallel ESSL Version 1.2.1.0                 09/1996          *
*               Parallel ESSL Version 2.1.0.0                 12/1997          *
*               Parallel ESSL Version 2.1.1.0                 11/1998          *
*               Parallel ESSL Version 2.1.2.0                 10/1999          *
*               Parallel ESSL Version 2.2.0.0                 07/2000          *
*               Parallel ESSL Version 2.3.0.0                 12/2001          *
*               Parallel ESSL Version 3.1.0.0                 09/2003          *
*               Parallel ESSL Version 3.2.0.0                 04/2005          *
*               Parallel ESSL Version 3.3.0.0                 10/2006          *
*               Parallel ESSL Version 3.3.0.1                 08/2007          *
*               Replace new.h with new on Linux               01/2009          *
*                                                                              *
********************************************************************************
*                                                                              *
*     SUBPROGRAMS CALLED BY MAIN :                                             *
*                                                                              *
*       PBLAS   =  CALL TO ALL  Parallel Basic Linear Algebra                  *
*                                      Subprograms                             *
*       PLEQ    =  CALL TO ALL  Parallel Linear Algebraic Equations            *
*       PSPRS   =  CALL TO ALL  Parallel Sparse Algebraic Equations            *
*       PEIG    =  CALL TO ALL  Parallel Eigen System Analysis                 *
*       PFFT2   =  CALL TO ALL  Parallel Two dimensional FFT                   *
*       PFFT3   =  CALL TO ALL  Parallel Three dimensional FFT                 *
*       PRND    =  CALL TO ALL  Parallel Random Number Generation              *
*                                      Subroutine                              *
*       PUTIL   =  CALL TO ALL  Parallel Utility Subroutines                   *
*                                                                              *
*******************************************************************************/
#ifdef _ESVSTL
#include <complex>
#include <iostream>
#else
#ifdef _ESVOCL
#include <complex.h>
#include <iostream.h>
#endif
#endif
#include <math.h>
#include <pessl.h>
 
#ifndef __linux
#include <new.h>
#else
#include <new>
#endif
 
#ifndef __linux
#if defined(_COMPLEX_) && !defined(_IOSTREAM_)
#include <iostream>
#endif
#else
#if defined(_CPP_COMPLEX) && !defined(_CPP_IOSTREAM)
#include <iostream>
#endif
#if !defined(_GLIBCXX_IOSTREAM) && !defined(_CPP_IOSTREAM)
#include <iostream>
#endif
#endif
 
#define MAX(a,b) ((a > b) ? (a) : (b))
#define MIN(a,b) ((a < b) ? (a) : (b))
#define ABSP(x) ( (x) < 0 ? (x) * -1 : (x) )
 
#define DLEN_    9
 
#define DT_      0
#define CTXT_    1
#define M_       2
#define N_       3
#define MB_      4
#define NB_      5
#define RSRC_    6
#define CSRC_    7
#define LLD_     8
#define BLOCK_CYCLIC_2D    1
 
#define DLEN_1D_    7
 
#define DT_1D_      0
#define CTXT_1D_    1
#define M_1D_       2
#define N_1D_       2
#define MB_1D_      3
#define NB_1D_      3
#define RSRC_1D_    4
#define CSRC_1D_    4
#define LLD_1D_     5
#define RSVD_1D_    6
#define BLOCK_ROW_1D        502
#define BLOCK_COLUMN_1D     501
 
void pblas(int *), pleq(int *), psprs(int *), peig(int *);
void psfft2(int *), psfft3(int *), pfft2(int *), pfft3(int *);
void prnd(int *), putil(int *);
 
extern "FORTRAN" {
 
  void blacs_pinfo( const int &, const int &);
  void blacs_exit( const int &);
  void blacs_abort( const int &, const int &);
  void blacs_gridexit( const int &);
  void blacs_get( const int &, const int &, const int &);
  void blacs_gridinit( const int &, char *, const int &, const int &,
                       size_t);
  void blacs_gridinfo( const int &, const int &, const int &, const int &,
                       const int &);
  void igamx2d(const int &, char *, char *, const int &, const int &, int *,
               const int &, int *, int *, const int &, const int &,
               const int &, size_t, size_t);
 
  void pdvtrcpy(const int &, const int &, const int &, double *, int *,
                double *, const int &);
  void pzvtrcpy(const int &,const int &,const int &, _ESVPCOM *, int *,
                void   *,  const int &);
  void psmtxcpy(const int &, const int &, const int &, const int &, void *,
                int *, void *);
  void pdmtxcpy(const int &, const int &, const int &, const int &, void *,
                int *, void *);
  void pcmtxcpy(const int &, const int &, const int &, const int &, void *,
                int *, void *);
  void pzmtxcpy(const int &, const int &, const int &, const int &, void *,
                int *, void *);
 
  void pdvtchk(char *, const int &, const int &, const int &, double *, int *,
               double *, const int &, const double &, const int &,
               const int &, const double &, int *, size_t);
  void pzvtchk(char *,const int &,const int &,const int &, _ESVPCOM *,  int *,
               _ESVPCOM *,const int &,const double &,const  int &,
                const int &,  const double &, int *, size_t);
  void psmtchk(char *,const int &, const int &, const int &, const int &,
               void *, int *, void *, const double &, const int &,
               const int &, const double &, int *, size_t);
  void pdmtchk(char *,const int &, const int &, const int &, const int &,
               void *, int *, void *, const double &, const int &,
               const int &, const double &, int *, size_t);
  void pcmtchk(char *, const int &, const int &, const int &, const int &,
               void *, int *, void *, const double &, const int &,const int &,
               const double &, int *,size_t);
  void pzmtchk(char *, const int &, const int &, const int &, const int &,
               void *, int *, void *, const double &, const int &,const int &,
               const double &, int *,size_t);
  void pdgspchk(const int &,const int &,const int &,void *,const int &,
               const int &, int *, void *, const int &, const int &,
               int *,const double &,void *,const int &,const int &,
               int *,void *,const int &,const int &,int *,double *,
               double *,const int &, const double &,const int &);
 
}
 
  void part_block(int *,int *,int *, int *,int *);
main()
{
    int ra[1],ca[1];
    int ifail[1],rc,ictx1,nprow,npcol,myrow,mycol,nnodes,iam;
    char order[]="R";
    char scope[]="A";
    char   top[]="1";
    ifail[0] = 0;
 
    blacs_pinfo(iam,nnodes);
    nprow = 1;
    npcol = nnodes;
    blacs_get(0,0,ictx1);
    blacs_gridinit(ictx1,order,nprow,npcol,strlen(order));
    blacs_gridinfo(ictx1,nprow,npcol,myrow,mycol);
 
    pblas(ifail);
    pleq(ifail);
    psprs(ifail);
    peig(ifail);
    psfft2(ifail);
    pfft2(ifail);
    psfft3(ifail);
    pfft3(ifail);
    prnd(ifail);
    putil(ifail);
 
    igamx2d(ictx1, scope, top, 1, 1, ifail, 1, ra, ca, 1, 0, 0,
            strlen(scope),strlen(top));
    rc = 0;
    if(iam == 0)
    {
      if(ifail[0] == 0) {
          cout << " PARALLEL ESSL installed successfully\n";
      }
      else {
         cout << " NUMBER OF FAILURES = " << ifail[0] << "\n";
          rc = 8;
      }
    }
    blacs_gridexit(ictx1);
    blacs_exit(0);
    return(rc);
}
 
/*********************************************************************
*          PARALLEL BLAS SUBROUTINES                                 *
*********************************************************************/
 
void pblas(int *ifail)
{
// PDTRMV, PZTRMV and PDTRSV, PZTRSV START
  int iam, nnodes,myrow,mycol;
  int desc_a[DLEN_], desc_x[DLEN_], il_lda;
  int ictx, nprow, npcol,ierr=5,ilerr;
  char modn[9], *modname=modn;
  modname = "PDTRMV  ";
  char order[]="R";
  double dta1[144], dtx1[12], error=0.0, tol=0.0;
  double dtas1[12][12] = { 1.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
                           2.0,3.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
                           1.0,2.0,3.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
                           2.0,3.0,1.0,1.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
                           1.0,1.0,3.0,2.0,2.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
                           1.0,2.0,2.0,2.0,1.0,1.0,0.0,0.0,0.0,0.0,0.0,0.0,
                           3.0,3.0,1.0,1.0,2.0,2.0,2.0,0.0,0.0,0.0,0.0,0.0,
                           1.0,1.0,2.0,1.0,2.0,1.0,1.0,2.0,0.0,0.0,0.0,0.0,
                           1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,3.0,0.0,0.0,0.0,
                           2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,1.0,2.0,0.0,0.0,
                           3.0,3.0,3.0,3.0,3.0,3.0,3.0,3.0,3.0,2.0,1.0,0.0,
                           2.0,3.0,1.0,2.0,3.0,1.0,2.0,3.0,1.0,2.0,3.0,1.0 };
 
  double dtxs1[12] = {2.0,3.0,1.0,2.0,3.0,1.0,2.0,3.0,1.0,2.0,3.0,1.0 };
  double dresx1[12] = {42.0,48.0,39.0,31.0,34.0,23.0,
                     23.0,23.0,15.0,12.0,6.0,1.0 };
 
      _ESVPCOM    zta1[16], ztx1[4];
      _ESVPCOM ztas1[16] = {
      _ESVPCOM(0.0,0.0),_ESVPCOM(1.0,1.0),_ESVPCOM(1.0,1.0),
      _ESVPCOM(3.0,3.0),
      _ESVPCOM(0.0,0.0),_ESVPCOM(0.0,0.0),_ESVPCOM(3.0,3.0),
      _ESVPCOM(4.0,4.0),
      _ESVPCOM(0.0,0.0),_ESVPCOM(0.0,0.0),_ESVPCOM(0.0,0.0),
      _ESVPCOM(3.0,3.0),
      _ESVPCOM(0.0,0.0),_ESVPCOM(0.0,0.0),_ESVPCOM(0.0,0.0),
      _ESVPCOM(0.0,0.0)};
 
      _ESVPCOM ztxs1[4]  = {
      _ESVPCOM(1.0,1.0), _ESVPCOM(2.0,2.0), _ESVPCOM(3.0,3.0),
      _ESVPCOM(4.0,4.0)};
      _ESVPCOM zresx1[4] = {
      _ESVPCOM(1.0,1.0), _ESVPCOM(2.0,4.0), _ESVPCOM(3.0,17.0),
      _ESVPCOM(4.0,44.0)};
 
  blacs_pinfo(iam,nnodes);
  nprow = 1;
  npcol = nnodes;
  blacs_get(0,0,ictx);
  blacs_gridinit(ictx,order,nprow,npcol,strlen(order));
  blacs_gridinfo(ictx,nprow,npcol,myrow,mycol);
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 12;
  desc_a[N_]    = 12;
  desc_a[MB_]   = 3;
  desc_a[NB_]   = 3;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(12,3,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 12;
  desc_x[N_]    = 1;
  desc_x[MB_]   = 3;
  desc_x[NB_]   = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(12,3,myrow,0,nprow);
  desc_x[LLD_]  = MAX(1,il_lda);
 
  pdmtxcpy(12, 12, 1, 1, dta1, desc_a, dtas1);
  pdvtrcpy(12, 1, 1, dtx1, desc_x, dtxs1, 1);
 
  pdtrmv("U","N","N", 12, dta1, 1, 1, desc_a, dtx1, 1, 1, desc_x, 1);
  pdvtchk(modname,12, 1, 1, dtx1, desc_x, dresx1, 1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  modname = "PDTRSV  ";
  pdvtrcpy(12, 1, 1, dtx1, desc_x, dresx1, 1);
  pdtrsv("U","N","N", 12, dta1, 1, 1, desc_a, dtx1, 1, 1, desc_x, 1);
  pdvtchk(modname,12, 1, 1, dtx1, desc_x, dtxs1, 1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  modname = "PZTRMV  ";
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 4;
  desc_a[N_] = 4;
  desc_a[MB_] = 3;
  desc_a[NB_] = 3;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(4,3,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 4;
  desc_x[N_] = 1;
  desc_x[MB_] = 3;
  desc_x[NB_] = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(4,3,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  pzmtxcpy(4, 4, 1, 1, zta1, desc_a, ztas1);
  pzvtrcpy(4, 1, 1, ztx1, desc_x, ztxs1, 1);
 
  pztrmv("L","N","U", 4, zta1, 1, 1, desc_a, ztx1, 1, 1, desc_x, 1);
  pzvtchk(modname,4, 1, 1, ztx1, desc_x, zresx1, 1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  modname = "PZTRSV  ";
  pzvtrcpy(4, 1, 1, ztx1, desc_x, zresx1, 1);
  pztrsv("L","N","U", 4,  zta1, 1, 1, desc_a, ztx1, 1, 1, desc_x, 1);
  pzvtchk(modname,4, 1, 1, ztx1, desc_x, ztxs1, 1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
// PDTRMV, PZTRMV and PDTRSV, PZTRSV END
 
// PDTRAN, PZTRANU, PZTRANC START
 
   double dta2[72], dtc2[72];
   double dtas2[9][8] = { 0.,  0.,  0.,  0.,  1.,  1.,  0.,  -1.,
                         -1.,  1.,  0.,  1.,  0.,  0.,  0.,   0.,
                         -1.,  0., -1.,  0.,  0.,  0., -1.,   0.,
                          0.,  1., -1., -1.,  0.,  0.,  0.,   0.,
                          0.,  0.,  0.,  1., -1.,  1., -1.,   0.,
                          0.,  1.,  0.,  1.,  0.,  1.,  0.,   0.,
                          0.,  0.,  1.,  0.,  0.,  0.,  0.,  -1.,
                          0.,  1.,  0.,  1.,  0.,  0.,  0.,   0.,
                          1.,  1.,  1.,  1.,  1.,  1.,  1.,   1. };
 
   double dtcs2[8][9] = { 0.,  0.,  0.,  0., -1., -1.,  0.,   1.,  1.,
                          1., -1.,  0., -1.,  2.,  3.,  4.,   5.,  2.,
                          1.,  0.,  1.,  0.,  0.,  0.,  1.,   0.,  3.,
                          5., -1.,  1.,  1.,  0.,  0.,  0.,   0.,  4.,
                          6.,  0.,  0., -1.,  1., -1.,  1.,   0.,  1.,
                          7., -1.,  0., -1.,  0., -1.,  0.,   0.,  1.,
                          8.,  0., -1.,  0.,  0.,  0.,  0.,   1.,  1.,
                          9.,  1.,  0.,  1.,  0.,  0.,  0.,   0.,  1. };
 
   double dresc2[8][9] = { 0., -1., -1.,  0., -1., -1.,  0.,   1.,  2.,
                          1.,  0.,  0.,  0.,  2.,  4.,  4.,   6.,  3.,
                          1.,  0.,  0., -1.,  0.,  0.,  2.,   0.,  4.,
                          5.,  0.,  1.,  0.,  1.,  1.,  0.,   1.,  5.,
                          7.,  0.,  0., -1.,  0., -1.,  1.,   0.,  2.,
                          8., -1.,  0., -1.,  1.,  0.,  0.,   0.,  2.,
                          8.,  0., -2.,  0., -1.,  0.,  0.,   1.,  2.,
                          8.,  1.,  0.,  1.,  0.,  0., -1.,   0.,  2. };
 
      _ESVPCOM zta2[56], ztc2[56],alpha,beta;
      _ESVPCOM ztas2[56] = {
      _ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM(0., 1.),_ESVPCOM(0.,1.),
      _ESVPCOM( 1., 2.),_ESVPCOM( 1., 2.),_ESVPCOM(0., 1.),_ESVPCOM(-1.,0.),
      _ESVPCOM(-1., 0.),_ESVPCOM( 1., 2.),_ESVPCOM(0., 1.),_ESVPCOM(1.,2.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM(0., 1.),_ESVPCOM(0.,1.),
      _ESVPCOM(-1., 0.),_ESVPCOM( 0., 1.),_ESVPCOM(-1., 0.),_ESVPCOM(0.,1.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM(-1.,0.),_ESVPCOM( 0.,1.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 1., 2.),_ESVPCOM(-1.,0.),_ESVPCOM(-1.,0.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM(0.,1.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM(1.,2.),
      _ESVPCOM(-1., 0.),_ESVPCOM( 1., 2.),_ESVPCOM(-1., 0.),_ESVPCOM(0.,1.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 1., 2.),_ESVPCOM( 0., 1.),_ESVPCOM(1.,2.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 1., 2.),_ESVPCOM( 0., 1.),_ESVPCOM(0.,1.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM( 1.,2.),_ESVPCOM(0.,1.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 0.,1.),_ESVPCOM( 0.,1.),_ESVPCOM(-1.,0.)};
 
      _ESVPCOM ztcs2[56] = {
      _ESVPCOM( 0.,-2.),_ESVPCOM( 1.,-1.),_ESVPCOM(2., 0.),_ESVPCOM(3.,1.),
      _ESVPCOM(-1.,-3.),_ESVPCOM(-1.,-3.),_ESVPCOM(5., 3.),
     _ESVPCOM( 1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),
      _ESVPCOM( 2., 0.),_ESVPCOM( 3., 1.),_ESVPCOM(4., 2.),
     _ESVPCOM( 1.,-1.),_ESVPCOM( 0.,-2.),_ESVPCOM( 1.,-1.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 0.,-2.),_ESVPCOM( 0.,-2.),_ESVPCOM( 1.,-1.),
      _ESVPCOM( 5., 3.),_ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(1.,-1.),
      _ESVPCOM( 0.,-2.),_ESVPCOM( 0.,-2.),_ESVPCOM( 0.,-2.),
     _ESVPCOM( 6., 4.),_ESVPCOM( 0.,-2.),_ESVPCOM( 2.,0.),_ESVPCOM(-1.,-3.),
     _ESVPCOM( 1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM( 1.,-1.),
     _ESVPCOM( 7., 5.),_ESVPCOM(-1.,-3.),_ESVPCOM( 3.,1.),_ESVPCOM(-1.,-3.),
      _ESVPCOM( 1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM( 0.,-2.),
      _ESVPCOM( 8., 6.),_ESVPCOM( 0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM(3.,1.),
      _ESVPCOM( 2., 0.),_ESVPCOM( 0.,-2.),_ESVPCOM( 0.,-2.),
      _ESVPCOM( 9., 7.),_ESVPCOM( 1.,-1.),_ESVPCOM(0.,-2.),_ESVPCOM(1.,-1.),
      _ESVPCOM( 0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.)};
 
      _ESVPCOM zresc2u[56] = {
     _ESVPCOM( 0.,- 1.),_ESVPCOM( 0.,- 1.),_ESVPCOM(1., 0.),_ESVPCOM(3.,2.),
      _ESVPCOM(-1.,-2.),_ESVPCOM(-1.,-2.),_ESVPCOM(5., 4.),
      _ESVPCOM( 1., 0.),_ESVPCOM( 0.,-1.),_ESVPCOM(0.,-1.),_ESVPCOM(0.,-1.),
      _ESVPCOM( 2., 1.),_ESVPCOM( 4., 3.),_ESVPCOM(4., 3.),
      _ESVPCOM( 1., 0.),_ESVPCOM(0.,-1.),_ESVPCOM(0.,-1.),_ESVPCOM(-1.,-2.),
      _ESVPCOM( 0.,-1.),_ESVPCOM( 0.,-1.),_ESVPCOM( 2.,1.),
      _ESVPCOM( 5., 4.),_ESVPCOM( 0.,-1.),_ESVPCOM( 1.,0.),_ESVPCOM(0.,-1.),
      _ESVPCOM( 1., 0.),_ESVPCOM( 1., 0.),_ESVPCOM( 0.,-1.),
     _ESVPCOM( 7., 6.),_ESVPCOM( 0.,-1.),_ESVPCOM( 2.,1.),_ESVPCOM(-1.,-2.),
      _ESVPCOM( 0.,-1.),_ESVPCOM(-1.,-2.),_ESVPCOM( 1., 0.),
     _ESVPCOM( 8., 7.),_ESVPCOM(-1.,-2.),_ESVPCOM( 3.,2.),_ESVPCOM(-1.,-2.),
      _ESVPCOM( 2., 1.),_ESVPCOM( 0.,-1.),_ESVPCOM( 0.,-1.),
      _ESVPCOM( 8., 7.),_ESVPCOM( 0.,-1.),_ESVPCOM(-2.,-3.),_ESVPCOM(3.,2.),
      _ESVPCOM( 1., 0.),_ESVPCOM( 0.,-1.),_ESVPCOM( 0.,-1.),
      _ESVPCOM( 8., 7.),_ESVPCOM( 1., 0.),_ESVPCOM(0.,-1.),_ESVPCOM(1.,0.),
      _ESVPCOM( 0.,-1.),_ESVPCOM( 0.,-1.),_ESVPCOM(-1.,-2.)};
 
      _ESVPCOM zresc2c[56] = {
     _ESVPCOM( 0.,- 3.),_ESVPCOM( 0.,- 1.),_ESVPCOM(1., 0.),_ESVPCOM(3.,0.),
      _ESVPCOM(-1.,-4.),_ESVPCOM(-1.,-4.),_ESVPCOM(5., 2.),
      _ESVPCOM( 1.,-2.),_ESVPCOM( 0.,-5.),_ESVPCOM(0.,-3.),_ESVPCOM(0.,-5.),
      _ESVPCOM( 2.,-1.),_ESVPCOM( 4.,-1.),_ESVPCOM(4., 1.),
      _ESVPCOM( 1.,-2.),_ESVPCOM(0.,-3.),_ESVPCOM(0.,-1.),_ESVPCOM(-1.,-2.),
      _ESVPCOM( 0.,-3.),_ESVPCOM( 0.,-3.),_ESVPCOM( 2.,-3.),
      _ESVPCOM( 5., 2.),_ESVPCOM( 0.,-5.),_ESVPCOM(1.,-2.),_ESVPCOM(0.,-1.),
      _ESVPCOM( 1.,-4.),_ESVPCOM( 1.,-4.),_ESVPCOM( 0.,-3.),
     _ESVPCOM( 7., 2.),_ESVPCOM( 0.,-3.),_ESVPCOM(2.,-1.),_ESVPCOM(-1.,-4.),
      _ESVPCOM( 0.,-1.),_ESVPCOM(-1.,-4.),_ESVPCOM( 1.,-2.),
     _ESVPCOM( 8., 3.),_ESVPCOM(-1.,-4.),_ESVPCOM( 3.,0.),_ESVPCOM(-1.,-4.),
      _ESVPCOM( 2.,-3.),_ESVPCOM( 0.,-5.),_ESVPCOM( 0.,-3.),
      _ESVPCOM( 8., 5.),_ESVPCOM( 0.,-3.),_ESVPCOM(-2.,-3.),_ESVPCOM(3.,0.),
      _ESVPCOM( 1., 0.),_ESVPCOM( 0.,-3.),_ESVPCOM( 0.,-3.),
      _ESVPCOM( 8., 7.),_ESVPCOM( 1.,-2.),_ESVPCOM(0.,-3.),_ESVPCOM(1.,-2.),
      _ESVPCOM( 0.,-3.),_ESVPCOM( 0.,-3.),_ESVPCOM(-1.,-2.)};
 
  int desc_c[DLEN_];
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 8;
  desc_a[N_]    = 9;
  desc_a[MB_]   = 2;
  desc_a[NB_]   = 4;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_c[DT_]   = BLOCK_CYCLIC_2D;
  desc_c[M_]    = 9;
  desc_c[N_]    = 8;
  desc_c[MB_]   = 4;
  desc_c[NB_]   = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(9,4,myrow,0,nprow);
  desc_c[LLD_]  = MAX(1,il_lda);
 
  modname = "PDTRAN  ";
  pdmtxcpy(8, 9, 1, 1, dta2, desc_a, dtas2);
  pdmtxcpy(9, 8, 1, 1, dtc2, desc_c, dtcs2);
 
  pdtran(9, 8, 1.0, dta2, 1, 1, desc_a, 1.0, dtc2, 1, 1, desc_c);
  pdmtchk(modname,9, 8, 1, 1, dtc2, desc_c, dresc2, tol, ierr, ilerr,
          error, ifail,strlen(modname));
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 8;
  desc_a[N_]    = 7;
  desc_a[MB_]   = 2;
  desc_a[NB_]   = 3;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_c[DT_]   = BLOCK_CYCLIC_2D;
  desc_c[M_]    = 7;
  desc_c[N_]    = 8;
  desc_c[MB_]   = 3;
  desc_c[NB_]   = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(7,3,myrow,0,nprow);
  desc_c[LLD_]  = MAX(1,il_lda);
 
  modname = "PZTRANU ";
  alpha = _ESVPCOM(1.0,0.0);
  beta  = _ESVPCOM(1.0,0.0);
  pzmtxcpy(8, 7, 1, 1, zta2, desc_a, ztas2);
  pzmtxcpy(7, 8, 1, 1, ztc2, desc_c, ztcs2);
 
  pztranu(7, 8,alpha, zta2, 1, 1, desc_a, beta, ztc2, 1, 1, desc_c);
  pzmtchk(modname,7, 8, 1, 1, ztc2, desc_c, zresc2u, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  modname = "PZTRANC ";
  pzmtxcpy(8, 7, 1, 1, zta2, desc_a, ztas2);
  pzmtxcpy(7, 8, 1, 1, ztc2, desc_c, ztcs2);
 
  pztranc(7, 8,alpha, zta2, 1, 1, desc_a, beta, ztc2, 1, 1, desc_c);
  pzmtchk(modname,7, 8, 1, 1, ztc2, desc_c, zresc2c, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
// PDTRAN, PZTRANU, PZTRANC END
 
// PDSYR2K, PZSYR2K, PZHER2K   START
 
   int i, desc_b[DLEN_];
   double dta3[72], dtb3[72], dtc3[81];
 
   for (i=0; i<81; i++)
     dtc3[i] = 0.0;
 
   double dtas3[9][8] = { 0.,  0.,  0.,  0.,  1.,  1.,  0.,  -1.,
                         -1.,  1.,  0.,  1.,  0.,  0.,  0.,   0.,
                         -1.,  0., -1.,  0.,  0.,  0., -1.,   0.,
                          0.,  1., -1., -1.,  0.,  0.,  0.,   0.,
                          0.,  0.,  0.,  1., -1.,  1., -1.,   0.,
                          0.,  1.,  0.,  1.,  0.,  1.,  0.,   0.,
                          0.,  0.,  1.,  0.,  0.,  0.,  0.,  -1.,
                          0.,  1.,  0.,  1.,  0.,  0.,  0.,   0.,
                          1.,  1.,  1.,  1.,  1.,  1.,  1.,   1. };
 
 
   double dtbs3[9][8] = { 0.,  0.,  0.,  0., -1., -1.,  0.,   1.,
                          1., -1.,  0., -1.,  0.,  0.,  0.,   0.,
                          1.,  0.,  1.,  0.,  0.,  0.,  1.,   0.,
                          0., -1.,  1.,  1.,  0.,  0.,  0.,   0.,
                          0.,  0.,  0., -1.,  1., -1.,  1.,   0.,
                          0., -1.,  0., -1.,  0., -1.,  0.,   0.,
                          0.,  0., -1.,  0.,  0.,  0.,  0.,   1.,
                          0., -1.,  0., -1.,  0.,  0.,  0.,   0.,
                         -1., -1., -1., -1., -1., -1., -1.,  -1. };
 
   double dresc3[9][9] = {-6.,  0.,  0.,  0.,  0.,  0.,  0.,   0., 0.,
                          0., -6.,  0.,  0.,  0.,  0.,  0.,   0.,  0.,
                          0., -2., -6.,  0.,  0.,  0.,  0.,   0.,  0.,
                          0.,  0., -2., -6.,  0.,  0.,  0.,   0.,  0.,
                          0., -2., -2.,  2., -8.,  0.,  0.,   0.,  0.,
                         -2., -4.,  0.,  0., -4., -6.,  0.,   0.,  0.,
                         -2.,  0.,  2.,  2.,  0.,  0., -4.,   0.,  0.,
                          0., -4.,  0.,  0., -2., -4.,  0.,  -4.,  0.,
                         -2., -2.,  6.,  2.,  0., -6.,  0.,  -4., -16. };
 
      _ESVPCOM    zta3[56], ztb3[56], ztc3[49];
      _ESVPCOM ztas3[56] = {
      _ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM(0., 1.),_ESVPCOM(0.,1.),
      _ESVPCOM( 1., 2.),_ESVPCOM( 1., 2.),_ESVPCOM(0., 1.),_ESVPCOM(-1.,0.),
      _ESVPCOM(-1., 0.),_ESVPCOM( 1., 2.),_ESVPCOM(0., 1.),_ESVPCOM(1.,2.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM(0., 1.),_ESVPCOM(0.,1.),
      _ESVPCOM(-1., 0.),_ESVPCOM( 0., 1.),_ESVPCOM(-1., 0.),_ESVPCOM(0.,1.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM(-1.,0.),_ESVPCOM( 0.,1.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 1., 2.),_ESVPCOM(-1.,0.),_ESVPCOM(-1.,0.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM(0.,1.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM(1.,2.),
      _ESVPCOM(-1., 0.),_ESVPCOM( 1., 2.),_ESVPCOM(-1., 0.),_ESVPCOM(0.,1.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 1., 2.),_ESVPCOM( 0., 1.),_ESVPCOM(1.,2.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 1., 2.),_ESVPCOM( 0., 1.),_ESVPCOM(0.,1.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 0., 1.),_ESVPCOM( 1.,2.),_ESVPCOM(0.,1.),
      _ESVPCOM( 0., 1.),_ESVPCOM( 0.,1.),_ESVPCOM( 0.,1.),_ESVPCOM(-1.,0.)};
 
      _ESVPCOM ztbs3[56] = {
      _ESVPCOM(0.,-2.),_ESVPCOM( 0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),
      _ESVPCOM(-1.,-3.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(1.,-1.),
      _ESVPCOM(1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),
      _ESVPCOM( 0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM( 0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM(0.,-2.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 0.,-2.),_ESVPCOM( 0.,-2.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(1.,-1.),
      _ESVPCOM( 0.,-2.),_ESVPCOM( 0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),
      _ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM( 0.,-2.),_ESVPCOM(-1.,-3.),
      _ESVPCOM( 1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),
      _ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),
      _ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(1.,-1.)};
 
      _ESVPCOM zresc3[49] = {
      _ESVPCOM(38.,-18.),_ESVPCOM(0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM( 0.,  0.),
      _ESVPCOM( 0.,  0.),_ESVPCOM(  0.,  0.),_ESVPCOM( 0.,  0.),
      _ESVPCOM(38.,- 6.),_ESVPCOM( 38.,-18.),_ESVPCOM( 0.,  0.),
      _ESVPCOM(0.,0.),
      _ESVPCOM ( 0.,  0.),_ESVPCOM(  0.,  0.),_ESVPCOM( 0.,  0.),
      _ESVPCOM(26.,6.),_ESVPCOM(26.,2.),_ESVPCOM(14.,6.),_ESVPCOM(0.,0.),
      _ESVPCOM ( 0.,  0.),_ESVPCOM(  0.,  0.),_ESVPCOM( 0.,  0.),
      _ESVPCOM(32.,0.),_ESVPCOM(32.,0.),_ESVPCOM(20.,8.),_ESVPCOM(26.,-6.),
      _ESVPCOM (0.,  0.),_ESVPCOM(  0.,  0.),_ESVPCOM( 0.,  0.),
      _ESVPCOM(35.,- 3.),_ESVPCOM(35.,- 7.),_ESVPCOM(23.,5.),
      _ESVPCOM(29.,7.),
      _ESVPCOM  (32.,-16.),_ESVPCOM(  0.,  0.),_ESVPCOM( 0.,  0.),
      _ESVPCOM (44.,-16.),_ESVPCOM( 44.,-20.),_ESVPCOM(32.,  0.),
      _ESVPCOM(38.,-6.),
      _ESVPCOM  (41.,-17.),_ESVPCOM( 50.,-30.),_ESVPCOM( 0.,  0.),
      _ESVPCOM(35.,- 7.),_ESVPCOM( 35.,- 3.),_ESVPCOM(23., 13.),
      _ESVPCOM (29.,7.),
      _ESVPCOM (32., 0.),_ESVPCOM( 41.,- 9.),_ESVPCOM(32., -8.)};
 
      _ESVPCOM zresc3h[49] = {
      _ESVPCOM(-50.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),
      _ESVPCOM(  0.,  0.),_ESVPCOM(  0.,  0.),_ESVPCOM(0.,0.),
      _ESVPCOM(-38.,0.),_ESVPCOM(-50.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),
      _ESVPCOM (  0.,  0.),_ESVPCOM(  0.,  0.),_ESVPCOM( 0.,  0.),
      _ESVPCOM(-26.,-12.),_ESVPCOM(-30.,-12.),_ESVPCOM(-26.,0.),
      _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),
      _ESVPCOM(-32.,-6.),_ESVPCOM(-32., -6.),_ESVPCOM(-24., 6.),
      _ESVPCOM(-38.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),
      _ESVPCOM(-35.,- 3.),_ESVPCOM(-39.,- 3.),_ESVPCOM(-27., 9.),
      _ESVPCOM(-25.,3.),_ESVPCOM(-48.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),
      _ESVPCOM(-48.,  6.),_ESVPCOM(-52.,  6.),_ESVPCOM(-32.,18.),
      _ESVPCOM(-38.,12.),_ESVPCOM(-49.,9.),_ESVPCOM(-62.,0.),_ESVPCOM(0.,0.),
      _ESVPCOM(-39.,- 3.),_ESVPCOM(-35.,- 3.),_ESVPCOM(-19., 9.),
      _ESVPCOM(-25.,3.),_ESVPCOM(-32.,0.),_ESVPCOM(-41.,- 9.),
      _ESVPCOM(-40.,0.)};
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 8;
  desc_a[N_]    = 9;
  desc_a[MB_]   = 2;
  desc_a[NB_]   = 4;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_b[DT_]   = BLOCK_CYCLIC_2D;
  desc_b[M_]    = 8;
  desc_b[N_]    = 9;
  desc_b[MB_]   = 2;
  desc_b[NB_]   = 4;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_b[LLD_]  = MAX(1,il_lda);
 
  desc_c[DT_]   = BLOCK_CYCLIC_2D;
  desc_c[M_]    = 9;
  desc_c[N_]    = 9;
  desc_c[MB_]   = 4;
  desc_c[NB_]   = 4;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(9,4,myrow,0,nprow);
  desc_c[LLD_]  = MAX(1,il_lda);
 
  modname = "PDSYR2K ";
  pdmtxcpy(8, 9, 1, 1, dta3, desc_a, dtas3);
  pdmtxcpy(8, 9, 1, 1, dtb3, desc_b, dtbs3);
 
  pdsyr2k("U","T", 9, 8, 1.0, dta3, 1, 1, desc_a, dtb3, 1, 1, desc_b, 0.0,
          dtc3, 1, 1, desc_c);
  pdmtchk(modname,9, 9, 1, 1, dtc3, desc_c, dresc3, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 8;
  desc_a[N_] = 7;
  desc_a[MB_] = 2;
  desc_a[NB_] = 3;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_b[M_] = 8;
  desc_b[N_] = 7;
  desc_b[MB_] = 2;
  desc_b[NB_] = 3;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_b[LLD_] = MAX(1,il_lda);
 
  desc_c[DT_] = BLOCK_CYCLIC_2D;
  desc_c[M_] = 7;
  desc_c[N_] = 7;
  desc_c[MB_] = 3;
  desc_c[NB_] = 3;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(7,3,myrow,0,nprow);
  desc_c[LLD_] = MAX(1,il_lda);
 
   for (i=0; i<49; i++)
   ztc3[i]=_ESVPCOM(0.0,0.0);
  modname = "PZSYR2K ";
  alpha=_ESVPCOM(1.0,0.0);
  beta=_ESVPCOM(0.0,0.0);
  pzmtxcpy(8, 7, 1, 1, zta3, desc_a, ztas3);
  pzmtxcpy(8, 7, 1, 1, ztb3, desc_b, ztbs3);
 
  pzsyr2k("U","T", 7, 8,alpha, zta3,1,1,desc_a,ztb3, 1,1, desc_b,beta,
          ztc3, 1, 1, desc_c);
  pzmtchk(modname,7, 7, 1, 1, ztc3, desc_c, zresc3, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
   for (i=0; i<49; i++)
   ztc3[i]=_ESVPCOM(0.0,0.0);
  modname = "PZHER2K ";
  pzmtxcpy(8, 7, 1, 1, zta3, desc_a, ztas3);
  pzmtxcpy(8, 7, 1, 1, ztb3, desc_b, ztbs3);
 
  pzher2k("U","C", 7, 8,alpha, zta3,1,1,desc_a,ztb3, 1,1, desc_b, 0.0,
          ztc3, 1, 1, desc_c);
  pzmtchk(modname,7, 7, 1, 1, ztc3, desc_c, zresc3h, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
// PDSYR2K, PZSYR2K, PZHER2K  END
 
// PDSYMM, PZSYMM, PZHEMM START
 
   double dta4[64], dtb4[128], dtc4[128];
   double dtas4[8][8] = { 0.,  0.,  0.,  0.,  0.,  0.,  0.,   0.,
                         -1.,  1.,  0.,  0.,  0.,  0.,  0.,   0.,
                         -1.,  0., -1.,  0.,  0.,  0.,  0.,   0.,
                          0.,  1., -1., -1.,  0.,  0.,  0.,   0.,
                          0.,  0.,  0.,  1., -1.,  0.,  0.,   0.,
                          0.,  1.,  0.,  1.,  0.,  1.,  0.,   0.,
                          0.,  0.,  1.,  0.,  0.,  0.,  0.,   0.,
                          0.,  1.,  0.,  1.,  0.,  0.,  0.,   0. };
 
   double dtbs4[8][16] = {-1., -1.,  1.,  0.,  0.,  0.,  1.,   0.,
                           0., -1.,  0.,  0.,  1.,  0.,  0.,  -1.,
                           0., -1.,  1., -1.,  1.,  0.,  1.,   0.,
                           0., -1.,  0.,  0.,  1.,  0.,  1.,   0.,
                           1.,  1., -1.,  0.,  0.,  1.,  0.,  -1.,
                           0.,  1.,  0.,  1., -1.,  0.,  0.,  -1.,
                          -1.,  0.,  0.,  0.,  1.,  0.,  0.,   0.,
                          -1.,  0.,  1.,  1.,  0.,  0.,  0.,   0.,
                           1.,  1., -1.,  0.,  0., -1.,  1.,   0.,
                           1.,  0.,  1.,  0., -1.,  1.,  0.,   0.,
                           1., -1.,  0.,  0.,  1., -1.,  1.,   1.,
                           1., -1.,  0., -1., -1.,  0.,  0.,   1.,
                          -1., -1.,  1.,  0.,  1.,  0.,  0.,   0.,
                           0.,  0.,  0.,  0.,  1.,  0.,  0.,   1.,
                          -1.,  1.,  0., -1.,  0.,  0., -1.,   1.,
                           1.,  1.,  0.,  0.,  1., -1.,  0.,   0. };
 
   double dresc4[8][16] = {-1.,  0.,  0.,  1., -1., -1., -1.,   1.,
                            0.,  0.,  0., -1.,  0.,  0., -1.,   1.,
                            0.,  0.,  0., -2.,  3., -1.,  0.,   2.,
                            1.,  0.,  1.,  0.,  0., -1.,  1.,   2.,
                            0., -1.,  1.,  0.,  0., -1., -1.,   1.,
                            1.,  0., -1., -2.,  1.,  0.,  0.,   3.,
                            1., -1.,  1., -2.,  1., -3.,  2.,   3.,
                            4., -2.,  0., -3.,  1.,  0.,  1.,   2.,
                           -2., -1.,  1.,  0.,  1.,  1., -1.,   0.,
                           -2.,  0.,  0.,  1.,  1., -1.,  0.,   0.,
                            0., -2.,  1., -1.,  3., -1.,  2.,   1.,
                            0., -2.,  1.,  0.,  0.,  0.,  1.,   1.,
                            1.,  1., -1.,  0.,  0.,  1.,  0.,  -1.,
                            0.,  1.,  0.,  1., -1.,  0.,  0.,  -1.,
                           -1., -1.,  1., -1.,  2.,  0.,  1.,   0.,
                           -1., -1.,  1.,  1.,  1.,  0.,  1.,   0. };
 
      _ESVPCOM zta4[64], ztb4[128], ztc4[128];
      _ESVPCOM ztas4[64] = {
      _ESVPCOM(0.,1.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),
      _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM(-1.,0.),_ESVPCOM(1., 2.),_ESVPCOM(0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM(0.,0.),_ESVPCOM(0., 0.),_ESVPCOM(0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM(-1.,0.),_ESVPCOM(0., 1.),_ESVPCOM(-1., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM(0.,0.),_ESVPCOM(0.,0.), _ESVPCOM(0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM (0.,1.),_ESVPCOM(1., 2.),_ESVPCOM(-1., 0.),_ESVPCOM(-1.,0.),
      _ESVPCOM (0.,0.),_ESVPCOM(0., 0.),_ESVPCOM (0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM (0.,1.),_ESVPCOM(0., 1.),_ESVPCOM (0., 1.),_ESVPCOM (1.,2.),
      _ESVPCOM (-1.,0.),_ESVPCOM (0., 0.),_ESVPCOM(0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM (0.,1.),_ESVPCOM(1., 2.),_ESVPCOM (0., 1.),_ESVPCOM (1.,2.),
      _ESVPCOM (0.,1.),_ESVPCOM(1., 2.),_ESVPCOM (0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM (0.,1.),_ESVPCOM(0., 1.),_ESVPCOM (1., 2.),_ESVPCOM (0.,1.),
      _ESVPCOM (0.,1.),_ESVPCOM(0.,1.),_ESVPCOM(0., 1.),_ESVPCOM(0., 0.),
      _ESVPCOM (0.,1.),_ESVPCOM(1., 2.),_ESVPCOM (0., 1.),_ESVPCOM (1.,2.),
      _ESVPCOM (0.,1.),_ESVPCOM(0., 1.),_ESVPCOM (0., 1.),_ESVPCOM(0., 1.)};
 
      _ESVPCOM ztbs4[128] = {
      _ESVPCOM(-1.,-3.),_ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 0.,-2.),_ESVPCOM( 0.,-2.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),
      _ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(-1.,-3.),
      _ESVPCOM(1.,-1.),_ESVPCOM( 0.,-2.),_ESVPCOM(1.,-1.),_ESVPCOM( 0.,-2.),
      _ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM( 0.,-2.),
      _ESVPCOM(1.,-1.),_ESVPCOM( 0.,-2.),_ESVPCOM(1.,-1.),_ESVPCOM( 0.,-2.),
      _ESVPCOM(1.,-1.),_ESVPCOM( 1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),
      _ESVPCOM(0.,-2.),_ESVPCOM( 1.,-1.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),
      _ESVPCOM(0.,-2.),_ESVPCOM( 1.,-1.),_ESVPCOM(0.,-2.),_ESVPCOM( 1.,-1.),
      _ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),
      _ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM( 0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM( 0.,-2.),
      _ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(1.,-1.),_ESVPCOM( 1.,-1.),
      _ESVPCOM( 0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM( 0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM(1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM( 0.,-2.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),
      _ESVPCOM(-1.,-3.),_ESVPCOM( 1.,-1.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(1.,-1.),
      _ESVPCOM(1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),
      _ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM( 1.,-1.),
      _ESVPCOM(-1.,-3.),_ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM( 0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 0.,-2.),_ESVPCOM( 0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM( 0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(1.,-1.),
      _ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),
      _ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM( 1.,-1.),
      _ESVPCOM(1.,-1.),_ESVPCOM(1.,-1.),_ESVPCOM( 0.,-2.),_ESVPCOM( 0.,-2.),
      _ESVPCOM(1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.)};
 
      _ESVPCOM zresc4[128] = {
      _ESVPCOM(11.,27.),_ESVPCOM( 7.,29.),_ESVPCOM(1.,27.),_ESVPCOM(6.,32.),
      _ESVPCOM(-4.,22.),_ESVPCOM(11.,27.),_ESVPCOM(-1.,23.),_ESVPCOM(-3.,29.),
      _ESVPCOM(-2.,26.),_ESVPCOM( 7.,29.),_ESVPCOM(-2.,26.),_ESVPCOM(5.,25),
      _ESVPCOM( 1.,27.),_ESVPCOM(4.,28.),_ESVPCOM(5.,25.),_ESVPCOM(0.,30.),
      _ESVPCOM(37.,39.),_ESVPCOM(37.,39.),_ESVPCOM(31.,37.),_ESVPCOM(48.,36.),
      _ESVPCOM(10.,40.),_ESVPCOM(41.,37.),_ESVPCOM(25.,35.),_ESVPCOM(23.,41.),
      _ESVPCOM(24.,38.),_ESVPCOM(37.,39.),_ESVPCOM(24.,38.),_ESVPCOM(31.,37.),
      _ESVPCOM(31.,37.),_ESVPCOM(38.,36.),_ESVPCOM(27.,39.),_ESVPCOM(26.,42.),
      _ESVPCOM( 7.,29.),_ESVPCOM(11.,27.),_ESVPCOM(-3.,29.),_ESVPCOM(10.,30.),
      _ESVPCOM(-8.,24.),_ESVPCOM(11.,27.),_ESVPCOM(-1.,23.),_ESVPCOM(-3.,29.),
      _ESVPCOM(-6.,28.),_ESVPCOM( 7.,29.),_ESVPCOM(2.,24.),_ESVPCOM(9.,23.),
      _ESVPCOM(-3.,29.),_ESVPCOM(4.,28.),_ESVPCOM(1.,27.),_ESVPCOM(-8.,34.),
      _ESVPCOM(27.,39.),_ESVPCOM(35.,35.),_ESVPCOM(21.,37.),_ESVPCOM(42.,34.),
      _ESVPCOM(12.,34.),_ESVPCOM(43.,31.),_ESVPCOM(11.,37.),_ESVPCOM(13.,41.),
      _ESVPCOM( 6.,42.),_ESVPCOM(39.,33.),_ESVPCOM(22.,34.),_ESVPCOM(37.,29.),
      _ESVPCOM(21.,37.),_ESVPCOM(28.,36.),_ESVPCOM(21.,37.),_ESVPCOM(20.,40.),
      _ESVPCOM(27.,29.),_ESVPCOM(23.,31.),_ESVPCOM( 9.,33.),_ESVPCOM(22.,34.),
      _ESVPCOM( 0.,30.),_ESVPCOM(15.,35.),_ESVPCOM(11.,27.),_ESVPCOM(13.,31.),
      _ESVPCOM(18.,26.),_ESVPCOM(19.,33.),_ESVPCOM(10.,30.),_ESVPCOM( 9.,33.),
      _ESVPCOM( 9.,33.),_ESVPCOM(20.,30.),_ESVPCOM(13.,31.),_ESVPCOM(16.,32.),
      _ESVPCOM(37.,39.),_ESVPCOM(45.,35.),_ESVPCOM(27.,39.),_ESVPCOM(44.,38.),
      _ESVPCOM(10.,40.),_ESVPCOM(41.,37.),_ESVPCOM(17.,39.),_ESVPCOM(27.,39.),
      _ESVPCOM(28.,36.),_ESVPCOM(45.,35.),_ESVPCOM(24.,38.),_ESVPCOM(31.,37.),
      _ESVPCOM(31.,37.),_ESVPCOM(34.,38.),_ESVPCOM(27.,39.),_ESVPCOM(30.,40.),
      _ESVPCOM(21.,37.),_ESVPCOM(21.,37.),_ESVPCOM(23.,31.),_ESVPCOM(28.,36.),
      _ESVPCOM(10.,30.),_ESVPCOM(21.,37.),_ESVPCOM(13.,31.),_ESVPCOM(23.,31.),
      _ESVPCOM(16.,32.),_ESVPCOM(21.,37.),_ESVPCOM(16.,32.),_ESVPCOM(15.,35.),
      _ESVPCOM(23.,31.),_ESVPCOM(22.,34.),_ESVPCOM(19.,33.),_ESVPCOM(26.,32.),
      _ESVPCOM(35.,35.),_ESVPCOM(35.,35.),_ESVPCOM(21.,37.),_ESVPCOM(38.,36.),
      _ESVPCOM( 8.,36.),_ESVPCOM(31.,37.),_ESVPCOM(15.,35.),_ESVPCOM(25.,35.),
      _ESVPCOM(26.,32.),_ESVPCOM(35.,35.),_ESVPCOM(18.,36.),_ESVPCOM(21.,37.),
      _ESVPCOM(21.,37.),_ESVPCOM(28.,36.),_ESVPCOM(21.,37.),_ESVPCOM(28.,36.)};
 
      _ESVPCOM    zta4h[64], ztb4h[128], ztc4h[128];
      _ESVPCOM ztas4h[64] = {
      _ESVPCOM(0.,0.),_ESVPCOM(0., 0.),_ESVPCOM(0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM(0.,0.), _ESVPCOM(0., 0.),_ESVPCOM(0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM(-1.,0.),_ESVPCOM (1., 0.),_ESVPCOM (0.,0.),_ESVPCOM(0., 0.),
      _ESVPCOM(0.,0.),_ESVPCOM (0., 0.),_ESVPCOM(0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM(-1.,0.),_ESVPCOM(0., 1.),_ESVPCOM(-1., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM(0.,0.),_ESVPCOM (0., 0.),_ESVPCOM (0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM(0.,1.),_ESVPCOM (1., 2.),_ESVPCOM(-1., 0.),_ESVPCOM(-1.,0.),
      _ESVPCOM(0.,0.),_ESVPCOM (0., 0.),_ESVPCOM (0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM(0.,1.),_ESVPCOM (0., 1.),_ESVPCOM(0., 1.),_ESVPCOM(1.,2.),
      _ESVPCOM(-1.,0.),_ESVPCOM (0., 0.),_ESVPCOM(0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM(0.,1.),_ESVPCOM(1., 2.),_ESVPCOM(0., 1.),_ESVPCOM(1.,2.),
      _ESVPCOM(0.,1.),_ESVPCOM(1., 0.),_ESVPCOM(0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM(0.,1.),_ESVPCOM(0., 1.),_ESVPCOM(1., 2.),_ESVPCOM(0.,1.),
      _ESVPCOM(0.,1.),_ESVPCOM(0., 1.),_ESVPCOM (0., 0.),_ESVPCOM(0., 0.),
      _ESVPCOM(0.,1.),_ESVPCOM(1., 2.),_ESVPCOM(0., 1.),_ESVPCOM(1.,2.),
      _ESVPCOM(0.,1.),_ESVPCOM(0., 1.),_ESVPCOM(0., 1.),_ESVPCOM(0., 0.)};
 
      _ESVPCOM ztbs4h[128] = {
      _ESVPCOM(-1.,-3.),_ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 0.,-2.),_ESVPCOM( 0.,-2.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),
      _ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(-1.,-3.),
      _ESVPCOM(1.,-1.),_ESVPCOM( 0.,-2.),_ESVPCOM(1.,-1.),_ESVPCOM( 0.,-2.),
      _ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM( 0.,-2.),
      _ESVPCOM(1.,-1.),_ESVPCOM( 0.,-2.),_ESVPCOM(1.,-1.),_ESVPCOM( 0.,-2.),
      _ESVPCOM(1.,-1.),_ESVPCOM( 1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),
      _ESVPCOM(0.,-2.),_ESVPCOM( 1.,-1.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),
      _ESVPCOM(0.,-2.),_ESVPCOM( 1.,-1.),_ESVPCOM(0.,-2.),_ESVPCOM( 1.,-1.),
      _ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),
      _ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM( 0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM( 0.,-2.),
      _ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(1.,-1.),_ESVPCOM( 1.,-1.),
      _ESVPCOM( 0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM( 0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM(1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM( 0.,-2.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),
      _ESVPCOM(-1.,-3.),_ESVPCOM( 1.,-1.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(1.,-1.),
      _ESVPCOM(1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),
      _ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM( 1.,-1.),
      _ESVPCOM(-1.,-3.),_ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM( 0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 0.,-2.),_ESVPCOM( 0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),
      _ESVPCOM( 1.,-1.),_ESVPCOM( 0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(1.,-1.),
      _ESVPCOM(-1.,-3.),_ESVPCOM(1.,-1.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),
      _ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.),_ESVPCOM(-1.,-3.),_ESVPCOM( 1.,-1.),
      _ESVPCOM(1.,-1.),_ESVPCOM(1.,-1.),_ESVPCOM( 0.,-2.),_ESVPCOM( 0.,-2.),
      _ESVPCOM(1.,-1.),_ESVPCOM(-1.,-3.),_ESVPCOM(0.,-2.),_ESVPCOM(0.,-2.)};
 
      _ESVPCOM zresc4h[128] = {
      _ESVPCOM(-12.,4.),_ESVPCOM(-10.,4.),_ESVPCOM(-10.,4.),_ESVPCOM(-10.,6.),
      _ESVPCOM(-8.,0.),_ESVPCOM(-13.,5.),_ESVPCOM(-10.,2.),_ESVPCOM(-7., 3.),
      _ESVPCOM(-8.,2.),_ESVPCOM(-10.,4.),_ESVPCOM(-8.,2.),_ESVPCOM(-11.,3.),
      _ESVPCOM(-10.,4.),_ESVPCOM(-10.,4.),_ESVPCOM(-11.,3.),_ESVPCOM(-7.,3.),
      _ESVPCOM(-19.,-5.),_ESVPCOM(-17.,-7.),_ESVPCOM(-19.,-5.),_ESVPCOM(-22.,-6.),
      _ESVPCOM(-10.,-8.),_ESVPCOM(-21.,-5.),_ESVPCOM(-17.,-7.),_ESVPCOM(-13.,-7.),
      _ESVPCOM(-14.,-8.),_ESVPCOM(-17.,-7.),_ESVPCOM(-14.,-8.),_ESVPCOM(-17.,-7.),
      _ESVPCOM(-19.,-5.),_ESVPCOM(-20.,-6.),_ESVPCOM(-17.,-5.),_ESVPCOM(-14.,-6.),
      _ESVPCOM(-9.,5.),_ESVPCOM(- 9.,3.),_ESVPCOM(- 7.,5.),_ESVPCOM(- 8.,4.),
      _ESVPCOM(-6.,2.),_ESVPCOM(-11.,5.),_ESVPCOM(- 9.,3.),_ESVPCOM(- 5.,3.),
      _ESVPCOM(-4.,2.),_ESVPCOM(- 7.,3.),_ESVPCOM(- 8.,2.),_ESVPCOM(-11.,3.),
      _ESVPCOM(-7.,5.),_ESVPCOM(- 8.,4.),_ESVPCOM(- 9.,5.),_ESVPCOM(- 2.,4.),
      _ESVPCOM(-5.,-5.),_ESVPCOM(- 5.,-9.),_ESVPCOM(-11.,1.),_ESVPCOM(-10.,-6.),
      _ESVPCOM(-6.,-4.),_ESVPCOM(-15.,-3.),_ESVPCOM(- 7.,-1.),_ESVPCOM(- 1.,-5.),
      _ESVPCOM(2.,-6.),_ESVPCOM(- 7.,-9.),_ESVPCOM(- 6.,-6.),_ESVPCOM(-13.,-5.),
      _ESVPCOM(-11., 1.),_ESVPCOM(-8.,-4.),_ESVPCOM(-9.,-1.),_ESVPCOM(-2.,-6.),
      _ESVPCOM( 3.,-3.),_ESVPCOM(3.,-1.),_ESVPCOM(5.,1.),_ESVPCOM(4.,0.),
      _ESVPCOM(4., 2.),_ESVPCOM(3.,3.),_ESVPCOM(1.,1.),_ESVPCOM(7.,-3.),
      _ESVPCOM(6.,-6.),_ESVPCOM(  5.,-1.),_ESVPCOM(2.,2.),_ESVPCOM(1.,5.),
      _ESVPCOM(5., 1.),_ESVPCOM(2., 0.),_ESVPCOM(3.,1.),_ESVPCOM(8.,-4.),
      _ESVPCOM(9.,-5.),_ESVPCOM(9.,-9.),_ESVPCOM(11.,-5.),_ESVPCOM( 10.,-8.),
      _ESVPCOM(10., 0.),_ESVPCOM(9.,-7.),_ESVPCOM(7., 1.),_ESVPCOM( 13.,-7.),
      _ESVPCOM(12.,-8.),_ESVPCOM(11.,-11.),_ESVPCOM(8.,-2.),_ESVPCOM(7.,-3.),
      _ESVPCOM(11.,-7.),_ESVPCOM(8.,-4.),_ESVPCOM(9.,-3.),_ESVPCOM(14.,-8.),
      _ESVPCOM(10., 2.),_ESVPCOM(14.,-2.),_ESVPCOM(12.,-4.),_ESVPCOM(12.,-2.),
      _ESVPCOM(9., 1.),_ESVPCOM(13.,-1.),_ESVPCOM(7.,3.),_ESVPCOM(13.,-5.),
      _ESVPCOM(12.,-2.),_ESVPCOM(15., -3.),_ESVPCOM(10.,0.),_ESVPCOM(11.,1.),
      _ESVPCOM(14.,-6.),_ESVPCOM(10.,0.),_ESVPCOM(11.,-1.),_ESVPCOM(13.,-5.),
      _ESVPCOM(18.,-6.),_ESVPCOM(20.,-8.),_ESVPCOM(17.,-1.),_ESVPCOM(19.,-7.),
      _ESVPCOM(14.,4.),_ESVPCOM(19.,-5.),_ESVPCOM(14.,2.),_ESVPCOM(18.,-4.),
      _ESVPCOM(17.,-5.),_ESVPCOM(20.,-8.),_ESVPCOM(16.,0.),_ESVPCOM(17.,-1.),
      _ESVPCOM(18.,-2.),_ESVPCOM(17.,-3.),_ESVPCOM(17.,-1.),_ESVPCOM(18.,-4.)};
 
   for (i=0; i<128; i++)
     dtc4[i] = 0.0;
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 8;
  desc_a[N_]    = 8;
  desc_a[MB_]   = 2;
  desc_a[NB_]   = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_b[DT_]   = BLOCK_CYCLIC_2D;
  desc_b[M_]    = 16;
  desc_b[N_]    = 8;
  desc_b[MB_]   = 4;
  desc_b[NB_]   = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(16,4,myrow,0,nprow);
  desc_b[LLD_]  = MAX(1,il_lda);
 
  desc_c[DT_]   = BLOCK_CYCLIC_2D;
  desc_c[M_]    = 16;
  desc_c[N_]    = 8;
  desc_c[MB_]   = 4;
  desc_c[NB_]   = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(16,4,myrow,0,nprow);
  desc_c[LLD_]  = MAX(1,il_lda);
 
  modname = "PDSYMM  ";
  pdmtxcpy(8, 8, 1, 1, dta4, desc_a, dtas4);
  pdmtxcpy(16, 8, 1, 1, dtb4, desc_b, dtbs4);
 
  pdsymm("R","U", 16, 8, 1.0, dta4, 1, 1, desc_a, dtb4, 1, 1, desc_b, 0.0,
          dtc4, 1, 1, desc_c);
  pdmtchk(modname,16, 8, 1, 1, dtc4, desc_c, dresc4, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
   for (i=0; i<128; i++)
     ztc4[i] = _ESVPCOM(0.0,0.0);
     alpha =_ESVPCOM(1.0,2.0);
     beta = _ESVPCOM( 0.0,0.0);
 
  modname = "PZSYMM  ";
  pzmtxcpy(8, 8, 1, 1, zta4, desc_a, ztas4);
  pzmtxcpy(16, 8, 1, 1, ztb4, desc_b, ztbs4);
 
  pzsymm("R","U", 16, 8,alpha, zta4, 1, 1, desc_a, ztb4, 1, 1, desc_b,
          beta, ztc4, 1, 1, desc_c);
  pzmtchk(modname,16, 8, 1, 1, ztc4, desc_c, zresc4, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
   for (i=0; i<128; i++)
     ztc4h[i] = _ESVPCOM(0.0,0.0);
     alpha =_ESVPCOM(1.0,0.0);
 
  modname = "PZHEMM  ";
  pzmtxcpy(8, 8, 1, 1, zta4h, desc_a, ztas4h);
  pzmtxcpy(16, 8, 1, 1, ztb4h, desc_b, ztbs4h);
 
  pzhemm("R","U", 16, 8,alpha, zta4h, 1, 1, desc_a,ztb4h, 1, 1, desc_b,
          beta, ztc4h, 1, 1, desc_c);
  pzmtchk(modname,16, 8, 1, 1, ztc4h, desc_c, zresc4h, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
// PDSYMM, PZSYMM, PZHEMM END
 
// PDGER, PZGERU, PZGERC START : 5
 
   int desc_y[DLEN_];
   double dta5[100], dtx5[11], dty5[11];
   double dtas5[100];
   double dtxs5[11] = {1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1. };
   double dtys5[11] = {1., 2., 3., 4., 5., 6., 7., 8., 9., 10., 11. };
   for (i=1; i<101; i++)
     dtas5[i-1] = (double)i;
   double dresa5[100] = {
                   1.,   2.,   3.,   4.,   5.,   6.,   7.,   8.,   9.,   10.,
                  11.,  14.,  15.,  16.,  17.,  18.,  19.,  20.,  21.,   22.,
                  21.,  25.,  26.,  27.,  28.,  29.,  30.,  31.,  32.,   33.,
                  31.,  36.,  37.,  38.,  39.,  40.,  41.,  42.,  43.,   44.,
                  41.,  47.,  48.,  49.,  50.,  51.,  52.,  53.,  54.,   55.,
                  51.,  58.,  59.,  60.,  61.,  62.,  63.,  64.,  65.,   66.,
                  61.,  69.,  70.,  71.,  72.,  73.,  74.,  75.,  76.,   77.,
                  71.,  80.,  81.,  82.,  83.,  84.,  85.,  86.,  87.,   88.,
                  81.,  91.,  92.,  93.,  94.,  95.,  96.,  97.,  98.,   99.,
                  91., 102., 103., 104., 105., 106., 107., 108., 109.,  110.};
 
      _ESVPCOM  zta5[100], ztx5[11], zty5[11];
      _ESVPCOM  ztxs5[11] = {
      _ESVPCOM(1., 4.),_ESVPCOM(1.,4.),_ESVPCOM(1.,3.),_ESVPCOM(1.,2.),
      _ESVPCOM(1., 1.),_ESVPCOM(1., 0.),_ESVPCOM(1.,-1.),_ESVPCOM(1.,-2.),
      _ESVPCOM(1.,-3.),_ESVPCOM(1.,-4.),_ESVPCOM(1.,-4.)};
      _ESVPCOM  ztys5[11] = {
      _ESVPCOM(2., 1.),_ESVPCOM(2., 1.),_ESVPCOM(3.,1.),_ESVPCOM(4.,1.),
      _ESVPCOM(5., 1.),_ESVPCOM(6., 1.),_ESVPCOM(7.,1.0),_ESVPCOM(8.,1.),
      _ESVPCOM(9., 1.),_ESVPCOM(10., 1.),_ESVPCOM(10., 1.)};
      _ESVPCOM  ztas5[100] = {
      _ESVPCOM(11.,2.),_ESVPCOM(12.,2.),_ESVPCOM(13.,2.0),_ESVPCOM(14.,2.),
      _ESVPCOM(15.,2.),_ESVPCOM(16.,2.),_ESVPCOM(17.,2.0),_ESVPCOM(18.,2.),
      _ESVPCOM(19.,2.),_ESVPCOM(20.,2.),
      _ESVPCOM(11.,2.),_ESVPCOM(12.,2.),_ESVPCOM(13.,2.0),_ESVPCOM(14.,2.),
      _ESVPCOM(15.,2.),_ESVPCOM(16.,2.),_ESVPCOM(17.,2.0),_ESVPCOM(18.,2.),
      _ESVPCOM (19.,2.),_ESVPCOM(20.,2.),
      _ESVPCOM(21.,1.),_ESVPCOM(22.,1.),_ESVPCOM(23.,1.0),_ESVPCOM(24.,1.),
      _ESVPCOM(25.,1.),_ESVPCOM(26.,1.),_ESVPCOM(27.,1.0),_ESVPCOM(28.,1.),
      _ESVPCOM(29.,1.),_ESVPCOM(30.,1.),
      _ESVPCOM(31.,0.),_ESVPCOM(32.,0.),_ESVPCOM(33.,0.0),_ESVPCOM(34.,0.),
      _ESVPCOM(35.,0.),_ESVPCOM(36.,0.),_ESVPCOM(37.,0.0),_ESVPCOM(38.,0.),
      _ESVPCOM(39.,0.),_ESVPCOM(40.,0.),
      _ESVPCOM(41.,-1.),_ESVPCOM(42.,-1.),_ESVPCOM(43.,-1.0),_ESVPCOM(44.,-1.),
      _ESVPCOM(45.,-1.),_ESVPCOM(46.,-1.),_ESVPCOM(47.,-1.0),_ESVPCOM(48.,-1.),
      _ESVPCOM(49.,-1.),_ESVPCOM(50.,-1.),
      _ESVPCOM(51.,-2.),_ESVPCOM(52.,-2.),_ESVPCOM(53.,-2.0),_ESVPCOM(54.,-2.),
      _ESVPCOM(55.,-2.),_ESVPCOM(56.,-2.),_ESVPCOM(57.,-2.0),_ESVPCOM(58.,-2.),
      _ESVPCOM(59.,-2.),_ESVPCOM(60.,-2.),
      _ESVPCOM(61.,2.),_ESVPCOM(62.,2.),_ESVPCOM(63.,2.0),_ESVPCOM(64.,2.),
      _ESVPCOM(65.,2.),_ESVPCOM(66.,2.),_ESVPCOM(67.,2.0),_ESVPCOM(68.,2.),
      _ESVPCOM(69.,2.),_ESVPCOM(70.,2.),
      _ESVPCOM(71.,1.),_ESVPCOM(72.,1.),_ESVPCOM(73.,1.0),_ESVPCOM(74.,1.),
      _ESVPCOM(75.,1.),_ESVPCOM(76.,1.),_ESVPCOM(77.,1.0),_ESVPCOM(78.,1.),
      _ESVPCOM(79.,1.),_ESVPCOM(80.,1.),
      _ESVPCOM(81.,0.),_ESVPCOM(82.,0.),_ESVPCOM(83.,0.0),_ESVPCOM(84.,0.),
      _ESVPCOM(85.,0.),_ESVPCOM(86.,0.),_ESVPCOM(87.,0.0),_ESVPCOM(88.,0.),
      _ESVPCOM(89.,0.),_ESVPCOM(90.,0.),
      _ESVPCOM(91.,-1.),_ESVPCOM(92.,-1.),_ESVPCOM(93.,-1.0),_ESVPCOM(94.,-1.),
      _ESVPCOM(95.,-1.),_ESVPCOM(96.,-1.),_ESVPCOM(97.,-1.0),_ESVPCOM(98.,-1.),
      _ESVPCOM(99.,-1.),_ESVPCOM(100.,-1.)};
      _ESVPCOM zresa5u[100] = {
      _ESVPCOM(19.,13.),_ESVPCOM(19.,13.),_ESVPCOM(19.,10.0),_ESVPCOM(19.,7.),
      _ESVPCOM(19., 4.),_ESVPCOM(19.,1.),_ESVPCOM(19.,-2.0),_ESVPCOM(19.,-5.),
      _ESVPCOM(19.,-8.),_ESVPCOM(19.,-11.),
      _ESVPCOM(19.,13.),_ESVPCOM(19.,13.),_ESVPCOM(19.,10.0),_ESVPCOM(19.,7.),
      _ESVPCOM(19.,4.),_ESVPCOM(19.,1.),_ESVPCOM(19.,-2.0),_ESVPCOM(19.,-5.),
      _ESVPCOM(19.,-8.),_ESVPCOM(19.,-11.),
      _ESVPCOM(34.,15.),_ESVPCOM(34.,15.),_ESVPCOM(33.,11.0),_ESVPCOM(32.,7.),
      _ESVPCOM(31., 3.),_ESVPCOM(30.,-1.),_ESVPCOM(29.,-5.0),_ESVPCOM(28.,-9.),
      _ESVPCOM(27.,-13.),_ESVPCOM(26.,-17.),
      _ESVPCOM(49.,17.),_ESVPCOM(49.,17.),_ESVPCOM(47.,12.0),_ESVPCOM(45.,7.),
      _ESVPCOM(43., 2.),_ESVPCOM(41.,-3.),_ESVPCOM(39.,-8.0),_ESVPCOM(37.,-13.),
      _ESVPCOM(35.,-18.),_ESVPCOM(33.,-23.),
      _ESVPCOM(64.,19.),_ESVPCOM(64.,19.),_ESVPCOM(61.,13.0),_ESVPCOM(58.,7.),
      _ESVPCOM(55., 1.),_ESVPCOM(52.,-5.),_ESVPCOM(49.,-11.0),_ESVPCOM(46.,-17.),
      _ESVPCOM(43.,-23.),_ESVPCOM(40.,-29.),
      _ESVPCOM(79.,21.),_ESVPCOM(79.,21.),_ESVPCOM(75.,14.0),_ESVPCOM(71.,7.),
      _ESVPCOM(67.,0.),_ESVPCOM(63.,-7.),_ESVPCOM(59.,-14.0),_ESVPCOM(55.,-21.),
      _ESVPCOM(51.,-28.),_ESVPCOM(47.,-35.),
      _ESVPCOM(94.,28.),_ESVPCOM(94.,28.),_ESVPCOM(89.,20.0),_ESVPCOM(84.,12.),
      _ESVPCOM(79.,4.),_ESVPCOM(74.,-4.),_ESVPCOM(69.,-12.0),_ESVPCOM(64.,-20.),
      _ESVPCOM(59.,-28.),_ESVPCOM(54.,-36.),
      _ESVPCOM(109.,30.),_ESVPCOM(109.,30.),_ESVPCOM(103.,21.0),_ESVPCOM(97.,12.),
      _ESVPCOM(91.,3.),_ESVPCOM(85.,-6.),_ESVPCOM(79.,-15.0),_ESVPCOM(73.,-24.),
      _ESVPCOM(67.,-33.),_ESVPCOM(61.,-42.),
      _ESVPCOM(124.,32.),_ESVPCOM(124., 32.),_ESVPCOM(117.,22.0),_ESVPCOM(110.,12.),
      _ESVPCOM(103.,2.),_ESVPCOM(96.,-8.),_ESVPCOM(89.,-18.0),_ESVPCOM(82.,-28.),
      _ESVPCOM(75.,-38.),_ESVPCOM(68.,-48.),
      _ESVPCOM(139.,34.),_ESVPCOM(139.,34.),_ESVPCOM(131.,23.0),_ESVPCOM(123.,12.),
      _ESVPCOM(115.,1.),_ESVPCOM(107.,-10.),_ESVPCOM(99.,-21.0),_ESVPCOM(91.,-32.),
      _ESVPCOM(83.,-43.),_ESVPCOM(75.,-54.)};
 
      _ESVPCOM zresa5c[100] = {
      _ESVPCOM(25.,3.),_ESVPCOM(25.,3.),_ESVPCOM(23.,2.0),_ESVPCOM(21.,1.),
      _ESVPCOM(19., 0.),_ESVPCOM(17.,-1.),_ESVPCOM(5.,-2.0),_ESVPCOM(13.,-3.),
      _ESVPCOM(11.,-4.),_ESVPCOM(9.,-5.),
      _ESVPCOM(25.,3.),_ESVPCOM(25.,3.),_ESVPCOM(23.,2.0),_ESVPCOM(21., 1.),
      _ESVPCOM(19., 0.),_ESVPCOM(17.,-1.),_ESVPCOM(15.,-2.0),_ESVPCOM(13.,-3.),
      _ESVPCOM(11.,-4.),_ESVPCOM(9.,-5.),
      _ESVPCOM(40.,5.),_ESVPCOM(40.,5.),_ESVPCOM(37.,3.0),_ESVPCOM(34., 1.),
      _ESVPCOM(31.,-1.),_ESVPCOM(28.,-3.),_ESVPCOM(25.,-5.0),_ESVPCOM(22.,-7.),
      _ESVPCOM(19.,-9.),_ESVPCOM(16.,-11.),
      _ESVPCOM(55.,7.),_ESVPCOM(55.,7.),_ESVPCOM(51.,4.0),_ESVPCOM(47., 1.),
      _ESVPCOM(43.,-2.),_ESVPCOM(39.,-5.),_ESVPCOM(35.,-8.0),_ESVPCOM(31.,-11.),
      _ESVPCOM(27.,-14.),_ESVPCOM(23.,-17.),
      _ESVPCOM(70.,9.),_ESVPCOM(70.,9.),_ESVPCOM(65.,5.0),_ESVPCOM(60., 1.),
      _ESVPCOM(55.,-3.),_ESVPCOM(50.,-7.),_ESVPCOM(45.,-11.0),_ESVPCOM(40.,-15.),
      _ESVPCOM(35.,-19.),_ESVPCOM(30.,-23.),
      _ESVPCOM(85.,11.),_ESVPCOM(85.,11.),_ESVPCOM(79.,6.0),_ESVPCOM(73., 1.),
      _ESVPCOM(67.,-4.),_ESVPCOM(61.,-9.),_ESVPCOM(55.,-14.0),_ESVPCOM(49.,-19.),
      _ESVPCOM(43.,-24.),_ESVPCOM(37.,-29.),
      _ESVPCOM(100.,18.),_ESVPCOM(100.,18.),_ESVPCOM(93.,12.0),_ESVPCOM(86.,6.),
      _ESVPCOM(79., 0.),_ESVPCOM(72.,-6.),_ESVPCOM(65.,-12.0),_ESVPCOM(58.,-18.),
      _ESVPCOM(51.,-24.),_ESVPCOM(44.,-30.),
      _ESVPCOM(115.,20.),_ESVPCOM(115.,20.),_ESVPCOM(107.,13.0),_ESVPCOM(99.,6.),
      _ESVPCOM(91.,-1.),_ESVPCOM(83.,-8.),_ESVPCOM(75.,-15.0),_ESVPCOM(67.,-22.),
      _ESVPCOM(59.,-29.),_ESVPCOM(51.,-36.),
      _ESVPCOM(130., 22.),_ESVPCOM(130., 22.),_ESVPCOM(121.,14.0),_ESVPCOM(112.,6.),
      _ESVPCOM(103.,-2.),_ESVPCOM(94.,-10.),_ESVPCOM(85.,-18.0),_ESVPCOM(76.,-26.),
      _ESVPCOM(67.,-34.),_ESVPCOM(58.,-42.),
      _ESVPCOM(145.,24.),_ESVPCOM(145.,24.),_ESVPCOM(135.,15.0),_ESVPCOM(125.,6.),
      _ESVPCOM(115.,-3.),_ESVPCOM(105.,-12.),_ESVPCOM(95.,-21.0),_ESVPCOM(85.,-30.),
      _ESVPCOM(75.,-39.),_ESVPCOM(65.,-48.)};
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 10;
  desc_a[N_]    = 10;
  desc_a[MB_]   = 4;
  desc_a[NB_]   = 4;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(10,4,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 11;
  desc_x[N_]    = 1;
  desc_x[MB_]   = 4;
  desc_x[NB_]   = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(11,4,myrow,0,nprow);
  desc_x[LLD_]  = MAX(1,il_lda);
 
  desc_y[DT_]   = BLOCK_CYCLIC_2D;
  desc_y[M_]    = 1;
  desc_y[N_]    = 11;
  desc_y[MB_]   = 1;
  desc_y[NB_]   = 4;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(1,1,myrow,0,nprow);
  desc_y[LLD_]  = MAX(1,il_lda);
 
  modname = "PDGER   ";
  pdmtxcpy(10, 10, 1, 1, dta5, desc_a, dtas5);
  pdvtrcpy(11, 1, 1, dtx5, desc_x, dtxs5, 1);
  pdmtxcpy(1, 11, 1, 1, dty5, desc_y, dtys5);
 
  pdger(9, 9, 1., dtx5, 2, 1, desc_x, 1, dty5, 1, 2, desc_y, 1, dta5, 2, 2,
        desc_a);
  pdmtchk(modname,9, 9, 2, 2, dta5, desc_a, dresa5, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  modname = "PZGERU  ";
  alpha = _ESVPCOM(1.,-1.);
  pzmtxcpy(10, 10, 1, 1, zta5, desc_a, ztas5);
  pzvtrcpy(11, 1, 1, ztx5, desc_x, ztxs5, 1);
  pzmtxcpy(1, 11, 1, 1, zty5, desc_y, ztys5);
 
  pzgeru(9,9,alpha, ztx5, 2, 1, desc_x, 1, zty5, 1,2,desc_y,1,zta5, 2,2,
        desc_a);
  pzmtchk(modname, 9, 9, 2, 2, zta5, desc_a, zresa5u, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  modname = "PZGERC  ";
  pzmtxcpy(10, 10, 1, 1, zta5, desc_a, ztas5);
  pzvtrcpy(11, 1, 1, ztx5, desc_x, ztxs5, 1);
  pzmtxcpy(1, 11, 1, 1, zty5, desc_y, ztys5);
 
  pzgerc(9,9,alpha, ztx5, 2, 1, desc_x, 1, zty5, 1,2,desc_y,1,zta5, 2,2,
        desc_a);
  pzmtchk(modname,9, 9, 2, 2, zta5, desc_a, zresa5c, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
// PDGER, PZGERU, PZGERC END                           */
 
// PDSYRK, PZSYRK, PZHERK    START : 8, PDTRMM, PZTRMM : 9
 
   double dta8[40], dtc8[64];
   for (i=0; i<64; i++)
     dtc8[i] = 0.0;
 
   double dtas8[40] = {  0.0,  1.0,  2.0,  3.0,  4.0,  5.0,  6.0,  7.0,
                         8.0,  9.0, 10.0, 11.0, 12.0, 13.0, 14.0, 15.0,
                        16.0, 17.0, 18.0, 19.0, 20.0, 21.0, 22.0, 23.0,
                        24.0, 25.0, 26.0, 27.0, 28.0, 29.0, 30.0, 31.0,
                        32.0, 33.0, 34.0, 35.0, 36.0, 37.0, 38.0, 39.0 };
 
   double dtcs8[64] = {  0.0,  1.0,  2.0,  3.0,  4.0,  5.0,  6.0,  7.0,
                         0.0,  8.0,  9.0, 10.0, 11.0, 12.0, 13.0, 14.0,
                         0.0,  0.0, 15.0, 16.0, 17.0, 18.0, 19.0, 20.0,
                         0.,   0.,   0.,  21.0, 22.0, 23.0, 24.0, 25.0,
                         0.,   0.,   0.,   0.,  26.0, 27.0, 28.0, 29.0,
                         0.,   0.,   0.,   0.,   0.,  30.0, 31.0, 32.0,
                         0.,   0.,   0.,   0.,   0.,   0.,  33.0, 34.0,
                         0.,   0.,   0.,   0.,   0.,   0.,   0.,  35.0 };
 
   double dresc8[64] = {1920., 2001., 2082., 2163., 2244., 2325., 2406., 2487.,
                           0., 2093., 2179., 2265., 2351., 2437., 2523., 2609.,
                           0.,    0., 2275., 2366., 2457., 2548., 2639., 2730.,
                           0.,    0.,    0., 2466., 2562., 2658., 2754., 2850.,
                           0.,    0.,    0.,    0., 2666., 2767., 2868., 2969.,
                           0.,    0.,    0.,    0.,    0., 2875., 2981., 3087.,
                           0.,    0.,    0.,    0.,    0.,    0., 3093., 3204.,
                           0.,    0.,    0.,    0.,    0.,    0.,    0., 3320.};
 
// PZSYRK, PZHERK START : 8
      _ESVPCOM   zta8[15],ztc8[9];
 
      _ESVPCOM ztas8[15] = {
      _ESVPCOM(2.0, 0.0),_ESVPCOM(3.0, 3.0),_ESVPCOM (1.0, 3.0),
      _ESVPCOM(3.0, 2.0),_ESVPCOM (8.0, 0.0),_ESVPCOM (2.0, 1.0),
      _ESVPCOM(4.0, 1.0),_ESVPCOM (2.0, 5.0),_ESVPCOM (6.0, 0.0),
      _ESVPCOM(1.0, 7.0),_ESVPCOM (2.0, 4.0),_ESVPCOM (3.0, 2.0),
      _ESVPCOM(0.0, 0.0), _ESVPCOM(1.0, 2.0), _ESVPCOM(2.0, 2.0)};
 
      _ESVPCOM ztcs8[9] = {
      _ESVPCOM (2.0, 1.0),_ESVPCOM (0.0, 0.0),_ESVPCOM (0.0, 0.0),
      _ESVPCOM (1.0, 9.0),_ESVPCOM (3.0, 1.0),_ESVPCOM (0.0, 0.0),
      _ESVPCOM (4.0, 5.0), _ESVPCOM(6.0, 7.0),_ESVPCOM (8.0, 1.0)};
 
      _ESVPCOM zresc8[9] ={
      _ESVPCOM (-57.0, 13.0),_ESVPCOM (  0.0,  0.0),_ESVPCOM (0.0,  0.0),
      _ESVPCOM (-63.0, 79.0),_ESVPCOM (-28.0, 90.0),_ESVPCOM (0.0,  0.0),
      _ESVPCOM (-24.0, 70.0),_ESVPCOM (-55.0,103.0),_ESVPCOM (13.0,75.0)};
 
      _ESVPCOM    zta8h[15], ztc8h[9];
 
      _ESVPCOM ztas8h[15] = {
      _ESVPCOM (2.0,0.0),_ESVPCOM(3.0,3.0),_ESVPCOM(1.,3.),_ESVPCOM(3.,3.),
      _ESVPCOM(1.0,9.0),
      _ESVPCOM(3.0,2.0),_ESVPCOM(8.0,0.0),_ESVPCOM(2.0,1.0),
      _ESVPCOM (8.0,0.0),_ESVPCOM(3.0,0.0),
      _ESVPCOM(4.0,1.0),_ESVPCOM(2.0,5.0),_ESVPCOM(6.0,0.0),
      _ESVPCOM (2.0,5.0),_ESVPCOM(6.0,7.0)};
 
      _ESVPCOM ztcs8h[9] = {
      _ESVPCOM (6.0, 0.0),_ESVPCOM(3.0,4.0), _ESVPCOM(9.0, 1.0),
      _ESVPCOM (0.0, 0.0),_ESVPCOM (10.0,0.0),_ESVPCOM (12.0,2.0),
      _ESVPCOM (0.0, 0.0),_ESVPCOM (0.0, 0.0),_ESVPCOM (3.0, 0.0)};
 
      _ESVPCOM zresc8h[9] = {
      _ESVPCOM (138.0,  0.0),_ESVPCOM( 65.0,80.0),_ESVPCOM (134.0, 46.0),
      _ESVPCOM (  0.0,  0.0),_ESVPCOM(165.0,0.0),_ESVPCOM ( 88.0,-88.0),
      _ESVPCOM (  0.0,  0.0),_ESVPCOM(  0.0,0.0),_ESVPCOM (199.0, 0.0)};
 
  double dta9[25],dtb9[15];
  double dtas9[25] = {  3.,   0.,   0.,   0.,   0.,
                       -1.,  -2.,   0.,   0.,   0.,
                        2.,   4.,  -3.,   0.,   0.,
                        2.,  -1.,   0.,   4.,   0.,
                        1.,   3.,   2.,  -2.,   1. };
 
  double dtbs9[15] = {  2.,   5.,   0.,   3.,  -1.,
                        3.,   5.,   1.,   1.,   2.,
                        1.,   4.,   2.,  -3.,   1. };
 
  double dresb9[15] = {  6., -16.,  -2.,  14.,  -1.,
                        10.,  -1.,   1.,   0.,   2.,
                        -2.,   6.,  -4., -14.,   1. };
 
/*  PZTRMM : 9     */
  _ESVPCOM zta9[25],ztb9[5];
  _ESVPCOM ztas9[25] = {
  _ESVPCOM(-4.0, 1.0),_ESVPCOM( 0.0, 0.0),_ESVPCOM(0.0,0.0),
  _ESVPCOM(0.0,0.0),_ESVPCOM(0.0,0.0),
  _ESVPCOM( 4.0,-3.0),_ESVPCOM(-2.0, 0.0),_ESVPCOM( 0.0, 0.0),
  _ESVPCOM(0.0,0.0),_ESVPCOM(0.0,0.0),
  _ESVPCOM(-1.0, 3.0),_ESVPCOM(-3.0,-1.0),_ESVPCOM(-5.0, 3.0),
  _ESVPCOM(0.0,0.0),_ESVPCOM(0.0,0.0),
  _ESVPCOM( 0.0, 0.0),_ESVPCOM(-2.0,-1.0),_ESVPCOM(-3.0,-3.0),
  _ESVPCOM(4.0,-4.0),_ESVPCOM(0.0,0.0),
  _ESVPCOM(-1.0, 0.0),_ESVPCOM( 4.0, 3.0),_ESVPCOM(-5.0,-5.0),
  _ESVPCOM(2.0,0.0),_ESVPCOM(2.0,-1.0)};
 
  _ESVPCOM ztbs9[ 5] = {
  _ESVPCOM ( 3.0, 4.0), _ESVPCOM(-4.0, 2.0),_ESVPCOM (-5.0, 0.0),
  _ESVPCOM (1.0, 3.0), _ESVPCOM(3.0, 1.0)};
 
  _ESVPCOM zresb9[ 5] = {
  _ESVPCOM (-8.0,-19.0),_ESVPCOM( 8.0, 21.0),_ESVPCOM(44.0,-8.0),
  _ESVPCOM (13.0,-7.0),_ESVPCOM(19.0, 2.0)};
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 8;
  desc_a[N_]    = 5;
  desc_a[MB_]   = 2;
  desc_a[NB_]   = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_c[DT_]   = BLOCK_CYCLIC_2D;
  desc_c[M_]    = 8;
  desc_c[N_]    = 8;
  desc_c[MB_]   = 2;
  desc_c[NB_]   = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_c[LLD_]  = MAX(1,il_lda);
 
 
  modname = "PDSYRK  ";
  pdmtxcpy(8, 5, 1, 1, dta8, desc_a, dtas8);
  pdmtxcpy(8, 8, 1, 1, dtc8, desc_c, dtcs8);
  pdsyrk("L","N",8, 5, 1.0, dta8, 1, 1, desc_a, 1.0, dtc8, 1, 1, desc_c);
 
  pdmtchk(modname,8, 8, 1, 1, dtc8, desc_a, dresc8, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
   for (i=0; i<9; i++)
    ztc8[i] = _ESVPCOM(0., 0.);
    alpha = _ESVPCOM(1., 1.);
    beta = _ESVPCOM(1., 1.);
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 3;
  desc_a[N_] = 5;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_c[DT_] = BLOCK_CYCLIC_2D;
  desc_c[M_] = 3;
  desc_c[N_] = 3;
  desc_c[MB_] = 2;
  desc_c[NB_] = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_c[LLD_] = MAX(1,il_lda);
 
 
  modname = "PZSYRK  ";
  pzmtxcpy(3, 5, 1, 1, zta8, desc_a, ztas8);
  pzmtxcpy(3, 3, 1, 1, ztc8, desc_c, ztcs8);
  pzsyrk("U","N",3,5,alpha,zta8,1,1, desc_a, beta, ztc8, 1, 1, desc_c);
 
  pzmtchk(modname,3, 3, 1, 1, ztc8, desc_c, zresc8, tol, ierr, ilerr,
         error, ifail, strlen(modname));
 
/* PZSYRK    END     */
 
 
   for (i=0; i<9; i++)
    ztc8h[i] = _ESVPCOM(0., 0.);
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 5;
  desc_a[N_] = 3;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_c[DT_] = BLOCK_CYCLIC_2D;
  desc_c[M_] = 3;
  desc_c[N_] = 3;
  desc_c[MB_] = 2;
  desc_c[NB_] = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_c[LLD_] = MAX(1,il_lda);
 
 
  modname = "PZHERK  ";
  pzmtxcpy(5, 3, 1, 1, zta8h, desc_a, ztas8h);
  pzmtxcpy(3, 3, 1, 1, ztc8h, desc_c, ztcs8h);
  pzherk("L","C",3,5, 1.0 ,zta8h,1,1, desc_a, 1.0, ztc8h, 1, 1,desc_c);
 
  pzmtchk(modname,3, 3, 1, 1, ztc8h, desc_c, zresc8h, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
// PDSYRK, PZSYRK, PZHERK    END
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 5;
  desc_a[N_]    = 5;
  desc_a[MB_]   = 2;
  desc_a[NB_]   = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_b[DT_]   = BLOCK_CYCLIC_2D;
  desc_b[M_]    = 5;
  desc_b[N_]    = 3;
  desc_b[MB_]   = 2;
  desc_b[NB_]   = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_b[LLD_]  = MAX(1,il_lda);
 
  modname = "PDTRMM  ";
  pdmtxcpy(5, 5, 1, 1, dta9, desc_a, dtas9);
  pdmtxcpy(5, 3, 1, 1, dtb9, desc_b, dtbs9);
  pdtrmm("L","U","N", "N", 5, 3, 1.0, dta9, 1, 1, desc_a, dtb9, 1, 1, desc_b);
 
  pdmtchk(modname,5, 3, 1, 1, dtb9, desc_b, dresb9, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 5;
  desc_a[N_] = 5;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_b[M_] = 5;
  desc_b[N_] = 1;
  desc_b[MB_] = 2;
  desc_b[NB_] = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_b[LLD_] = MAX(1,il_lda);
 
  modname = "PZTRMM  ";
  alpha = _ESVPCOM(1., 0.);
  pzmtxcpy(5, 5, 1, 1, zta9, desc_a, ztas9);
  pzmtxcpy(5, 1, 1, 1, ztb9, desc_b, ztbs9);
  pztrmm("L","U","C", "N", 5, 1,alpha,zta9,1,1,desc_a,ztb9,1,1,desc_b);
 
  pzmtchk(modname,5, 1, 1, 1, ztb9, desc_b, zresb9, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
// PDTRMM, PZTRMM    END
 
// PDSYMV,PZHEMV    START :10
 
   double dta10[64], dtx10[8];
   double dtas10[64] = {  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,
                         -1.0,  1.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,
                         -1.0,  0.0, -1.0,  0.0,  0.0,  0.0,  0.0,  0.0,
                          0.0,  1.0, -1.0, -1.0,  0.0,  0.0,  0.0,  0.0,
                          0.0,  0.0,  0.0,  1.0, -1.0,  0.0,  0.0,  0.0,
                          0.0,  1.0,  0.0,  1.0,  0.0,  1.0,  0.0,  0.0,
                          0.0,  0.0,  1.0,  0.0,  0.0,  0.0,  0.0,  0.0,
                          0.0,  1.0,  0.0,  1.0,  0.0,  0.0,  0.0,  0.0 };
 
   double dtxs10[8] = {  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0 };
   double dty10[8] = {  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0 };
   double dresy10[8] = { -2.0,  3.0, -2.0,  2.0,  0.0,  3.0,  1.0,  2.0 };
 
      _ESVPCOM  zta10[36], ztx10[6];
      _ESVPCOM ztas10[36] = {
      _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),
      _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),
      _ESVPCOM (-1.,1.),_ESVPCOM(2.,-1.),_ESVPCOM(0.,0.),_ESVPCOM(0., 0.),
      _ESVPCOM (0., 0.),_ESVPCOM(0.,0.),
      _ESVPCOM (-1.,-2.),_ESVPCOM(5., 4.),_ESVPCOM(0.,0.),_ESVPCOM(0., 0.),
      _ESVPCOM (0., 0.),_ESVPCOM(0.,0.),
      _ESVPCOM ( 0., 3.),_ESVPCOM(2., 0.),_ESVPCOM(-1.,0.),_ESVPCOM(4.,2.),
      _ESVPCOM (0., 0.),_ESVPCOM(0.,0.),
      _ESVPCOM (2., 1.),_ESVPCOM(-1.,-1.),_ESVPCOM(0.,0.),_ESVPCOM(2.,-1.),
      _ESVPCOM (-1.,2.),_ESVPCOM(0.,0.),
      _ESVPCOM (1., 0.),_ESVPCOM( 0., 2.),_ESVPCOM(1.,1.),_ESVPCOM(-1.,-1.),
      _ESVPCOM (0.,2.),_ESVPCOM(1.,0.)};
 
      _ESVPCOM ztxs10[6] = {
      _ESVPCOM(-1.,1.),_ESVPCOM(2.,1.),_ESVPCOM(1.,2.),_ESVPCOM(-2.,-3.),
      _ESVPCOM(0., 1.),_ESVPCOM(1.,0.)};
      _ESVPCOM zty10[6] = {
      _ESVPCOM( 0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM( 0., 0.),
      _ESVPCOM(0., 0.),_ESVPCOM(0.,0.)};
      _ESVPCOM  zresy10 [6] = {
      _ESVPCOM(9.,-7.),_ESVPCOM(0.,11.),_ESVPCOM(16.,-2.),
      _ESVPCOM(-2.,-8.),_ESVPCOM(-5.,-3.),_ESVPCOM(12.,-1.)};
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 8;
  desc_a[N_]    = 8;
  desc_a[MB_]   = 2;
  desc_a[NB_]   = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 8;
  desc_x[N_]    = 1;
  desc_x[MB_]   = 2;
  desc_x[NB_]   = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_x[LLD_]  = MAX(1,il_lda);
 
  desc_y[DT_]   = BLOCK_CYCLIC_2D;
  desc_y[M_]    = 8;
  desc_y[N_]    = 1;
  desc_y[MB_]   = 2;
  desc_y[NB_]   = 1;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_y[LLD_]  = MAX(1,il_lda);
 
  modname = "PDSYMV  ";
  pdmtxcpy(8, 8, 1, 1, dta10, desc_a, dtas10);
  pdvtrcpy(8, 1, 1, dtx10, desc_x, dtxs10, 1);
 
  pdsymv("U", 8, 1.0, dta10, 1, 1, desc_a, dtx10, 1, 1, desc_x, 1, 0.0,
          dty10, 1, 1, desc_y, 1);
  pdvtchk(modname,8, 1, 1, dty10, desc_y, dresy10, 1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 6;
  desc_a[N_] = 6;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(6,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 6;
  desc_x[N_] = 1;
  desc_x[MB_] = 2;
  desc_x[NB_] = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(6,2,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 6;
  desc_y[N_] = 1;
  desc_y[MB_] = 2;
  desc_y[NB_] = 1;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(6,2,myrow,0,nprow);
  desc_y[LLD_] = MAX(1,il_lda);
 
  modname = "PZHEMV  ";
  alpha = _ESVPCOM(1.0,0.0);
  beta = _ESVPCOM(0.0,0.0);
  pzmtxcpy(6, 6, 1, 1, zta10, desc_a, ztas10);
  pzvtrcpy(6, 1, 1, ztx10, desc_x, ztxs10, 1);
 
  pzhemv("U", 6,alpha,zta10, 1, 1, desc_a, ztx10, 1, 1, desc_x, 1,beta,
          zty10, 1, 1, desc_y, 1);
  pzvtchk(modname,6, 1, 1, zty10, desc_y, zresy10, 1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
// PDSYMV, PZHEMV    END
 
// PDGEMM    START :11
 
   double dta11[30], dtb11[20], dtc11[24];
   double dtas11[30] = {  1.0,  2.0,  1.0, -3.0,  4.0, -1.0,
                          2.0,  0.0, -1.0,  2.0,  0.0, -1.0,
                         -1.0,  1.0, -1.0,  2.0, -2.0,  1.0,
                         -1.0,  1.0,  1.0,  2.0,  1.0, -3.0,
                          4.0, -1.0,  2.0,  0.0, -1.0,  2.0 };
 
   double dtbs11[20] = {  1.0,  2.0,  1.0, -3.0,  4.0,
                         -1.0,  2.0,  0.0, -1.0,  2.0,
                          0.0, -1.0, -1.0,  1.0, -1.0,
                          2.0, -2.0,  1.0, -1.0,  1.0 };
 
   double dtcs11[24] = {  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                          0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                          0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                          0.5,  0.5,  0.5,  0.5,  0.5,  0.5 };
 
   double dresc11[24] = { 24.0, -3.0,  4.0, -2.0, -4.0, 16.0,
                          13.0, -4.0,  1.0,  6.0, -6.0,  7.0,
                          -5.0,  2.0,  2.0, -1.0,  5.0, -4.0,
                           3.0,  4.0,  5.0, -9.0,  5.0,  7.0 };
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 6;
  desc_a[N_]    = 5;
  desc_a[MB_]   = 3;
  desc_a[NB_]   = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(6,3,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_b[DT_]   = BLOCK_CYCLIC_2D;
  desc_b[M_]    = 5;
  desc_b[N_]    = 4;
  desc_b[MB_]   = 2;
  desc_b[NB_]   = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_b[LLD_]  = MAX(1,il_lda);
 
  desc_c[DT_]   = BLOCK_CYCLIC_2D;
  desc_c[M_]    = 6;
  desc_c[N_]    = 4;
  desc_c[MB_]   = 3;
  desc_c[NB_]   = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(6,3,myrow,0,nprow);
  desc_c[LLD_]  = MAX(1,il_lda);
 
  modname = "PDGEMM  ";
  pdmtxcpy(6, 5, 1, 1, dta11, desc_a, dtas11);
  pdmtxcpy(5, 4, 1, 1, dtb11, desc_b, dtbs11);
  pdmtxcpy(6, 4, 1, 1, dtc11, desc_c, dtcs11);
 
  pdgemm("N", "N", 6, 4, 5, 1.0, dta11, 1, 1, desc_a, dtb11, 1, 1, desc_b,
          2.0, dtc11, 1, 1, desc_c);
  pdmtchk(modname,6, 4, 1, 1, dtc11, desc_c, dresc11, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
// PDGEMM    END
// PZGEMM    START
 
  _ESVPCOM zta11[18], ztb11[6], ztc11[12];
  _ESVPCOM ztas11[18] = {
       _ESVPCOM(1.0, 5.0), _ESVPCOM(2.0, 4.0), _ESVPCOM(3.0, 3.0),
       _ESVPCOM(4.0, 2.0), _ESVPCOM(5.0, 1.0), _ESVPCOM(6.0, 6.0),
       _ESVPCOM(9.0, 2.0), _ESVPCOM(8.0, 3.0), _ESVPCOM(7.0, 5.0),
       _ESVPCOM(4.0, 7.0), _ESVPCOM(5.0, 1.0), _ESVPCOM(3.0, 6.0),
       _ESVPCOM(1.0, 9.0), _ESVPCOM(1.0, 8.0), _ESVPCOM(1.0, 7.0),
       _ESVPCOM(1.0, 5.0), _ESVPCOM(1.0, 6.0), _ESVPCOM(1.0, 4.0)};
 
  _ESVPCOM ztbs11[6] = {
       _ESVPCOM(1.0, 8.0), _ESVPCOM(4.0, 4.0), _ESVPCOM(6.0, 2.0),
       _ESVPCOM(2.0, 7.0), _ESVPCOM(6.0, 8.0), _ESVPCOM(4.0, 5.0)};
 
  _ESVPCOM ztcs11[12] = {
       _ESVPCOM(0.5, 0.0), _ESVPCOM(0.5, 0.0), _ESVPCOM(0.5, 0.0),
       _ESVPCOM(0.5, 0.0), _ESVPCOM(0.5, 0.0), _ESVPCOM(0.5, 0.0),
       _ESVPCOM(0.5, 0.0), _ESVPCOM(0.5, 0.0), _ESVPCOM(0.5, 0.0),
       _ESVPCOM(0.5, 0.0), _ESVPCOM(0.5, 0.0), _ESVPCOM(0.5, 0.0)};
 
 
  _ESVPCOM zresc11[12] = {
      _ESVPCOM(-22.0,113.0), _ESVPCOM(-19.0,114.0), _ESVPCOM(-20.0,119.0),
      _ESVPCOM(-27.0,110.0), _ESVPCOM(  8.0,103.0), _ESVPCOM(-55.0,116.0),
      _ESVPCOM(-35.0,142.0), _ESVPCOM(-35.0,141.0), _ESVPCOM(-43.0,146.0),
      _ESVPCOM(-58.0,131.0), _ESVPCOM(  0.0,112.0), _ESVPCOM(-75.0,135.0)};
 
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 6;
  desc_a[N_]    = 3;
  desc_a[MB_]   = 2;
  desc_a[NB_]   = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(6,2,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_b[DT_]   = BLOCK_CYCLIC_2D;
  desc_b[M_]    = 3;
  desc_b[N_]    = 2;
  desc_b[MB_]   = 2;
  desc_b[NB_]   = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_b[LLD_]  = MAX(1,il_lda);
 
  desc_c[DT_]   = BLOCK_CYCLIC_2D;
  desc_c[M_]    = 6;
  desc_c[N_]    = 2;
  desc_c[MB_]   = 2;
  desc_c[NB_]   = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(6,2,myrow,0,nprow);
  desc_c[LLD_]  = MAX(1,il_lda);
 
  modname = "PZGEMM  ";
  pzmtxcpy(6, 3, 1, 1, zta11, desc_a, ztas11);
  pzmtxcpy(3, 2, 1, 1, ztb11, desc_b, ztbs11);
  pzmtxcpy(6, 2, 1, 1, ztc11, desc_c, ztcs11);
 
  alpha = _ESVPCOM(1.0,0.0);
  beta  = _ESVPCOM(2.0,0.0);
 
  pzgemm("N", "N", 6, 2, 3, alpha, zta11, 1, 1, desc_a, ztb11,
          1, 1, desc_b, beta, ztc11, 1, 1, desc_c);
  pzmtchk(modname,6, 2, 1, 1, ztc11, desc_c, zresc11, tol,
          ierr,ilerr, error, ifail, strlen(modname));
 
// PZGEMM    END
 
// PDGEMV, PZGEMV    START :12
 
   double dta12[30], dtx12[20], dty12[24];
   double dtas12[30] = {  1.0,  2.0,  1.0, -3.0,  4.0, -1.0,
                          2.0,  0.0, -1.0,  2.0,  0.0, -1.0,
                         -1.0,  1.0, -1.0,  2.0, -2.0,  1.0,
                         -1.0,  1.0,  1.0,  2.0,  1.0, -3.0,
                          4.0, -1.0,  2.0,  0.0, -1.0,  2.0 };
 
   double dtxs12[20] = {  1.0,  2.0,  1.0, -3.0,  4.0,
                         -1.0,  2.0,  0.0, -1.0,  2.0,
                          0.0, -1.0, -1.0,  1.0, -1.0,
                          2.0, -2.0,  1.0, -1.0,  1.0 };
 
   double dtys12[24] = {  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                          0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                          0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                          0.5,  0.5,  0.5,  0.5,  0.5,  0.5 };
 
   double dresy12[24] = {  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                           0.5,  0.5,  1.0,  6.0, -6.0,  7.0,
                           0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                           0.5,  0.5,  0.5,  0.5,  0.5,  0.5 };
 
      _ESVPCOM    zta12[18], ztx12[6], zty12[12];
      _ESVPCOM ztas12[18] = {
      _ESVPCOM(1.,5.),_ESVPCOM(2.,4.),_ESVPCOM(3.,3.),_ESVPCOM(4.,2.),
      _ESVPCOM (5.,1.),_ESVPCOM(6.,6.),
      _ESVPCOM (9.,2.),_ESVPCOM(8.,3.),_ESVPCOM(7.,5.),_ESVPCOM(4.,7.),
      _ESVPCOM (5.,1.),_ESVPCOM(3.,6.),
      _ESVPCOM (1.,9.),_ESVPCOM(1.,8.),_ESVPCOM(1.,7.),_ESVPCOM(1.,5.),
      _ESVPCOM (1.,6.),_ESVPCOM(1.,4.)};
 
      _ESVPCOM ztxs12[6] = {
      _ESVPCOM(1., 8.),_ESVPCOM(4., 4.),_ESVPCOM(6., 2.),
      _ESVPCOM(2., 7.),_ESVPCOM(6., 8.),_ESVPCOM(4., 5.)};
 
      _ESVPCOM ztys12[12] = {
      _ESVPCOM (0.5, 0.),_ESVPCOM(0.5,0.),_ESVPCOM(0.5,0.),_ESVPCOM(0.5,0.),
      _ESVPCOM (0.5, 0.),_ESVPCOM (0.5, 0.),
      _ESVPCOM(0.5,0.),_ESVPCOM(0.5,0.),_ESVPCOM(0.5,0.),_ESVPCOM(0.5,0.),
      _ESVPCOM (0.5, 0.),_ESVPCOM(0.5, 0.)};
 
      _ESVPCOM zresy12[12] = {
      _ESVPCOM(0.5,0.5),_ESVPCOM(0.5,0.),_ESVPCOM(0.5,0.),_ESVPCOM(0.5,0.),
      _ESVPCOM (0.5, 0.),_ESVPCOM(0.5, 0.),
      _ESVPCOM(-35.,142.),_ESVPCOM(-35.,141.),_ESVPCOM(-43.,146.),
      _ESVPCOM (-58.,131.),_ESVPCOM(0.,112.),_ESVPCOM(-75., 135.)};
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 6;
  desc_a[N_]    = 5;
  desc_a[MB_]   = 3;
  desc_a[NB_]   = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(6,3,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 5;
  desc_x[N_]    = 4;
  desc_x[MB_]   = 2;
  desc_x[NB_]   = 2;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_x[LLD_]  = MAX(1,il_lda);
 
  desc_y[DT_]   = BLOCK_CYCLIC_2D;
  desc_y[M_]    = 6;
  desc_y[N_]    = 4;
  desc_y[MB_]   = 3;
  desc_y[NB_]   = 2;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(6,3,myrow,0,nprow);
  desc_y[LLD_]  = MAX(1,il_lda);
 
  modname = "PDGEMV  ";
  pdmtxcpy(4, 5, 3, 1, dta12, desc_a, dtas12);
  pdmtxcpy(5, 3, 1, 2, dtx12, desc_x, dtxs12);
  pdmtxcpy(4, 3, 3, 2, dty12, desc_y, dtys12);
 
  pdgemv("N", 4, 5, 1.0, dta12, 3, 1, desc_a, dtx12, 1, 2, desc_x, 1,
          2.0, dty12, 3, 2, desc_y, 1);
  pdmtchk(modname,4, 3, 3, 2, dty12, desc_y, dresy12, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 6;
  desc_a[N_] = 3;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(6,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 3;
  desc_x[N_] = 2;
  desc_x[MB_] = 2;
  desc_x[NB_] = 2;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 6;
  desc_y[N_] = 2;
  desc_y[MB_] = 2;
  desc_y[NB_] = 2;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(6,2,myrow,0,nprow);
  desc_y[LLD_] = MAX(1,il_lda);
 
  modname = "PZGEMV  ";
  alpha = _ESVPCOM(1.0,0.0);
  beta = _ESVPCOM(2.0,0.0);
 
  pzmtxcpy(6, 3, 1, 1, zta12, desc_a, ztas12);
  pzmtxcpy(3, 1, 1, 2, ztx12, desc_x, ztxs12);
  pzmtxcpy(6, 1, 1, 2, zty12, desc_y, ztys12);
 
  pzgemv("N", 6, 3,alpha, zta12, 1, 1, desc_a, ztx12, 1, 2, desc_x, 1,
         beta, zty12, 1, 2, desc_y, 1);
  pzmtchk(modname,6, 1, 1, 2, zty12, desc_y, zresy12, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
// PDGEMV, PZGEMV    END
 
// PDSYR, PZHER     START :13
 
   double dta13[81], dtx13[9];
   double dtxs13[9]  = { 1., 1., 1., 1., 1., 1., 1., 1., 1. };
   double dtas13[81] = { 1.,   2.,   3.,   4.,   5.,   6.,   7.,   8.,   9.,
                        11.,  12.,  13.,  14.,  15.,  16.,  17.,  18.,  19.,
                        21.,  22.,  23.,  24.,  25.,  26.,  27.,  28.,  29.,
                        31.,  32.,  33.,  34.,  35.,  36.,  37.,  38.,  39.,
                        41.,  42.,  43.,  44.,  45.,  46.,  47.,  48.,  49.,
                        51.,  52.,  53.,  54.,  55.,  56.,  57.,  58.,  59.,
                        61.,  62.,  63.,  64.,  65.,  66.,  67.,  68.,  69.,
                        71.,  72.,  73.,  74.,  75.,  76.,  77.,  78.,  79.,
                        81.,  82.,  83.,  84.,  85.,  86.,  87.,  88.,  89. };
 
   double dresa13[81] = {2.,   3.,   4.,   5.,   6.,   7.,   8.,   9.,  10.,
                        11.,  13.,  14.,  15.,  16.,  17.,  18.,  19.,  20.,
                        21.,  22.,  24.,  25.,  26.,  27.,  28.,  29.,  30.,
                        31.,  32.,  33.,  35.,  36.,  37.,  38.,  39.,  40.,
                        41.,  42.,  43.,  44.,  46.,  47.,  48.,  49.,  50.,
                        51.,  52.,  53.,  54.,  55.,  57.,  58.,  59.,  60.,
                        61.,  62.,  63.,  64.,  65.,  66.,  68.,  69.,  70.,
                        71.,  72.,  73.,  74.,  75.,  76.,  77.,  79.,  80.,
                        81.,  82.,  83.,  84.,  85.,  86.,  87.,  88.,  90. };
 
      _ESVPCOM    zta13[9], ztx13[3];
      _ESVPCOM ztxs13[3] = {
      _ESVPCOM(1., 2.),_ESVPCOM(4., 0.),_ESVPCOM(3., 4.)};
      _ESVPCOM ztas13[9] = {
      _ESVPCOM (1., 0.),_ESVPCOM (3., -5.),_ESVPCOM (2., 3.),
      _ESVPCOM (0., 0.),_ESVPCOM (7.,  0.),_ESVPCOM (4., 8.),
      _ESVPCOM (0., 0.),_ESVPCOM (0.,  0.),_ESVPCOM (6., 0.)};
 
      _ESVPCOM zresa13[9] = {
      _ESVPCOM (6., 0.),_ESVPCOM (7., -13.),_ESVPCOM (13., 1.),
      _ESVPCOM (0., 0.),_ESVPCOM (23.,  0.),_ESVPCOM (16., 24.),
      _ESVPCOM (0., 0.),_ESVPCOM (0.,  0.), _ESVPCOM (31., 0.)};
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 9;
  desc_a[N_]    = 9;
  desc_a[MB_]   = 4;
  desc_a[NB_]   = 4;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(9,4,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 9;
  desc_x[N_]    = 1;
  desc_x[MB_]   = 4;
  desc_x[NB_]   = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(9,4,myrow,0,nprow);
  desc_x[LLD_]  = MAX(1,il_lda);
 
  modname = "PDSYR   ";
  pdmtxcpy(9, 9, 1, 1, dta13, desc_a, dtas13);
  pdvtrcpy(9, 1, 1, dtx13, desc_x, dtxs13, 1);
 
  pdsyr("L",9, 1., dtx13, 1, 1, desc_x, 1, dta13, 1, 1, desc_a);
  pdmtchk(modname,9, 9, 1, 1, dta13, desc_a, dresa13, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 3;
  desc_a[N_] = 3;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 3;
  desc_x[N_] = 1;
  desc_x[MB_] = 2;
  desc_x[NB_] = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  modname = "PZHER   ";
  pzmtxcpy(3, 3, 1, 1, zta13, desc_a, ztas13);
  pzvtrcpy(3, 1, 1, ztx13, desc_x, ztxs13, 1);
 
  pzher("L",3, 1., ztx13, 1, 1, desc_x, 1, zta13, 1, 1, desc_a);
  pzmtchk(modname,3, 3, 1, 1, zta13, desc_a, zresa13, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
// PDSYR, PZHER END
 
// PDSYR2,PZHER2    START :14
 
   double dta14[81], dtx14[9], dty14[9];
   double dtxs14[9]  = { 1., 1., 1., 1., 1., 1., 1., 1., 1. };
   double dtys14[9]  = { 2., 2., 2., 2., 2., 2., 2., 2., 2. };
   double dtas14[81] = { 1.,   2.,   3.,   4.,   5.,   6.,   7.,   8.,   9.,
                        11.,  12.,  13.,  14.,  15.,  16.,  17.,  18.,  19.,
                        21.,  22.,  23.,  24.,  25.,  26.,  27.,  28.,  29.,
                        31.,  32.,  33.,  34.,  35.,  36.,  37.,  38.,  39.,
                        41.,  42.,  43.,  44.,  45.,  46.,  47.,  48.,  49.,
                        51.,  52.,  53.,  54.,  55.,  56.,  57.,  58.,  59.,
                        61.,  62.,  63.,  64.,  65.,  66.,  67.,  68.,  69.,
                        71.,  72.,  73.,  74.,  75.,  76.,  77.,  78.,  79.,
                        81.,  82.,  83.,  84.,  85.,  86.,  87.,  88.,  89. };
 
   double dresa14[81] = { 5.,   6.,   7.,   8.,   9.,  10., 11.,  12.,  13.,
                         11.,  16.,  17.,  18.,  19.,  20., 21.,  22.,  23.,
                         21.,  22.,  27.,  28.,  29.,  30., 31.,  32.,  33.,
                         31.,  32.,  33.,  38.,  39.,  40., 41.,  42.,  43.,
                         41.,  42.,  43.,  44.,  49.,  50., 51.,  52.,  53.,
                         51.,  52.,  53.,  54.,  55.,  60., 61.,  62.,  63.,
                         61.,  62.,  63.,  64.,  65.,  66., 71.,  72.,  73.,
                         71.,  72.,  73.,  74.,  75.,  76., 77.,  82.,  83.,
                         81.,  82.,  83.,  84.,  85.,  86., 87.,  88.,  93. };
 
   _ESVPCOM zta14[9], ztx14[3], zty14[3];
   _ESVPCOM ztxs14[3] = {
   _ESVPCOM(1., 2.),_ESVPCOM(4., 0.),_ESVPCOM(3., 4.)};
   _ESVPCOM ztys14[3] = {
   _ESVPCOM(1., 0.),_ESVPCOM(2.,-1.),_ESVPCOM(2., 1.)};
   _ESVPCOM ztas14 [9] = {
   _ESVPCOM(1., 0.),_ESVPCOM (3., -5.),_ESVPCOM (2., 3.),
   _ESVPCOM(0., 0.),_ESVPCOM (7.,  0.),_ESVPCOM (4., 8.),
   _ESVPCOM(0., 0.),_ESVPCOM (0.,  0.),_ESVPCOM (6., 0.)};
 
   _ESVPCOM zresa14 [9] = {
   _ESVPCOM (3., 0.),_ESVPCOM (7., -10.),_ESVPCOM (9., 4.),
   _ESVPCOM (0., 0.),_ESVPCOM (23.,  0.),_ESVPCOM (14., 23.),
   _ESVPCOM (0., 0.),_ESVPCOM (0.,  0.),_ESVPCOM  (26., 0.)};
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 9;
  desc_a[N_]    = 9;
  desc_a[MB_]   = 4;
  desc_a[NB_]   = 4;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(9,4,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 9;
  desc_x[N_]    = 1;
  desc_x[MB_]   = 4;
  desc_x[NB_]   = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(9,4,myrow,0,nprow);
  desc_x[LLD_]  = MAX(1,il_lda);
 
  desc_y[DT_]   = BLOCK_CYCLIC_2D;
  desc_y[M_]    = 9;
  desc_y[N_]    = 1;
  desc_y[MB_]   = 4;
  desc_y[NB_]   = 1;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(9,4,myrow,0,nprow);
  desc_y[LLD_]  = MAX(1,il_lda);
 
  modname = "PDSYR2  ";
  pdmtxcpy(9, 9, 1, 1, dta14, desc_a, dtas14);
  pdvtrcpy(9, 1, 1, dtx14, desc_x, dtxs14, 1);
  pdvtrcpy(9, 1, 1, dty14, desc_x, dtys14, 1);
 
  pdsyr2("L",9, 1., dtx14, 1, 1, desc_x, 1, dty14, 1, 1, desc_y, 1,
         dta14, 1, 1, desc_a);
  pdmtchk(modname,9, 9, 1, 1, dta14, desc_a, dresa14, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 3;
  desc_a[N_] = 3;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 3;
  desc_x[N_] = 1;
  desc_x[MB_] = 2;
  desc_x[NB_] = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 3;
  desc_y[N_] = 1;
  desc_y[MB_] = 2;
  desc_y[NB_] = 1;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_y[LLD_] = MAX(1,il_lda);
 
  modname = "PZHER2  ";
  alpha =_ESVPCOM(1., 0.);
  pzmtxcpy(3, 3, 1, 1, zta14, desc_a, ztas14);
  pzvtrcpy(3, 1, 1, ztx14, desc_x, ztxs14, 1);
  pzvtrcpy(3, 1, 1, zty14, desc_x, ztys14, 1);
 
  pzher2("L",3, alpha, ztx14, 1, 1, desc_x, 1, zty14, 1, 1, desc_y, 1,
         zta14, 1, 1, desc_a);
  pzmtchk(modname,3, 3, 1, 1, zta14, desc_a, zresa14, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
// PDSYR2, PZHER2 END
 
// PDTRSM, PZTRSM    START :15
 
   double dta15[25], dtb15[15];
   double dtas15[25] = {  3.,   0.,   0.,   0.,   0.,
                         -1.,  -2.,   0.,   0.,   0.,
                          2.,   4.,  -3.,   0.,   0.,
                          2.,  -1.,   0.,   4.,   0.,
                          1.,   3.,   2.,  -2.,   1. };
 
   double dtbs15[15] = {  6., -16.,  -2.,  14.,  -1.,
                         10.,  -1.,   1.,   0.,   2.,
                         -2.,   6.,  -4., -14.,   1. };
 
   double dresb15[15] = {  2.,   5.,   0.,   3.,  -1.,
                           3.,   5.,   1.,   1.,   2.,
                           1.,   4.,   2.,  -3.,   1. };
 
      _ESVPCOM  zta15[25], ztb15[10];
      _ESVPCOM  ztas15[25] = {
      _ESVPCOM(-4.,1.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),
      _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),
      _ESVPCOM( 4.,-3.),_ESVPCOM(-2., 0.),_ESVPCOM(0.,0.),
      _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),
      _ESVPCOM(-1., 3.),_ESVPCOM(-3.,-1.),_ESVPCOM(-5.,3.),
      _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),
      _ESVPCOM(0., 0.),_ESVPCOM(-2.,-1.),_ESVPCOM(-3.,-3.),
      _ESVPCOM(4.,-4.),_ESVPCOM(0.,0.),
      _ESVPCOM(-1.,0.),_ESVPCOM( 4., 3.),_ESVPCOM(-5.,-5.),
      _ESVPCOM(2.,0.),_ESVPCOM(2.,-1.)};
 
      _ESVPCOM ztbs15[10] = {
      _ESVPCOM (5.5,-13.50),_ESVPCOM(-6.5,14.5),_ESVPCOM(26.0,18.0),
      _ESVPCOM (10.0,3.0),_ESVPCOM(8.5,10.5),
      _ESVPCOM (-3.0,-5.0),_ESVPCOM(-3.0, 5.0),_ESVPCOM( 4.0,-3.0),
      _ESVPCOM  ( 6.0,-6.0),_ESVPCOM(13.0,-12.0)};
 
      _ESVPCOM zresb15[10] = {
      _ESVPCOM ( 3.,4.),_ESVPCOM(-4.,2.),_ESVPCOM(-5.,0.),
      _ESVPCOM (1.,3.),_ESVPCOM(3.,1.),
      _ESVPCOM ( 2.,0.),_ESVPCOM(3.,-1.),_ESVPCOM(-1.,2.),
      _ESVPCOM (0.,-2.),_ESVPCOM(1.,3.)};
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 5;
  desc_a[N_]    = 5;
  desc_a[MB_]   = 2;
  desc_a[NB_]   = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_b[DT_]   = BLOCK_CYCLIC_2D;
  desc_b[M_]    = 5;
  desc_b[N_]    = 3;
  desc_b[MB_]   = 2;
  desc_b[NB_]   = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_b[LLD_]  = MAX(1,il_lda);
 
  modname = "PDTRSM  ";
  pdmtxcpy(5, 5, 1, 1, dta15, desc_a, dtas15);
  pdmtxcpy(5, 3, 1, 1, dtb15, desc_b, dtbs15);
 
  pdtrsm("L", "U", "N", "N", 5, 3, 1.0, dta15, 1, 1, desc_a,
          dtb15, 1, 1, desc_b);
  pdmtchk(modname,5, 3, 1, 1, dtb15, desc_b, dresb15, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_b[M_] = 5;
  desc_b[N_] = 2;
  desc_b[MB_] = 2;
  desc_b[NB_] = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_b[LLD_] = MAX(1,il_lda);
 
  modname = "PZTRSM  ";
  alpha = _ESVPCOM(1.0,-1.0);
  tol = 0.1;
  pzmtxcpy(5, 5, 1, 1, zta15, desc_a, ztas15);
  pzmtxcpy(5, 2, 1, 1, ztb15, desc_b, ztbs15);
 
  pztrsm("L", "U", "C", "N", 5, 2,alpha, zta15, 1, 1, desc_a,
          ztb15, 1, 1, desc_b);
  pzmtchk(modname,5, 2, 1, 1, ztb15, desc_b, zresb15, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
// PDTRSM, PZTRSM  END
 
 
  blacs_gridexit(ictx);
}
 
/*********************************************************************
*          PARALLEL EIGEN SYSTEM ANALYSIS SUBROUTINES                *
*********************************************************************/
 
void peig(int *ifail)
{
// PDSYTRD  PZHETRD START : 16
 
   int iam, nnodes,myrow,mycol,istor[1],ra[1],ca[1],ix;
   int desc_a[DLEN_], desc_x[DLEN_], il_lda, desc_b[DLEN_];
   int desc_u[DLEN_],  desc_vt[DLEN_];
   int   lrwork=0;
   double rwork[1];
   double dts18[3],dtvt18[12],dtu18[12];
   double ztd18[3],zte18[2],zts18[3];
   _ESVPCOM zta18[12],ztvt18[12],ztu18[9];
   _ESVPCOM zttp18[3],zttq18[12];
 
  _ESVPCOM ztvts18[12] = {
         _ESVPCOM(   .02,  -.08), _ESVPCOM(  -.42,  -.48), _ESVPCOM(   .20,   .16),
         _ESVPCOM(   .03,   .03), _ESVPCOM(  -.41,   .11), _ESVPCOM(   .17,   .78),
         _ESVPCOM(   .39,   .51), _ESVPCOM(  -.32,   .39), _ESVPCOM(  -.39,  -.11),
         _ESVPCOM(  -.48,  -.59), _ESVPCOM(  -.21,   .33), _ESVPCOM(  -.35,  -.03)
   };
 
   _ESVPCOM zresa18[12] = {
          _ESVPCOM(-3.46,  .00), _ESVPCOM(-2.08,  .00), _ESVPCOM(  .11, -.40),
          _ESVPCOM(  .17, -.26), _ESVPCOM( 4.51,  .00), _ESVPCOM(-1.73,  .00),
          _ESVPCOM(  .43, -.10), _ESVPCOM(  .10, -.59), _ESVPCOM( 2.51,  .00),
          _ESVPCOM( -.10, -.43), _ESVPCOM(  .33,  .52), _ESVPCOM(  .01,  .11)
   };
 
   _ESVPCOM ztas18[12] = {
          _ESVPCOM( 1.0, 1.0), _ESVPCOM(-1.0, 1.0), _ESVPCOM( 0.0,-1.0),
          _ESVPCOM( 1.0,-1.0), _ESVPCOM(-1.0, 0.0), _ESVPCOM(-1.0,-1.0),
          _ESVPCOM( 2.0, 0.0), _ESVPCOM( 0.0,-2.0), _ESVPCOM(-2.0, 2.0),
          _ESVPCOM( 0.0,-2.0), _ESVPCOM( 2.0, 2.0), _ESVPCOM( 2.0,-2.0)
   };
 
   double zresd18[3] = { -3.46, 4.51, 2.51 };
   double zrese18[2] = { -2.08, -1.73 };
   _ESVPCOM zrestp18[3] = { _ESVPCOM(1.29,-0.29),_ESVPCOM(1.03,0.36),
                           _ESVPCOM(1.88,-0.43) };
   _ESVPCOM zrestq18[3] = { _ESVPCOM(1.69,-0.14),_ESVPCOM(1.91,0.41),
                           _ESVPCOM(0.0,0.0) };
   double dtvts18[12] = {
          -0.25, 0.26, 0.93, 0.00,
          -0.54, -0.84, 0.09, 0.00,
          -0.81, 0.48, -0.35, 0.00
   };
 
   double dtus18[12] = {
          -0.37, -0.51, -0.64, -0.44,
          0.28, -0.82,  0.50, 0.00,
          0.86, -0.02, -0.51, 0.03
   };
 
   _ESVPCOM ztus18[12] = {
          _ESVPCOM(   .34,   .00),_ESVPCOM(  -.60,  -.11),_ESVPCOM(   .08,   .71),
          _ESVPCOM(  -.82,   .00),_ESVPCOM(  -.06,  -.37),_ESVPCOM(   .34,   .25),
          _ESVPCOM(  -.45,   .00),_ESVPCOM(  -.35,   .60),_ESVPCOM(  -.55,   .08),
          _ESVPCOM(   .00,   .00),_ESVPCOM(   .00,   .00),_ESVPCOM(   .00,   .00)
   };
 
   double dtss18[3] = {33.64, 9.89, 7.70};
   double ztss18[3] = {5.51, 3.31, 2.15};
 
   int ictx, nprow, npcol,ierr=5, ilerr, i, INCX, info=0, lwork=0;
   double error=0.0, tol=0.01;  // will not pass if tol is lower than 0.01
   char modn[9], *modname=modn;
   char order[]="R";
   char scope[]="A";
   char   top[]="1";
   double dta16[16], dtd16[4], dte16[4], dtt16[4], work[5000];
   _ESVPCOM zta16[16], ztt16[4], workz[1],rworkz[1];
   for (i=0; i< 16; i++)
    dta16[i] = 0.0;
   for (i=0; i< 16; i++)
    zta16[i] = _ESVPCOM(0.0,0.0);
   double dtas16[16] = { 5.,  0.,  0.,  0.,
                         4.,  5.,  0.,  0.,
                         1.,  1.,  4.,  0.,
                         1.,  1.,  2.,  4. };
 
   double dresa16[16] = { 1.,   0.,    0.,   0.,
                          0.,   6.,    0.,   0.,
                          0.41, 2.83,  7.0,  0.,
                          0.22, 0.22, -2.45, 4.0 };
 
   double dresd16[4] = { 1.,  6.,  7.,    4. };
   double drese16[4] = { 0.,  0.,  2.83, -2.45 };
   double drest16[4] = { 0.,  0.,  1.71,  1.82 };
 
     _ESVPCOM ztas16[16] = {
     _ESVPCOM(5.,0.),_ESVPCOM(4.,1.),_ESVPCOM(1.,2.),_ESVPCOM(2.,3.),
     _ESVPCOM(0.,0.),_ESVPCOM(5.,0.),_ESVPCOM(1.,0.),_ESVPCOM(3.,2.),
     _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(4.,0.),_ESVPCOM(5.,1.),
     _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(4.,0.)};
 
     _ESVPCOM zresa16[16] = {
     _ESVPCOM(5.,0.),_ESVPCOM(-5.92,0.00),_ESVPCOM(0.12,0.19),
     _ESVPCOM(0.23,0.28),
     _ESVPCOM(0.,0.),_ESVPCOM(10.09,0.00),_ESVPCOM(2.36,0.00),
     _ESVPCOM(0.14,0.19),
     _ESVPCOM(0.,0.),_ESVPCOM( 0.00,0.00),_ESVPCOM(4.16,0.00),
     _ESVPCOM(1.62,0.00),
     _ESVPCOM(0.,0.),_ESVPCOM( 0.00,0.00),_ESVPCOM(0.00,0.00),
     _ESVPCOM(-1.25,0.00)};
 
      double dresd161[4] = { 5., 10.09,  4.16, -1.25 };
      double drese161[4] = {-5.92,  2.36,  1.62,  0.00 };
     _ESVPCOM zrest16[4] = {
     _ESVPCOM(1.68,.17),_ESVPCOM(1.87,.21),_ESVPCOM(1.96,-.27),
     _ESVPCOM(0.,0.)};
 
  blacs_pinfo(iam,nnodes);
  nprow = 1;
  npcol = nnodes;
  blacs_get(0,0,ictx);
  blacs_gridinit(ictx,order,nprow,npcol,strlen(order));
  blacs_gridinfo(ictx,nprow,npcol,myrow,mycol);
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 4;
  desc_a[N_]    = 4;
  desc_a[MB_]   = 1;
  desc_a[NB_]   = 1;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(4,1,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
// following desc_x is created to check vectors d, e, and t
// IX = 1
// JX = JA
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 1;
  INCX = desc_x[M_];
  desc_x[N_]    = desc_a[N_];
  desc_x[MB_]   = 1;
  desc_x[NB_]   = desc_a[NB_];
  desc_x[CSRC_] = desc_a[CSRC_];
  desc_x[CTXT_] = ictx;
  desc_x[LLD_]  = INCX;
 
  modname = "PDSYTRD ";
  pdmtxcpy(4, 4, 1, 1, dta16, desc_a, dtas16);
  pdsytrd("U", 4, dta16, 1, 1, desc_a, dtd16, dte16, dtt16, work, lwork, info);
  if (info != 0)
  {
   if (iam == 0)
    cout << " INFO from PDSYTRD = " << info;
    *ifail = *ifail + 1;
  }
  else
  {
  pdmtchk(modname, 4, 4, 1, 1, dta16, desc_a, dresa16, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  for (i=0; i< nprow; i++)
   {
    desc_x[RSRC_] = i;
    pdvtchk(modname,4, 1, 1, dtd16, desc_x, dresd16, 1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
    pdvtchk(modname,4, 1, 1, dte16, desc_x, drese16, 1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
    pdvtchk(modname,4, 1, 1, dtt16, desc_x, drest16, 1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
   }
  }
 
  modname = "PZHETRD ";
  pzmtxcpy(4, 4, 1, 1, zta16, desc_a, ztas16);
  pzhetrd("L", 4, zta16, 1, 1,desc_a,dtd16,dte16,ztt16,workz,lwork,info);
  if (info != 0)
  {
   if (iam == 0)
    cout << " INFO from PZHETRD = " << info;
    *ifail = *ifail + 1;
  }
  else
  {
  pzmtchk(modname, 4, 4, 1, 1, zta16, desc_a, zresa16, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  for (i=0; i< nprow; i++)
   {
    desc_x[RSRC_] = i;
    pdvtchk(modname,4, 1, 1, dtd16, desc_x, dresd161,1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
    pdvtchk(modname,4, 1, 1, dte16, desc_x, drese161,1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
    pzvtchk(modname,4, 1, 1, ztt16, desc_x, zrest16,1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
   }
  }
 
// PDSYTRD   PZHETRD END
 
// PDGEHRD   START : 17
 
   double dta17[9], dtt17[2];
   double dtas17[9] = {  33.00, -24.00,  -8.00,
                         16.00, -10.00,  -4.00,
                         72.00, -57.00, -17.00 };
 
   double dresa17[9] = {  33.00,  25.30,  0.16,
                         -37.95, -29.00,  0.00,
                          63.25,  53.00,  2.00 };
 
   double drest17[2] = { 1.95, 0.0 };
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 3;
  desc_a[N_]    = 3;
  desc_a[MB_]   = 1;
  desc_a[NB_]   = 1;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(3,1,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
// following desc_x is created to check vectors d, e, and t
// IX = 1
// JX = JA
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 1;
  INCX = desc_x[M_];
  desc_x[N_]    = desc_a[N_]-1;
  desc_x[MB_]   = 1;
  desc_x[NB_]   = desc_a[NB_];
  desc_x[CSRC_] = desc_a[CSRC_];
  desc_x[CTXT_] = ictx;
  desc_x[LLD_]  = INCX;
 
  modname = "PDGEHRD ";
  pdmtxcpy(3, 3, 1, 1, dta17, desc_a, dtas17);
  pdgehrd(3, 1, 3, dta17, 1, 1, desc_a, dtt17, work, lwork, info);
  if (info != 0)
  {
   if (iam == 0)
    cout << " INFO from PDGEHRD = " << info;
    *ifail = *ifail + 1;
  }
  else
  {
  pdmtchk(modname, 3, 3, 1, 1, dta17, desc_a, dresa17, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  for (i=0; i< nprow; i++)
   {
    desc_x[RSRC_] = i;
    pdvtchk(modname,2, 1, 1, dtt17, desc_x, drest17, 1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
   }
  }
 
// PDGEHRD   END
 
// PDGEBRD PZGEBRD PDGESVD   START : 18
 
   int desc_xr[DLEN_], desc_xc[DLEN_];
   double dta18[12], dtd18[3], dte18[2], dttq18[3], dttp18[3];
   double dtas18[12] = { 10.00,   2.00,   3.00,  4.0,
                          5.00,  16.00,   7.00,  8.0,
                          9.00,  10.00,  21.00, 12.0  };
 
   double dresa18[12] = { -11.36,   0.09,  0.14, 0.19,
                           22.80,  23.32,  0.46, 0.22,
                            0.56,   1.67, -9.68, 0.08 };
 
   double dresd18[3] = {-11.36, 23.32, -9.68  };
   double drese18[2] = { 22.80,  1.67 };
   double drestq18[3] = { 1.88,  1.59,  1.99  };
   double drestp18[3] = { 1.52,  0.0 ,  0.0   };
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 4;
  desc_a[N_]    = 3;
  desc_a[MB_]   = 2;
  desc_a[NB_]   = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(4,2,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
// following desc_xr, and desc_xc are created to check vectors d, e, tp, and tq
 
  desc_xr[DT_]   = BLOCK_CYCLIC_2D;
  desc_xr[M_]    = 1;
  INCX = desc_xr[M_];
  desc_xr[N_]    = MAX(desc_a[M_],desc_a[N_]);
  desc_xr[MB_]   = desc_a[MB_];
  desc_xr[NB_]   = desc_a[NB_];
  desc_xr[RSRC_] = desc_a[RSRC_];
  desc_xr[CSRC_] = desc_a[CSRC_];
  desc_xr[CTXT_] = desc_a[CTXT_];
  desc_xr[LLD_]  = INCX;
 
  desc_xc[DT_]   = BLOCK_CYCLIC_2D;
  desc_xc[M_]    = MAX(desc_a[M_],desc_a[N_]);
  desc_xc[N_]    = 1;
  desc_xc[MB_]   = desc_a[MB_];
  desc_xc[NB_]   = desc_a[NB_];
  desc_xc[RSRC_] = desc_a[RSRC_];
  desc_xc[CSRC_] = desc_a[CSRC_];
  desc_xc[CTXT_] = desc_a[CTXT_];
  il_lda = numroc(desc_a[N_],desc_a[NB_],mycol,desc_a[CSRC_],npcol);
  desc_xc[LLD_]  = MAX(desc_a[LLD_],il_lda);
 
  modname = "PDGEBRD ";
  pdmtxcpy(4, 3, 1, 1, dta18, desc_a, dtas18);
  pdgebrd(4, 3, dta18, 1, 1, desc_a, dtd18, dte18, dttq18, dttp18, work,
          lwork, info);
  if (info != 0)
  {
   if (iam == 0)
    cout << " INFO from PDGEBRD = " << info;
    *ifail = *ifail + 1;
  }
  else
  {
  pdmtchk(modname, 4, 3, 1, 1, dta18, desc_a, dresa18, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  for (i=0; i< nprow; i++)
   {
     // as M >= N, only vector D and Tq are neede to check as row vector
    desc_xr[N_] = 3;
    desc_xr[RSRC_] = i;
    pdvtchk(modname,3, 1, 1, dtd18, desc_xr, dresd18, 1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
    pdvtchk(modname,3, 1, 1, dttq18, desc_xr, drestq18, 1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
   }
 
  for (i=0; i< npcol; i++)
   {
     // as M >= N, only vector E and Tp are neede to check as column vector
    desc_xc[M_] = 2;
    desc_xc[CSRC_] = i;
    pdvtchk(modname,2, 1, 1, dte18, desc_xc, drese18, 1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
    desc_xc[M_] = 3;
    pdvtchk(modname,3, 1, 1, dttp18, desc_xc, drestp18, 1, tol, ierr, ilerr,
          error, ifail, strlen(modname));
   }
  }
 
// PDGEBRD   END
 
 
// PDGESVD   START : 18
 
  desc_a[DT_]    = BLOCK_CYCLIC_2D;
  desc_a[M_]     = 4;
  desc_a[N_]     = 3;
  desc_a[MB_]    = 2;
  desc_a[NB_]    = 2;
  desc_a[RSRC_]  = 0;
  desc_a[CSRC_]  = 0;
  desc_a[CTXT_]  = ictx;
  il_lda         = numroc(4,2,myrow,0,nprow);
  desc_a[LLD_]   = MAX(1,il_lda);
 
// following desc_u and desc_vt are created to check vectors u, vt
 
  desc_vt[DT_]   = BLOCK_CYCLIC_2D;
  desc_vt[M_]    = 4;
  desc_vt[N_]    = 3;
  desc_vt[MB_]   = 2;
  desc_vt[NB_]   = 2;
  desc_vt[RSRC_] = 0;
  desc_vt[CSRC_] = 0;
  desc_vt[CTXT_] = desc_a[CTXT_];
  il_lda         = numroc(4,2,myrow,0,nprow);
  desc_vt[LLD_]  = MAX(1,il_lda);
 
  desc_u[DT_]    = BLOCK_CYCLIC_2D;
  desc_u[M_]     = 4;
  desc_u[N_]     = 3;
  desc_u[MB_]    = 2;
  desc_u[NB_]    = 2;
  desc_u[RSRC_]  = 0;
  desc_u[CSRC_]  = 0;
  desc_u[CTXT_]  = desc_a[CTXT_];
  il_lda         = numroc(4,2,myrow,0,nprow);
  desc_u[LLD_]   = MAX(1,il_lda);
 
  desc_xr[DT_]   = BLOCK_CYCLIC_2D;
  desc_xr[M_]    = 1;
  INCX           = desc_xr[M_];
  desc_xr[N_]    = 3;
  desc_xr[MB_]   = desc_a[MB_];
  desc_xr[NB_]   = 3;
  desc_xr[RSRC_] = desc_a[RSRC_];
  desc_xr[CSRC_] = desc_a[CSRC_];
  desc_xr[CTXT_] = desc_a[CTXT_];
  desc_xr[LLD_]  = INCX;
 
  // Zero out array so that extra space has no garbage during the error check
  for(ix=0;ix<12;ix++) dtvt18[ix] = 0.00;
 
  modname = "PDGESVD ";
  pdmtxcpy(4, 3, 1, 1, dta18, desc_a, dtas18);
 
  pdgesvd ("V", "V", 4, 3, dta18, 1, 1, desc_a, dts18,
                dtu18, 1, 1, desc_u, dtvt18, 1, 1, desc_vt,
                work, lwork, info);
 
  if (info != 0) {
   if (iam == 0) cout << " INFO from PDGESVD = " << info << endl;
   *ifail = *ifail + 1;
  } else {
   pdmtchk(modname, 4, 3, 1, 1, dtu18, desc_u, dtus18, tol,
           ierr,ilerr, error, ifail,strlen(modname));
   pdmtchk(modname, 4, 3, 1, 1, dtvt18, desc_vt, dtvts18, tol,
           ierr,ilerr, error, ifail,strlen(modname));
   for (i=0; i < 3; i++)
       if ( ABSP(dts18[i] - dtss18[i]) > tol)
        {
          *ifail = *ifail + 1;
         cout << " PDGESVD dts18[" << i << "] = " << dts18[i];
         cout << "  dtss18[" << i << "] = " << dtss18[i] << "\n";
        }
  }
 
// PDGESVD   END
// PZGEBRD   START : 18
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 3;
  desc_a[N_] = 4;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
// following desc_xr, and desc_xc are created to check vectors d,e,tp,tq
 
  desc_xr[DT_] = BLOCK_CYCLIC_2D;
  desc_xr[M_] = 1;
  INCX = desc_xr[M_];
  desc_xr[N_] = MAX(desc_a[M_],desc_a[N_]);
  desc_xr[MB_] = desc_a[MB_];
  desc_xr[NB_] = desc_a[NB_];
  desc_xr[RSRC_] = desc_a[RSRC_];
  desc_xr[CSRC_] = desc_a[CSRC_];
  desc_xr[CTXT_] = desc_a[CTXT_];
  desc_xr[LLD_] = INCX;
 
  desc_xc[DT_] = BLOCK_CYCLIC_2D;
  desc_xc[M_] = MAX(desc_a[M_],desc_a[N_]);
  desc_xc[N_] = 1;
  desc_xc[MB_] = desc_a[MB_];
  desc_xc[NB_] = desc_a[NB_];
  desc_xc[RSRC_] = desc_a[RSRC_];
  desc_xc[CSRC_] = desc_a[CSRC_];
  desc_xc[CTXT_] = desc_a[CTXT_];
  il_lda = numroc(desc_a[N_],desc_a[NB_],mycol,desc_a[CSRC_],npcol);
  desc_xc[LLD_] = MAX(desc_a[LLD_],il_lda);
 
  modname = "PZGEBRD ";
  pzmtxcpy(3, 4, 1, 1, zta18, desc_a, ztas18);
  pzgebrd(3, 4, zta18, 1, 1, desc_a, ztd18, zte18, zttq18, zttp18, workz,
          lwork,info);
  if (info != 0) {
   if (iam == 0) cout << " INFO from PZGEBRD = " << info << endl;
   *ifail = *ifail + 1;
  } else {
   pzmtchk(modname, 3, 4, 1, 1, zta18, desc_a, zresa18, tol, ierr,ilerr,
           error, ifail,strlen(modname));
 
   for (i=0; i< nprow; i++) {
    // as M >= N, only vector D and Tq are neede to check as row vector
    desc_xr[N_] = 2;
    desc_xr[RSRC_] = i;
    pdvtchk(modname,2, 1, 1, zte18, desc_xr, zrese18, 1, tol, ierr,ilerr,
            error, ifail,strlen(modname));
    desc_xr[N_] = 3;
    pzvtchk(modname,3, 1, 1, zttq18, desc_xr, zrestq18, 1, tol, ierr,ilerr,
            error, ifail,strlen(modname));
   }
 
   for (i=0; i< npcol; i++) {
    // as M >= N, only vector E and Tp are neede to check as column vector
    desc_xc[M_] = 3;
    desc_xc[CSRC_] = i;
    pdvtchk(modname,3, 1, 1, ztd18, desc_xc, zresd18, 1, tol, ierr,ilerr,
            error, ifail,strlen(modname));
    pzvtchk(modname,3, 1, 1, zttp18, desc_xc, zrestp18, 1, tol, ierr,ilerr,
            error, ifail,strlen(modname));
   }
  }
 
// PZGEBRD   END
// PZGESVD   START : 18
 
  desc_a[DT_]    = BLOCK_CYCLIC_2D;
  desc_a[M_]     = 3;
  desc_a[N_]     = 4;
  desc_a[MB_]    = 2;
  desc_a[NB_]    = 2;
  desc_a[RSRC_]  = 0;
  desc_a[CSRC_]  = 0;
  desc_a[CTXT_]  = ictx;
  il_lda         = numroc(3,2,myrow,0,nprow);
  desc_a[LLD_]   = MAX(1,il_lda);
 
// following desc_u and desc_vt are created to check vectors u, vt
 
  desc_xr[DT_]   = BLOCK_CYCLIC_2D;
  desc_xr[M_]    = 1;
  INCX           = desc_xr[M_];
  desc_xr[N_]    = 3;
  desc_xr[MB_]   = desc_a[MB_];
  desc_xr[NB_]   = 3;
  desc_xr[RSRC_] = desc_a[RSRC_];
  desc_xr[CSRC_] = desc_a[CSRC_];
  desc_xr[CTXT_] = desc_a[CTXT_];
  desc_xr[LLD_]  = INCX;
 
  desc_vt[DT_]   = BLOCK_CYCLIC_2D;
  desc_vt[M_]    = 3;
  desc_vt[N_]    = 4;
  desc_vt[MB_]   = 2;
  desc_vt[NB_]   = 2;
  desc_vt[RSRC_] = 0;
  desc_vt[CSRC_] = 0;
  desc_vt[CTXT_] = desc_a[CTXT_];
  il_lda         = numroc(3,2,myrow,0,nprow);
  desc_vt[LLD_]  = MAX(1,il_lda);
 
  desc_u[DT_]   = BLOCK_CYCLIC_2D;
  desc_u[M_]    = 3;
  desc_u[N_]    = 3;
  desc_u[MB_]   = 2;
  desc_u[NB_]   = 2;
  desc_u[RSRC_] = 0;
  desc_u[CSRC_] = 0;
  desc_u[CTXT_] = desc_a[CTXT_];
  il_lda        = numroc(3,2,myrow,0,nprow);
  desc_u[LLD_]  = MAX(1,il_lda);
 
  // Zero out array so that extra space has no garbage during the error check
  for(ix=0;ix<12;ix++) ztu18[ix] = _ESVPCOM(0.00, 0.00);
 
  modname = "PZGESVD ";
  pzmtxcpy(3, 4, 1, 1, zta18, desc_a, ztas18);
 
  pzgesvd ("V", "V", 3, 4, zta18, 1, 1, desc_a, zts18,
           ztu18, 1, 1, desc_u, ztvt18, 1, 1, desc_vt,
           workz, lwork, rwork, lrwork, info);
 
  if (info != 0) {
   if (iam == 0) cout << " INFO from PZGESVD = " << info << endl;
   *ifail = *ifail + 1;
  } else {
   pzmtchk(modname, 3, 4, 1, 1, ztu18, desc_u, ztus18, tol,
           ierr,ilerr, error, ifail,strlen(modname));
   pzmtchk(modname, 3, 4, 1, 1, ztvt18, desc_vt, ztvts18, tol,
           ierr,ilerr, error, ifail,strlen(modname));
   for (i=0; i < 3; i++)
       if ( ABSP(zts18[i] - ztss18[i]) > tol)
        {
          *ifail = *ifail + 1;
         cout << " PZGESVD zts18[" << i << "] = " << zts18[i];
         cout << "  ztss18[" << i << "] = " << ztss18[i] << "\n";
        }
  }
 
// PZGESVD   END
// PDSYEVX   PZHEEVX START : 19
 
   int M=0,NZ=0,ifaile[4],desc_z[DLEN_],iwork[1],liwork=0;
   double dta19[16], dtz19[16], dtw19[4];
   _ESVPCOM zta19[16], ztz19[16];
   int ifailres[4] = { 0, 0, 0, 0 };
   double *dtg19 = new double[nprow*npcol];
   double *dtg19res = new double[nprow*npcol];
   int    *iclustr  = new int[2*nprow*npcol];
   int    *iclustres= new int[2*nprow*npcol];
   istor[0] = 0;
   if (!dtg19 || !dtg19res || !iclustr || !iclustres)
     istor[0] = 1;
     igamx2d(ictx,scope, top, 1, 1, istor,
                     1, ra, ca, 1, 0, 0,strlen(scope),strlen(top));
   if (istor[0]==1)
    {
     if (iam == 0)
     cout << "Memory allocation error for PDSYEVX\n";
     delete [] dtg19;         // free memory
     delete [] dtg19res;
     delete [] iclustr;
     delete [] iclustres;
     *ifail = *ifail + 1;
     blacs_gridexit(ictx);
     return;
    }
   for (i=0; i< nprow*npcol; i++)
     dtg19res[i] = -1.0;
   for (i=0; i< 2*nprow*npcol; i++)
     iclustres[i] = 0;
 
   double dtas19[16] = { 5.,  0.,  0.,  0.,
                         4.,  5.,  0.,  0.,
                         1.,  1.,  4.,  0.,
                         1.,  1.,  2.,  4. };
 
   double dresz19[16] = { 0.7071, -0.7071,  0.0,     0.0,
                          0.0,     0.0,    -0.7071,  0.7071,
                         -0.3162, -0.3162,  0.6325,  0.6325,
                         -0.6325, -0.6325, -0.3162, -0.3162  };
 
   double dresw19[4] = { 1.0,  2.0,  5.0,   10.0 };
 
     _ESVPCOM ztas19[16] =  {
     _ESVPCOM(5.,0.),_ESVPCOM(4.,1.),_ESVPCOM(1.,2.),_ESVPCOM(2.,3.),
     _ESVPCOM(0.,0.),_ESVPCOM(5.,0.),_ESVPCOM(1.,0.),_ESVPCOM(3.,2.),
     _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(4.,0.),_ESVPCOM(5.,1.),
     _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(4.,0.)};
 
     _ESVPCOM zresz19[16] = {
      _ESVPCOM( .0926, .0000),_ESVPCOM( .0840,-.2873),
      _ESVPCOM( .5013,-.3461),_ESVPCOM(-.6703, .2854),
      _ESVPCOM( .7591, .0000),_ESVPCOM(-.5874, .0177),
      _ESVPCOM( .0526,-.0674),_ESVPCOM(-.0155,-.2662),
      _ESVPCOM( .3758, .0000),_ESVPCOM( .5370,-.2067),
      _ESVPCOM(-.6287,-.2149),_ESVPCOM(-.2294,-.1834),
      _ESVPCOM(-.5234, .0000),_ESVPCOM(-.4515,-.1735),
      _ESVPCOM(-.2864,-.3134),_ESVPCOM(-.3059,-.4672)};
 
      double zresw19[4] = { -1.765, 0.727, 4.664, 14.374 };
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 4;
  desc_a[N_]    = 4;
  desc_a[MB_]   = 1;
  desc_a[NB_]   = 1;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(4,1,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_z[DT_]   = BLOCK_CYCLIC_2D;
  desc_z[M_]    = 4;
  desc_z[N_]    = 4;
  desc_z[MB_]   = 1;
  desc_z[NB_]   = 1;
  desc_z[RSRC_] = 0;
  desc_z[CSRC_] = 0;
  desc_z[CTXT_] = ictx;
  il_lda = numroc(4,1,myrow,0,nprow);
  desc_z[LLD_]  = MAX(1,il_lda);
 
  modname = "PDSYEVX ";
  pdmtxcpy(4, 4, 1, 1, dta19, desc_a, dtas19);
  pdsyevx("V","A","U", 4, dta19, 1, 1, desc_a, 0.0, 0.0, 0, 0, -1.0, M, NZ,
          dtw19, -1.0, dtz19, 1, 1, desc_z, work, lwork, iwork, liwork, ifaile,
          iclustr, dtg19, info);
  if (info != 0)
  {
   if (iam == 0)
    cout << " INFO from PDSYEVX = " << info << "\n";
    *ifail = *ifail + 1;
  }
  if (M != 4)
  {
   if (iam == 0)
    cout << " M from PDSYEVX = " << M << "\n";
    *ifail = *ifail + 1;
  }
  if (NZ != 4)
  {
   if (iam == 0)
    cout << " NZ from PDSYEVX = " << NZ << "\n";
    *ifail = *ifail + 1;
  }
  if ( info != 0 || M != 4 || NZ != 4) *ifail = *ifail + 1;
  pdmtchk(modname, 4, 4, 1, 1, dtz19, desc_z, dresz19, tol, ierr, ilerr,
          error, ifail, strlen(modname));
   for (i=0; i< nprow*npcol; i++)
       if ((ABSP(dtg19[i]) - ABSP(dtg19res[i])) > tol)
        {
          *ifail = *ifail + 1;
         cout << " PDSYEVX dtg19[" << i << "] = " << dtg19[i];
         cout << "  dtg19res[" << i << "] = " << dtg19res[i] << "\n";
        }
   for (i=0; i< 2*nprow*npcol; i++)
       if (iclustr[i] != iclustres[i])
        {
          *ifail = *ifail + 1;
         cout << " PDSYEVX iclustr[" << i << "] = " << iclustr[i];
         cout << "  iclustres[" << i << "] = " << iclustres[i] << "\n";
        }
   for (i=0; i< 4; i++)
       if (ifaile[i] != ifailres[i])
        {
          *ifail = *ifail + 1;
         cout << " PDSYEVX ifaile[" << i << "] = " << ifaile[i];
         cout << " ifailres[" << i << "] = " << ifailres[i] << "\n";
        }
   for (i=0; i< 4; i++)
       if ((ABSP(dtw19[i]) - ABSP(dresw19[i])) > tol)
        {
          *ifail = *ifail + 1;
         cout << " PDSYEVX dtw19[" << i << "] = " << dtw19[i];
         cout << "  dresw19[" << i << "] = " << dresw19[i] << "\n";
        }
 
  modname = "PZHEEVX ";
  pzmtxcpy(4, 4, 1, 1, zta19, desc_a, ztas19);
  pzheevx("V","A","L", 4, zta19, 1, 1, desc_a, 0.0, 0.0, 0, 0, -1.0, M, NZ,
          dtw19,-1.0,ztz19,1,1,desc_z,workz,0, work,lwork,iwork,liwork, ifaile,
          iclustr, dtg19, info);
  if (info != 0)
  {
   if (iam == 0)
    cout << " INFO from PZHEEVX = " << info << "\n";
    *ifail = *ifail + 1;
  }
  if (M != 4)
  {
   if (iam == 0)
    cout << " M from PZHEEVX = " << M << "\n";
    *ifail = *ifail + 1;
  }
  if (NZ != 4)
  {
   if (iam == 0)
    cout << " NZ from PZHEEVX = " << NZ << "\n";
    *ifail = *ifail + 1;
  }
  if ( info != 0 || M != 4 || NZ != 4) *ifail = *ifail + 1;
  pzmtchk(modname, 4, 4, 1, 1, ztz19, desc_z, zresz19, tol, ierr, ilerr,
          error, ifail, strlen(modname));
   for (i=0; i< nprow*npcol; i++)
       if ((ABSP(dtg19[i]) - ABSP(dtg19res[i])) > tol)
        {
          *ifail = *ifail + 1;
         cout << " PZHEEVX dtg19[" << i << "] = " << dtg19[i];
         cout << "  dtg19res[" << i << "] = " << dtg19res[i] << "\n";
        }
   for (i=0; i< 2*nprow*npcol; i++)
       if (iclustr[i] != iclustres[i])
        {
          *ifail = *ifail + 1;
         cout << " PZHEEVX iclustr[" << i << "] = " << iclustr[i];
         cout << "  iclustres[" << i << "] = " << iclustres[i] << "\n";
        }
   for (i=0; i< 4; i++)
       if (ifaile[i] != ifailres[i])
        {
          *ifail = *ifail + 1;
         cout << " PZHEEVX ifaile[" << i << "] = " << ifaile[i];
         cout << " ifailres[" << i << "] = " << ifailres[i] << "\n";
        }
   for (i=0; i< 4; i++)
       if ((ABSP(dtw19[i]) - ABSP(zresw19[i])) > tol)
        {
          *ifail = *ifail + 1;
         cout << " PZHEEVX dtw19[" << i << "] = " << dtw19[i];
         cout << "  zresw19[" << i << "] = " << zresw19[i] << "\n";
        }
 
// PDSYEVX  PZHEEVX  END
 
// PDSYGST  PZHEGST START : 20
 
      int     ibtype,iret,il_ldb;
      double dta20[16], dtb20[16], scale, scaler=1.0,dtw20[4],dtz20[16];
      double dtc20[16],tstnrm;
   _ESVPCOM zta20[16], ztb20[16], ztz20[16];
      double ztw20[4];
      double  dtas20[16] =  { -1.,  1., -1.,  1.,
                               0.,  1., -1.,  1.,
                               0.,  0.,  1., -1.,
                               0.,  0.,  0.,  1.};
 
      double dtbs20[16] =  {   2.,  1.,  0.,  0.,
                               0.,  2.,  1.,  0.,
                               0.,  0.,  2.,  1.,
                               0.,  0.,  0.,  2.};
 
      double  dtac20[16] =  { -1.,  1., -1.,  1.,
                               1.,  1., -1.,  1.,
                              -1., -1.,  1., -1.,
                               1.,  1., -1.,  1.};
 
      double dtbc20[16] =  {   2.,  1.,  0.,  0.,
                               1.,  2.,  1.,  0.,
                               0.,  1.,  2.,  1.,
                               0.,  0.,  1.,  2.};
 
      double dresa20[16]= { -0.50000, 0.86603,-1.22474, 1.58114,
                             0.00000,-0.16667,-0.23570, 0.54772,
                             0.00000, 0.00000, 1.16667,-1.93649,
                            0.00000, 0.00000, 0.00000, 3.10000};
 
      double dresz20[16]= { 0.88420,-0.56190, 0.30720,-0.11980,
                            0.00000, 0.36340, 0.42660, 0.06310,
                            0.00000,-0.40980, 0.13430, 0.54410,
                            0.13490,-0.76430, 0.95160,-0.69690};
 
      double dresw20[4]= { -1.55261, 0.00000, 0.00000, 5.15261};
 
     _ESVPCOM ztas20[16] =  {
     _ESVPCOM(1.,0.),_ESVPCOM(5.,-2.),_ESVPCOM(7.,4.),_ESVPCOM(9.,-6.),
     _ESVPCOM(0.,0.),_ESVPCOM(10.,0.),_ESVPCOM(15.,6.),_ESVPCOM(20.,-4.),
     _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(20.,0.),_ESVPCOM(25.,-9.),
     _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(30.,0.)};
 
     _ESVPCOM ztbs20[16] =  {
     _ESVPCOM(9.,0.),_ESVPCOM(3.,-3.),_ESVPCOM(3.,3.),_ESVPCOM(3.,-3.),
     _ESVPCOM(0.,0.),_ESVPCOM(18.,0.),_ESVPCOM(8.,6.),_ESVPCOM(8.,-6.),
     _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(27.,0.),_ESVPCOM(12.,-9.),
     _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(61.,0.)};
 
     _ESVPCOM zresa20[16] =  {
     _ESVPCOM(.1111,0.),_ESVPCOM(.3889,-.1389),_ESVPCOM(.2919,.2485),
     _ESVPCOM(.2422,-.2175),
     _ESVPCOM(0.,0.),_ESVPCOM(.3472,0.),_ESVPCOM(.5714,-.0652),
     _ESVPCOM(.2001,-.0009),
     _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(.0757,0.),
     _ESVPCOM(.4003,.2645),
     _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(-.0488,.0)};
 
     _ESVPCOM zresb20[16] =  {
     _ESVPCOM(3.,0.),_ESVPCOM(1.,-1.),_ESVPCOM(1.,1.),_ESVPCOM(1.,-1.),
     _ESVPCOM(0.,0.),_ESVPCOM(4.,0.),_ESVPCOM(2.,1.),_ESVPCOM(1.5,-1.5),
     _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(4.4721,0.),
     _ESVPCOM(2.3479,-.5590),
     _ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(0.,0.),_ESVPCOM(6.9767,0.)};
 
     _ESVPCOM zresz20[16] =  {
     _ESVPCOM(.2784,-.0218),_ESVPCOM(.0601, .1506),_ESVPCOM(0.,-.1675),
     _ESVPCOM(-.0406, .0627),
     _ESVPCOM(-.1300,-.0562),_ESVPCOM(.1844,.1016),_ESVPCOM(.0004,-.0082),
     _ESVPCOM(-.0726,-.0362),
     _ESVPCOM(.1842,-.0064),_ESVPCOM(-.0064,-.0516),_ESVPCOM(-.0591,.1230),
     _ESVPCOM(-.0414,-.0635),
     _ESVPCOM(-.0692,.0118),_ESVPCOM(-.0948,.0008),_ESVPCOM(-.0946,.0178),
     _ESVPCOM(-.0513,.0023)};
 
      double zresw20[4]= { -0.7969, -0.1152, 0.1669, 1.2304};
 
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 4;
      desc_a[N_]    = 4;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
      desc_b[DT_]   = BLOCK_CYCLIC_2D;
      desc_b[M_]    = 4;
      desc_b[N_]    = 4;
      desc_b[MB_]   = 1;
      desc_b[NB_]   = 1;
      desc_b[RSRC_] = 0;
      desc_b[CSRC_] = 0;
      desc_b[CTXT_] = ictx;
      il_ldb = numroc(4,1,myrow,0,nprow);
      desc_b[LLD_]  = MAX(1,il_ldb);
 
      modname = "PDSYGST ";
      pdmtxcpy(4, 4, 1, 1, dta20, desc_a, dtas20);
      pdmtxcpy(4, 4, 1, 1, dtb20, desc_b, dtbs20);
      pdpotrf("L", 4, dtb20, 1, 1, desc_b, iret);
      pdsygst(1,"L", 4, dta20, 1, 1, desc_a, dtb20, 1, 1, desc_b,
                    scale,  info);
      if (info != 0)
      {
       if (iam == 0)
    cout << " INFO from PDSYGST = " << info << "\n";
       *ifail = *ifail + 1;
      }
      else
      {
        pdmtchk(modname, 4, 4, 1, 1, dta20, desc_a, dresa20, tol,
                    ierr, ilerr, error, ifail,strlen(modname));
 
       if ((ABSP(scale) - ABSP(scaler)) > tol)
        {
          *ifail = *ifail + 1;
         cout << " PDSYGST scale = " << scale, scaler;
        }
      }
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 4;
      desc_a[N_]    = 4;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
      desc_b[DT_]   = BLOCK_CYCLIC_2D;
      desc_b[M_]    = 4;
      desc_b[N_]    = 4;
      desc_b[MB_]   = 1;
      desc_b[NB_]   = 1;
      desc_b[RSRC_] = 0;
      desc_b[CSRC_] = 0;
      desc_b[CTXT_] = ictx;
      il_ldb = numroc(4,1,myrow,0,nprow);
      desc_b[LLD_]  = MAX(1,il_ldb);
 
      modname = "PZHEGST ";
      pzmtxcpy(4, 4, 1, 1, zta20, desc_a, ztas20);
      pzmtxcpy(4, 4, 1, 1, ztb20, desc_b, ztbs20);
      pzpotrf("L", 4, ztb20, 1, 1, desc_b, iret);
      pzhegst(1,"L", 4, zta20, 1, 1, desc_a, ztb20, 1, 1, desc_b,
                    scale,  info);
      if (info != 0)
      {
       if (iam == 0)
    cout << " INFO from PZHEGST = " << info << "\n";
       *ifail = *ifail + 1;
      }
      else
      {
        pzmtchk(modname, 4, 4, 1, 1, zta20, desc_a, zresa20, tol,
                    ierr, ilerr, error, ifail,strlen(modname));
 
       if ((ABSP(scale) - ABSP(scaler)) > tol)
        {
          *ifail = *ifail + 1;
         cout << " PZHEGST scale = " << scale, scaler;
        }
      }
/* PDSYGVX PZHEGVX         START : 20   */
 
   for (i=0; i< nprow*npcol; i++)
     dtg19res[i] = -1.0;
   for (i=0; i< 2*nprow*npcol; i++)
     iclustres[i] = 0;
 
      M = 0;
      NZ = 0;
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 4;
      desc_a[N_]    = 4;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
      desc_b[DT_]   = BLOCK_CYCLIC_2D;
      desc_b[M_]    = 4;
      desc_b[N_]    = 4;
      desc_b[MB_]   = 1;
      desc_b[NB_]   = 1;
      desc_b[RSRC_] = 0;
      desc_b[CSRC_] = 0;
      desc_b[CTXT_] = ictx;
      il_ldb = numroc(4,1,myrow,0,nprow);
      desc_b[LLD_]  = MAX(1,il_ldb);
 
      desc_z[DT_]   = BLOCK_CYCLIC_2D;
      desc_z[M_]    = 4;
      desc_z[N_]    = 4;
      desc_z[MB_]   = 1;
      desc_z[NB_]   = 1;
      desc_z[RSRC_] = 0;
      desc_z[CSRC_] = 0;
      desc_z[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_z[LLD_]  = MAX(1,il_lda);
 
      modname = "PDSYGVX ";
      lwork = 5000;
      pdmtxcpy(4, 4, 1, 1, dta20, desc_a, dtas20);
      pdmtxcpy(4, 4, 1, 1, dtb20, desc_b, dtbs20);
      pdsygvx(1,"V","A","L", 4, dta20, 1, 1, desc_a,
        dtb20, 1, 1, desc_b,  0.0, 0.0,
        0, 0, -1.0, M, NZ, dtw20, -1.0, dtz20, 1, 1, desc_z,work,
        lwork, iwork, liwork, ifaile, iclustr, dtg19, info);
  if (info != 0) {
   if (iam == 0)
    cout << " INFO from PDSYGVX = " << info << "\n";
  }
  if (M != 4) {
   if (iam == 0)
    cout << " M from PDSYGVX = " << M << "\n";
  }
  if (NZ != 4) {
   if (iam == 0)
    cout << " NZ from PDSYGVX = " << NZ << "\n";
  }
  if ( info != 0 || M != 4 || NZ != 4) *ifail = *ifail + 1;
  if (info == 0) {
  pdmtxcpy(4, 4, 1, 1, dta20, desc_a, dtac20);
  pdmtxcpy(4, 4, 1, 1, dtb20, desc_b, dtbc20);
  pdgspchk(1, 4, 4, dta20, 1, 1, desc_a, dtb20, 1, 1, desc_b,
           500.0, dtz20, 1, 1, desc_z, dtc20, 1, 1, desc_a,
           dtw20, work, lwork, tstnrm, iret);
       if (iret != 0 || (tstnrm-tstnrm) != 0.0)
        {
         *ifail = *ifail + 1;
    cout << " PDSYGVX dtz20 error = " << tstnrm << iret << "\n";
        }
   for (i=0; i< nprow*npcol; i++)
       if ((ABSP(dtg19[i]) - ABSP(dtg19res[i])) > tol)
        {
          *ifail = *ifail + 1;
         cout << " PDSYGVX dtg19[" << i << "] = " << dtg19[i];
         cout << "  dtg19res[" << i << "] = " << dtg19res[i] << "\n";
        }
   for (i=0; i< 2*nprow*npcol; i++)
       if (iclustr[i] != iclustres[i])
        {
          *ifail = *ifail + 1;
         cout << " PDSYGVX iclustr[" << i << "] = " << iclustr[i];
         cout << "  iclustres[" << i << "] = " << iclustres[i] << "\n";
        }
   for (i=0; i< 4; i++)
       if (ifaile[i] != ifailres[i])
        {
          *ifail = *ifail + 1;
         cout << " PDSYGVX ifaile[" << i << "] = " << ifaile[i];
         cout << " ifailres[" << i << "] = " << ifailres[i] << "\n";
        }
   for (i=0; i< 4; i++)
       if ((ABSP(dtw20[i]) - ABSP(dresw20[i])) > tol)
        {
          *ifail = *ifail + 1;
         cout << " PDSYGVX dtw20[" << i << "] = " << dtw20[i];
         cout << "  dresw20[" << i << "] = " << dresw20[i] << "\n";
        }
   }
   for (i=0; i< nprow*npcol; i++)
     dtg19res[i] = -1.0;
   for (i=0; i< 2*nprow*npcol; i++)
     iclustres[i] = 0;
 
      M = 0;
      NZ = 0;
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 4;
      desc_a[N_]    = 4;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
      desc_b[DT_]   = BLOCK_CYCLIC_2D;
      desc_b[M_]    = 4;
      desc_b[N_]    = 4;
      desc_b[MB_]   = 1;
      desc_b[NB_]   = 1;
      desc_b[RSRC_] = 0;
      desc_b[CSRC_] = 0;
      desc_b[CTXT_] = ictx;
      il_ldb = numroc(4,1,myrow,0,nprow);
      desc_b[LLD_]  = MAX(1,il_ldb);
 
      desc_z[DT_]   = BLOCK_CYCLIC_2D;
      desc_z[M_]    = 4;
      desc_z[N_]    = 4;
      desc_z[MB_]   = 1;
      desc_z[NB_]   = 1;
      desc_z[RSRC_] = 0;
      desc_z[CSRC_] = 0;
      desc_z[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_z[LLD_]  = MAX(1,il_lda);
 
      modname = "PZHEGVX ";
      pzmtxcpy(4, 4, 1, 1, zta20, desc_a, ztas20);
      pzmtxcpy(4, 4, 1, 1, ztb20, desc_b, ztbs20);
      pzhegvx(1,"V","A","L", 4, zta20, 1, 1, desc_a,
        ztb20, 1, 1, desc_b,  0.0, 0.0,
        0, 0, -1.0, M, NZ, ztw20, -1.0, ztz20, 1, 1, desc_z,workz,0,
        work, 0, iwork, 0, ifaile, iclustr, dtg19, info);
  if (info != 0) {
   if (iam == 0)
    cout << " INFO from PZHEGVX = " << info << "\n";
  }
  if (M != 4) {
   if (iam == 0)
    cout << " M from PZHEGVX = " << M << "\n";
  }
  if (NZ != 4) {
   if (iam == 0)
    cout << " NZ from PZHEGVX = " << NZ << "\n";
  }
  if ( info != 0 || M != 4 || NZ != 4) *ifail = *ifail + 1;
  if (info == 0) {
        pzmtchk(modname, 4, 4, 1, 1, ztz20, desc_z, zresz20, tol,
                    ierr, ilerr, error, ifail,strlen(modname));
   for (i=0; i< nprow*npcol; i++)
       if ((ABSP(dtg19[i]) - ABSP(dtg19res[i])) > tol)
        {
          *ifail = *ifail + 1;
         cout << " PZHEGVX dtg19[" << i << "] = " << dtg19[i];
         cout << "  dtg19res[" << i << "] = " << dtg19res[i] << "\n";
        }
   for (i=0; i< 2*nprow*npcol; i++)
       if (iclustr[i] != iclustres[i])
        {
          *ifail = *ifail + 1;
         cout << " PZHEGVX iclustr[" << i << "] = " << iclustr[i];
         cout << "  iclustres[" << i << "] = " << iclustres[i] << "\n";
        }
   for (i=0; i< 4; i++)
       if (ifaile[i] != ifailres[i])
        {
          *ifail = *ifail + 1;
         cout << " PZHEGVX ifaile[" << i << "] = " << ifaile[i];
         cout << " ifailres[" << i << "] = " << ifailres[i] << "\n";
        }
   for (i=0; i< 4; i++)
       if ((ABSP(ztw20[i]) - ABSP(zresw20[i])) > tol)
        {
          *ifail = *ifail + 1;
         cout << " PZHEGVX ztw20[" << i << "] = " << ztw20[i];
         cout << "  zresw20[" << i << "] = " << zresw20[i] << "\n";
        }
   }
// PDSYGST   PZHEGVX END : 20
// PDSYNTRD  START : 21
 
   double dta21[16], dtd21[4], dte21[4], dtt21[4];
 
   double dtas21[16] = { 5.,  0.,  0.,  0.,
                         4.,  5.,  0.,  0.,
                         1.,  1.,  4.,  0.,
                         1.,  1.,  2.,  4. };
 
   double dresa21[16] = { 1.,   0.,    0.,   0.,
                          0.,   6.,    0.,   0.,
                          0.41, 2.83,  7.0,  0.,
                          0.22, 0.22, -2.45, 4.0 };
 
   double dresd21[4] = { 1.,  6.,  7.,    4. };
   double drese21[4] = { 0.,  0.,  2.83, -2.45 };
   double drest21[4] = { 0.,  0.,  1.71,  1.82 };
 
 
   desc_a[DT_]   = BLOCK_CYCLIC_2D;
   desc_a[M_]    = 4;
   desc_a[N_]    = 4;
   desc_a[MB_]   = 1;
   desc_a[NB_]   = 1;
   desc_a[RSRC_] = 0;
   desc_a[CSRC_] = 0;
   desc_a[CTXT_] = ictx;
   il_lda = numroc(4,1,myrow,0,nprow);
   desc_a[LLD_]  = MAX(1,il_lda);
 
   // following desc_x is created to check vectors d, e, and t
   // IX = 1
   // JX = JA
 
   desc_x[DT_]   = BLOCK_CYCLIC_2D;
   desc_x[M_]    = 1;
   INCX = desc_x[M_];
   desc_x[N_]    = desc_a[N_];
   desc_x[MB_]   = 1;
   desc_x[NB_]   = desc_a[NB_];
   desc_x[CSRC_] = desc_a[CSRC_];
   desc_x[CTXT_] = ictx;
   desc_x[LLD_]  = INCX;
 
   modname = "PDSYNTRD ";
 
   for (i=0; i< 16; i++) dta21[i] = 0.0;
 
   pdmtxcpy(4, 4, 1, 1, dta21, desc_a, dtas21);
 
   pdsyntrd("U", 4, dta21, 1, 1, desc_a, dtd21, dte21, dtt21, work, 0, info);
 
  if (info != 0) {
      if (iam == 0) cout << " INFO from PDSYNTRD = " << info;
      *ifail = *ifail + 1;
  } else {
      pdmtchk(modname, 4, 4, 1, 1, dta21, desc_a, dresa21, tol, ierr, ilerr,
              error, ifail, strlen(modname));
 
      for (i=0; i< nprow; i++) {
          desc_x[RSRC_] = i;
          pdvtchk(modname,4, 1, 1, dtd21, desc_x, dresd21, 1, tol, ierr, ilerr,
                  error, ifail, strlen(modname));
          pdvtchk(modname,4, 1, 1, dte21, desc_x, drese21, 1, tol, ierr, ilerr,
                  error, ifail, strlen(modname));
          pdvtchk(modname,4, 1, 1, dtt21, desc_x, drest21, 1, tol, ierr, ilerr,
                  error, ifail, strlen(modname));
      }
  }
 
// PDSYNTRD  END   : 21
// PDSYNGST  START : 22
 
   double dta22[16], dtb22[16];
 
   double  dtas22[16] =  { -1.,  1., -1.,  1.,
                            0.,  1., -1.,  1.,
                            0.,  0.,  1., -1.,
                            0.,  0.,  0.,  1.};
 
   double dtbs22[16] =  {   2.,  1.,  0.,  0.,
                            0.,  2.,  1.,  0.,
                            0.,  0.,  2.,  1.,
                            0.,  0.,  0.,  2.};
 
   double dresa22[16]= { -0.50000, 0.86603,-1.22474, 1.58114,
                          0.00000,-0.16667,-0.23570, 0.54772,
                          0.00000, 0.00000, 1.16667,-1.93649,
                          0.00000, 0.00000, 0.00000, 3.10000};
 
   desc_a[DT_]   = BLOCK_CYCLIC_2D;
   desc_a[M_]    = 4;
   desc_a[N_]    = 4;
   desc_a[MB_]   = 1;
   desc_a[NB_]   = 1;
   desc_a[RSRC_] = 0;
   desc_a[CSRC_] = 0;
   desc_a[CTXT_] = ictx;
   il_lda = numroc(4,1,myrow,0,nprow);
   desc_a[LLD_]  = MAX(1,il_lda);
 
   desc_b[DT_]   = BLOCK_CYCLIC_2D;
   desc_b[M_]    = 4;
   desc_b[N_]    = 4;
   desc_b[MB_]   = 1;
   desc_b[NB_]   = 1;
   desc_b[RSRC_] = 0;
   desc_b[CSRC_] = 0;
   desc_b[CTXT_] = ictx;
   il_ldb = numroc(4,1,myrow,0,nprow);
   desc_b[LLD_]  = MAX(1,il_ldb);
 
   modname = "PDSYNGST ";
 
   pdmtxcpy(4, 4, 1, 1, dta22, desc_a, dtas22);
   pdmtxcpy(4, 4, 1, 1, dtb22, desc_b, dtbs22);
 
   pdpotrf("L", 4, dtb22, 1, 1, desc_b, iret);
 
   pdsyngst(1,"L", 4, dta22, 1, 1, desc_a, dtb22, 1, 1, desc_b,
            scale, work, 0, info);
 
   if (info != 0) {
       if (iam == 0) cout << " INFO from PDSYNGST = " << info << "\n";
       *ifail = *ifail + 1;
   } else {
       pdmtchk(modname, 4, 4, 1, 1, dta22, desc_a, dresa22, tol,
               ierr, ilerr, error, ifail,strlen(modname));
 
       if ((ABSP(scale) - ABSP(scaler)) > tol) {
           *ifail = *ifail + 1;
           cout << " PDSYNGST scale = " << scale, scaler;
       }
   }
 
// PDSYNGST END :   22
// PDSYEVD  START : 23
 
   double dta23[16], dtz23[16], dtw23[4];
 
   double dtas23[16] = { 5.,  0.,  0.,  0.,
                         4.,  5.,  0.,  0.,
                         1.,  1.,  4.,  0.,
                         1.,  1.,  2.,  4. };
 
   double dresz23[16] = { 0.7071, -0.7071,  0.0,     0.0,
                          0.0,     0.0,     0.7071, -0.7071,
                          0.3162,  0.3162, -0.6325, -0.6325,
                          0.6325,  0.6325,  0.3162,  0.3162  };
 
   double dresw23[4] = { 1.0,  2.0,  5.0,   10.0 };
 
 
   desc_a[DT_]   = BLOCK_CYCLIC_2D;
   desc_a[M_]    = 4;
   desc_a[N_]    = 4;
   desc_a[MB_]   = 1;
   desc_a[NB_]   = 1;
   desc_a[RSRC_] = 0;
   desc_a[CSRC_] = 0;
   desc_a[CTXT_] = ictx;
   il_lda = numroc(4,1,myrow,0,nprow);
   desc_a[LLD_]  = MAX(1,il_lda);
 
   desc_z[DT_]   = BLOCK_CYCLIC_2D;
   desc_z[M_]    = 4;
   desc_z[N_]    = 4;
   desc_z[MB_]   = 1;
   desc_z[NB_]   = 1;
   desc_z[RSRC_] = 0;
   desc_z[CSRC_] = 0;
   desc_z[CTXT_] = ictx;
   il_lda = numroc(4,1,myrow,0,nprow);
   desc_z[LLD_]  = MAX(1,il_lda);
 
   modname = "PDSYEVD ";
 
   pdmtxcpy(4, 4, 1, 1, dta23, desc_a, dtas23);
 
   pdsyevd("V","U", 4, dta23, 1, 1, desc_a, dtw23, dtz23, 1, 1, desc_z,
            work, 0, iwork, 0, info);
 
   if (info != 0) {
       if (iam == 0) cout << " INFO from PDSYEVD = " << info << "\n";
       *ifail = *ifail + 1;
   } else {
 
       pdmtchk(modname, 4, 4, 1, 1, dtz23, desc_z, dresz23, tol, ierr, ilerr,
               error, ifail, strlen(modname));
        for (i=0; i< 4; i++)
            if ((ABSP(dtw23[i]) - ABSP(dresw23[i])) > tol) {
                *ifail = *ifail + 1;
                cout << " PDSYEVD dtw23[" << i << "] = " << dtw23[i];
                cout << "  dresw23[" << i << "] = " << dresw23[i] << "\n";
            }
   }
 
// PDSYEVD   END : 23
 
   if(dtg19)      delete [] dtg19;         // free memory
   if(dtg19res)   delete [] dtg19res;
   if(iclustr)    delete [] iclustr;
   if(iclustres)  delete [] iclustres;
 
   blacs_gridexit(ictx);
 
}
 
/*********************************************************************
*          PARALLEL LINEAR ALGEBRAIC EQUATIONS SUBROUTINES           *
*********************************************************************/
 
void pleq(int *ifail)
{
// PDPOTRF/S  PDPOSV START : 6
 
   int iam, nnodes,myrow,mycol,istor[1],ra[1],ca[1];
   int desc_a[DLEN_], desc_b[DLEN_], il_lda, i, desc_x[DLEN_],INCX;
   int ictx, nprow, npcol,ierr=5, ilerr;
   double error=0.0, tol=0.0,rcond,anorm,rcondd=0.03053,rcondz=0.006173;
   double rwork[100],ntol=0.01;
   double dta7[81], dtb7[45],work[100];
   double rconds2=0.02773;
   double ztol = 0.1;
   int iwork[10],liwork=0,lwork=0;
   _ESVPCOM workz[100];
   int ipvt[9], iret, lrwork=0;
   char modn[9], *modname=modn;
   char order[]="R";
   char scope[]="A";
   char   top[]="1";
   double dta6[81], dtb6[54];
   double dtas6[81] = { 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,
                        0.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,
                        0.,  0.,  3.,  3.,  3.,  3.,  3.,  3.,  3.,
                        0.,  0.,  0.,  4.,  4.,  4.,  4.,  4.,  4.,
                        0.,  0.,  0.,  0.,  5.,  5.,  5.,  5.,  5.,
                        0.,  0.,  0.,  0.,  0.,  6.,  6.,  6.,  6.,
                        0.,  0.,  0.,  0.,  0.,  0.,  7.,  7.,  7.,
                        0.,  0.,  0.,  0.,  0.,  0.,  0.,  8.,  8.,
                        0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  9. };
 
   double dresa6[81] = { 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,
                         0.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,
                         0.,  0.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,
                         0.,  0.,  0.,  1.,  1.,  1.,  1.,  1.,  1.,
                         0.,  0.,  0.,  0.,  1.,  1.,  1.,  1.,  1.,
                         0.,  0.,  0.,  0.,  0.,  1.,  1.,  1.,  1.,
                         0.,  0.,  0.,  0.,  0.,  0.,  1.,  1.,  1.,
                         0.,  0.,  0.,  0.,  0.,  0.,  0.,  1.,  1.,
                         0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  1. };
 
   double dresa61[81] = {
        2., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,
        0.,  2., -1.,  0.,  0.,  0.,  0.,  0.,  0.,
        0.,  0.,  2., -1.,  0.,  0.,  0.,  0.,  0.,
        0.,  0.,  0.,  2., -1.,  0.,  0.,  0.,  0.,
        0.,  0.,  0.,  0.,  2., -1.,  0.,  0.,  0.,
        0.,  0.,  0.,  0.,  0.,  2., -1.,  0.,  0.,
        0.,  0.,  0.,  0.,  0.,  0.,  2., -1.,  0.,
        0.,  0.,  0.,  0.,  0.,  0.,  0.,  2., -1.,
        0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  1. };
 
 
   double dtbs6[54] = { 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,
                       18., 34., 48., 60., 70., 78., 84., 88., 90.,
                       27., 51., 72., 90.,105.,117.,126.,132.,135.,
                       36., 68., 96.,120.,140.,156.,168.,176.,180.,
                       45., 85.,120.,150.,175.,195.,210.,220.,225.,
                        9., 17., 24., 30., 35., 39., 42., 44., 45. };
 
   double dresb6[54] = { 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,
                         2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,
                         3.,  3.,  3.,  3.,  3.,  3.,  3.,  3.,  3.,
                         4.,  4.,  4.,  4.,  4.,  4.,  4.,  4.,  4.,
                         5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,
                         1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1. };
 
// PZPOTRF/S PZPOSV
  _ESVPCOM zta6[81], ztb6[54];
  _ESVPCOM ztas6[81] = {
       _ESVPCOM(18.0,0.0), _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0),
       _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0),
       _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0),   // 1st col end
       _ESVPCOM(0.0, 0.0), _ESVPCOM(18.0,0.0), _ESVPCOM(3.0, 1.0),
       _ESVPCOM(3.0, 1.0), _ESVPCOM(3.0, 1.0), _ESVPCOM(3.0, 1.0),
       _ESVPCOM(3.0, 1.0), _ESVPCOM(3.0, 1.0), _ESVPCOM(3.0, 1.0),   // 2nd col end
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(18.0,0.0),
       _ESVPCOM(5.0, 1.0), _ESVPCOM(5.0, 1.0), _ESVPCOM(5.0, 1.0),
       _ESVPCOM(5.0, 1.0), _ESVPCOM(5.0, 1.0), _ESVPCOM(5.0, 1.0),   // 3rd col end
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(18.0,0.0), _ESVPCOM(7.0, 1.0), _ESVPCOM(7.0, 1.0),
       _ESVPCOM(7.0, 1.0), _ESVPCOM(7.0, 1.0), _ESVPCOM(7.0, 1.0),   // 4th col end
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(18.0,0.0), _ESVPCOM(9.0, 1.0),
       _ESVPCOM(9.0, 1.0), _ESVPCOM(9.0, 1.0), _ESVPCOM(9.0, 1.0),   // 5th col end
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(18.0,0.0),
       _ESVPCOM(11.0,1.0), _ESVPCOM(11.0,1.0), _ESVPCOM(11.0,1.0),   // 6th col end
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(18.0,0.0), _ESVPCOM(13.0,1.0), _ESVPCOM(13.0,1.0),   // 7th col end
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(18.0,0.0), _ESVPCOM(15.0,1.0),   // 8th col end
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(18.0,0.0)};  // 9th col end
 
  _ESVPCOM ztbs6[54] = {
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),  // 1st col end
       _ESVPCOM( 60.0, 10.0),_ESVPCOM( 86.0, 28.0),_ESVPCOM(108.0, 44.0),
       _ESVPCOM(126.0, 58.0),_ESVPCOM(140.0, 70.0),_ESVPCOM(150.0, 80.0),
       _ESVPCOM(156.0, 88.0),_ESVPCOM(158.0, 94.0),_ESVPCOM(156.0, 98.0),  //2nd
       _ESVPCOM( 86.0,  2.0),_ESVPCOM(126.0, 22.0),_ESVPCOM(160.0, 40.0),
       _ESVPCOM(188.0, 56.0),_ESVPCOM(210.0, 70.0),_ESVPCOM(226.0, 82.0),
       _ESVPCOM(236.0, 92.0),_ESVPCOM(240.0,100.0),_ESVPCOM(238.0,106.0),  //3rd
       _ESVPCOM(112.0, -6.0),_ESVPCOM(166.0, 16.0),_ESVPCOM(212.0, 36.0),
       _ESVPCOM(250.0, 54.0),_ESVPCOM(280.0, 70.0),_ESVPCOM(302.0, 84.0),
       _ESVPCOM(316.0, 96.0),_ESVPCOM(322.0,106.0),_ESVPCOM(320.0,114.0),  //4th
       _ESVPCOM(138.0,-14.0),_ESVPCOM(206.0, 10.0),_ESVPCOM(264.0, 32.0),
       _ESVPCOM(312.0, 52.0),_ESVPCOM(350.0, 70.0),_ESVPCOM(378.0, 86.0),
       _ESVPCOM(396.0,100.0),_ESVPCOM(404.0,112.0),_ESVPCOM(402.0,122.0),  //5th
       _ESVPCOM( 34.0, 18.0),_ESVPCOM( 46.0, 34.0),_ESVPCOM( 56.0, 48.0),
       _ESVPCOM( 64.0, 60.0),_ESVPCOM( 70.0, 70.0),_ESVPCOM( 74.0, 78.0),
       _ESVPCOM( 76.0, 84.0),_ESVPCOM( 76.0, 88.0),_ESVPCOM( 74.0, 90.0)}; //6th
 
  _ESVPCOM zresa6[81] = {
       _ESVPCOM( 4.20,  0.0),_ESVPCOM( 0.24, 0.24),_ESVPCOM( 0.24, 0.24),
       _ESVPCOM( 0.24, 0.24),_ESVPCOM( 0.24, 0.24),_ESVPCOM( 0.24, 0.24),
       _ESVPCOM( 0.24, 0.24),_ESVPCOM( 0.24, 0.24),_ESVPCOM( 0.24, 0.24),
       _ESVPCOM( 0.00,  0.0),_ESVPCOM( 4.20, 0.00),_ESVPCOM( 0.68, 0.24),
       _ESVPCOM( 0.68, 0.24),_ESVPCOM( 0.68, 0.24),_ESVPCOM( 0.68, 0.24),
       _ESVPCOM( 0.68, 0.24),_ESVPCOM( 0.68, 0.24),_ESVPCOM( 0.68, 0.24),
       _ESVPCOM( 0.00,  0.0),_ESVPCOM( 0.00, 0.00),_ESVPCOM( 4.20, 0.00),
       _ESVPCOM( 1.10, 0.24),_ESVPCOM( 1.10, 0.24),_ESVPCOM( 1.10, 0.24),
       _ESVPCOM( 1.10, 0.24),_ESVPCOM( 1.10, 0.24),_ESVPCOM( 1.10, 0.24),
       _ESVPCOM( 0.00,  0.0),_ESVPCOM( 0.00, 0.00),_ESVPCOM( 0.00, 0.00),
       _ESVPCOM( 4.00, 0.00),_ESVPCOM( 1.30, 0.25),_ESVPCOM( 1.30, 0.25),
       _ESVPCOM( 1.30, 0.25),_ESVPCOM( 1.30, 0.25),_ESVPCOM( 1.30, 0.25),
       _ESVPCOM( 0.00,  0.0),_ESVPCOM( 0.00, 0.00),_ESVPCOM( 0.00, 0.00),
       _ESVPCOM( 0.00, 0.00),_ESVPCOM( 3.80, 0.00),_ESVPCOM( 1.40, 0.26),
       _ESVPCOM( 1.40, 0.26),_ESVPCOM( 1.40, 0.26),_ESVPCOM( 1.40, 0.26),
       _ESVPCOM( 0.00,  0.0),_ESVPCOM( 0.00, 0.00),_ESVPCOM( 0.00, 0.00),
       _ESVPCOM( 0.00, 0.00),_ESVPCOM( 0.00, 0.00),_ESVPCOM( 3.50, 0.00),
       _ESVPCOM( 1.50, 0.28),_ESVPCOM( 1.50, 0.28),_ESVPCOM( 1.50, 0.28),
       _ESVPCOM( 0.00,  0.0),_ESVPCOM( 0.00, 0.00),_ESVPCOM( 0.00, 0.00),
       _ESVPCOM( 0.00, 0.00),_ESVPCOM( 0.00, 0.00),_ESVPCOM( 0.00, 0.00),
       _ESVPCOM( 3.20, 0.00),_ESVPCOM( 1.60, 0.32),_ESVPCOM( 1.60, 0.32),
       _ESVPCOM( 0.00,  0.0),_ESVPCOM( 0.00, 0.00),_ESVPCOM( 0.00, 0.00),
       _ESVPCOM( 0.00, 0.00),_ESVPCOM( 0.00, 0.00),_ESVPCOM( 0.00, 0.00),
       _ESVPCOM( 0.00, 0.00),_ESVPCOM( 2.70, 0.00),_ESVPCOM( 1.60, 0.37),
       _ESVPCOM( 0.00,  0.0),_ESVPCOM( 0.00, 0.00),_ESVPCOM( 0.00, 0.00),
       _ESVPCOM( 0.00, 0.00),_ESVPCOM( 0.00, 0.00),_ESVPCOM( 0.00, 0.00),
       _ESVPCOM( 0.00, 0.00),_ESVPCOM( 0.00, 0.00),_ESVPCOM( 2.20, 0.00)};
 
 
  _ESVPCOM ztas6b[81] = {
       _ESVPCOM(18.0,1.0), _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0),
       _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0),
       _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0), /* end col 1 */
       _ESVPCOM(0.0, 0.0), _ESVPCOM(18.0,2.0), _ESVPCOM(3.0, 1.0),
       _ESVPCOM(3.0, 1.0), _ESVPCOM(3.0, 1.0), _ESVPCOM(3.0, 1.0),
       _ESVPCOM(3.0, 1.0), _ESVPCOM(3.0, 1.0), _ESVPCOM(3.0, 1.0), /* end col 2 */
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(18.0,3.0),
       _ESVPCOM(5.0, 1.0), _ESVPCOM(5.0, 1.0), _ESVPCOM(5.0, 1.0),
       _ESVPCOM(5.0, 1.0), _ESVPCOM(5.0, 1.0), _ESVPCOM(5.0, 1.0), /* end col 3 */
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(18.0,4.0), _ESVPCOM(7.0, 1.0), _ESVPCOM(7.0, 1.0),
       _ESVPCOM(7.0, 1.0), _ESVPCOM(7.0, 1.0), _ESVPCOM(7.0, 1.0), /* end col 4 */
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(18.0,5.0), _ESVPCOM(9.0, 1.0),
       _ESVPCOM(9.0, 1.0), _ESVPCOM(9.0, 1.0), _ESVPCOM(9.0, 1.0), /* end col 5 */
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(18.0,6.0),
       _ESVPCOM(11.0,1.0), _ESVPCOM(11.0,1.0), _ESVPCOM(11.0,1.0), /* end col 6 */
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(18.0,7.0), _ESVPCOM(13.0,1.0), _ESVPCOM(13.0,1.0), /* end col 7 */
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(18.0,8.0), _ESVPCOM(15.0,1.0), /* end col 8 */
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(18.0,9.0)  /* end col 9 */
  };
 
 
 
 
  _ESVPCOM zresa61[81] = {
       _ESVPCOM( 0.06, 0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.06, 0.0), _ESVPCOM(-0.01, 0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.07, 0.0),
       _ESVPCOM(-0.01, 0.0), _ESVPCOM(-0.01, 0.0), _ESVPCOM(-0.01, 0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.07, 0.0), _ESVPCOM(-0.01,-0.01), _ESVPCOM(-0.01, 0.0),
       _ESVPCOM(-0.01, 0.0), _ESVPCOM(-0.01, 0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.09, 0.0), _ESVPCOM(-0.02,-0.01),
       _ESVPCOM(-0.01, 0.0), _ESVPCOM(-0.01, 0.0), _ESVPCOM(-0.01, 0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.11, 0.0),
       _ESVPCOM(-0.02,-0.01), _ESVPCOM(-0.02,-0.01), _ESVPCOM(-0.02, 0.01),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.15, 0.0), _ESVPCOM(-0.04,-0.02), _ESVPCOM(-0.05, 0.01),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.22, 0.0), _ESVPCOM(-0.13,-0.03),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0),
       _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.0,  0.0), _ESVPCOM( 0.22, 0.0)
 };
 
  _ESVPCOM zresb6[54] = {
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),  // 1st col end
       _ESVPCOM(2.0, 1.0), _ESVPCOM(2.0, 1.0), _ESVPCOM(2.0, 1.0),
       _ESVPCOM(2.0, 1.0), _ESVPCOM(2.0, 1.0), _ESVPCOM(2.0, 1.0),
       _ESVPCOM(2.0, 1.0), _ESVPCOM(2.0, 1.0), _ESVPCOM(2.0, 1.0),  // 2nd col end
       _ESVPCOM(3.0, 1.0), _ESVPCOM(3.0, 1.0), _ESVPCOM(3.0, 1.0),
       _ESVPCOM(3.0, 1.0), _ESVPCOM(3.0, 1.0), _ESVPCOM(3.0, 1.0),
       _ESVPCOM(3.0, 1.0), _ESVPCOM(3.0, 1.0), _ESVPCOM(3.0, 1.0),  // 3rd col end
       _ESVPCOM(4.0, 1.0), _ESVPCOM(4.0, 1.0), _ESVPCOM(4.0, 1.0),
       _ESVPCOM(4.0, 1.0), _ESVPCOM(4.0, 1.0), _ESVPCOM(4.0, 1.0),
       _ESVPCOM(4.0, 1.0), _ESVPCOM(4.0, 1.0), _ESVPCOM(4.0, 1.0),  // 4th col end
       _ESVPCOM(5.0, 1.0), _ESVPCOM(5.0, 1.0), _ESVPCOM(5.0, 1.0),
       _ESVPCOM(5.0, 1.0), _ESVPCOM(5.0, 1.0), _ESVPCOM(5.0, 1.0),
       _ESVPCOM(5.0, 1.0), _ESVPCOM(5.0, 1.0), _ESVPCOM(5.0, 1.0),  // 5th col end
       _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0),
       _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0),
       _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0), _ESVPCOM(1.0, 1.0)}; // 6th col end
 
// PDPOTRF/S  PDPOTRI  PDLANSY         PDPOCON PDPOSV START : 6
// PZPOTRF/S  PZPOTRI  PZLANSY PZLANHE PZPOCON PZPOSV START : 6
 
  blacs_pinfo(iam,nnodes);
  nprow = 1;
  npcol = nnodes;
  blacs_get(0,0,ictx);
  blacs_gridinit(ictx,order,nprow,npcol,strlen(order));
  blacs_gridinfo(ictx,nprow,npcol,myrow,mycol);
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 9;
  desc_a[N_]    = 9;
  desc_a[MB_]   = 3;
  desc_a[NB_]   = 3;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(9,3,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_b[DT_]   = BLOCK_CYCLIC_2D;
  desc_b[M_]    = 9;
  desc_b[N_]    = 6;
  desc_b[MB_]   = 3;
  desc_b[NB_]   = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(9,3,myrow,0,nprow);
  desc_b[LLD_]  = MAX(1,il_lda);
 
  modname = "PDPOTRF ";
  pdmtxcpy(9, 9, 1, 1, dta6, desc_a, dtas6);
  pdpotrf("L", 9, dta6, 1, 1, desc_a, iret);
  if (iret != 0)
  {
   if (iam == 0)
    cout << " IRET from PDPOTRF = " << iret;
    *ifail = *ifail + 1;
  }
  else
  {
  pdmtchk(modname,9, 9, 1, 1, dta6, desc_a, dresa6, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  modname = "PDPOTRS ";
  pdmtxcpy(9, 5, 1, 2, dtb6, desc_b, dtbs6);
  pdpotrs("L", 9, 5, dta6, 1, 1, desc_a, dtb6, 1, 2, desc_b, iret);
  if (iret != 0)
  {
   if (iam == 0)
    cout << " IRET from PDPOTRS = " << iret;
    *ifail = *ifail + 1;
  }
  pdmtchk(modname,9, 5, 1, 2, dtb6, desc_b, dresb6, tol, ierr, ilerr,
          error, ifail, strlen(modname));
  }
 
  modname = "PDPOTRI ";
  tol = 0.001;
  pdmtxcpy(9, 9, 1, 1, dta6, desc_a, dtas6);
  pdpotrf("L", 9, dta6, 1, 1, desc_a,iret);
  pdpotri("L", 9, dta6, 1, 1, desc_a,iret);
  if (iret != 0) {
   if (iam == 0) cout << " IRET from PDPOTRI = " << iret << endl;
   *ifail = *ifail + 1;
  } else {
   pdmtchk(modname,9, 9, 1, 1, dta6, desc_a, dresa61, tol,
           ierr,ilerr, error, ifail,strlen(modname));
  }
 
  modname = "PDPOCON ";
  pdmtxcpy(9, 9, 1, 1, dta6, desc_a, dtas6);
  anorm=pdlansy("I","L",9,dta6,1,1,desc_a, work);
  if ( ABSP(anorm-45.00) > ntol) {
   if (iam == 0) cout << " ANORM from PDLANSY  = " << anorm << endl;
   *ifail = *ifail + 1;
  }
  pdpotrf("L", 9, dta6, 1, 1, desc_a,iret);
  pdpocon("L",9,dta6,1,1,desc_a,anorm,rcond,work,0,iwork,0,iret);
  if (iret != 0) {
   if (iam == 0) cout << " IRET from PDPOCON = " << iret << endl;
   *ifail = *ifail + 1;
  } else {
   if(ABSP(rcond-rcondz) > tol) {
    if (iam == 0) cout << " RCOND from PDPOCON = "<<rcond<<" "<<rcondz<<endl;
    *ifail = *ifail + 1;
   }
  }
 
  modname = "PDPOSV  ";
  pdmtxcpy(9, 9, 1, 1, dta6, desc_a, dtas6);
  pdmtxcpy(9, 5, 1, 2, dtb6, desc_b, dtbs6);
  pdposv("L", 9, 5, dta6, 1, 1, desc_a, dtb6, 1, 2, desc_b, iret);
  if (iret != 0)
  {
   if (iam == 0)
    cout << " IRET from PDPOSV  = " << iret;
    *ifail = *ifail + 1;
  }
  pdmtchk(modname,9, 9, 1, 1, dta6, desc_a, dresa6, tol, ierr, ilerr,
          error, ifail, strlen(modname));
  pdmtchk(modname,9, 5, 1, 2, dtb6, desc_b, dresb6, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  tol = 0.1;
  modname = "PZPOTRF ";
  pzmtxcpy(9, 9, 1, 1, zta6, desc_a, ztas6);
  pzpotrf("L", 9, zta6, 1, 1, desc_a, iret);
  if (iret != 0)
  {
   if (iam == 0)
    cout << " IRET from PZPOTRF = " << iret;
    *ifail = *ifail + 1;
  }
  else
  {
  pzmtchk(modname,9, 9, 1, 1, zta6, desc_a, zresa6, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  modname = "PZPOTRS ";
  pzmtxcpy(9, 5, 1, 2, ztb6, desc_b, ztbs6);
  pzpotrs("L", 9, 5, zta6, 1, 1, desc_a, ztb6, 1, 2, desc_b, iret);
  if (iret != 0)
  {
   if (iam == 0)
    cout << " IRET from PZPOTRS = " << iret;
    *ifail = *ifail + 1;
  }
  pzmtchk(modname,9, 5, 1, 2, ztb6, desc_b, zresb6, tol, ierr, ilerr,
          error, ifail, strlen(modname));
  }
 
  modname = "PZPOTRI ";
  pzmtxcpy(9, 9, 1, 1, zta6, desc_a, ztas6);
  pzpotrf("L", 9, zta6, 1, 1, desc_a,iret);
 
  pzpotri("L", 9, zta6, 1, 1, desc_a,iret);
 
  if (iret != 0) {
   if (iam == 0) cout << " IRET from PZPOTRI = " << iret << endl;
   *ifail = *ifail + 1;
  } else {
   pzmtchk(modname,9, 9, 1, 1, zta6, desc_a, zresa61, ztol,
           ierr,ilerr, error, ifail,strlen(modname));
  }
 
  modname = "PZPOCON ";
  ntol = 0.1;
  pzmtxcpy(9, 9, 1, 1, zta6, desc_a, ztas6);
  anorm=pzlanhe("1","L",9,zta6,1,1,desc_a, work);
  if ( ABSP(anorm-82.92) > ntol) {
    if (iam == 0) cout << " ANORM from PZLANHE  = " << anorm << endl;
    *ifail = *ifail + 1;
  }
  pzpotrf("L", 9, zta6, 1, 1, desc_a,iret);
  pzpocon("L",9,zta6,1,1,desc_a,anorm,rcond,workz,0,work,0,iret);
  if (iret != 0) {
   if (iam == 0) cout << " IRET from PZPOCON = " << iret << endl;
   *ifail = *ifail + 1;
  } else {
   if(ABSP(rcond-rconds2) > tol) {
    if (iam == 0) cout << " RCOND from PZPOCON = "<<rcond<< " "<<rconds2<<endl;
    *ifail = *ifail + 1;
   }
  }
 
  modname = "PZPOSV  ";
  pzmtxcpy(9, 9, 1, 1, zta6, desc_a, ztas6);
  pzmtxcpy(9, 5, 1, 2, ztb6, desc_b, ztbs6);
  pzposv("L", 9, 5, zta6, 1, 1, desc_a, ztb6, 1, 2, desc_b, iret);
  if (iret != 0)
  {
   if (iam == 0)
    cout << " IRET from PZPOSV  = " << iret;
    *ifail = *ifail + 1;
  }
  pzmtchk(modname,9, 9, 1, 1, zta6, desc_a, zresa6, tol, ierr, ilerr,
          error, ifail, strlen(modname));
  pzmtchk(modname,9, 5, 1, 2, ztb6, desc_b, zresb6, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  modname = "PZLANSY ";
  ntol = 0.01;
  pzmtxcpy(9, 9, 1, 1, zta6, desc_a, ztas6b);
  anorm=pzlansy("F","L",9,zta6,1,1,desc_a, work);
  if ( ABSP(anorm-81.88) > ntol) {
    if (iam == 0) cout << " ANORM from PZLANSY  = " << anorm << endl;
    *ifail = *ifail + 1;
  }
 
// PDPOTRF/S  PDPOTRI  PDLANSY         PDPOCON PDPOSV END : 6
// PZPOTRF/S  PZPOTRI  PZLANSY PZLANHE PZPOCON PZPOSV END : 6
 
// PDGETRF/S & PZGETRF/S PDGESV & PZGESV START : 7
// PDGETRI & PZGETRI PDGECON & PZGECON PDLANGE & PZLANGE START : 7
 
  double dtas7[81] =  { 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6,
                        1.2, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4,
                        1.4, 1.2, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2,
                        1.6, 1.4, 1.2, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0,
                        1.8, 1.6, 1.4, 1.2, 1.0, 1.2, 1.4, 1.6, 1.8,
                        2.0, 1.8, 1.6, 1.4, 1.2, 1.0, 1.2, 1.4, 1.6,
                        2.2, 2.0, 1.8, 1.6, 1.4, 1.2, 1.0, 1.2, 1.4,
                        2.4, 2.2, 2.0, 1.8, 1.6, 1.4, 1.2, 1.0, 1.2,
                        2.6, 2.4, 2.2, 2.0, 1.8, 1.6, 1.4, 1.2, 1.0 };
 
  double dresa7[81] = {
   2.6000,   .3846,   .4615,   .5385,  .6154,  .6923,  .7692, .8462, .9231,
   2.4000,   .2769,  -.3889,  -.3333, -.2778, -.2222, -.1667,-.1111,-.0556,
   2.2000,   .5538,   .4000,   .0000,  .0000,  .0000,  .0000, .0000, .0000,
   2.0000,   .8308,   .8000,   .4000,  .0000,  .0000,  .0000, .0000, .0000,
   1.8000,  1.1077,  1.2000,   .8000,  .4000,  .0000,  .0000, .0000, .0000,
   1.6000,  1.3846,  1.6000,  1.2000,  .8000,  .4000,  .0000, .0000, .0000,
   1.4000,  1.6615,  2.0000,  1.6000, 1.2000,  .8000,  .4000, .0000, .0000,
   1.2000,  1.9385,  2.4000,  2.0000, 1.6000, 1.2000,  .8000, .4000, .0000,
   1.0000,  2.2154,  2.8000,  2.4000, 2.0000, 1.6000, 1.2000, .8000, .4000 };
 
    double   dresa71[81] = {
            -2.3611,  2.5000,   .0000,   .0000,  .0000,
              .0000,  .0000, .0000, .1389,
             2.5000, -5.0000,  2.5000,   .0000,  .0000,
              .0000,  .0000, .0000, .0000,
             0.0000,  2.5000, -5.0000,  2.5000,  .0000,
              .0000,  .0000, .0000, .0000,
             0.0000,   .0000,  2.5000, -5.0000, 2.5000,
              .0000,  .0000, .0000, .0000,
             0.0000,  0.0000,  0.0000,  2.5000,-5.0000,
             2.5000,  .0000, .0000, .0000,
             0.0000,  0.0000,  0.0000,  0.0000, 2.5000,
            -5.0000, 2.5000, .0000, .0000,
             0.0000,  0.0000,  0.0000,  0.0000, 0.0000,
             2.5000,-5.0000,2.5000, .0000,
             0.0000,  0.0000,  0.0000,  0.0000, 0.0000,
             0.0000, 2.5000,-5.0000,2.5000,
             0.1389,  0.0000,  0.0000,  0.0000, 0.0000,
             0.0000, 0.0000,2.5000,-2.3611 };
 
  double  dtbs7[45]= {
            93.,     84.4000, 76.6000, 70.0000, 65.0000,
            62.0000, 61.4000, 63.60,   69.0000,
           186.,    168.8000,153.2000, 140.0000,130.0000,
           124.000, 122.8000,127.2000, 138.0000,
           279.000, 253.2000,229.8000, 210.0000,195.0000,
           186.0000,184.2000,190.8000, 207.0000,
           372.000, 337.6000,306.4000, 280.0000,260.0000,
           248.0000,245.6000,254.4000, 276.0000,
           465.000, 422.0000,383.0000, 350.0000,325.0000,
           310.0000,307.0000,318.0000, 345.0000};
 
  double  dresb7[45]= {
          1.0000,  2.0000,  3.0000,  4.0000,  5.0000,
          6.0000, 7.0000, 8.0000,  9.0000,
          2.0000,  4.0000,  6.0000,  8.0000, 10.0000,
         12.0000,14.0000,16.0000, 18.0000,
          3.0000,  6.0000,  9.0000, 12.0000, 15.0000,
         18.0000,21.0000,24.0000, 27.0000,
          4.0000,  8.0000, 12.0000, 16.0000, 20.0000,
         24.0000,28.0000,32.0000, 36.0000,
          5.0000, 10.0000, 15.0000, 20.0000, 25.0000,
         30.0000,35.0000,40.0000, 45.0000};
 
  _ESVPCOM zta7[81], ztb7[54];
  _ESVPCOM ztas7[81] = {
       _ESVPCOM( 2.0,1.0), _ESVPCOM(2.4, 1.0), _ESVPCOM(2.8, 1.0),
       _ESVPCOM(3.2, 1.0), _ESVPCOM(3.6, 1.0), _ESVPCOM(4.0, 1.0),
       _ESVPCOM(4.4, 1.0), _ESVPCOM(4.8, 1.0), _ESVPCOM(5.2, 1.0),  // 1st col end
       _ESVPCOM(2.4,-1.0), _ESVPCOM(2.0, 1.0), _ESVPCOM(2.4, 1.0),
       _ESVPCOM(2.8, 1.0), _ESVPCOM(3.2, 1.0), _ESVPCOM(3.6, 1.0),
       _ESVPCOM(4.0, 1.0), _ESVPCOM(4.4, 1.0), _ESVPCOM(4.8, 1.0),  // 2nd col end
       _ESVPCOM(2.8,-1.0), _ESVPCOM(2.4,-1.0), _ESVPCOM(2.0, 1.0),
       _ESVPCOM(2.4, 1.0), _ESVPCOM(2.8, 1.0), _ESVPCOM(3.2, 1.0),
       _ESVPCOM(3.6, 1.0), _ESVPCOM(4.0, 1.0), _ESVPCOM(4.4, 1.0),  // 3rd col end
       _ESVPCOM(3.2,-1.0), _ESVPCOM(2.8,-1.0), _ESVPCOM(2.4,-1.0),
       _ESVPCOM(2.0, 1.0), _ESVPCOM(2.4, 1.0), _ESVPCOM(2.8, 1.0),
       _ESVPCOM(3.2, 1.0), _ESVPCOM(3.6, 1.0), _ESVPCOM(4.0, 1.0),  // 4th col end
       _ESVPCOM(3.6,-1.0), _ESVPCOM(3.2,-1.0), _ESVPCOM(2.8,-1.0),
       _ESVPCOM(2.4,-1.0), _ESVPCOM(2.0, 1.0), _ESVPCOM(2.4, 1.0),
       _ESVPCOM(2.8, 1.0), _ESVPCOM(3.2, 1.0), _ESVPCOM(3.6, 1.0),  // 5th col end
       _ESVPCOM(4.0,-1.0), _ESVPCOM(3.6,-1.0), _ESVPCOM(3.2,-1.0),
       _ESVPCOM(2.8,-1.0), _ESVPCOM(2.4,-1.0), _ESVPCOM(2.0, 1.0),
       _ESVPCOM(2.4, 1.0), _ESVPCOM(2.8, 1.0), _ESVPCOM(3.2, 1.0),  // 6th col end
       _ESVPCOM(4.4,-1.0), _ESVPCOM(4.0,-1.0), _ESVPCOM(3.6,-1.0),
       _ESVPCOM(3.2,-1.0), _ESVPCOM(2.8,-1.0), _ESVPCOM(2.4,-1.0),
       _ESVPCOM(2.0, 1.0), _ESVPCOM(2.4, 1.0), _ESVPCOM(2.8, 1.0),  // 7th col end
       _ESVPCOM(4.8,-1.0), _ESVPCOM(4.4,-1.0), _ESVPCOM(4.0,-1.0),
       _ESVPCOM(3.6,-1.0), _ESVPCOM(3.2,-1.0), _ESVPCOM(2.8,-1.0),
       _ESVPCOM(2.4,-1.0), _ESVPCOM(2.0, 1.0), _ESVPCOM(2.4, 1.0),  // 8th col end
       _ESVPCOM(5.2,-1.0), _ESVPCOM(4.8,-1.0), _ESVPCOM(4.4,-1.0),
       _ESVPCOM(4.0,-1.0), _ESVPCOM(3.6,-1.0), _ESVPCOM(3.2,-1.0),
       _ESVPCOM(2.8,-1.0), _ESVPCOM(2.4,-1.0), _ESVPCOM(2.0, 1.0)}; // 9th col end
 
  _ESVPCOM zresa7[81] = {
       _ESVPCOM(5.20,1.0), _ESVPCOM(0.4, 0.1), _ESVPCOM(0.5, 0.1),
       _ESVPCOM(0.6, 0.1), _ESVPCOM(0.6, 0.1), _ESVPCOM(0.7, 0.1),
       _ESVPCOM(0.8, 0.0), _ESVPCOM(0.9, 0.0), _ESVPCOM(0.9, 0.0),  // 1st col end
       _ESVPCOM(4.80,1.0), _ESVPCOM(0.6,-2.0), _ESVPCOM(0.0,-0.1),
       _ESVPCOM(0.0,-0.1), _ESVPCOM(0.0,-0.1), _ESVPCOM(0.0,-0.1),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),  // 2nd col end
       _ESVPCOM(4.40,1.0), _ESVPCOM(1.1,-1.9), _ESVPCOM(0.6,-1.9),
       _ESVPCOM(-0.1,-0.1),_ESVPCOM(-0.1,-0.1), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),  // 3rd col end
       _ESVPCOM(4.00,1.0), _ESVPCOM(1.7,-1.9), _ESVPCOM(1.2,-1.8),
       _ESVPCOM( 0.7,-1.9),_ESVPCOM(-0.1, 0.0),_ESVPCOM(-0.1, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),  // 4th col end
       _ESVPCOM(3.60,1.0), _ESVPCOM(2.3,-1.8), _ESVPCOM(1.8,-1.7),
       _ESVPCOM( 1.3,-1.7),_ESVPCOM( 0.7,-1.9),_ESVPCOM(-0.1, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),  // 5th col end
       _ESVPCOM(3.20,1.0), _ESVPCOM(2.8,-1.8), _ESVPCOM(2.5,-1.6),
       _ESVPCOM( 2.0,-1.6),_ESVPCOM( 1.5,-1.7),_ESVPCOM( 0.8,-1.9),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),  // 6th col end
       _ESVPCOM(2.80,1.0), _ESVPCOM(3.4,-1.7), _ESVPCOM(3.1,-1.5),
       _ESVPCOM( 2.7,-1.5),_ESVPCOM( 2.2,-1.6),_ESVPCOM( 1.6,-1.8),
       _ESVPCOM(0.8,-1.9), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),  // 7th col end
       _ESVPCOM(2.40,1.0), _ESVPCOM(3.9,-1.7), _ESVPCOM(3.7,-1.4),
       _ESVPCOM( 3.4,-1.4),_ESVPCOM( 2.9,-1.5),_ESVPCOM( 2.4,-1.6),
       _ESVPCOM(1.7,-1.8), _ESVPCOM(0.8,-2.0), _ESVPCOM(0.0, 0.0),  // 8th col end
       _ESVPCOM(2.00,1.0), _ESVPCOM(4.5,-1.6), _ESVPCOM(4.3,-1.3),
       _ESVPCOM( 4.0,-1.2),_ESVPCOM( 3.7,-1.3),_ESVPCOM( 3.2,-1.5),
       _ESVPCOM(2.5,-1.8), _ESVPCOM(1.7,-1.9), _ESVPCOM(0.8,-2.0)}; // 9th col end
 
  _ESVPCOM zresa71[81] = {
     _ESVPCOM(-.17,-.42),_ESVPCOM(0.18, 0.43),_ESVPCOM(0.01, 0.00),
     _ESVPCOM(0.01, 0.00),_ESVPCOM(0.00, 0.01),_ESVPCOM(0.00, 0.01),
     _ESVPCOM(0.00, 0.01),_ESVPCOM(0.00, 0.01),_ESVPCOM(-.01, 0.01),
     _ESVPCOM(-0.12,0.13),_ESVPCOM(-0.04,-0.55),_ESVPCOM(0.18, 0.43),
     _ESVPCOM(0.01, 0.00),_ESVPCOM(0.01, 0.00),_ESVPCOM(0.00, 0.01),
     _ESVPCOM(0.00, 0.01),_ESVPCOM(0.00, 0.01),_ESVPCOM(0.00, 0.01),
     _ESVPCOM(-0.06,0.15),_ESVPCOM(-0.06,-0.03),_ESVPCOM(-0.04,-0.55),
     _ESVPCOM( 0.18, 0.43),_ESVPCOM( 0.01, 0.00),_ESVPCOM(0.01, 0.00),
     _ESVPCOM(0.00, 0.01),_ESVPCOM(0.00, 0.01),_ESVPCOM(0.00, 0.01),
     _ESVPCOM(0.00, 0.15),_ESVPCOM(-0.06,-0.01),_ESVPCOM(-0.06,-0.03),
     _ESVPCOM(-0.04,-0.55),_ESVPCOM( 0.18, 0.43),_ESVPCOM( 0.01, 0.00),
     _ESVPCOM(0.01, 0.00),_ESVPCOM( 0.00, 0.01),_ESVPCOM(0.00, 0.01),
     _ESVPCOM(0.05,0.13),_ESVPCOM(-0.05, 0.01),_ESVPCOM(-0.06,-0.01),
     _ESVPCOM(-0.06,-0.03),_ESVPCOM(-0.04,-0.55),_ESVPCOM( 0.18, 0.43),
     _ESVPCOM(0.01, 0.00),_ESVPCOM(0.01, 0.00),_ESVPCOM(0.00, 0.01),
     _ESVPCOM(0.09,0.09),_ESVPCOM(-0.04, 0.03),_ESVPCOM(-0.05, 0.01),
     _ESVPCOM(-0.06,-0.01),_ESVPCOM(-0.06,-0.03),_ESVPCOM(-0.04,-0.55),
     _ESVPCOM(0.18, 0.43),_ESVPCOM(0.01, 0.00),_ESVPCOM(0.01, 0.00),
     _ESVPCOM(0.11,0.05),_ESVPCOM(-0.03, 0.04),_ESVPCOM(-0.04, 0.03),
     _ESVPCOM(-0.05, 0.01),_ESVPCOM(-0.06,-0.01),_ESVPCOM(-0.06,-0.03),
     _ESVPCOM(-0.04,-0.55),_ESVPCOM(0.18, 0.43),_ESVPCOM(0.01, 0.00),
     _ESVPCOM(0.11,0.00),_ESVPCOM(-0.01, 0.04),_ESVPCOM(-0.03, 0.04),
     _ESVPCOM(-0.04, 0.03),_ESVPCOM(-0.05, 0.01),_ESVPCOM(-0.06,-0.01),
     _ESVPCOM(-0.06,-0.03),_ESVPCOM(-0.04,-0.55),_ESVPCOM(0.18, 0.43),
     _ESVPCOM(0.04,-0.28),_ESVPCOM(0.11, 0.00),_ESVPCOM(0.11, 0.05),
     _ESVPCOM( 0.09, 0.09),_ESVPCOM( 0.05, 0.13),_ESVPCOM( 0.00, 0.15),
     _ESVPCOM(-.06, 0.15),_ESVPCOM(-0.12, 0.13),_ESVPCOM(-0.17,-0.42)};
 
  _ESVPCOM ztbs7[54] = {
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),  // 1st col end
       _ESVPCOM(193.0,-10.60), _ESVPCOM(173.8,-9.40), _ESVPCOM(156.2, -5.40),
       _ESVPCOM(141.0,  1.40), _ESVPCOM(129.0,11.00), _ESVPCOM(121.0, 23.40),
       _ESVPCOM(117.8, 38.60), _ESVPCOM(120.2,56.60), _ESVPCOM(129.0, 77.40),
       _ESVPCOM(200.0, 21.80), _ESVPCOM(178.8,20.20), _ESVPCOM(159.2, 22.20),
       _ESVPCOM(142.0, 27.80), _ESVPCOM(128.0,37.00), _ESVPCOM(118.0, 49.80),
       _ESVPCOM(112.8, 66.20), _ESVPCOM(113.2,86.20), _ESVPCOM(120.0,109.80),
       _ESVPCOM(207.0, 54.20), _ESVPCOM(183.8,49.80), _ESVPCOM(162.2, 49.80),
       _ESVPCOM(143.0, 54.20), _ESVPCOM(127.0,63.00), _ESVPCOM(115.0, 76.20),
       _ESVPCOM(107.8, 93.80), _ESVPCOM(106.2,115.8), _ESVPCOM(111.0,142.20),
       _ESVPCOM(214.0, 86.60), _ESVPCOM(188.8,79.40), _ESVPCOM(165.2, 77.40),
       _ESVPCOM(144.0, 80.60), _ESVPCOM(126.0,89.00), _ESVPCOM(112.0,102.60),
       _ESVPCOM(102.8,121.40), _ESVPCOM( 99.2,145.4), _ESVPCOM(102.0,174.60),
       _ESVPCOM(221.0,119.00), _ESVPCOM(193.8,109.0), _ESVPCOM(168.2,105.00),
       _ESVPCOM(145.0,107.00), _ESVPCOM(125.0,115.0), _ESVPCOM(109.0,129.00),
       _ESVPCOM( 97.8,149.00), _ESVPCOM( 92.2,175.0), _ESVPCOM( 93.0,207.00)};
 
 
  _ESVPCOM zresb7[54] = {
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),
       _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0), _ESVPCOM(0.0, 0.0),  // 1st col end
       _ESVPCOM(1.0, 1.0), _ESVPCOM(2.0, 1.0), _ESVPCOM(3.0, 1.0),
       _ESVPCOM(4.0, 1.0), _ESVPCOM(5.0, 1.0), _ESVPCOM(6.0, 1.0),
       _ESVPCOM(7.0, 1.0), _ESVPCOM(8.0, 1.0), _ESVPCOM(9.0, 1.0),  // 2nd col end
       _ESVPCOM(1.0, 2.0), _ESVPCOM(2.0, 2.0), _ESVPCOM(3.0, 2.0),
       _ESVPCOM(4.0, 2.0), _ESVPCOM(5.0, 2.0), _ESVPCOM(6.0, 2.0),
       _ESVPCOM(7.0, 2.0), _ESVPCOM(8.0, 2.0), _ESVPCOM(9.0, 2.0),  // 3rd col end
       _ESVPCOM(1.0, 3.0), _ESVPCOM(2.0, 3.0), _ESVPCOM(3.0, 3.0),
       _ESVPCOM(4.0, 3.0), _ESVPCOM(5.0, 3.0), _ESVPCOM(6.0, 3.0),
       _ESVPCOM(7.0, 3.0), _ESVPCOM(8.0, 3.0), _ESVPCOM(9.0, 3.0),  // 4th col end
       _ESVPCOM(1.0, 4.0), _ESVPCOM(2.0, 4.0), _ESVPCOM(3.0, 4.0),
       _ESVPCOM(4.0, 4.0), _ESVPCOM(5.0, 4.0), _ESVPCOM(6.0, 4.0),
       _ESVPCOM(7.0, 4.0), _ESVPCOM(8.0, 4.0), _ESVPCOM(9.0, 4.0),  // 5th col end
       _ESVPCOM(1.0, 5.0), _ESVPCOM(2.0, 5.0), _ESVPCOM(3.0, 5.0),
       _ESVPCOM(4.0, 5.0), _ESVPCOM(5.0, 5.0), _ESVPCOM(6.0, 5.0),
       _ESVPCOM(7.0, 5.0), _ESVPCOM(8.0, 5.0), _ESVPCOM(9.0, 5.0)}; // 6th col end
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 9;
  desc_a[N_]    = 9;
  desc_a[MB_]   = 3;
  desc_a[NB_]   = 3;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(9,3,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_b[DT_]   = BLOCK_CYCLIC_2D;
  desc_b[M_]    = 9;
  desc_b[N_]    = 5;
  desc_b[MB_]   = 3;
  desc_b[NB_]   = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(9,3,myrow,0,nprow);
  desc_b[LLD_]  = MAX(1,il_lda);
 
  modname = "PDGETRF ";
  tol = 0.001;
  pdmtxcpy(9, 9, 1, 1, dta7, desc_a, dtas7);
  pdgetrf(9, 9, dta7, 1, 1, desc_a, ipvt, iret);
  if (iret != 0)
  {
   if (iam == 0)
    cout << " IRET from PDGETRF = " << iret;
    *ifail = *ifail + 1;
  }
  else
  {
  pdmtchk(modname,9, 9, 1, 1, dta7, desc_a, dresa7, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  modname = "PDGETRS ";
  pdmtxcpy(9, 5, 1, 1, dtb7, desc_b, dtbs7);
  pdgetrs("N", 9, 5, dta7, 1, 1, desc_a, ipvt, dtb7, 1, 1, desc_b, iret);
  if (iret != 0)
  {
   if (iam == 0)
    cout << " IRET from PDGETRS = " << iret;
    *ifail = *ifail + 1;
  }
  pdmtchk(modname,9, 5, 1, 1, dtb7, desc_b, dresb7, tol, ierr, ilerr,
          error, ifail, strlen(modname));
  }
 
  modname = "PDGESV  ";
  pdmtxcpy(9, 9, 1, 1, dta7, desc_a, dtas7);
  pdmtxcpy(9, 5, 1, 1, dtb7, desc_b, dtbs7);
  pdgesv(9, 5, dta7, 1, 1, desc_a, ipvt, dtb7, 1, 1, desc_b, iret)      ;
  if (iret != 0)
  {
   if (iam == 0)
    cout << " IRET from PDGESV  = " << iret;
    *ifail = *ifail + 1;
  }
  pdmtchk(modname,9, 9, 1, 1, dta7, desc_a, dresa7, tol, ierr, ilerr,
          error, ifail, strlen(modname));
  pdmtchk(modname,9, 5, 1, 1, dtb7, desc_b, dresb7, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  modname = "PDGETRI ";
  tol = 0.001;
  pdmtxcpy(9, 9, 1, 1, dta7, desc_a, dtas7);
  pdgetrf(9, 9, dta7, 1, 1, desc_a, ipvt, iret);
  pdgetri(9, dta7, 1, 1, desc_a, ipvt, work,lwork,iwork,liwork, iret);
  if (iret != 0)
  {
   if (iam == 0)
    cout << " IRET from PDGETRI = " << iret;
    *ifail = *ifail + 1;
  }
  pdmtchk(modname,9, 9, 1, 1, dta7, desc_a, dresa71, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  modname = "PDGECON ";
  tol = 0.001;
  pdmtxcpy(9, 9, 1, 1, dta7, desc_a, dtas7);
  anorm = pdlange("O", 9, 9, dta7, 1, 1, desc_a, work);
  pdgetrf(9, 9, dta7, 1, 1, desc_a, ipvt, iret);
  pdgecon("O", 9, dta7, 1, 1, desc_a, anorm, rcond, work, lwork,
          iwork, liwork, iret);
  if (iret != 0)
   {
   if(iam == 0)
    cout << " IRET from PDGECON = " << iret;
   *ifail = *ifail + 1;
  }
 
       if ((ABSP(rcond) - ABSP(rcondd)) > tol)
        {
          *ifail = *ifail + 1;
    cout << " RCOND from PDGECON = " << rcond, rcondd;
        }
 
  modname = "PZGETRF ";
  tol = 0.1;
  pzmtxcpy(9, 9, 1, 1, zta7, desc_a, ztas7);
  pzgetrf(9, 9, zta7, 1, 1, desc_a, ipvt, iret);
  if (iret != 0)
  {
   if (iam == 0)
    cout << " IRET from PZGETRF = " << iret;
    *ifail = *ifail + 1;
  }
  else
  {
    pzmtchk(modname,9, 9, 1, 1, zta7, desc_a, zresa7, tol, ierr, ilerr,
            error, ifail, strlen(modname));
 
  desc_b[DT_]   = BLOCK_CYCLIC_2D;
  desc_b[M_]    = 9;
  desc_b[N_]    = 6;
  desc_b[MB_]   = 3;
  desc_b[NB_]   = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(9,3,myrow,0,nprow);
  desc_b[LLD_]  = MAX(1,il_lda);
 
  modname = "PZGETRS ";
  pzmtxcpy(9, 6, 1, 2, ztb7, desc_b, ztbs7);
  pzgetrs("N", 9, 5, zta7, 1, 1, desc_a, ipvt, ztb7, 1, 2, desc_b, iret);
  if (iret != 0)
  {
   if (iam == 0)
    cout << " IRET from PZGETRS = " << iret;
    *ifail = *ifail + 1;
  }
  pzmtchk(modname,9, 6, 1, 2, ztb7, desc_b, zresb7, tol, ierr, ilerr,
          error, ifail, strlen(modname));
  }
 
  modname = "PZGESV  ";
  pzmtxcpy(9, 9, 1, 1, zta7, desc_a, ztas7);
  pzmtxcpy(9, 6, 1, 2, ztb7, desc_b, ztbs7);
  pzgesv(9, 5, zta7, 1, 1, desc_a, ipvt, ztb7, 1, 2, desc_b, iret);
  if (iret != 0)
  {
   if (iam == 0)
    cout << " IRET from PZGESV  = " << iret;
    *ifail = *ifail + 1;
  }
   pzmtchk(modname,9, 9, 1, 1, zta7, desc_a, zresa7, tol, ierr, ilerr,
            error, ifail, strlen(modname));
  pzmtchk(modname,9, 6, 1, 2, ztb7, desc_b, zresb7, tol, ierr, ilerr,
          error, ifail, strlen(modname));
 
  modname = "PZGETRI ";
  tol = 0.1;
  pzmtxcpy(9, 9, 1, 1, zta7, desc_a, ztas7);
  pzgetrf(9, 9, zta7, 1, 1, desc_a, ipvt, iret);
  pzgetri(9, zta7, 1, 1, desc_a, ipvt,workz,lwork,iwork,liwork, iret);
  if (iret != 0)
  {
   if (iam == 0)
    cout << " IRET from PZGETRI = " << iret;
    *ifail = *ifail + 1;
  }
    pzmtchk(modname,9, 9, 1, 1, zta7, desc_a, zresa71,tol, ierr, ilerr,
            error, ifail, strlen(modname));
 
  modname = "PZGECON ";
  tol = 0.1;
  pzmtxcpy(9, 9, 1, 1, zta7, desc_a, ztas7);
  anorm = pzlange("O", 9, 9, zta7, 1, 1, desc_a, rwork);
  pzgetrf(9, 9, zta7, 1, 1, desc_a, ipvt, iret);
  pzgecon("O", 9, zta7, 1, 1, desc_a, anorm,  rcond, workz, lwork,
           rwork, lrwork, iret);
  if (iret != 0)
   {
   if(iam == 0)
    cout << " IRET from PZGECON = " << iret;
   *ifail = *ifail + 1;
  }
       if ((ABSP(rcond) - ABSP(rcondz)) > tol)
        {
          *ifail = *ifail + 1;
    cout << " RCOND from PZGECON = " << rcond, rcondz;
        }
 
// PDGETRF/S & PZGETRF/S PDGESV, PZGESV END : 7
// PDGETRI   & PZGETRI PDLANGE PZLANGE  END : 7
// PDGECON   & PZGECON                  END : 7
 
// PDGEQRF PZGEQRF
 
      double dga11[12], dgt11[3];
      _ESVPCOM zga11[12], zgt11[3];
      double dgas11[12] = {  1.0,  2.0,  2.0,  4.0,
                           -2.0,  0.0, -4.0,  0.0,
                           -1.0,  1.0,  2.0,  0.0};
 
      double dresa11[12] = { -5.0,  0.33,  0.33,  0.67,
                            2.0, -4.00, -0.50,  0.50,
                           -1.00, 1.00, -2.00,  0.00};
 
      double drest11[3] = {  1.2,  1.33,  2.00};
 
  _ESVPCOM zgas11[12] = {
      _ESVPCOM (1.0, 0.0),_ESVPCOM (1.0,-1.0),_ESVPCOM(1.0,-2.0),
      _ESVPCOM (-2.0,1.0),_ESVPCOM(2.0, 2.0),_ESVPCOM (-2.0, 3.0),
      _ESVPCOM (-3.0,-1.0),_ESVPCOM (-3.0, 0.0),_ESVPCOM(-3.0, 1.0),
      _ESVPCOM (4.0,-3.0),_ESVPCOM(-4.0,-2.0),_ESVPCOM (4.0,-1.0)};
 
  _ESVPCOM zresa11[12] = {
       _ESVPCOM(-2.83, 0.0),_ESVPCOM (0.26,-0.26),_ESVPCOM(0.26,-0.52),
       _ESVPCOM( 3.54,-1.41),_ESVPCOM(-3.39, 0.0),_ESVPCOM (-0.29,-0.09),
       _ESVPCOM( 3.89, 3.18), _ESVPCOM( 0.37,-0.37),_ESVPCOM(-1.87, 0.0),
       _ESVPCOM(-2.83, 0.71),_ESVPCOM( 6.78, 0.74),_ESVPCOM (1.87,-1.87)};
 
  _ESVPCOM zrest11[3] = {
       _ESVPCOM( 1.35, 0.0),_ESVPCOM (1.83,-0.02),_ESVPCOM(1.47,-0.88)};
 
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 4;
      desc_a[N_]    = 3;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
/* following desc_x is created to check vector tau    */
 
      desc_x[DT_]   = BLOCK_CYCLIC_2D;
      desc_x[M_]    = 1;
      INCX          = desc_x[M_];
      desc_x[N_]    = desc_a[N_];
      desc_x[MB_]   = 1;
      desc_x[NB_]   = desc_a[NB_];
      desc_x[CSRC_] = desc_a[CSRC_];
      desc_x[CTXT_] = ictx;
      desc_x[LLD_]  = INCX;
 
      modname = "PDGEQRF ";
      tol = 0.1;
      pdmtxcpy(4, 3, 1, 1, dga11, desc_a, dgas11);
      pdgeqrf(4, 3, dga11, 1, 1, desc_a, dgt11, work, 0, iret);
      if (iret != 0)
  {
    if(iam == 0)
    cout << " IRET from PDGEQRF = " << iret;
   *ifail = *ifail + 1;
  }
      else
  {
    pdmtchk(modname, 4, 3, 1, 1, dga11, desc_a, dresa11, tol,
                    ierr, ilerr,  error, ifail, strlen(modname));
    for (i=0; i< nprow; i++)
      {
         desc_x[RSRC_] = i;
         pdvtchk(modname,3, 1, 1, dgt11, desc_x, drest11, 1, tol,
                      ierr, ilerr, error, ifail, strlen(modname));
       }
   }
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 3;
      desc_a[N_]    = 4;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(3,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
/* following desc_x is created to check vector tau    */
 
      desc_x[DT_]   = BLOCK_CYCLIC_2D;
      desc_x[M_]    = 1;
      INCX          = desc_x[M_];
      desc_x[N_]    = desc_a[M_];
      desc_x[MB_]   = 1;
      desc_x[NB_]   = desc_a[NB_];
      desc_x[CSRC_] = desc_a[CSRC_];
      desc_x[CTXT_] = ictx;
      desc_x[LLD_]  = INCX;
 
      modname = "PZGEQRF ";
      tol = 0.1;
      pzmtxcpy(3, 4, 1, 1, zga11, desc_a, zgas11);
      pzgeqrf(3, 4, zga11, 1, 1, desc_a, zgt11, workz, 0, iret);
      if (iret != 0)
  {
    if(iam == 0)
    cout << " IRET from PZGEQRF = " << iret;
   *ifail = *ifail + 1;
  }
      else
  {
    pzmtchk(modname, 3, 4, 1, 1, zga11, desc_a, zresa11, tol,
                    ierr, ilerr,  error, ifail, strlen(modname));
    for (i=0; i< nprow; i++)
      {
         desc_x[RSRC_] = i;
         pzvtchk(modname,3, 1, 1, zgt11, desc_x, zrest11, 1, tol,
                      ierr, ilerr, error, ifail, strlen(modname));
       }
   }
/* PDGEQRF PZGEQRF  END : 11  */
/* PDGELS  PZGELS   START : 12  */
      double dga12[12], dgb12[20];
     _ESVPCOM zga12[12], zgb12[12];
      double dgas12[12]=  {  1.0,  2.0,  2.0,  4.0,
                            -2.0,  0.0, -4.0,  0.0,
                            -1.0,  1.0,  2.0,  0.0};
 
      double dgbs12[20]=  { -1.0,  1.0,  1.0, -2.0,
                            -2.0,  3.0,  0.0,  4.0,
                            -7.0,  4.0,  4.0,  4.0,
                             0.0,  3.0,  2.0,  0.0,
                            -5.0,  5.0,  2.0,  4.0};
 
      double dresb12[20]=  { -0.4,  0.00,  1.00, -1.00,
                              1.0,  1.00,  1.00,  0.00,
                              0.80, 1.50,  4.00,  2.00,
                              0.2,  0.00,  1.00, -2.00,
                              1.00, 1.50,  3.00,  0.00};
 
     _ESVPCOM zgas12[12] = {
       _ESVPCOM(1.0, 0.0),_ESVPCOM (1.0,-1.0),_ESVPCOM(1.0,-2.0),
       _ESVPCOM(-2.0,1.0),_ESVPCOM(2.0, 2.0),_ESVPCOM (-2.0, 3.0),
       _ESVPCOM(-3.0,-1.0),_ESVPCOM (-3.0, 0.0),_ESVPCOM(-3.0, 1.0),
       _ESVPCOM(4.0,-3.0),_ESVPCOM(-4.0,-2.0),_ESVPCOM (4.0,-1.0)};
 
     _ESVPCOM zgbs12[12] = {
      _ESVPCOM( 1.0,0.0),_ESVPCOM (-1.0,1.0), _ESVPCOM(2.0, 1.0),
      _ESVPCOM(0.0, 0.0),
      _ESVPCOM( 0.0,1.0),_ESVPCOM ( 1.0,-1.0),_ESVPCOM(1.0, 2.0),
      _ESVPCOM (0.0, 0.0),
      _ESVPCOM( 1.0,1.0),_ESVPCOM ( 0.0, 0.0),_ESVPCOM(-1.0,-1.0),
      _ESVPCOM(0.0, 0.0)};
 
     _ESVPCOM zresb12[12] = {
       _ESVPCOM(-.16,.15),_ESVPCOM( .11,.02),_ESVPCOM(-.13,-.32),
       _ESVPCOM(.37,-.05),
       _ESVPCOM(-.08,.18),_ESVPCOM( .21,-.50),_ESVPCOM(.16, .12),
       _ESVPCOM(.04, .06),
       _ESVPCOM( .16,-.31),_ESVPCOM(-.38,.65),_ESVPCOM(-.27,-.28),
       _ESVPCOM(-.19,.32)};
 
 
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 4;
      desc_a[N_]    = 3;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
/* following desc_b   */
 
      desc_b[DT_]   = BLOCK_CYCLIC_2D;
      desc_b[M_]    = 4;
      desc_b[N_]    = 5;
      desc_b[MB_]   = 1;
      desc_b[NB_]   = 2;
      desc_b[RSRC_] = 0;
      desc_b[CSRC_] = 0;
      desc_b[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_b[LLD_]  = MAX(1,il_lda);
 
      modname = "PDGELS  ";
      tol = 0.1;
      pdmtxcpy(4, 3, 1, 1, dga12, desc_a, dgas12);
      pdmtxcpy(4, 5, 1, 1, dgb12, desc_b, dgbs12);
      pdgels("N",4,3,5,dga12,1,1,desc_a,dgb12,1,1,desc_b, work,0,
                  iret);
      if (iret != 0)
  {
    if(iam == 0)
    cout << " IRET from PDGELS  = " << iret;
   *ifail = *ifail + 1;
  }
      else
  {
    pdmtchk(modname, 4, 5, 1, 1, dgb12, desc_b, dresb12, tol,
                    ierr, ilerr,  error, ifail,strlen(modname));
   }
 
 
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 3;
      desc_a[N_]    = 4;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(3,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
/* following desc_b   */
 
      desc_b[DT_]   = BLOCK_CYCLIC_2D;
      desc_b[M_]    = 4;
      desc_b[N_]    = 3;
      desc_b[MB_]   = 1;
      desc_b[NB_]   = 1;
      desc_b[RSRC_] = 0;
      desc_b[CSRC_] = 0;
      desc_b[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_b[LLD_]  = MAX(1,il_lda);
 
      modname = "PZGELS  ";
      tol = 0.1;
      pzmtxcpy(3, 4, 1, 1, zga12, desc_a, zgas12);
      pzmtxcpy(4, 3, 1, 1, zgb12, desc_b, zgbs12);
      pzgels("N",3,4,3,zga12,1,1,desc_a,zgb12,1,1,desc_b, workz,0,
                  iret);
      if (iret != 0)
  {
    if(iam == 0)
    cout << " IRET from PZGELS  = " << iret;
   *ifail = *ifail + 1;
  }
      else
  {
    pzmtchk(modname, 4, 3, 1, 1, zgb12, desc_b, zresb12, tol,
                    ierr, ilerr,  error, ifail,strlen(modname));
   }
 
/* PDGELS  PZGELS  END   : 12  */
 
//PDGTTRF/S START
 
      int   desc_r[DLEN_], ipiv[12],  info;
      int laf, nprows, npcols, ictxs, myrows, mycols;
      double dtl8[12], dtd8[12], dtu8[12], dtb8[36], dt2u8[12];
      double dtls8[12] = {0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
                          1.0, 1.0, 1.0, 1.0};
 
      double dtds8[12] = {2.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0,
                          3.0, 3.0, 3.0, 3.0};
 
      double dtus8[12] = {2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0,
                          2.0, 2.0, 2.0, 0.0};
 
      double dtbs8[36] = {46.0, 65.0, 59.0, 53.0,
                          47.0, 41.0, 35.0, 29.0,
                          23.0, 17.0, 11.0,  5.0,
                           6.0, 13.0, 19.0, 25.0,
                          31.0, 37.0, 43.0, 49.0,
                          55.0, 61.0, 67.0, 47.0,
                           4.0,  6.0,  6.0,  6.0,
                           6.0,  6.0,  6.0,  6.0,
                           6.0,  6.0,  6.0,  4.0};
 
      double dresb8[36] =  {12.0, 11.0, 10.0,  9.0,
                             8.0,  7.0,  6.0,  5.0,
                             4.0,  3.0,  2.0,  1.0,
                             1.0,  2.0,  3.0,  4.0,
                             5.0,  6.0,  7.0,  8.0,
                             9.0, 10.0, 11.0, 12.0,
                             1.0,  1.0,  1.0,  1.0,
                             1.0,  1.0,  1.0,  1.0,
                             1.0,  1.0,  1.0,  1.0};
// PDPTTRF/S, PDPTSV
      double dtd10[12], dtu10[12], dtb10[36];
      double dtds10[12] = {4.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0,
                           5.0, 5.0, 5.0, 5.0};
      double dtus10[12] = {2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0,
                           2.0, 2.0, 2.0, 0.0};
      double dtbs10[36] = { 70.0, 99.0, 90.0, 81.0,
                            72.0, 63.0, 54.0, 45.0,
                            36.0, 27.0, 18.0,  9.0,
                             8.0, 18.0, 27.0, 36.0,
                            45.0, 54.0, 63.0, 72.0,
                            81.0, 90.0, 99.0, 82.0,
                             6.0,  9.0,  9.0,  9.0,
                             9.0,  9.0,  9.0,  9.0,
                             9.0,  9.0,  9.0,  7.0};
 
  lwork = 0;
  tol = 0.00000001;
 
  nprows = nnodes;
  npcols = 1;
  blacs_get(0,0,ictxs);
  blacs_gridinit(ictxs,order,nprows,npcols,strlen(order));
  blacs_gridinfo(ictxs,nprows,npcols,myrows,mycols);
 
// following desc_x is used to distribute vector in 1xnpcol grid
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 1;
  desc_x[N_]    = 12;
  desc_x[MB_]   = 1;
  desc_x[NB_]   = (12+(nprow*npcol)-1)/(nprow*npcol);
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  desc_x[LLD_]  = 1;
 
// following desc_r is used to distribute RHS matrix in nprowsx1 grid
  desc_r[DT_]   = BLOCK_CYCLIC_2D;
  desc_r[M_]    = 12;
  desc_r[N_]    = 3;
  desc_r[MB_]   = (12+(nprow*npcol)-1)/(nprow*npcol);
  desc_r[NB_]   = 3;
  desc_r[RSRC_] = 0;
  desc_r[CSRC_] = 0;
  desc_r[CTXT_] = ictxs;
  il_lda = numroc(12,desc_r[MB_],myrows,0,nprows);
   desc_r[LLD_] = MAX(1,il_lda);
 
 
// following desc_a is used by DGTTRF/S, DDTTRF/S, DGTSV/DDTSV routines
  desc_a[DT_1D_]   = BLOCK_COLUMN_1D;
  desc_a[CTXT_1D_] = ictx;
  desc_a[N_1D_]    = 12;
  desc_a[NB_1D_]   = (12+(nprow*npcol)-1)/(nprow*npcol);
  desc_a[CSRC_1D_] = 0;
  desc_a[LLD_1D_]  = 0;
  desc_a[RSVD_1D_] = 0;
 
// following desc_b is used by DGTTRS, DDTTRS, DGTSV/DDTSV routine
  desc_b[DT_1D_]   = BLOCK_ROW_1D;
  desc_b[CTXT_1D_] = ictx;
  desc_b[M_1D_]    = 12;
  desc_b[MB_1D_]   = (12+(nprow*npcol)-1)/(nprow*npcol);
  desc_b[RSRC_1D_] = 0;
  desc_b[LLD_1D_]  = desc_r[LLD_];  // equal to 2D grid for check rtn
  desc_b[RSVD_1D_] = 0;
 
  laf = 12*nprow*npcol + 3*desc_a[NB_1D_];
  istor[0] = 0;
  double *af = new double[laf];
   if (!af)
    istor[0] = 1;
     igamx2d(ictx,scope, top, 1, 1,
                    istor, 1, ra, ca, 1, 0,0,strlen(scope),strlen(top));
   if (istor[0]==1)
    {
     if (iam == 0)
     cout << "Memory allocation error for PDGTTRF\n";
     delete [] af;         // free memory
     *ifail = *ifail + 1;
     blacs_gridexit(ictxs);
     blacs_gridexit(ictx);
     return;
    }
 
  modname = "PDGTTRF ";
  pdvtrcpy(12, 1, 1, dtl8, desc_x, dtls8, 1);
  pdvtrcpy(12, 1, 1, dtd8, desc_x, dtds8, 1);
  pdvtrcpy(12, 1, 1, dtu8, desc_x, dtus8, 1);
  pdgttrf(12,dtl8,dtd8,dtu8,dt2u8,1,desc_a,ipiv,af,laf,
           work, lwork, info);
 
  if (info != 0)
  {
   if (iam == 0)
    cout << " info from PDGTTRF = " << info;
    *ifail = *ifail + 1;
  }
  else
  {
 
  modname = "PDGTTRS ";
  pdmtxcpy(12, 3, 1, 1, dtb8, desc_r, dtbs8);
  pdgttrs("N",12,3,dtl8,dtd8,dtu8,dt2u8,1,desc_a, ipiv,
        dtb8,1,desc_b, af, laf, work,lwork,info);
  if (info != 0)
  {
   if (iam == 0)
    cout << " info from PDGTTRS = " << info;
    *ifail = *ifail + 1;
  }
 
  pdmtchk(modname, 12, 3, 1, 1, dtb8, desc_r, dresb8, tol,
                   ierr,ilerr, error, ifail, strlen(modname));
  }
 
    modname = "PDGTSV  ";
    pdvtrcpy(12, 1, 1, dtl8, desc_x, dtls8, 1);
    pdvtrcpy(12, 1, 1, dtd8, desc_x, dtds8, 1);
    pdvtrcpy(12, 1, 1, dtu8, desc_x, dtus8, 1);
    pdmtxcpy(12, 3, 1, 1, dtb8, desc_r, dtbs8);
    pdgtsv(12,3,dtl8,dtd8,dtu8,1,desc_a,
         dtb8,1,desc_b, work,lwork,info);
    if (info != 0)
  {
   if (iam == 0)
    cout << " info from PDGTSV = " << info;
    *ifail = *ifail + 1;
  }
 
    pdmtchk(modname, 12, 3, 1, 1, dtb8, desc_r, dresb8, tol,
                     ierr,ilerr, error, ifail, strlen(modname));
 
  modname = "PDDTTRF ";
  pdvtrcpy(12, 1, 1, dtl8, desc_x, dtls8, 1);
  pdvtrcpy(12, 1, 1, dtd8, desc_x, dtds8, 1);
  pdvtrcpy(12, 1, 1, dtu8, desc_x, dtus8, 1);
  pddttrf(12,dtl8,dtd8,dtu8,1,desc_a, af, laf,
         work,lwork, info);
  if (info != 0)
  {
   if (iam == 0)
    cout << " info from PDDTTRF = " << info;
    *ifail = *ifail + 1;
  }
  else
  {
 
  modname = "PDDTTRS ";
  pdmtxcpy(12, 3, 1, 1, dtb8, desc_r, dtbs8);
  pddttrs("N",12,3,dtl8,dtd8,dtu8,1,desc_a,
        dtb8,1,desc_b, af, laf, work, lwork, info);
  if (info != 0)
  {
   if (iam == 0)
    cout << " info from PDDTTRS = " << info;
    *ifail = *ifail + 1;
  }
 
  pdmtchk(modname, 12, 3, 1, 1, dtb8, desc_r, dresb8, tol,
                   ierr,ilerr, error, ifail, strlen(modname));
  }
 
  modname = "PDDTSV  ";
  pdvtrcpy(12, 1, 1, dtl8, desc_x, dtls8, 1);
  pdvtrcpy(12, 1, 1, dtd8, desc_x, dtds8, 1);
  pdvtrcpy(12, 1, 1, dtu8, desc_x, dtus8, 1);
  pdmtxcpy(12, 3, 1, 1, dtb8, desc_r, dtbs8);
  pddtsv(12,3,dtl8,dtd8,dtu8,1,desc_a,
              dtb8,1,desc_b, work,lwork,info);
  if (info != 0)
  {
   if (iam == 0)
    cout << " info from PDDTSV = " << info;
    *ifail = *ifail + 1;
  }
 
  pdmtchk(modname, 12, 3, 1, 1, dtb8, desc_r, dresb8, tol,
                   ierr,ilerr, error, ifail, strlen(modname));
 
// PDPTTRF/S PDPTSV
 
  modname = "PDPTTRF ";
  pdvtrcpy(12, 1, 1, dtd10, desc_x, dtds10, 1);
  pdvtrcpy(12, 1, 1, dtu10, desc_x, dtus10, 1);
  pdpttrf(12,dtd10,dtu10,1,desc_a, af, laf,
         work,lwork, info);
  if (info != 0)
  {
   if (iam == 0)
    cout << " info from PDPTTRF = " << info;
    *ifail = *ifail + 1;
  }
  else
  {
 
  modname = "PDPTTRS ";
  pdmtxcpy(12, 3, 1, 1, dtb10, desc_r, dtbs10);
  pdpttrs(12,3,dtd10,dtu10,1,desc_a,
        dtb10,1,desc_b, af, laf, work, lwork, info);
  if (info != 0)
  {
   if (iam == 0)
    cout << " info from PDPTTRS = " << info;
    *ifail = *ifail + 1;
  }
 
  pdmtchk(modname, 12, 3, 1, 1, dtb10, desc_r, dresb8, tol,
                   ierr,ilerr, error, ifail, strlen(modname));
  }
 
  modname = "PDPTSV  ";
  pdvtrcpy(12, 1, 1, dtd10, desc_x, dtds10, 1);
  pdvtrcpy(12, 1, 1, dtu10, desc_x, dtus10, 1);
  pdmtxcpy(12, 3, 1, 1, dtb10, desc_r, dtbs10);
  pdptsv(12,3,dtd10,dtu10,1,desc_a,
              dtb10,1,desc_b, work,lwork,info);
  if (info != 0)
  {
   if (iam == 0)
    cout << " info from PDPTSV = " << info;
    *ifail = *ifail + 1;
  }
 
  pdmtchk(modname, 12, 3, 1, 1, dtb10, desc_r, dresb8, tol,
                   ierr,ilerr, error, ifail, strlen(modname));
 
   delete [] af;         // free memory
 
//PDPBTRF/S  START : 10
 
// PDPBTRF/S START : 9
      int k;
      double dta9[72];
      double dtb9[27];
      double dtas9[72] = {
         1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
         2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 1.0,
         3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 2.0, 0.0,
         4.0, 4.0, 4.0, 4.0, 4.0, 3.0, 0.0, 0.0,
         5.0, 5.0, 5.0, 5.0, 4.0, 0.0, 0.0, 0.0,
         6.0, 6.0, 6.0, 5.0, 0.0, 0.0, 0.0, 0.0,
         7.0, 7.0, 6.0, 0.0, 0.0, 0.0, 0.0, 0.0,
         8.0, 7.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
         8.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
 
      double dtbs9[27] = {
         8.0, 16.0, 23.0, 29.0, 34.0, 38.0,  41.0, 43.0, 36.0,
        36.0, 80.0,122.0,161.0,196.0,226.0, 250.0,267.0,240.0,
        44.0, 80.0,108.0,129.0,144.0,154.0, 160.0,163.0,120.0 };
 
      double dresb9[27] = {
           1.00,  1.00,  1.00,  1.00,  1.00,  1.00, 1.00,  1.00,  1.00,
           1.00,  2.00,  3.00,  4.00,  5.00,  6.00, 7.00,  8.00,  9.00,
           9.00,  8.00,  7.00,  6.00,  5.00,  4.00, 3.00,  2.00,  1.00 };
 
// PDPBTRF/S  PDPBSV START : 9
//
// following desc_x is used to distribute matrix in 1xnpcol grid
      desc_x[DT_]   = BLOCK_CYCLIC_2D;
      desc_x[M_]    = 8;
      desc_x[N_]    = 9;
      desc_x[MB_]   = 8;
      desc_x[NB_]   = (9+(nprow*npcol)-1)/(nprow*npcol);
      desc_x[RSRC_] = 0;
      desc_x[CSRC_] = 0;
      desc_x[CTXT_] = ictx;
      desc_x[LLD_]  = 8;
 
// following desc_r is used to distribute RHS matrix in nprows x 1 grid
      desc_r[DT_]   = BLOCK_CYCLIC_2D;
      desc_r[M_]    = 9;
      desc_r[N_]    = 3;
      desc_r[MB_]   = (9+(nprow*npcol)-1)/(nprow*npcol);
      desc_r[NB_]   = 3;
      desc_r[RSRC_] = 0;
      desc_r[CSRC_] = 0;
      desc_r[CTXT_] = ictxs;
      il_lda = numroc(9,desc_r[MB_],myrows,0,nprows);
      desc_r[LLD_]  = MAX(1,il_lda);
 
// following desc_a is used by DPBTRF/S, DPBSV routines
      desc_a[DT_1D_]   = BLOCK_COLUMN_1D;
      desc_a[CTXT_1D_] = ictx;
      desc_a[N_1D_]    = 9;
      desc_a[NB_1D_]   = (9+(nprow*npcol)-1)/(nprow*npcol);
      desc_a[CSRC_1D_] = 0;
      desc_a[LLD_1D_]  = 8;
      desc_a[RSVD_1D_] = 0;
 
// following desc_b is used by DGTTRS, DPBSV routine
      desc_b[DT_1D_]   = BLOCK_ROW_1D;
      desc_b[CTXT_1D_] = ictx;
      desc_b[M_1D_]    = 9;
      desc_b[MB_1D_]   = (9+(nprow*npcol)-1)/(nprow*npcol);
      desc_b[RSRC_1D_] = 0;
      desc_b[LLD_1D_]  = desc_r[LLD_];  // equal to 2D grid for check rtn
      desc_b[RSVD_1D_] = 0;
 
      k = 7;
      laf = k*(desc_a[NB_1D_]+2*k);
      af = new double[laf];
      if (!af)
       {
        cout << "Memory allocation error\n";
        *ifail = *ifail + 1;
        return;
       }
 
      modname = "PDPBTRF ";
      pdmtxcpy(8, 9, 1, 1, dta9, desc_x, dtas9);
      pdpbtrf("L",9,7,dta9,1,desc_a, af, laf, work,lwork,info);
 
      if (info != 0)
  {
   if (iam == 0)
      cout << " info from PDPBTRF = " << info;
    *ifail = *ifail + 1;
  }
  else
  {
 
      modname = "PDPBTRS ";
      pdmtxcpy(9, 3, 1, 1, dtb9, desc_r, dtbs9);
      pdpbtrs("L", 9, 7, 3, dta9, 1, desc_a, dtb9,1,desc_b, af,
               laf, work,lwork,info);
 
      if (info != 0)
  {
       if (iam == 0)
        cout << " info from PDPBTRS = " << info;
        *ifail = *ifail + 1;
  }
 
      pdmtchk(modname, 9, 3, 1, 1, dtb9, desc_r, dresb9, tol,
              ierr,ilerr, error, ifail, strlen(modname));
  }
 
      modname = "PDPBSV  ";
      pdmtxcpy(8, 9, 1, 1, dta9, desc_x, dtas9);
      pdmtxcpy(9, 3, 1, 1, dtb9, desc_r, dtbs9);
      pdpbsv("L",9,7,3,dta9,1,desc_a, dtb9,1,desc_b, work, lwork,info);
 
      if (info != 0)
  {
       if (iam == 0)
       cout << " info from PDPBSV  = " << info;
    *ifail = *ifail + 1;
  }
 
      pdmtchk(modname, 9, 3, 1, 1, dtb9, desc_r, dresb9, tol,
              ierr,ilerr, error, ifail, strlen(modname));
 
   delete [] af;         // free memory
 
// PDPBTRF/S END   : 9
 
 
  blacs_gridexit(ictxs);
  blacs_gridexit(ictx);
}
 
/*********************************************************************
*          PARALLEL SPARSE LINEAR ALGEBRAIC EQUATIONS SUBROUTINES    *
*********************************************************************/
 
void psprs(int *ifail)
/**********************************************************************
!
! This program illustrates how to use the Parallel ESSL F77 Sparse Iterative
! Solver and its supporting utility subroutines.  A very simple problem
! (DSRIS Example 1 from the ESSL Guide and Reference) using an
! HPF BLOCK data distribution is solved.
!
***********************************************************************/
 
{
 
    int  izero=0, ione=1, dupflag=2, n=9, nnz=22;
    int  ierr=5, ilerr,istor[1],ra[1],ca[1];
    double error=0.0, tol=0.0001;
    int desc_r[DLEN_],il_lda;
    char modn[9], *modname=modn;
    char order[]="R";
    char stor[]="DEF  ";
    char mtype[]="GEN  ";
    char scope[]="A";
    char   top[]="1";
 
/**********************************************************************
! Sparse Matrices and related information
!  What size should these be? I dont find that in the prolog
***********************************************************************/
    double            as[22];
    int               ia1[50],ia2[50];
    int               infoa[30];
 
    double            bs[22];
    int               ib1[10],ib2[22];
    int               infob[30];
 
/**********************************************************************
! Preconditioner Data Structure
***********************************************************************/
    double                  prcs[110];
    int                     lprcs;
 
/**********************************************************************
! Dense Vectors
***********************************************************************/
    double                  b[9],x[9];
 
/**********************************************************************
! BLACS parameters
***********************************************************************/
    int            ictxt,nprow,npcol,myrow,mycol,nnodes,iam,ercode=-1;
 
/**********************************************************************
! Solver parameters
***********************************************************************/
      int                   iter,itmax,info,itrace;
      int                   iprec,methd,istopc;
      double                err,eps;
 
/**********************************************************************
! We will not have duplicates so PV used by the PARTS subroutine
! PART_BLOCK only needs to be of length 1.
***********************************************************************/
 
      int                   pv[1];
 
// Other variables
      int                   nb, ldb, ldbg;
      int                   nx, ldx, ldxg;
      int                   nrhs ;
      int                   i,j;
      int                   global_indx, nv_count;
      int                   global_indx_owner, nv;
      int                   local_indx;
 
/**********************************************************************
!
!     Global Problem
!      DSRIS Example 1 from the ESSL Guide and Reference
!
***********************************************************************/
      double a_global[22] = {2.0,2.0,-1.0,1.0,2.0,1.0,2.0,-1.0,
                         1.0,2.0,-1.0,1.0,2.0,-1.0,1.0,2.0,
                        -1.0,1.0,2.0,-1.0,1.0,2.0};
/**********************************************************************
!
!     ia and ja arrays used by Fortran 77 subrouitines
!      that why their elements are 1 origin
!
***********************************************************************/
      int  ja[22] = {1,2,3,2,3,1,4,5,4,5,6,5,6,7,6,7,8,
                         7,8,9,8,9};
      int  ia[10] = {1,2,4,6,9,12,15,18,21,23};
 
      int  iparm[20] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
 
      double  rparm[20] = {0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,
                           0.,0.,0.,0.,0.,0.,0.,0.,0.};
      double b_global[9] = {2.0,1.0,3.0,2.0,2.0,2.0,2.0,2.0,
                           3.0};
      double xinit_global[9] = {0.0,0.0,0.0,0.0,0.0,0.0,0.0,
                               0.0, 0.0};
      double xres_global[9] = {1.0,1.0,1.0,1.0,1.0,1.0,1.0,
                               1.0, 1.0};
 
/**********************************************************************
! Initialize BLACS
! Define a P x 1 Process Grid
***********************************************************************/
 
    blacs_pinfo(iam,nnodes);
    nprow = nnodes;
    npcol = 1;
    blacs_get(0,0,ictxt);
    blacs_gridinit(ictxt,order,nprow,npcol,strlen(order));
    blacs_gridinfo(ictxt,nprow,npcol,myrow,mycol);
 
/**********************************************************************
!
! Initialize some elements of the sparse matrix A
! and its descriptor vector, DESC_A
!
***********************************************************************/
      istor[0] = 0;
      int *desc_a=new int[33+3*nnodes+4*n];
      if(!desc_a)
        istor[0] = 1;
      igamx2d(ictxt,scope, top, 1, 1,
               istor, 1, ra, ca, 1, 0, 0,strlen(scope),strlen(top));
      if(istor[0]==1)
       {
        if (iam == 0)
        cout << "Memory allocation error for SPARSE\n";
        *ifail = *ifail + 1;
        delete [] desc_a;
        blacs_gridexit(ictxt);
        return;
       }
      desc_a[10] = 33+3*nnodes+4*n;
      padinit(n,part_block,desc_a,ictxt);
 
      infoa[0] = sizeof(as)/sizeof(double);
      infoa[1] = sizeof(ia1)/sizeof(int);
      infoa[2] = sizeof(ia2)/sizeof(int);
      pdspinit(as,ia1,ia2,infoa,desc_a);
 
/**********************************************************************
!
! In this simple example, all processes have a copy of
! the global sparse matrix, A
!
! Only the owning process will insert the row into the
! local part of the sparse matrix
!
! global_indx used by c and Fortran 77 subroutines
!  that why global_indx is 1 origin
!
***********************************************************************/
      nb = 1;
      ldb = sizeof(b)/sizeof(double);
      ldbg = sizeof(b_global)/sizeof(double);
      nx = 1;
      ldx = sizeof(x)/sizeof(double);
      ldxg = sizeof(xinit_global)/sizeof(double);
 
      for(global_indx=1;global_indx<=n;++global_indx)
       {
        part_block(&global_indx,&n,&nnodes,pv,&nv);
/**********************************************************************
!
! In this simple example, NV will always be 1
! since there will not be duplicate coefficents
!
***********************************************************************/
        for(nv_count=0;nv_count<nv;++nv_count)
         {
          global_indx_owner = pv[nv_count];
          if (global_indx_owner == myrow) {
           ib2[0] = 1;
           ib2[1] = 1;
           for(j=ia[global_indx-1]-1;j<ia[global_indx]-1;j++)
            {
             bs[ib2[1]-1] = a_global[j];
             ib1[ib2[1]-1]= ja[j];
             ib2[1] = ib2[1] + 1;
            }
           infob[0] = ib2[1] - 1;
           infob[1] = ib2[1] - 1;
           infob[2] = 2;
           infob[3] = 1;
           infob[4] = 1;
           infob[5] = 1;
           infob[6] = n;
           pdspins(as,ia1,ia2,infoa,desc_a,global_indx, 1,
                      bs,ib1,ib2,infob);
           pdgeins(nb,b,ldb,global_indx,1,1,1,
                      &b_global[global_indx-1],ldbg,desc_a);
           pdgeins(nx,x,ldx,global_indx,1,1,1,
                      &xinit_global[global_indx-1],ldxg,desc_a);
          }
         }
       }
 
/*! Assemble A and DESC_A       */
      pdspasb(as,ia1,ia2,infoa,desc_a,
              mtype,stor,dupflag,info);
 
      if (info != 0) {
       if (iam == 0) {
        cout << "Error in assembly: " << info << "\n";
        delete [] desc_a;
        *ifail = *ifail + 1;
        blacs_gridexit(ictxt);
        return;
       }
      }
 
/**********************************************************************
!
! In this simple example, all processes have a copy of
! the global rhs vector, B, and the global initial guess vector, X
!
! Only the owning process will insert the element into the
! local part of the vector(s)
!
***********************************************************************/
 
/*! Assemble B and X     */
 
      pdgeasb(nb,b,ldb,desc_a);
      pdgeasb(nx,x,ldx,desc_a);
 
/**********************************************************************
!
!  Preconditioning
!
!  We are using ILU for the preconditioner
!
***********************************************************************/
      iprec = 2;
      lprcs = sizeof(prcs)/sizeof(double);
 
      pdspgpr(iprec,as,ia1,ia2,infoa,prcs,
              lprcs,desc_a,info);
 
      if (info != 0) {
       if (iam == 0) {
        cout << "Error in preconditioner: " << info << "\n";
        delete [] desc_a;
        *ifail = *ifail + 1;
        blacs_gridexit(ictxt);
        return;
       }
      }
 
/**********************************************************************
!
!  Iterative Solver - use the BICGSTAB method
!
***********************************************************************/
      nrhs  = 1;
      itmax = 1000;
      eps   = 1.e-8;
      methd = 1;
      istopc = 1;
      itrace = 0;
      iparm[0] = methd;
      iparm[1] = istopc;
      iparm[2] = itmax;
      iparm[3] = itrace;
      rparm[0] = eps;
 
      pdspgis(as,ia1,ia2,infoa,nrhs,b,ldb,x,ldx,prcs,
                   desc_a,iparm,rparm,info);
 
 
      if (info != 0) {
       if (iam == 0) {
        cout << "Error in solver: " << info << "\n";
        delete [] desc_a;
        *ifail = *ifail + 1;
        blacs_gridexit(ictxt);
        return;
       }
      }
 
/*    err = rparm[1];
      iter = iparm[4];
      if (iam == 0) {
        cout << "Number of iteration: " << iter << "\n";
        cout << "Error on exit: " << err << "\n";
      }                                         */
 
/* following desc_r  RHS vector in nprowsx1 grid*/
  desc_r[DT_] = BLOCK_CYCLIC_2D;
  desc_r[M_] = 9;
  desc_r[N_] = 1;
  desc_r[MB_] = (9+(nprow*npcol)-1)/(nprow*npcol);
  desc_r[NB_] = 1;
  desc_r[RSRC_] = 0;
  desc_r[CSRC_] = 0;
  desc_r[CTXT_] = ictxt;
  il_lda = numroc(9,desc_r[MB_],myrow,0,nprow);
   desc_r[LLD_] = MAX(1,il_lda);
 
  modname = "PDSPGIS ";
  pdmtchk(modname, 9, 1, 1, 1, x, desc_r, xres_global, tol,
                   ierr,ilerr, error, ifail, strlen(modname));
 
/**********************************************************************
!
!  Terminate the process grid and the BLACS
!
***********************************************************************/
       delete [] desc_a;
       blacs_gridexit(ictxt);
}
 
void part_block(int *global_indx,int *n,int *nnodes,int *pv,int *nv)
{
  int dim_block;
     dim_block = (*n + *nnodes -1)/(*nnodes);
     *nv = 1;
     pv[*nv-1] = (*global_indx - 1)/dim_block;
}
 
 
/*********************************************************************
*          PARALLEL TWO DIMENSIONAL SHORT PRECISION FFT SUBROUTINES  *
*********************************************************************/
 
void psfft2(int *ifail)
{
// PSCFT2    START : 1,
 
   int i, iam, nnodes,myrow,mycol,ip[40];
   int desc_x[DLEN_], desc_y[DLEN_], il_lda;
   int ictx, nprow, npcol,ierr=5,ilerr;
   char modn[9], *modname=modn;
   char order[]="R";
   double error=0.0, tol=0.0;
   _ESVPCM cx1[48], cy1[48],cxs1[48],cresy1[48];
   _ESVPCM cx3[16];
   float   sy3[48], sys3[48], sresy3[48];
 
   ip[0] = 1;
   ip[1] = 1;
   ip[19] = 0;
   ip[20] = 0;
 
   cxs1[0] =_ESVPCM(48.0,0.0);
   for (i=1; i<48; i++)
    cxs1[i] =_ESVPCM(0.0,0.0);
 
   for (i=0; i<48; i++)
    cresy1[i] =_ESVPCM(1.0,0.0);
 
   _ESVPCM cxs3[16] = {
     _ESVPCM(1.0,1.0), _ESVPCM(1.0,0.0), _ESVPCM(1.0,1.0), _ESVPCM(1.0,0.0),
     _ESVPCM(1.0,0.0), _ESVPCM(1.0,0.0), _ESVPCM(1.0,0.0), _ESVPCM(1.0,0.0),
     _ESVPCM(1.0,0.0), _ESVPCM(1.0,0.0), _ESVPCM(1.0,0.0), _ESVPCM(1.0,0.0),
     _ESVPCM(1.0,0.0), _ESVPCM(1.0,0.0), _ESVPCM(1.0,0.0), _ESVPCM(1.0,0.0) };
 
    sresy3[0] = 1.0;
   for (i=1; i<48; i++)
    sresy3[i] = 0.0;
 
  blacs_pinfo(iam,nnodes);
  if (nnodes > 1)
   {
     nprow = 1;
     npcol = 2;
   }
  else
   {
     nprow = 1;
     npcol = 1;
   }
  blacs_get(0,0,ictx);
  blacs_gridinit(ictx,order,nprow,npcol,strlen(order));
  blacs_gridinfo(ictx,nprow,npcol,myrow,mycol);
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 8;
  desc_x[N_]    = 6;
  desc_x[MB_]   = 8;
  desc_x[NB_]   = 6/npcol;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
 
  desc_y[DT_]   = BLOCK_CYCLIC_2D;
  desc_y[M_]    = 8;
  desc_y[N_]    = 6;
  desc_y[MB_]   = 8;
  desc_y[NB_]   = 6/npcol;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
 
  modname = "PSCFT2  ";
  if (iam < 2 && nnodes > 1 )
  {
    il_lda = numroc(8,8,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(8,8,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    pcmtxcpy(8, 6, 1, 1, cx1,  desc_x, cxs1);
    pscft2(cx1, cy1, 8, 6, -1, 1.0/48.0, ictx, ip);
    pcmtchk(modname,8, 6, 1, 1, cy1, desc_y, cresy1, tol, ierr, ilerr,
            error, ifail, strlen(modname));
  }
  else if (iam == 0)
  {
    il_lda = numroc(8,8,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(8,8,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    pcmtxcpy(8, 6, 1, 1, cx1,  desc_x, cxs1);
    pscft2(cx1, cy1, 8, 6, -1, 1.0/48.0, ictx, ip);
    pcmtchk(modname,8, 6, 1, 1, cy1, desc_y, cresy1, tol, ierr, ilerr,
            error, ifail, strlen(modname));
  }
 
   ip[0] = 1;
   ip[19] = 4;
   ip[20] = 12;
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 4;
  desc_x[N_]    = 4;
  desc_x[MB_]   = 4;
  desc_x[NB_]   = 8/(2*npcol);
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  desc_x[LLD_]  = 4;
 
  desc_y[DT_]   = BLOCK_CYCLIC_2D;
  desc_y[M_]    = 8;
  desc_y[N_]    = 4;
  desc_y[MB_]   = 8;
  desc_y[NB_]   = 4/npcol;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  desc_y[LLD_]  = 12;
  if (iam < 2 && nnodes > 1 )
  {
    modname = "PSCRFT2 ";
    pcmtxcpy(4, 4, 1, 1, cx3,  desc_x, cxs3);
    pscrft2(cx3, sy3, 8, 4, -1, 1.0/32.0, ictx, ip);
    psmtchk(modname,8, 4, 1, 1, sy3, desc_y, sresy3, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    ip[19] = 12;
    ip[20] = 4;
    modname = "PSRCFT2 ";
    psrcft2(sy3, cx3, 8, 4, 1, 1.0, ictx, ip);
    pcmtchk(modname,4, 4, 1, 1, cx3, desc_x, cxs3, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    blacs_gridexit(ictx);
  }
  else if (iam == 0)
  {
    modname = "PSCRFT2 ";
    pcmtxcpy(4, 4, 1, 1, cx3,  desc_x, cxs3);
    pscrft2(cx3, sy3, 8, 4, -1, 1.0/32.0, ictx, ip);
    psmtchk(modname,12, 4, 1, 1, sy3, desc_y, sresy3, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    ip[19] = 12;
    ip[20] = 4;
    modname = "PSRCFT2 ";
    psrcft2(sy3, cx3, 8, 4, 1, 1.0, ictx, ip);
    pcmtchk(modname,4, 4, 1, 1, cx3, desc_x, cxs3, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    blacs_gridexit(ictx);
  }
 
}
 
/*********************************************************************
*          PARALLEL TWO DIMENSIONAL LONG PRECISION FFT SUBROUTINES   *
*********************************************************************/
 
void pfft2(int *ifail)
{
// PDCFT2    START : 1,
 
   int i, iam, nnodes,myrow,mycol,ip[40];
   int desc_x[DLEN_], desc_y[DLEN_], il_lda;
   int ictx, nprow, npcol,ierr=5,ilerr;
   char modn[9], *modname=modn;
   char order[]="R";
   double error=0.0, tol=0.0;
   _ESVPCOM zx1[48], zy1[48],zxs1[48],zresy1[48];
   _ESVPCOM zx3[16];
   double  dy3[48], dys3[48], dresy3[48];
 
   ip[0]  = 1;
   ip[1]  = 1;
   ip[19] = 0;
   ip[20] = 0;
 
   zxs1[0] = _ESVPCOM(48.0,0.0);
   for (i=1; i<48; i++)
    zxs1[i] = _ESVPCOM(0.0,0.0);
 
   for (i=0; i<48; i++)
    zresy1[i] = _ESVPCOM(1.0,0.0);
 
   _ESVPCOM zxs3[16] = {
     _ESVPCOM(1.0,1.0), _ESVPCOM(1.0,0.0), _ESVPCOM(1.0,1.0), _ESVPCOM(1.0,0.0),
     _ESVPCOM(1.0,0.0), _ESVPCOM(1.0,0.0), _ESVPCOM(1.0,0.0), _ESVPCOM(1.0,0.0),
     _ESVPCOM(1.0,0.0), _ESVPCOM(1.0,0.0), _ESVPCOM(1.0,0.0), _ESVPCOM(1.0,0.0),
     _ESVPCOM(1.0,0.0), _ESVPCOM(1.0,0.0), _ESVPCOM(1.0,0.0), _ESVPCOM(1.0,0.0) };
 
    dresy3[0] = 1.0;
   for (i=1; i<48; i++)
    dresy3[i] = 0.0;
 
  blacs_pinfo(iam,nnodes);
  if (nnodes > 1)
   {
     nprow = 1;
     npcol = 2;
   }
  else
   {
     nprow = 1;
     npcol = 1;
   }
  blacs_get(0,0,ictx);
  blacs_gridinit(ictx,order,nprow,npcol,strlen(order));
  blacs_gridinfo(ictx,nprow,npcol,myrow,mycol);
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 8;
  desc_x[N_]    = 6;
  desc_x[MB_]   = 8;
  desc_x[NB_]   = 6/npcol;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
 
  desc_y[DT_]   = BLOCK_CYCLIC_2D;
  desc_y[M_]    = 8;
  desc_y[N_]    = 6;
  desc_y[MB_]   = 8;
  desc_y[NB_]   = 6/npcol;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
 
  modname = "PDCFT2  ";
  if (iam < 2 && nnodes > 1 )
  {
    il_lda = numroc(8,8,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(8,8,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    pzmtxcpy(8, 6, 1, 1, zx1,  desc_x, zxs1);
    pdcft2(zx1, zy1, 8, 6, -1, 1.0/48.0, ictx, ip);
    pzmtchk(modname,8, 6, 1, 1, zy1, desc_y, zresy1, tol, ierr, ilerr,
            error, ifail, strlen(modname));
  }
  else if (iam == 0)
  {
    il_lda = numroc(8,8,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(8,8,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    pzmtxcpy(8, 6, 1, 1, zx1,  desc_x, zxs1);
    pdcft2(zx1, zy1, 8, 6, -1, 1.0/48.0, ictx, ip);
    pzmtchk(modname,8, 6, 1, 1, zy1, desc_y, zresy1, tol, ierr, ilerr,
            error, ifail, strlen(modname));
  }
 
   ip[0]  = 1;
   ip[19] = 4;
   ip[20] = 12;
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 4;
  desc_x[N_]    = 4;
  desc_x[MB_]   = 4;
  desc_x[NB_]   = 8/(2*npcol);
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  desc_x[LLD_]  = 4;
 
  desc_y[DT_]   = BLOCK_CYCLIC_2D;
  desc_y[M_]    = 8;
  desc_y[N_]    = 4;
  desc_y[MB_]   = 8;
  desc_y[NB_]   = 4/npcol;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  desc_y[LLD_]  = 12;
  if (iam < 2 && nnodes > 1 )
  {
    modname = "PDCRFT2 ";
    pzmtxcpy(4, 4, 1, 1, zx3,  desc_x, zxs3);
    pdcrft2(zx3, dy3, 8, 4, -1, 1.0/32.0, ictx, ip);
    pdmtchk(modname,8, 4, 1, 1, dy3, desc_y, dresy3, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    ip[19] = 12;
    ip[20] = 4;
    modname = "PDRCFT2 ";
    pdrcft2(dy3, zx3, 8, 4, 1, 1.0, ictx, ip);
    pzmtchk(modname,4, 4, 1, 1, zx3, desc_x, zxs3, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    blacs_gridexit(ictx);
  }
  else if (iam == 0)
  {
    modname = "PDCRFT2 ";
    pzmtxcpy(4, 4, 1, 1, zx3,  desc_x, zxs3);
    pdcrft2(zx3, dy3, 8, 4, -1, 1.0/32.0, ictx, ip);
    pdmtchk(modname,12, 4, 1, 1, dy3, desc_y, dresy3, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    ip[19] = 12;
    ip[20] = 4;
    modname = "PDRCFT2 ";
    pdrcft2(dy3, zx3, 8, 4, 1, 1.0, ictx, ip);
    pzmtchk(modname,4, 4, 1, 1, zx3, desc_x, zxs3, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    blacs_gridexit(ictx);
  }
 
}
 
/*********************************************************************
*          PARALLEL THREE DIMENSIONAL SHORT PRECISION FFT SUBROUTINES*
*********************************************************************/
 
void psfft3(int *ifail)
{
// PSCFT3    START : 2,
 
   int i, iam, nnodes,myrow,mycol,ip[40];
   int desc_x[DLEN_], desc_y[DLEN_], il_lda;
   int ictx, nprow, npcol,ierr=5,ilerr;
   double error=0.0, tol=0.0;
   char modn[9], *modname=modn;
   char order[]="R";
   _ESVPCM cx2[32], cy2[32],cxs2[32],cresy2[32];
   _ESVPCM cx4[64], cxs4[64];
   float  sy4[128], sresy4[128];
 
   ip[0]  = 1;
   ip[1]  = 1;
   ip[19] = 0;
   ip[20] = 0;
   ip[21] = 0;
   ip[22] = 0;
 
   cxs2[0] =_ESVPCM(1.0,0.0);
   for (i=1; i<32; i++)
    cxs2[i] =_ESVPCM(0.0,0.0);
 
   for (i=0; i<32; i++)
    cresy2[i] =_ESVPCM(1.0,0.0);
 
    sresy4[0] = 1.0;
   for (i=1; i<128; i++)
    sresy4[i] = 0.0;
 
   for (i=0; i<64; i++)
    cxs4[i] =_ESVPCM(1.0,0.0);
 
    cxs4[0] =_ESVPCM(1.0,1.0);
    cxs4[2] =_ESVPCM(1.0,1.0);
    cxs4[8] =_ESVPCM(1.0,1.0);
    cxs4[10] =_ESVPCM(1.0,1.0);
 
  blacs_pinfo(iam,nnodes);
  if (nnodes > 1)
   {
     nprow = 1;
     npcol = 2;
   }
  else
   {
     nprow = 1;
     npcol = 1;
   }
  blacs_get(0,0,ictx);
  blacs_gridinit(ictx,order,nprow,npcol,strlen(order));
  blacs_gridinfo(ictx,nprow,npcol,myrow,mycol);
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 4*4;     // n1*n2
  desc_x[N_]    = 2;       // n3
  desc_x[MB_]   = 4*4;     // n1*n2
  desc_x[NB_]   = 2/npcol; // n3/npcol
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
 
  desc_y[DT_]   = BLOCK_CYCLIC_2D;
  desc_y[M_]    = 4*4;     // n1*n2
  desc_y[N_]    = 2;       // n3
  desc_y[MB_]   = 4*4;     // n1*n2
  desc_y[NB_]   = 2/npcol; // n3/npcol
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
 
  modname = "PSCFT3  ";
  if (iam < 2 && nnodes > 1 )
  {
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    pcmtxcpy(16, 2, 1, 1, cx2,  desc_x, cxs2);
    pscft3(cx2, cy2, 4, 4, 2, 1, 1.0, ictx, ip);
    pcmtchk(modname,16, 2, 1, 1, cy2, desc_y, cresy2, tol, ierr, ilerr,
            error, ifail, strlen(modname));
  }
  else if (iam == 0)
  {
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    pcmtxcpy(16, 2, 1, 1, cx2,  desc_x, cxs2);
    pscft3(cx2, cy2, 4, 4, 2, 1, 1.0, ictx, ip);
    pcmtchk(modname,16, 2, 1, 1, cy2, desc_y, cresy2, tol, ierr, ilerr,
            error, ifail, strlen(modname));
  }
 
   ip[0]  = 1;
   ip[19] = 4;
   ip[20] = 4;
   ip[21] = 8;
   ip[22] = 4;
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 4*4;     // n3*n2
  desc_x[N_]    = 4;       // n1/2
  desc_x[MB_]   = 4*4;     // n3*n2
  desc_x[NB_]   = 4/npcol; // n1/(2*npcol)
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
 
  desc_y[DT_]   = BLOCK_CYCLIC_2D;
  desc_y[M_]    = 8*4;     // n1*n2
  desc_y[N_]    = 4;       // n3
  desc_y[MB_]   = 8*4;     // n1*n2
  desc_y[NB_]   = 4/npcol; // n3/npcol
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
 
  if (iam < 2 && nnodes > 1 )
  {
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(32,32,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    modname = "PSCRFT3 ";
    pcmtxcpy(16, 4, 1, 1, cx4, desc_x, cxs4);
    pscrft3(cx4, sy4, 8, 4, 4, -1, 1.0/(8*4*4), ictx, ip);
    psmtchk(modname,32, 4, 1, 1, sy4, desc_y, sresy4, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    ip[19] = 8;
    ip[21] = 4;
    modname = "PSRCFT3 ";
    psrcft3(sy4, cx4, 8, 4, 4, 1, 1.0, ictx, ip);
    pcmtchk(modname,16, 4, 1, 1, cx4, desc_x, cxs4, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    blacs_gridexit(ictx);
  }
  else if (iam == 0)
  {
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(32,32,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    modname = "PSCRFT3 ";
    pcmtxcpy(16, 4, 1, 1, cx4, desc_x, cxs4);
    pscrft3(cx4, sy4, 8, 4, 4, -1, 1.0/(8*4*4), ictx, ip);
    psmtchk(modname,32, 4, 1, 1, sy4, desc_y, sresy4, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    ip[19] = 8;
    ip[21] = 4;
    modname = "PSRCFT3 ";
    psrcft3(sy4, cx4, 8, 4, 4, 1, 1.0, ictx, ip);
    pcmtchk(modname,16, 4, 1, 1, cx4, desc_x, cxs4, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    blacs_gridexit(ictx);
  }
}
 
/*********************************************************************
*          PARALLEL THREE DIMENSIONAL LONG PRECISION FFT SUBROUTINES *
*********************************************************************/
 
void pfft3(int *ifail)
{
// PDCFT3    START : 2,
 
   int i, iam, nnodes,myrow,mycol,ip[40];
   int desc_x[DLEN_], desc_y[DLEN_], il_lda;
   int ictx, nprow, npcol,ierr=5,ilerr;
   double error=0.0, tol=0.0;
   char modn[9], *modname=modn;
   char order[]="R";
   _ESVPCOM zx2[32], zy2[32],zxs2[32],zresy2[32];
   _ESVPCOM zx4[64], zxs4[64];
   double dy4[128], dresy4[128];
 
   ip[0]  = 1;
   ip[1]  = 1;
   ip[19] = 0;
   ip[20] = 0;
   ip[21] = 0;
   ip[22] = 0;
 
   zxs2[0] = _ESVPCOM(1.0,0.0);
   for (i=1; i<32; i++)
    zxs2[i] = _ESVPCOM(0.0,0.0);
 
   for (i=0; i<32; i++)
    zresy2[i] = _ESVPCOM(1.0,0.0);
 
    dresy4[0] = 1.0;
   for (i=1; i<128; i++)
    dresy4[i] = 0.0;
 
   for (i=0; i<64; i++)
    zxs4[i] = _ESVPCOM(1.0,0.0);
 
    zxs4[0] = _ESVPCOM(1.0,1.0);
    zxs4[2] = _ESVPCOM(1.0,1.0);
    zxs4[8] = _ESVPCOM(1.0,1.0);
    zxs4[10] = _ESVPCOM(1.0,1.0);
 
  blacs_pinfo(iam,nnodes);
  if (nnodes > 1)
   {
     nprow = 1;
     npcol = 2;
   }
  else
   {
     nprow = 1;
     npcol = 1;
   }
  blacs_get(0,0,ictx);
  blacs_gridinit(ictx,order,nprow,npcol,strlen(order));
  blacs_gridinfo(ictx,nprow,npcol,myrow,mycol);
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 4*4;     // n1*n2
  desc_x[N_]    = 2;       // n3
  desc_x[MB_]   = 4*4;     // n1*n2
  desc_x[NB_]   = 2/npcol; // n3/npcol
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
 
  desc_y[DT_]   = BLOCK_CYCLIC_2D;
  desc_y[M_]    = 4*4;     // n1*n2
  desc_y[N_]    = 2;       // n3
  desc_y[MB_]   = 4*4;     // n1*n2
  desc_y[NB_]   = 2/npcol; // n3/npcol
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
 
  modname = "PDCFT3  ";
  if (iam < 2 && nnodes > 1 )
  {
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    pzmtxcpy(16, 2, 1, 1, zx2,  desc_x, zxs2);
    pdcft3(zx2, zy2, 4, 4, 2, 1, 1.0, ictx, ip);
    pzmtchk(modname,16, 2, 1, 1, zy2, desc_y, zresy2, tol, ierr, ilerr,
            error, ifail, strlen(modname));
  }
  else if (iam == 0)
  {
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    pzmtxcpy(16, 2, 1, 1, zx2,  desc_x, zxs2);
    pdcft3(zx2, zy2, 4, 4, 2, 1, 1.0, ictx, ip);
    pzmtchk(modname,16, 2, 1, 1, zy2, desc_y, zresy2, tol, ierr, ilerr,
            error, ifail, strlen(modname));
  }
 
   ip[0]  = 1;
   ip[19] = 4;
   ip[20] = 4;
   ip[21] = 8;
   ip[22] = 4;
 
  desc_x[DT_]   = BLOCK_CYCLIC_2D;
  desc_x[M_]    = 4*4;     // n3*n2
  desc_x[N_]    = 4;       // n1/2
  desc_x[MB_]   = 4*4;     // n3*n2
  desc_x[NB_]   = 4/npcol; // n1/(2*npcol)
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
 
  desc_y[DT_]   = BLOCK_CYCLIC_2D;
  desc_y[M_]    = 8*4;     // n1*n2
  desc_y[N_]    = 4;       // n3
  desc_y[MB_]   = 8*4;     // n1*n2
  desc_y[NB_]   = 4/npcol; // n3/npcol
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
 
  if (iam < 2 && nnodes > 1 )
  {
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(32,32,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    modname = "PDCRFT3 ";
    pzmtxcpy(16, 4, 1, 1, zx4, desc_x, zxs4);
    pdcrft3(zx4, dy4, 8, 4, 4, -1, 1.0/(8*4*4), ictx, ip);
    pdmtchk(modname,32, 4, 1, 1, dy4, desc_y, dresy4, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    ip[19] = 8;
    ip[21] = 4;
    modname = "PDRCFT3 ";
    pdrcft3(dy4, zx4, 8, 4, 4, 1, 1.0, ictx, ip);
    pzmtchk(modname,16, 4, 1, 1, zx4, desc_x, zxs4, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    blacs_gridexit(ictx);
  }
  else if (iam == 0)
  {
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(32,32,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    modname = "PDCRFT3 ";
    pzmtxcpy(16, 4, 1, 1, zx4, desc_x, zxs4);
    pdcrft3(zx4, dy4, 8, 4, 4, -1, 1.0/(8*4*4), ictx, ip);
    pdmtchk(modname,32, 4, 1, 1, dy4, desc_y, dresy4, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    ip[19] = 8;
    ip[21] = 4;
    modname = "PDRCFT3 ";
    pdrcft3(dy4, zx4, 8, 4, 4, 1, 1.0, ictx, ip);
    pzmtchk(modname,16, 4, 1, 1, zx4, desc_x, zxs4, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    blacs_gridexit(ictx);
  }
}
 
/*********************************************************************
*          PARALLEL RANDOM NUMBER GENERATION SUBROUTINE              *
*********************************************************************/
 
void prnd(int *ifail)
{
// PDURNG    START :
 
   int iam, nnodes,myrow,mycol;
   int desc_x[DLEN_], il_lda;
   int ictx, nprow, npcol,ierr=5, ilerr;
   char modn[9], *modname=modn;
   char order[]="R";
   double dx1r[30], seed, error=0.0, tol=0.0;
   double dresx1r[30]= {
     0.6838215161, 0.0588744078, 0.3918552508, 0.7559946530, 0.5577643014,
     0.0013338017, 0.0568559327, 0.3310630362, 0.3473397944, 0.6494290203,
     0.3861448762, 0.4572248550, 0.8925181341, 0.0745487482, 0.9123793668,
     0.1128094991, 0.8575476050, 0.7564809018, 0.0469933644, 0.8894576840,
     0.1677757661, 0.5049527226, 0.9997259245, 0.6962694873, 0.6718965980,
     0.2714721560, 0.5664184066, 0.4646848657, 0.9824425397, 0.0224404825 };
 
  blacs_pinfo(iam,nnodes);
  if (nnodes > 4)
   {
     nprow = 1;
     npcol = 5;
   }
  else
   {
     nprow = 1;
     npcol = 1;
   }
  blacs_get(0,0,ictx);
  blacs_gridinit(ictx,order,nprow,npcol,strlen(order));
  blacs_gridinfo(ictx,nprow,npcol,myrow,mycol);
 
  tol = 0.00000001;
  seed = 31415926535897.0;
  modname = "PDURNG  ";
  if ( iam < 5 && nnodes > 4 )
  {
    desc_x[DT_]   = BLOCK_CYCLIC_2D;
    desc_x[M_]    = 1;  // consider X is distributed row wise to Check numbers
    desc_x[N_]    = 30;
    desc_x[MB_]   = 1;
    desc_x[NB_]   = 3;
    desc_x[RSRC_] = 0;
    desc_x[CSRC_] = 0;
    desc_x[CTXT_] = ictx;
    il_lda = numroc(1,1,myrow,0,nprow);
    desc_x[LLD_]  = MAX(1,il_lda);
 
    pdurng(seed, 30, 3, dx1r, 0, ictx);
    pdvtchk(modname,30, 1, 1, dx1r, desc_x, dresx1r, 1, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    blacs_gridexit(ictx);
  }
  else if (iam == 0)
  {
    desc_x[DT_]   = BLOCK_CYCLIC_2D;
    desc_x[M_]    = 30; // consider X is ditributed column wise to check nos.
    desc_x[N_]    = 1;
    desc_x[MB_]   = 3;
    desc_x[NB_]   = 1;
    desc_x[RSRC_] = 0;
    desc_x[CSRC_] = 0;
    desc_x[CTXT_] = ictx;
    il_lda = numroc(30,3,myrow,0,nprow);
    desc_x[LLD_]  = MAX(1,il_lda);
 
    pdurng(seed, 30, 3, dx1r, 0, ictx);
    pdvtchk(modname,30, 1, 1, dx1r, desc_x, dresx1r, 1, tol, ierr, ilerr,
            error, ifail, strlen(modname));
    blacs_gridexit(ictx);
  }
}
 
/*********************************************************************
*          PARALLEL UTILITY SUBROUTINE                               *
*********************************************************************/
 
void putil(int *ifail)
{
 
  int iam, nnodes,myrow,mycol;
  int ictx, nprow, npcol,ipsl;
  char order[]="R";
  int m,n,mb,nb,irsrc,crsrc,lld,info,ifails;
  int ix,ceil,inum,idenom,i1,i2,lcm;
  int iic,jjc,ic,jc,m_c,mb_c,nb_c,rsrc_c,csrc_c,icrow,iccol;
  int testvalue,rc,li;
 
  int desc[DLEN_];
  int desc_a[DLEN_]={ 1,0,9,9,4,4,0,0,0 };
 
// IPESSL    START :
 
  blacs_pinfo(iam,nnodes);
 
  nprow = 1;
  npcol = nnodes;
 
  blacs_get(0,0,ictx);
  blacs_gridinit(ictx,order,1,nnodes,strlen(order));
  blacs_gridinfo(ictx,nprow,npcol,myrow,mycol);
 
  ipsl = ipessl();
  if (ipsl < 3010000) {
    cout << " IPESSL FAIL\n";
    *ifail = *ifail + 1;
  }
 
// IPESSL    END
// DESCINIT  START :
 
  m     = 9;
  n     = 9;
  mb    = 4;
  nb    = 4;
  irsrc = 0;
  crsrc = 0;
  lld   = numroc(m,mb,myrow,irsrc,nprow);
  lld   = MAX(1, lld);
 
  desc_a[1] = ictx;
  desc_a[8] = lld;
 
  ifails = 0;
 
  descinit(desc,m,n,mb,nb,irsrc,crsrc,ictx,lld,info);
 
  if( info != 0 ) {
    if (!iam) cout << " INFO from DESCINIT = " << info << endl;
    *ifail = *ifail + 1;
  } else {
    for(ix=0;ix<9;ix++) {
      if(desc[ix] != desc_a[ix] ) {
        cout << " ERROR IN DESCINIT:  ELEMENT " << ix << " = "
             << desc[ix] << " instead of " << desc_a[ix] << endl;
        ifails = 1;
      }
    }
  }
 
  if(ifails == 1) *ifail = *ifail + 1;
 
// DESCINIT  END
// DESCSET   START :
 
  ifails = 0;
 
  descset(desc,m,n,mb,nb,irsrc,crsrc,ictx,lld);
 
  for(ix=0;ix<9;ix++) {
    if(desc[ix] != desc_a[ix] ) {
      cout << " ERROR IN DESCSET:  ELEMENT " << ix << " = "
           << desc[ix] << " instead of " << desc_a[ix] << endl;
      ifails = 1;
    }
  }
 
  if(ifails == 1) *ifail = *ifail + 1;
 
// DESCSET   END
// ICEIL     START :
 
  inum   = 5;
  idenom = 3;
 
  ceil   = iceil(inum, idenom);
 
  if( ceil != 2 ) {
    if(!iam) cout << " ERROR IN ICEIL:  RESULT: " << ceil << " EXPECTED: 2"
                  << endl;
    *ifail = *ifail + 1;
  }
 
// ICEIL     END
// ILCM      START :
 
  i1  = 5;
  i2  = 3;
 
  lcm = ilcm(i1, i2);
 
  if( lcm != 15 ) {
    if(!iam) cout << " ERROR IN ILCM:  RESULT: " << lcm << " EXPECTED: 15"
                     << endl;
    *ifail = *ifail + 1;
  }
 
// ILCM      END
// INDXG2L   START :
 
  ic     = 6;
  jc     = 9;
  mb_c   = 4;
  nb_c   = 4;
  rsrc_c = 0;
  csrc_c = 0;
 
  ifails = 0;
 
  iic = indxg2l(ic,mb_c,0,rsrc_c,nprow);
  jjc = indxg2l(jc,nb_c,0,csrc_c,npcol);
 
  testvalue = mb_c*((ic-1)/(mb_c*nprow))+( (ic-1) % mb_c)+1;
 
  if(iic != testvalue) {
    if(!iam) cout << " ERROR IN INDXG2L:  RESULT: " << iic << " EXPECTED: " <<
                     testvalue << endl;
    ifails = 1;
  }
 
  testvalue = nb_c*((jc-1)/(nb_c*npcol))+( (jc-1) % nb_c)+1;
 
  if(jjc != testvalue) {
    if(!iam) cout << " ERROR IN INDXG2L:  RESULT: " << jjc << " EXPECTED: " <<
                     testvalue << endl;
    ifails = 1;
  }
 
  if(ifails == 1) *ifail = *ifail + 1;
 
// INDXG2L   END
// INDXG2P   START :
 
  ic     = 6;
  jc     = 9;
  mb_c   = 4;
  nb_c   = 4;
  rsrc_c = 0;
  csrc_c = 0;
 
  ifails = 0;
 
  iic = indxg2p(ic,mb_c,0,rsrc_c,nprow);
  jjc = indxg2p(jc,nb_c,0,csrc_c,npcol);
 
  testvalue = (rsrc_c + (ic - 1) / mb_c) % nprow;
 
  if(iic != testvalue) {
    if(!iam) cout << " ERROR IN INDXG2P:  RESULT: " << iic << " EXPECTED: " <<
                     testvalue << endl;
    ifails = 1;
  }
 
  testvalue = (csrc_c + (jc - 1) / nb_c) % npcol;
 
  if(jjc != testvalue) {
    if(!iam) cout << " ERROR IN INDXG2P:  RESULT: " << jjc << " EXPECTED: " <<
                     testvalue << endl;
    ifails = 1;
  }
 
  if(ifails == 1) *ifail = *ifail + 1;
 
// INDXG2P   END
// INDXL2G   START :
 
  ic        = 1;
  jc        = 1;
 
  mb_c      = 4;
  nb_c      = 4;
  rsrc_c    = 0;
  csrc_c    = 0;
 
  ifails    = 0;
 
  iic = indxl2g(ic,mb_c,myrow,rsrc_c,nprow);
  jjc = indxl2g(jc,nb_c,mycol,csrc_c,npcol);
 
  testvalue = nprow*mb_c*((ic-1)/mb_c) + (ic-1) % mb_c + ((nprow+myrow-rsrc_c) %
              nprow)*mb_c + 1;
 
  if(iic != testvalue ) {
    if(!iam) cout << " ERROR IN INDXL2G:  RESULT: " << iic
                  << " EXPECTED: " << testvalue << endl;
    ifails = 1;
  }
 
  testvalue = npcol*nb_c*((jc-1)/nb_c) + (jc-1) % nb_c + ((npcol+mycol-csrc_c) %
              npcol)*nb_c + 1;
 
  if(jjc != testvalue) {
    if(!iam) cout << " ERROR IN INDXL2G:  RESULT: " << jjc
                  << " EXPECTED: " << testvalue << endl;
    ifails = 1;
  }
 
  if(ifails == 1) *ifail = *ifail + 1;
 
// INDXL2G   END
// INFOG1L   START :
 
  ic          = 5;
  jc          = 7;
  m_c         = 9;
 
  desc[DT_]   = 1;
  desc[CTXT_] = ictx;
  desc[M_]    = 9;
  desc[N_]    = 9;
  desc[MB_]   = 4;
  desc[NB_]   = 4;
  desc[RSRC_] = 0;
  desc[CSRC_] = 0;
  desc[LLD_]  = MAX(1,numroc(m_c,mb_c,myrow,rsrc_c,nprow));
 
  ifails      = 0;
 
  infog1l(ic,desc[MB_],nprow,myrow,desc[RSRC_],iic,icrow);
  infog1l(jc,desc[NB_],npcol,mycol,desc[CSRC_],jjc,iccol);
 
  testvalue = ( ((ic-1) / desc[MB_]) + desc[RSRC_] ) % nprow;
 
  if(icrow != testvalue) {
    cout << " ERROR IN INFOG1L:  RESULT: " << icrow << " EXPECTED: " <<
            testvalue << endl;
    ifails = 1;
  }
 
  testvalue = ( ((jc-1) / desc[NB_]) + desc[CSRC_] ) % npcol;
 
  if(iccol != testvalue) {
    cout << " ERROR IN INFOG1L:  RESULT: " << iccol << " EXPECTED: " <<
            testvalue << endl;
    ifails = 1;
  }
 
  /*  Test iic  */
 
  rc = ( ((ic-1)/desc[MB_]) + desc[RSRC_]) % nprow;
  li = ( ((ic-1)/desc[MB_]) / nprow + 1) * desc[MB_] + 1;
 
  if( ((myrow+nprow-desc[RSRC_]) % nprow) >= ( ((ic-1)/desc[MB_]) % nprow) ) {
     if(myrow == rc) li = li + (ic-1) % desc[MB_];
     li = li - desc[MB_];
  }
 
  testvalue = li;
 
  if(iic != testvalue ) {
    cout << " ERROR IN INFOG1L:  RESULT: " << iic << " EXPECTED: " <<
            testvalue << endl;
    ifails = 1;
  }
 
  /*  Test jjc  */
 
  rc = ( ((jc-1)/desc[NB_]) + desc[CSRC_]) % npcol;
  li = ( ((jc-1)/desc[NB_]) / npcol +1) * desc[NB_] + 1;
 
  if( ((mycol+npcol-desc[CSRC_]) % npcol) >= ( ((jc-1)/desc[NB_]) % npcol) ) {
    if(mycol == rc) li = li + (jc-1) % desc[NB_];
    li = li - desc[NB_];
  }
 
  testvalue = li;
 
  if(jjc != testvalue) {
    cout << " ERROR IN INFOG1L:  RESULT: " << jjc << " EXPECTED: " <<
            testvalue << endl;
    ifails = 1;
  }
 
  if(ifails == 1) *ifail = *ifail + 1;
 
// INFOG1L   END
// INFOG2L   START :
 
  ic          = 5;
  jc          = 7;
 
  desc[DT_]   = 1;
  desc[CTXT_] = ictx;
  desc[M_]    = 9;
  desc[N_]    = 9;
  desc[MB_]   = 4;
  desc[NB_]   = 4;
  desc[RSRC_] = 0;
  desc[CSRC_] = 0;
  desc[LLD_]  = MAX(1,numroc(m_c,mb_c,myrow,rsrc_c,nprow));
 
  ifails      = 0;
 
  infog2l(ic,jc,desc,nprow,npcol,myrow,mycol,iic,jjc,icrow,iccol);
 
  testvalue =  ( ( (ic-1) / desc[MB_] ) + desc[RSRC_] ) % nprow;
 
  if(icrow != testvalue) {
    cout << " ERROR IN INFOG2L:  RESULT: " << icrow << " EXPECTED: " <<
            testvalue << endl;
    ifails = 1;
  }
 
  testvalue = ( ( (jc-1) / desc[NB_] ) + desc[CSRC_] ) % npcol;
 
  if(iccol != testvalue) {
    cout << " ERROR IN INFOG2L:  RESULT: " << iccol << " EXPECTED: " <<
            testvalue << endl;
    ifails = 1;
  }
 
  rc = ( ( (ic-1) / desc[MB_] ) + desc[RSRC_]) % nprow;
  testvalue = ( ( (ic-1) / desc[MB_] ) / nprow + 1 ) * desc[MB_] + 1;
 
  if( (myrow+nprow-desc[RSRC_]) % nprow >= ( ( (ic-1) / desc[MB_] ) % nprow) ) {
    if( myrow == rc ) testvalue = testvalue + (ic-1) % desc[MB_];
    testvalue = testvalue - desc[MB_];
  }
 
  if(iic != testvalue ) {
    cout << " ERROR IN INFOG2L:  RESULT: " << iic << " EXPECTED: " <<
            testvalue << endl;
    ifails = 1;
  }
 
  rc = ( ( (jc-1) / desc[NB_] ) + desc[CSRC_]) % npcol;
  testvalue = ( ( (jc-1) / desc[NB_] ) / npcol + 1 ) * desc[NB_] + 1;
 
  if( (mycol+npcol-desc[CSRC_]) % npcol >= ( ( (jc-1) / desc[NB_] ) % npcol) ) {
    if( mycol == rc ) testvalue = testvalue + (jc-1) % desc[NB_];
    testvalue = testvalue - desc[NB_];
  }
  if(jjc != testvalue ) {
    cout << " ERROR IN INFOG2L:  RESULT: " << jjc << " EXPECTED: " <<
            testvalue << endl;
    ifails = 1;
  }
 
  if(ifails == 1) *ifail = *ifail + 1;
 
// INFOG2L   END
 
  blacs_gridexit(ictx);
}
