/*******************************************************************************
*                                                                              *
*    LICENSED MATERIALS - PROPERTY OF IBM                                      *
*                                                                              *
*    "RESTRICTED MATERIALS OF IBM"                                             *
*                                                                              *
*    5799-FQC                                                                  *
*    5765-422                                                                  *
*    5765-C41                                                                  *
*    5765-F84                                                                  *
*    5765-G18                                                                  *
*                                                                              *
*    (C) COPYRIGHT IBM CORP. 1995, 2006.  All rights reserved.                 *
*                                                                              *
*    US Government Users Restricted Rights - Use, duplication or disclosure    *
*    restricted by GSA ADP Schedule Contract with IBM Corp.                    *
*                                                                              *
********************************************************************************
*                                                                              *
*                                                                              *
*     MODULE NAME = esvcpivp.c                                                 *
*                                                                              *
*     MODULE-DESCRIPTION :                                                     *
*                                                                              *
*   FUNCTION = Test example from Parallel ESSL library routines                *
*                                                                              *
*       This program runs one example for each Parallel ESSL library           *
*       user callable routine and checks the results as expected.              *
*       If results match as expected, it returns return code of zero,          *
*       else it prints which routine gives wrong answers and returns           *
*       return code of 8.                                                      *
*                                                                              *
*     NOTES =                                                                  *
*                                                                              *
*     PROCESSOR = C Compiler                                                   *
*                                                                              *
*     Revision:                                                                *
*               Parallel ESSL Version 1.1.0.0                 10/95            *
*               Parallel ESSL Version 1.2.0.0                 05/96            *
*               Parallel ESSL Version 1.2.1.0                 09/96            *
*               Parallel ESSL Version 2.1.0.0                 12/97            *
*               Parallel ESSL Version 2.1.1.0                 11/98            *
*               Parallel ESSL Version 2.1.2.0                 10/99            *
*               Parallel ESSL Version 2.2.0.0                 07/00            *
*               Parallel ESSL Version 2.3.0.0                 12/01            *
*               Parallel ESSL Version 3.1.0.0                 09/03            *
*               Parallel ESSL Version 3.3.0.0                 10/06            *
*                                                                              *
********************************************************************************
*                                                                              *
*     SUBPROGRAMS CALLED BY MAIN :                                             *
*                                                                              *
*       PBLAS   =  CALL TO ALL  Parallel Basic Linear Algebra                  *
*                                      Subprograms                             *
*       PLEQ    =  CALL TO ALL  Parallel Linear Algebraic Equations            *
*       PSPRS   =  CALL TO ALL  Parallel Sparse Algebraic Equations            *
*       PEIG    =  CALL TO ALL  Parallel Eigen System Analysis                 *
*       PFFT2   =  CALL TO ALL  Parallel Two dimensional FFT                   *
*       PFFT3   =  CALL TO ALL  Parallel Three dimensional FFT                 *
*       PRND    =  CALL TO ALL  Parallel Random Number Geneartion              *
*                                      Subroutine                              *
*       PUTIL   =  CALL TO ALL  Parallel Utility Subroutines                   *
*                                                                              *
*******************************************************************************/
#include <math.h>
#include <string.h>
#include <pessl.h>
 
#define MAX(a,b) ((a > b) ? (a) : (b))
#define MIN(a,b) ((a < b) ? (a) : (b))
#define CSET(x, a, b)   (RE(x) = a, IM(x) = b)
 
#define DLEN_    9
 
#define DT_      0
#define CTXT_    1
#define M_       2
#define N_       3
#define MB_      4
#define NB_      5
#define RSRC_    6
#define CSRC_    7
#define LLD_     8
#define BLOCK_CYCLIC_2D    1
 
#define DLEN_1D_    7
 
#define DT_1D_      0
#define CTXT_1D_    1
#define M_1D_       2
#define N_1D_       2
#define MB_1D_      3
#define NB_1D_      3
#define RSRC_1D_    4
#define CSRC_1D_    4
#define LLD_1D_     5
#define RSVD_1D_    6
#define BLOCK_ROW_1D        502
#define BLOCK_COLUMN_1D     501
 
 
void pblas(int *), pleq(int *), psprs(int *), peig(int *);
void psfft2(int *), psfft3(int *), pfft2(int *), pfft3(int *);
void prnd(int *), putil(int *);
 
/* The following temporary variables are used to create fortran linkage */
 
int  __esvi1,__esvi2,__esvi3,__esvi4,__esvi5;
int  __esvi6,__esvi7,__esvi8,__esvi9,__esvi10;
int  __esvi11,__esvi12;
double __esvd1,__esvd2;
 
 
  void * malloc(size_t);
  void blacs_pinfo(  int *,  int *);
  void blacs_exit(  int *);
  void blacs_abort(  int *,  int *);
  void blacs_gridexit(  int *);
  void blacs_get(  int *,  int *,  int *);
  void blacs_gridinit(  int *,  char *, int *,  int *, size_t);
  void blacs_gridinfo(  int *,  int *,  int *,  int *,
                             int *);
  void igamx2d( int *, char *, char *,  int *,  int *, int *,
                int *, int *, int *,  int *, int *,
                     int *, size_t, size_t);
 
  void pdvtrcpy( int *,  int *,  int *, double *, int *,
                double *,  int *);
  void pzvtrcpy( int *,  int *,  int *, dcmplx *, int *,
                void   *,  int *);
  void psmtxcpy( int *,  int *,  int *,  int *, void *,
                int *, void *);
  void pdmtxcpy( int *,  int *,  int *,  int *, void *,
                int *, void *);
  void pcmtxcpy( int *,  int *,  int *,  int *, void *,
                int *, void *);
  void pzmtxcpy( int *,  int *,  int *,  int *, void *,
                int *, void *);
  void pdvtchk(char *,  int *,  int *,  int *, double *, int *,
               double *,  int *,  double *,  int *,
                int *,  double *, int *, size_t);
  void pzvtchk(char *,  int *,  int *,  int *, dcmplx *, int *,
               dcmplx *,  int *,  double *,  int *,
                int *,  double *, int *, size_t);
  void psmtchk(char *, int *,  int *,  int *,  int *,
               void *, int *, void *,  double *,  int *,
                int *, double *, int *, size_t);
  void pdmtchk(char *, int *,  int *,  int *,  int *,
               void *, int *, void *,  double *,  int *,
                int *,  double *, int *, size_t);
  void pcmtchk(char *,  int *,  int *,  int *,  int *,
               void *, int *, void *,  double *,  int *, int *,
                double *, int *, size_t);
  void pzmtchk(char *,  int *,  int *,  int *,  int *,
               void *, int *, void *,  double *,  int *, int *,
                double *, int *, size_t);
  void pdgspchk(int *,  int *,  int *, void *, int *, int *, int *,
                void *,int *,int *,int *,double *,void *,int *,int *,
                int *,void *,int *,int *,int *,
                double *, double *, int *, double *, int *);
 
#define blacs_pinfo(x1,x2)         \
                           (blacs_pinfo(x1,x2))
#define blacs_exit(i1)     (__esvi1=i1,  \
                            blacs_exit(&__esvi1))
#define blacs_abort(i1,x1) (__esvi1=i1,  \
                            blacs_abort(&__esvi1,x1))
#define blacs_gridexit(i1) (__esvi1=i1,   \
                            blacs_gridexit(&__esvi1))
#define blacs_get(i1,i2,x1) (__esvi1=i1,__esvi2=i2,  \
                            blacs_get(&__esvi1,&__esvi2,x1))
#define blacs_gridinit(x1,ch1,i1,i2) (__esvi1=i1,__esvi2=i2,  \
              blacs_gridinit(x1,ch1,&__esvi1,&__esvi2,strlen(ch1)))
#define blacs_gridinfo(i1,x1,x2,x3,x4) (__esvi1=i1,     \
                            blacs_gridinfo(&__esvi1,x1,x2,x3,x4))
#define igamx2d(i1,ch1,ch2,i2,i3,x1,i4,x2,x3,i5,i6,i7)       \
                           (__esvi1=i1,__esvi2 =i2,__esvi3=i3, \
                            __esvi4=i4,__esvi5 =i5,__esvi5=i5, \
                            __esvi6=i6,__esvi7 =i7,            \
        igamx2d(&__esvi1,ch1,ch2,&__esvi2,&__esvi3,x1,&__esvi4,x2,x3,\
                &__esvi5,&__esvi6,&__esvi7,strlen(ch1),strlen(ch2)))
#define pdvtrcpy(i1,i2,i3,x1,x2,x3,i4)  (__esvi1=i1,__esvi2=i2, \
                                         __esvi3=i3,__esvi4=i4, \
          pdvtrcpy(&__esvi1,&__esvi2,&__esvi3,x1,x2,x3,&__esvi4))
#define pzvtrcpy(i1,i2,i3,x1,x2,x3,i4)  (__esvi1=i1,__esvi2=i2, \
                                         __esvi3=i3,__esvi4=i4, \
          pzvtrcpy(&__esvi1,&__esvi2,&__esvi3,x1,x2,x3,&__esvi4))
#define psmtxcpy(i1,i2,i3,i4,x1,x2,x3)  (__esvi1=i1,__esvi2=i2, \
                                         __esvi3=i3,__esvi4=i4, \
          psmtxcpy(&__esvi1,&__esvi2,&__esvi3,&__esvi4,x1,x2,x3))
#define pdmtxcpy(i1,i2,i3,i4,x1,x2,x3)  (__esvi1=i1,__esvi2=i2, \
                                         __esvi3=i3,__esvi4=i4, \
          pdmtxcpy(&__esvi1,&__esvi2,&__esvi3,&__esvi4,x1,x2,x3))
#define pcmtxcpy(i1,i2,i3,i4,x1,x2,x3)  (__esvi1=i1,__esvi2=i2, \
                                         __esvi3=i3,__esvi4=i4, \
          pcmtxcpy(&__esvi1,&__esvi2,&__esvi3,&__esvi4,x1,x2,x3))
#define pzmtxcpy(i1,i2,i3,i4,x1,x2,x3)  (__esvi1=i1,__esvi2=i2, \
                                         __esvi3=i3,__esvi4=i4, \
          pzmtxcpy(&__esvi1,&__esvi2,&__esvi3,&__esvi4,x1,x2,x3))
#define pdvtchk(ch1,i1,i2,i3,x1,x2,x3,i4,d1,i5,x4,x5,x6)         \
                           (__esvi1=i1,__esvi2 =i2,__esvi3=i3, \
                            __esvi4=i4,__esvi5 =i5,__esvd1=d1, \
        pdvtchk(ch1,&__esvi1,&__esvi2,&__esvi3,x1,x2,x3,&__esvi4, \
                           &__esvd1,&__esvi5,x4,x5,x6,strlen(ch1)))
#define pzvtchk(ch1,i1,i2,i3,x1,x2,x3,i4,d1,i5,x4,x5,x6)         \
                           (__esvi1=i1,__esvi2 =i2,__esvi3=i3, \
                            __esvi4=i4,__esvi5 =i5,__esvd1=d1, \
        pzvtchk(ch1,&__esvi1,&__esvi2,&__esvi3,x1,x2,x3,&__esvi4, \
                           &__esvd1,&__esvi5,x4,x5,x6,strlen(ch1)))
#define psmtchk(ch1,i1,i2,i3,i4,x1,x2,x3,d1,i5,x4,x5,x6)         \
                           (__esvi1=i1,__esvi2 =i2,__esvi3=i3, \
                            __esvi4=i4,__esvi5 =i5,__esvd1=d1, \
        psmtchk(ch1,&__esvi1,&__esvi2,&__esvi3,&__esvi4,x1,x2,x3, \
                           &__esvd1,&__esvi5,x4,x5,x6,strlen(ch1)))
#define pdmtchk(ch1,i1,i2,i3,i4,x1,x2,x3,d1,i5,x4,x5,x6)         \
                           (__esvi1=i1,__esvi2 =i2,__esvi3=i3, \
                            __esvi4=i4,__esvi5 =i5,__esvd1=d1, \
        pdmtchk(ch1,&__esvi1,&__esvi2,&__esvi3,&__esvi4,x1,x2,x3, \
                           &__esvd1,&__esvi5,x4,x5,x6,strlen(ch1)))
#define pcmtchk(ch1,i1,i2,i3,i4,x1,x2,x3,d1,i5,x4,x5,x6)         \
                           (__esvi1=i1,__esvi2 =i2,__esvi3=i3, \
                            __esvi4=i4,__esvi5 =i5,__esvd1=d1, \
        pcmtchk(ch1,&__esvi1,&__esvi2,&__esvi3,&__esvi4,x1,x2,x3, \
                           &__esvd1,&__esvi5,x4,x5,x6,strlen(ch1)))
#define pzmtchk(ch1,i1,i2,i3,i4,x1,x2,x3,d1,i5,x4,x5,x6)         \
                           (__esvi1=i1,__esvi2 =i2,__esvi3=i3, \
                            __esvi4=i4,__esvi5 =i5,__esvd1=d1, \
        pzmtchk(ch1,&__esvi1,&__esvi2,&__esvi3,&__esvi4,x1,x2,x3, \
                           &__esvd1,&__esvi5,x4,x5,x6,strlen(ch1)))
#define pdgspchk(i1,i2,i3,x1,i4,i5,x2,x3,i6,i7,x4,d1,x5,i8,i9,x6, \
                 x7,i10,i11,x8,x9,x10,i12,x11,x12)               \
                           (__esvi1=i1,__esvi2 =i2,__esvi3=i3, \
                            __esvi4=i4,__esvi5 =i5,__esvi6=i6, \
                            __esvi7=i7,__esvi8 =i8,__esvi9=i9, \
                            __esvi10=i10,__esvi11 =i11,__esvi12=i12, \
                            __esvd1=d1, \
       pdgspchk(&__esvi1,&__esvi2,&__esvi3,x1,&__esvi4,&__esvi5,x2,x3, \
           &__esvi6,&__esvi7,x4,&__esvd1,x5,&__esvi8,&__esvi9,x6,x7,   \
           &__esvi10,&__esvi11,x8,x9,x10,&__esvi12,x11,x12))
  void part_block();
main()
{
    int ra[1],ca[1];
    int ifail[1],rc,ictx1,nprow,npcol,myrow,mycol,nnodes,iam;
    ifail[0] = 0;
 
    blacs_pinfo(&iam,&nnodes);
    nprow = 1;
    npcol = nnodes;
    blacs_get(0,0,&ictx1);
    blacs_gridinit(&ictx1,"R",nprow,npcol);
    blacs_gridinfo(ictx1,&nprow,&npcol,&myrow,&mycol);
 
    pblas(ifail);
    pleq(ifail);
    psprs(ifail);
    peig(ifail);
    psfft2(ifail);
    pfft2(ifail);
    psfft3(ifail);
    pfft3(ifail);
    prnd(ifail);
    putil(ifail);
 
    igamx2d(ictx1, "ALL", "1-tree", 1, 1, ifail, 1, ra, ca, 1, 0, 0);
    rc = 0;
    if(iam == 0)
    {
      if(ifail[0] == 0) {
          printf( " PARALLEL ESSL installed successfully\n");
      }
      else {
          printf( " NUMBER OF FAILURES = %4d\n",ifail[0]);
          rc = 8;
      }
    }
    blacs_gridexit(ictx1);
    blacs_exit(0);
    return(rc);
}
 
/*********************************************************************
*          PARALLEL BLAS SUBROUTINES                                 *
*********************************************************************/
 
void pblas(int *ifail)
{
/* PDTRMV, PZTRMV and PDTRSV, PZTRSV START  */
  int iam, nnodes,myrow,mycol;
  int desc_a[DLEN_], desc_x[DLEN_], il_lda;
  int ictx, nprow, npcol,ierr=5,ilerr;
  double dta1[144], dtx1[12], error=0.0, tol=0.0;
  double dtas1[12][12] = {1.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
                           2.0,3.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
                           1.0,2.0,3.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
                           2.0,3.0,1.0,1.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
                           1.0,1.0,3.0,2.0,2.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
                           1.0,2.0,2.0,2.0,1.0,1.0,0.0,0.0,0.0,0.0,0.0,0.0,
                           3.0,3.0,1.0,1.0,2.0,2.0,2.0,0.0,0.0,0.0,0.0,0.0,
                           1.0,1.0,2.0,1.0,2.0,1.0,1.0,2.0,0.0,0.0,0.0,0.0,
                           1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,3.0,0.0,0.0,0.0,
                           2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,1.0,2.0,0.0,0.0,
                           3.0,3.0,3.0,3.0,3.0,3.0,3.0,3.0,3.0,2.0,1.0,0.0,
                           2.0,3.0,1.0,2.0,3.0,1.0,2.0,3.0,1.0,2.0,3.0,1.0 };
 
  double dtxs1[12] = {2.0,3.0,1.0,2.0,3.0,1.0,2.0,3.0,1.0,2.0,3.0,1.0 };
  double dresx1[12] = {42.0,48.0,39.0,31.0,34.0,23.0,
                     23.0,23.0,15.0,12.0,6.0,1.0 };
   double dta2[72], dtc2[72];
   double dtas2[9][8] = { 0.,  0.,  0.,  0.,  1.,  1.,  0.,  -1.,
                         -1.,  1.,  0.,  1.,  0.,  0.,  0.,   0.,
                         -1.,  0., -1.,  0.,  0.,  0., -1.,   0.,
                          0.,  1., -1., -1.,  0.,  0.,  0.,   0.,
                          0.,  0.,  0.,  1., -1.,  1., -1.,   0.,
                          0.,  1.,  0.,  1.,  0.,  1.,  0.,   0.,
                          0.,  0.,  1.,  0.,  0.,  0.,   0.,  -1,
                          0.,  1.,  0.,  1.,  0.,  0.,  0.,   0.,
                          1.,  1.,  1.,  1.,  1.,  1.,  1.,   1. };
 
   double dtcs2[8][9] = { 0.,  0.,  0.,  0., -1., -1.,  0.,   1.,  1.,
                          1., -1.,  0., -1.,  2.,  3.,  4.,   5.,  2.,
                          1.,  0.,  1.,  0.,  0.,  0.,  1.,   0.,  3.,
                          5., -1.,  1.,  1.,  0.,  0.,  0.,   0.,  4.,
                          6.,  0.,  0., -1.,  1., -1.,  1.,   0.,  1.,
                          7., -1.,  0., -1.,  0., -1.,  0.,   0.,  1.,
                          8.,  0., -1.,  0.,  0.,  0.,   0.,   1.,  1.,
                          9.,  1.,  0.,  1.,  0.,  0.,  0.,   0.,  1. };
 
   double dresc2[8][9] = { 0., -1., -1.,  0., -1., -1.,  0.,   1.,  2.,
                          1.,  0.,  0.,  0.,  2.,  4.,  4.,   6.,  3.,
                          1.,  0.,  0., -1.,  0.,  0.,  2.,   0.,  4.,
                          5.,  0.,  1.,  0.,  1.,  1.,  0.,   1.,  5.,
                          7.,  0.,  0., -1.,  0., -1.,  1.,   0.,  2.,
                          8., -1.,  0., -1.,  1.,  0.,  0.,   0.,  2.,
                          8.,  0., -2.,  0., -1.,  0.,   0.,   1.,  2.,
                          8.,  1.,  0.,  1.,  0.,  0., -1.,   0.,  2. };
 
  int desc_c[DLEN_];
 
      dcmplx     zta1[16], ztx1[4];
      dcmplx ztas1[16] = {
                  {0.0,0.0},{1.0,1.0},{1.0,1.0},{3.0,3.0},
                  {0.0,0.0},{0.0,0.0},{3.0,3.0},{4.0,4.0},
                  {0.0,0.0},{0.0,0.0},{0.0,0.0},{3.0,3.0},
                  {0.0,0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0}};
 
      dcmplx ztxs1[4]  = { {1.0,1.0},{2.0,2.0},{3.0,3.0},{4.0,4.0}};
      dcmplx zresx1[4] = { {1.0,1.0},{2.0,4.0},{3.0,17.0},{4.0,44.0}};
 
      dcmplx   zta2[56], ztc2[56],alpha,beta;
      dcmplx ztas2[56] =  {
                  {0.,1.},{0.,1.},{0.,1.},{0.,1.},
                  {1.,2.},{1.,2.},{0.,1.},{-1.,0.},
                 {-1.,0.},{1.,2.},{0.,1.},{1.,2.},
                  {0.,1.},{0.,1.},{0.,1.},{ 0.,1.},
                 {-1.,0.},{0.,1.},{-1.,0.},{0.,1.},
                 { 0.,1.},{0.,1.},{-1.,0.},{0.,1.},
                 { 0.,1.},{1.,2.},{-1.,0.},{-1.,0.},
                 { 0.,1.},{0.,1.},{ 0.,1.},{ 0.,1.},
                 { 0.,1.},{0.,1.},{ 0.,1.},{ 1.,2.},
                 {-1.,0.},{1.,2.},{-1.,0.},{ 0.,1.},
                 { 0.,1.},{1.,2.},{ 0.,1.},{ 1.,2.},
                 { 0.,1.},{1.,2.},{ 0.,1.},{ 0.,1.},
                 { 0.,1.},{0.,1.},{ 1.,2.},{ 0.,1.},
                 { 0.,1.},{0.,1.},{ 0.,1.},{-1.,0.}};
 
      dcmplx ztcs2[56] = {
                 {0.,-2.},{1.,-1.},{2.,0.},{3.,1.},
                 {-1.,-3.},{-1.,-3.},{5.,3.},
                 { 1.,-1.},{-1.,-3.},{0.,-2.},{-1.,-3.},
                 {2.,0.},{3.,1.},{4.,2.},
                 {1.,-1.},{0.,-2.},{1.,-1.},{0.,-2.},
                 { 0.,-2.},{0.,-2.},{ 1.,-1.},
                 { 5.,3.},{-1.,-3.},{ 1.,-1.},{ 1.,-1.},
                 { 0.,-2.},{0.,-2.},{ 0.,-2.},
                 { 6.,4.},{0.,-2.},{ 2.,0.},{-1.,-3.},
                 { 1.,-1.},{-1.,-3.},{ 1.,-1.},
                 { 7.,5.},{-1.,-3.},{ 3.,1.},{-1.,-3.},
                 { 1.,-1.},{-1.,-3.},{ 0.,-2.},
                 { 8.,6.},{0.,-2.},{-1.,-3.},{ 3.,1.},
                 { 2.,0.},{0.,-2.},{ 0.,-2.},
                 { 9.,7.},{1.,-1.},{ 0.,-2.},{ 1.,-1.},
                 { 0.,-2.},{0.,-2.},{ 0.,-2.}};
      dcmplx zresc2u[56] = {
                 {0.,-1.},{0.,-1.},{1.,0.},{3.,2.},
                 {-1.,-2.},{-1.,-2.},{5.,4.},
                 { 1., 0.},{ 0.,-1.},{0.,-1.},{ 0.,-1.},
                 {2.,1.},{4.,3.},{4.,3.},
                 {1., 0.},{0.,-1.},{0.,-1.},{-1.,-2.},
                 { 0.,-1.},{0.,-1.},{ 2., 1.},
                 { 5.,4.},{ 0.,-1.},{ 1., 0.},{ 0.,-1.},
                 { 1., 0.},{1., 0.},{ 0.,-1.},
                 { 7.,6.},{0.,-1.},{ 2.,1.},{-1.,-2.},
                 { 0.,-1.},{-1.,-2.},{ 1., 0.},
                 { 8.,7.},{-1.,-2.},{ 3.,2.},{-1.,-2.},
                 { 2., 1.},{ 0.,-1.},{ 0.,-1.},
                 { 8.,7.},{0.,-1.},{-2.,-3.},{ 3.,2.},
                 { 1.,0.},{0.,-1.},{ 0.,-1.},
                 { 8.,7.},{1., 0.},{ 0.,-1.},{ 1., 0.},
                 { 0.,-1.},{0.,-1.},{-1.,-2.}};
      dcmplx    zresc2c[56] = {
                 {0.,-3.},{0.,-1.},{1.,0.},{3.,0.},
                 {-1.,-4.},{-1.,-4.},{5.,2.},
                 { 1.,-2.},{ 0.,-5.},{0.,-3.},{ 0.,-5.},
                 {2.,-1.},{4.,-1.},{4.,1.},
                 {1.,-2.},{0.,-3.},{0.,-1.},{-1.,-2.},
                 { 0.,-3.},{0.,-3.},{ 2.,-3.},
                 { 5.,2.},{ 0.,-5.},{ 1.,-2.},{ 0.,-1.},
                 { 1.,-4.},{1.,-4.},{ 0.,-3.},
                 { 7.,2.},{0.,-3.},{ 2.,-1.},{-1.,-4.},
                 { 0.,-1.},{-1.,-4.},{ 1.,-2.},
                 { 8.,3.},{-1.,-4.},{ 3.,0.},{-1.,-4.},
                 { 2.,-3.},{ 0.,-5.},{ 0.,-3.},
                 { 8.,5.},{0.,-3.},{-2.,-3.},{ 3.,0.},
                 { 1.,0.},{0.,-3.},{ 0.,-3.},
                 { 8.,7.},{1.,-2.},{ 0.,-3.},{ 1.,-2.},
                 { 0.,-3.},{0.,-3.},{-1.,-2.}};
 
/* PDSYR2K, PZSYR2K, PZHER2K START  */
 
   int i, desc_b[DLEN_];
   double dta3[72], dtb3[72], dtc3[81];
 
 
   double dtas3[9][8] = { 0.,  0.,  0.,  0.,  1.,  1.,  0.,  -1.,
                         -1.,  1.,  0.,  1.,  0.,  0.,  0.,   0.,
                         -1.,  0., -1.,  0.,  0.,  0., -1.,   0.,
                          0.,  1., -1., -1.,  0.,  0.,  0.,   0.,
                          0.,  0.,  0.,  1., -1.,  1., -1.,   0.,
                          0.,  1.,  0.,  1.,  0.,  1.,  0.,   0.,
                          0.,  0.,  1.,  0.,  0.,  0.,   0.,  -1.,
                          0.,  1.,  0.,  1.,  0.,  0.,  0.,   0.,
                          1.,  1.,  1.,  1.,  1.,  1.,  1.,   1. };
 
 
   double dtbs3[9][8] = { 0.,  0.,  0.,  0., -1., -1.,  0.,   1.,
                          1., -1.,  0., -1.,  0.,  0.,  0.,   0.,
                          1.,  0.,  1.,  0.,  0.,  0.,  1.,   0.,
                          0., -1.,  1.,  1.,  0.,  0.,  0.,   0.,
                          0.,  0.,  0., -1.,  1., -1.,  1.,   0.,
                          0., -1.,  0., -1.,  0., -1.,  0.,   0.,
                          0.,  0., -1.,  0.,  0.,  0.,   0.,   1.,
                          0., -1.,  0., -1.,  0.,  0.,  0.,   0.,
                         -1., -1., -1., -1., -1., -1., -1.,  -1. };
 
   double dresc3[9][9] = {-6.,  0.,  0.,  0.,  0.,  0.,  0.,   0., 0.,
                          0., -6.,  0.,  0.,  0.,  0.,  0.,   0.,  0.,
                          0., -2., -6.,  0.,  0.,  0.,  0.,   0.,  0.,
                          0.,  0., -2., -6.,  0.,  0.,  0.,   0.,  0.,
                          0., -2., -2.,  2., -8.,  0.,  0.,   0.,  0.,
                         -2., -4.,  0.,  0., -4., -6.,  0.,   0.,  0.,
                         -2.,  0.,  2.,  2.,  0.,  0.,  -4.,   0.,  0.,
                          0., -4.,  0.,  0., -2., -4.,  0.,  -4.,  0.,
                         -2., -2.,  6.,  2.,  0., -6.,  0.,  -4., -16. };
 
      dcmplx     zta3[56], ztb3[56], ztc3[49];
      dcmplx ztas3[56] = {
              { 0., 1.},{ 0., 1.},{0., 1.},{0., 1.},
              { 1., 2.},{ 1., 2.},{0., 1.},{-1.,0.},
              {-1., 0.},{ 1., 2.},{0., 1.},{1., 2.},
              { 0., 1.},{ 0., 1.},{0., 1.},{ 0.,1.},
              {-1., 0.},{ 0., 1.},{-1., 0.},{0., 1.},
              { 0., 1.},{ 0., 1.},{-1., 0.},{ 0.,1.},
              { 0., 1.},{ 1., 2.},{-1., 0.},{-1., 0.},
              { 0., 1.},{ 0., 1.},{ 0., 1.},{ 0.,1.},
              { 0., 1.},{ 0., 1.},{ 0., 1.},{ 1., 2.},
              {-1., 0.},{ 1., 2.},{-1., 0.},{ 0.,1.},
              { 0., 1.},{ 1., 2.},{ 0., 1.},{ 1., 2.},
              { 0., 1.},{ 1., 2.},{ 0., 1.},{ 0.,1.},
              { 0., 1.},{ 0., 1.},{ 1., 2.},{ 0., 1.},
              { 0., 1.},{ 0., 1.},{ 0., 1.},{-1.,0.}};
 
      dcmplx ztbs3[56] = {
              { 0.,-2.},{ 0.,-2.},{0.,-2.},{0.,-2.},
              {-1.,-3.},{-1.,-3.},{0.,-2.},{ 1.,-1.},
              { 1.,-1.},{-1.,-3.},{0.,-2.},{-1.,-3.},
              { 0.,-2.},{ 0.,-2.},{0.,-2.},{ 0.,-2.},
              { 1.,-1.},{ 0.,-2.},{ 1.,-1.},{0.,-2.},
              { 0.,-2.},{ 0.,-2.},{ 1.,-1.},{ 0.,-2.},
              { 0.,-2.},{-1.,-3.},{ 1.,-1.},{ 1.,-1.},
              { 0.,-2.},{ 0.,-2.},{ 0.,-2.},{ 0.,-2.},
              { 0.,-2.},{ 0.,-2.},{ 0.,-2.},{-1.,-3.},
              { 1.,-1.},{-1.,-3.},{ 1.,-1.},{ 0.,-2.},
              { 0.,-2.},{-1.,-3.},{ 0.,-2.},{-1.,-3.},
              { 0.,-2.},{-1.,-3.},{ 0.,-2.},{ 0.,-2.},
              { 0.,-2.},{ 0.,-2.},{-1.,-3.},{ 0.,-2.},
              { 0.,-2.},{ 0.,-2.},{ 0.,-2.},{ 1.,-1.}};
 
      dcmplx zresc3[49] = {
        { 38.,-18.},{  0.,  0.},{ 0.,  0.},{  0.,  0.},
        {  0.,  0.},{  0.,  0.},{ 0.,  0.},
        { 38.,- 6.},{ 38.,-18.},{ 0.,  0.},{  0.,  0.},
        {  0.,  0.},{  0.,  0.},{ 0.,  0.},
        { 26.,  6.},{ 26.,  2.},{14.,  6.},{  0.,  0.},
        {  0.,  0.},{  0.,  0.},{ 0.,  0.},
        { 32.,  0.},{ 32.,  0.},{20.,  8.},{ 26., -6.},
        {  0.,  0.},{  0.,  0.},{ 0.,  0.},
        { 35.,- 3.},{ 35.,- 7.},{23.,  5.},{ 29.,  7.},
        { 32.,-16.},{  0.,  0.},{ 0.,  0.},
        { 44.,-16.},{ 44.,-20.},{32.,  0.},{ 38., -6.},
        { 41.,-17.},{ 50.,-30.},{ 0.,  0.},
        { 35.,- 7.},{ 35.,- 3.},{23., 13.},{ 29.,  7.},
        { 32.,  0.},{ 41.,- 9.},{32., -8.}};
 
      dcmplx zresc3h[49] = {
        {-50.,  0.},{  0.,  0.},{ 0.,  0.},{  0.,  0.},
        {  0.,  0.},{  0.,  0.},{ 0.,  0.},
        {-38.,  0.},{-50.,  0.},{ 0.,  0.},{  0.,  0.},
        {  0.,  0.},{  0.,  0.},{ 0.,  0.},
        {-26.,-12.},{-30.,-12.},{-26., 0.},{  0.,  0.},
        {  0.,  0.},{  0.,  0.},{ 0.,  0.},
        {-32., -6.},{-32., -6.},{-24., 6.},{-38.,  0.},
        {  0.,  0.},{  0.,  0.},{ 0.,  0.},
        {-35.,- 3.},{-39.,- 3.},{-27., 9.},{-25.,  3.},
        {-48.,  0.},{  0.,  0.},{ 0.,  0.},
        {-48.,  6.},{-52.,  6.},{-32.,18.},{-38., 12.},
        {-49.,  9.},{-62.,  0.},{ 0.,  0.},
        {-39.,- 3.},{-35.,- 3.},{-19., 9.},{-25.,  3.},
        {-32.,  0.},{-41.,- 9.},{-40., 0.}};
 
/* PDSYMM, PZSYMM, PZHEMM START  */
 
   double dta4[64], dtb4[128], dtc4[128];
   double dtas4[8][8] = { 0.,  0.,  0.,  0.,  0.,  0.,  0.,   0.,
                         -1.,  1.,  0.,  0.,  0.,  0.,  0.,   0.,
                         -1.,  0., -1.,  0.,  0.,  0.,  0.,   0.,
                          0.,  1., -1., -1.,  0.,  0.,  0.,   0.,
                          0.,  0.,  0.,  1., -1.,  0.,  0.,   0.,
                          0.,  1.,  0.,  1.,  0.,  1.,  0.,   0.,
                          0.,  0.,  1.,  0.,  0.,  0.,  0.,   0.,
                          0.,  1.,  0.,  1.,  0.,  0.,  0.,   0. };
 
   double dtbs4[8][16] = {-1., -1.,  1.,  0.,  0.,  0.,  1.,   0.,
                           0., -1.,  0.,  0.,  1.,  0.,  0.,  -1.,
                           0., -1.,  1., -1.,  1.,  0.,  1.,   0.,
                           0., -1.,  0.,  0.,  1.,  0.,  1.,   0.,
                           1.,  1., -1.,  0.,  0.,  1.,  0.,  -1.,
                           0.,  1.,  0.,  1., -1.,  0.,  0.,  -1.,
                          -1.,  0.,  0.,  0.,  1.,  0.,  0.,   0.,
                          -1.,  0.,  1.,  1.,  0.,  0.,  0.,   0.,
                           1.,  1., -1.,  0.,  0., -1.,  1.,   0.,
                           1.,  0.,  1.,  0., -1.,  1.,  0.,   0.,
                           1., -1.,  0.,  0.,  1., -1.,  1.,   1.,
                           1., -1.,  0., -1., -1.,  0.,  0.,   1.,
                          -1., -1.,  1.,  0.,  1.,  0.,  0.,   0.,
                           0.,  0.,  0.,  0.,  1.,  0.,  0.,   1.,
                          -1.,  1.,  0., -1.,  0.,  0., -1.,   1.,
                           1.,  1.,  0.,  0.,  1., -1.,  0.,   0. };
 
   double dresc4[8][16] = {-1.,  0.,  0.,  1., -1., -1., -1.,   1.,
                            0.,  0.,  0., -1.,  0.,  0., -1.,   1.,
                            0.,  0.,  0., -2.,  3., -1.,  0.,   2.,
                            1.,  0.,  1.,  0.,  0., -1.,  1.,   2.,
                            0., -1.,  1.,  0.,  0., -1., -1.,   1.,
                            1.,  0., -1., -2.,  1.,  0.,  0.,   3.,
                            1., -1.,  1., -2.,  1., -3.,  2.,   3.,
                            4., -2.,  0., -3.,  1.,  0.,  1.,   2.,
                           -2., -1.,  1.,  0.,  1.,  1., -1.,   0.,
                           -2.,  0.,  0.,  1.,  1., -1.,  0.,   0.,
                            0., -2.,  1., -1.,  3., -1.,  2.,   1.,
                            0., -2.,  1.,  0.,  0.,  0.,  1.,   1.,
                            1.,  1., -1.,  0.,  0.,  1.,  0.,  -1.,
                            0.,  1.,  0.,  1., -1.,  0.,  0.,  -1.,
                           -1., -1.,  1., -1.,  2.,  0.,  1.,   0.,
                           -1., -1.,  1.,  1.,  1.,  0.,  1.,   0. };
 
      dcmplx zta4[64], ztb4[128], ztc4[128];
      dcmplx ztas4[64] = {
                   {0.,1.},{0.,  0.}, {0., 0.},{0., 0.},
                   {0.,0.}, {0., 0.}, {0., 0.},{0., 0.},
                  {-1.,0.}, {1., 2.}, {0., 0.},{0., 0.},
                   {0.,0.}, {0., 0.}, {0., 0.},{0., 0.},
                  {-1.,0.}, {0., 1.},{-1., 0.},{0., 0.},
                   {0.,0.}, {0., 0.}, {0., 0.},{0., 0.},
                   {0.,1.}, {1., 2.},{-1., 0.},{-1.,0.},
                   {0.,0.}, {0., 0.}, {0., 0.},{0., 0.},
                   {0.,1.}, {0., 1.}, {0., 1.}, {1.,2.},
                  {-1.,0.}, {0., 0.}, {0., 0.},{0., 0.},
                   {0.,1.}, {1., 2.}, {0., 1.}, {1.,2.},
                   {0.,1.}, {1., 2.}, {0., 0.},{0., 0.},
                   {0.,1.}, {0., 1.}, {1., 2.}, {0.,1.},
                   {0.,1.}, {0., 1.}, {0., 1.},{0., 0.},
                   {0.,1.}, {1., 2.}, {0., 1.}, {1.,2.},
                   {0.,1.}, {0., 1.}, {0., 1.},{0., 1.}};
 
      dcmplx ztbs4[128] = {
                  {-1.,-3.},{-1.,-3.},{1.,-1.},{0.,-2.},
                  { 0.,-2.},{ 0.,-2.},{1.,-1.},{0.,-2.},
                  { 0.,-2.},{-1.,-3.},{0.,-2.},{0.,-2.},
                  { 1.,-1.},{0.,-2.},{0.,-2.},{-1.,-3.},
                  {0.,-2.},{-1.,-3.},{1.,-1.},{-1.,-3.},
                  {1.,-1.},{ 0.,-2.},{1.,-1.},{ 0.,-2.},
                  {0.,-2.},{-1.,-3.},{0.,-2.},{ 0.,-2.},
                  {1.,-1.},{ 0.,-2.},{1.,-1.},{ 0.,-2.},
                  {1.,-1.},{ 1.,-1.},{-1.,-3.},{0.,-2.},
                  {0.,-2.},{ 1.,-1.},{0.,-2.},{-1.,-3.},
                  {0.,-2.},{ 1.,-1.},{0.,-2.},{ 1.,-1.},
                  {-1.,-3.},{0.,-2.},{0.,-2.},{-1.,-3.},
                  {-1.,-3.},{0.,-2.},{0.,-2.},{ 0.,-2.},
                  { 1.,-1.},{0.,-2.},{0.,-2.},{ 0.,-2.},
                  {-1.,-3.},{0.,-2.},{1.,-1.},{ 1.,-1.},
                  { 0.,-2.},{0.,-2.},{0.,-2.},{ 0.,-2.},
                  { 1.,-1.},{1.,-1.},{-1.,-3.},{0.,-2.},
                  { 0.,-2.},{-1.,-3.},{1.,-1.},{0.,-2.},
                  { 1.,-1.},{ 0.,-2.},{1.,-1.},{0.,-2.},
                  {-1.,-3.},{ 1.,-1.},{0.,-2.},{0.,-2.},
                  { 1.,-1.},{-1.,-3.},{0.,-2.},{0.,-2.},
                  { 1.,-1.},{-1.,-3.},{1.,-1.},{1.,-1.},
                  {1.,-1.},{-1.,-3.},{0.,-2.},{-1.,-3.},
                  {-1.,-3.},{0.,-2.},{0.,-2.},{ 1.,-1.},
                  {-1.,-3.},{-1.,-3.},{1.,-1.},{0.,-2.},
                  { 1.,-1.},{ 0.,-2.},{0.,-2.},{0.,-2.},
                  { 0.,-2.},{ 0.,-2.},{0.,-2.},{0.,-2.},
                  { 1.,-1.},{ 0.,-2.},{0.,-2.},{1.,-1.},
                  {-1.,-3.},{1.,-1.},{0.,-2.},{-1.,-3.},
                  {0.,-2.},{0.,-2.},{-1.,-3.},{ 1.,-1.},
                  {1.,-1.},{1.,-1.},{ 0.,-2.},{ 0.,-2.},
                  {1.,-1.},{-1.,-3.},{0.,-2.},{ 0.,-2.}};
 
      dcmplx zresc4[128] = {
                   {11.,27.},{ 7.,29.},{1.,27.},{6.,32.},
                {-4.,22.},{11.,27.},{-1.,23.},{-3.,29.},
                 {-2.,26.},{ 7.,29.},{-2.,26.},{5.,25.},
                  { 1.,27.},{4.,28.},{5.,25.},{0.,30.},
                {37.,39.},{37.,39.},{31.,37.},{48.,36.},
                {10.,40.},{41.,37.},{25.,35.},{23.,41.},
                {24.,38.},{37.,39.},{24.,38.},{31.,37.},
                {31.,37.},{38.,36.},{27.,39.},{26.,42.},
                { 7.,29.},{11.,27.},{-3.,29.},{10.,30.},
                {-8.,24.},{11.,27.},{-1.,23.},{-3.,29.},
                {-6.,28.},{ 7.,29.},{2.,24.},{ 9.,23.},
                {-3.,29.},{4.,28.},{1.,27.},{-8.,34.},
                {27.,39.},{35.,35.},{21.,37.},{42.,34.},
                {12.,34.},{43.,31.},{11.,37.},{13.,41.},
                { 6.,42.},{39.,33.},{22.,34.},{37.,29.},
                {21.,37.},{28.,36.},{21.,37.},{20.,40.},
                {27.,29.},{23.,31.},{ 9.,33.},{22.,34.},
                { 0.,30.},{15.,35.},{11.,27.},{13.,31.},
                {18.,26.},{19.,33.},{10.,30.},{ 9.,33.},
                { 9.,33.},{20.,30.},{13.,31.},{16.,32.},
                {37.,39.},{45.,35.},{27.,39.},{44.,38.},
                {10.,40.},{41.,37.},{17.,39.},{27.,39.},
                {28.,36.},{45.,35.},{24.,38.},{31.,37.},
                {31.,37.},{34.,38.},{27.,39.},{30.,40.},
                {21.,37.},{21.,37.},{23.,31.},{28.,36.},
                {10.,30.},{21.,37.},{13.,31.},{23.,31.},
                {16.,32.},{21.,37.},{16.,32.},{15.,35.},
                {23.,31.},{22.,34.},{19.,33.},{26.,32.},
                {35.,35.},{35.,35.},{21.,37.},{38.,36.},
                { 8.,36.},{31.,37.},{15.,35.},{25.,35.},
                {26.,32.},{35.,35.},{18.,36.},{21.,37.},
                {21.,37.},{28.,36.},{21.,37.},{28.,36.}};
 
      dcmplx     zta4h[64], ztb4h[128], ztc4h[128];
      dcmplx ztas4h[64] = {
                   {0.,0.}, {0., 0.}, {0., 0.},{0., 0.},
                   {0.,0.}, {0., 0.}, {0., 0.},{0., 0.},
                  {-1.,0.}, {1., 0.}, {0., 0.},{0., 0.},
                   {0.,0.}, {0., 0.}, {0., 0.},{0., 0.},
                  {-1.,0.}, {0., 1.},{-1., 0.},{0., 0.},
                   {0.,0.}, {0., 0.}, {0., 0.},{0., 0.},
                   {0.,1.}, {1., 2.},{-1., 0.},{-1.,0.},
                   {0.,0.}, {0., 0.}, {0., 0.},{0., 0.},
                   {0.,1.}, {0., 1.}, {0., 1.}, {1.,2.},
                  {-1.,0.}, {0., 0.}, {0., 0.},{0., 0.},
                   {0.,1.}, {1., 2.}, {0., 1.}, {1.,2.},
                   {0.,1.}, {1., 0.}, {0., 0.},{0., 0.},
                   {0.,1.}, {0., 1.}, {1., 2.}, {0.,1.},
                   {0.,1.}, {0., 1.}, {0., 0.},{0., 0.},
                   {0.,1.}, {1., 2.}, {0., 1.}, {1.,2.},
                   {0.,1.}, {0., 1.}, {0., 1.},{0., 0.}};
 
      dcmplx ztbs4h[128] = {
                  {-1.,-3.},{-1.,-3.},{1.,-1.},{0.,-2.},
                  { 0.,-2.},{ 0.,-2.},{1.,-1.},{0.,-2.},
                  { 0.,-2.},{-1.,-3.},{0.,-2.},{0.,-2.},
                  { 1.,-1.},{0.,-2.},{0.,-2.},{-1.,-3.},
                  {0.,-2.},{-1.,-3.},{1.,-1.},{-1.,-3.},
                  {1.,-1.},{ 0.,-2.},{1.,-1.},{ 0.,-2.},
                  {0.,-2.},{-1.,-3.},{0.,-2.},{ 0.,-2.},
                  {1.,-1.},{ 0.,-2.},{1.,-1.},{ 0.,-2.},
                  {1.,-1.},{ 1.,-1.},{-1.,-3.},{0.,-2.},
                  {0.,-2.},{ 1.,-1.},{0.,-2.},{-1.,-3.},
                  {0.,-2.},{ 1.,-1.},{0.,-2.},{ 1.,-1.},
                  {-1.,-3.},{0.,-2.},{0.,-2.},{-1.,-3.},
                  {-1.,-3.},{0.,-2.},{0.,-2.},{ 0.,-2.},
                  { 1.,-1.},{0.,-2.},{0.,-2.},{ 0.,-2.},
                  {-1.,-3.},{0.,-2.},{1.,-1.},{ 1.,-1.},
                  { 0.,-2.},{0.,-2.},{0.,-2.},{ 0.,-2.},
                  { 1.,-1.},{1.,-1.},{-1.,-3.},{0.,-2.},
                  { 0.,-2.},{-1.,-3.},{1.,-1.},{0.,-2.},
                  { 1.,-1.},{ 0.,-2.},{1.,-1.},{0.,-2.},
                  {-1.,-3.},{ 1.,-1.},{0.,-2.},{0.,-2.},
                  { 1.,-1.},{-1.,-3.},{0.,-2.},{0.,-2.},
                  { 1.,-1.},{-1.,-3.},{1.,-1.},{1.,-1.},
                  {1.,-1.},{-1.,-3.},{0.,-2.},{-1.,-3.},
                  {-1.,-3.},{0.,-2.},{0.,-2.},{ 1.,-1.},
                  {-1.,-3.},{-1.,-3.},{1.,-1.},{0.,-2.},
                  { 1.,-1.},{ 0.,-2.},{0.,-2.},{0.,-2.},
                  { 0.,-2.},{ 0.,-2.},{0.,-2.},{0.,-2.},
                  { 1.,-1.},{ 0.,-2.},{0.,-2.},{1.,-1.},
                  {-1.,-3.},{1.,-1.},{0.,-2.},{-1.,-3.},
                  {0.,-2.},{0.,-2.},{-1.,-3.},{ 1.,-1.},
                  {1.,-1.},{1.,-1.},{ 0.,-2.},{ 0.,-2.},
                  {1.,-1.},{-1.,-3.},{0.,-2.},{ 0.,-2.}};
 
      dcmplx zresc4h[128] = {
                {-12.,4.},{-10.,4.},{-10.,4.},{-10.,6.},
                {-8., 0.},{-13.,5.},{-10.,2.},{-7., 3.},
                {-8., 2.},{-10.,4.},{-8., 2.},{-11.,3.},
                {-10., 4.},{-10.,4.},{-11.,3.},{-7.,3.},
            {-19.,-5.},{-17.,-7.},{-19.,-5.},{-22.,-6.},
            {-10.,-8.},{-21.,-5.},{-17.,-7.},{-13.,-7.},
            {-14.,-8.},{-17.,-7.},{-14.,-8.},{-17.,-7.},
            {-19.,-5.},{-20.,-6.},{-17.,-5.},{-14.,-6.},
            {- 9., 5.},{- 9., 3.},{- 7., 5.},{- 8., 4.},
            {- 6., 2.},{-11., 5.},{- 9., 3.},{- 5., 3.},
            {- 4., 2.},{- 7., 3.},{- 8., 2.},{-11., 3.},
            {- 7., 5.},{- 8., 4.},{- 9., 5.},{- 2., 4.},
            {- 5.,-5.},{- 5.,-9.},{-11., 1.},{-10.,-6.},
            {- 6.,-4.},{-15.,-3.},{- 7.,-1.},{- 1.,-5.},
            {  2.,-6.},{- 7.,-9.},{- 6.,-6.},{-13.,-5.},
            {-11., 1.},{- 8.,-4.},{- 9.,-1.},{- 2.,-6.},
            {  3.,-3.},{  3.,-1.},{  5., 1.},{ 4., 0.},
            {  4., 2.},{  3., 3.},{  1., 1.},{  7.,-3.},
            {  6.,-6.},{  5.,-1.},{  2., 2.},{  1., 5.},
            {  5., 1.},{  2., 0.},{  3., 1.},{  8.,-4.},
            {  9.,-5.},{  9.,-9.},{ 11.,-5.},{ 10.,-8.},
            { 10., 0.},{  9.,-7.},{  7., 1.},{ 13.,-7.},
            { 12.,-8.},{11.,-11.},{  8.,-2.},{  7.,-3.},
            { 11.,-7.},{ 8., -4.},{  9.,-3.},{ 14.,-8.},
            { 10., 2.},{14., -2.},{ 12.,-4.},{ 12.,-2.},
            {  9., 1.},{13., -1.},{  7., 3.},{ 13.,-5.},
            { 12.,-2.},{15., -3.},{ 10., 0.},{ 11., 1.},
            { 14.,-6.},{10.,  0.},{ 11.,-1.},{ 13.,-5.},
            { 18.,-6.},{20., -8.},{ 17.,-1.},{ 19.,-7.},
            { 14., 4.},{19., -5.},{ 14., 2.},{ 18.,-4.},
            { 17.,-5.},{20., -8.},{ 16., 0.},{ 17.,-1.},
            { 18.,-2.},{17., -3.},{ 17.,-1.},{ 18.,-4.}};
 
/* PDGER, PZGERU, PZGERC START : 5 */
 
   int desc_y[DLEN_];
   double dta5[100], dtx5[11], dty5[11];
   double dtas5[100];
   double dtxs5[11] = {1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1. };
   double dtys5[11] = {1., 2., 3., 4., 5., 6., 7., 8., 9., 10., 11. };
   double dresa5[100] = {
                   1.,   2.,   3.,   4.,   5.,   6.,   7.,   8.,   9.,   10.,
                  11.,  14.,  15.,  16.,  17.,  18.,  19.,  20.,  21.,   22.,
                  21.,  25.,  26.,  27.,  28.,  29.,  30.,  31.,  32.,   33.,
                  31.,  36.,  37.,  38.,  39.,  40.,  41.,  42.,  43.,   44.,
                  41.,  47.,  48.,  49.,  50.,  51.,  52.,  53.,  54.,   55.,
                  51.,  58.,  59.,  60.,  61.,  62.,  63.,  64.,  65.,   66.,
                  61.,  69.,  70.,  71.,  72.,  73.,  74.,  75.,  76.,   77.,
                  71.,  80.,  81.,  82.,  83.,  84.,  85.,  86.,  87.,   88.,
                  81.,  91.,  92.,  93.,  94.,  95.,  96.,  97.,  98.,   99.,
                  91., 102., 103., 104., 105., 106., 107., 108., 109.,  110.};
 
      dcmplx   zta5[100], ztx5[11], zty5[11];
      dcmplx   ztxs5[11] = {
                   {1., 4.},{1., 4.},{1.,3.},{1.,2.},
                   {1., 1.},{1., 0.},{1.,-1.},{1.,-2.},
                   {1.,-3.},{1.,-4.},{1.,-4.}};
      dcmplx   ztys5[11] = {
                   {2., 1.},{2., 1.},{3.,1.},{4.,1.},
                   {5., 1.},{6., 1.},{7., 1.0},{8., 1.},
                   {9., 1.},{10., 1.},{10., 1.}};
      dcmplx ztas5[100] = {
       {11., 2.},{12., 2.},{13., 2.0},{14., 2.},
       {15., 2.},{16., 2.},{17., 2.0},{18., 2.},
       {19., 2.},{20., 2.},
       {11., 2.},{12., 2.},{13., 2.0},{14., 2.},
       {15., 2.},{16., 2.},{17., 2.0},{18., 2.},
       {19., 2.},{20., 2.},
       {21., 1.},{22., 1.},{23., 1.0},{24., 1.},
       {25., 1.},{26., 1.},{27., 1.0},{28., 1.},
       {29., 1.},{30., 1.},
       {31., 0.},{32., 0.},{33., 0.0},{34., 0.},
       {35., 0.},{36., 0.},{37., 0.0},{38., 0.},
       {39., 0.},{40., 0.},
       {41.,-1.},{42.,-1.},{43.,-1.0},{44.,-1.},
       {45.,-1.},{46.,-1.},{47.,-1.0},{48.,-1.},
       {49.,-1.},{50.,-1.},
       {51.,-2.},{52.,-2.},{53.,-2.0},{54.,-2.},
       {55.,-2.},{56.,-2.},{57.,-2.0},{58.,-2.},
       {59.,-2.},{60.,-2.},
       {61., 2.},{62., 2.},{63., 2.0},{64., 2.},
       {65., 2.},{66., 2.},{67., 2.0},{68., 2.},
       {69., 2.},{70., 2.},
       {71., 1.},{72., 1.},{73., 1.0},{74., 1.},
       {75., 1.},{76., 1.},{77., 1.0},{78., 1.},
       {79., 1.},{80., 1.},
       {81., 0.},{82., 0.},{83., 0.0},{84., 0.},
       {85., 0.},{86., 0.},{87., 0.0},{88., 0.},
       {89., 0.},{90., 0.},
       {91.,-1.},{92.,-1.},{93.,-1.0},{94.,-1.},
       {95.,-1.},{96.,-1.},{97.,-1.0},{98.,-1.},
       {99.,-1.},{100.,-1.}};
      dcmplx zresa5u[100] = {
       {19.,13.},{19.,13.},{19.,10.0},{19., 7.},
       {19., 4.},{19., 1.},{19.,-2.0},{19.,-5.},
       {19.,-8.},{19.,-11.},
       {19.,13.},{19.,13.},{19.,10.0},{19., 7.},
       {19., 4.},{19., 1.},{19.,-2.0},{19.,-5.},
       {19.,-8.},{19.,-11.},
       {34.,15.},{34.,15.},{33.,11.0},{32., 7.},
       {31., 3.},{30.,-1.},{29.,-5.0},{28.,-9.},
       {27.,-13.},{26.,-17.},
       {49.,17.},{49.,17.},{47.,12.0},{45., 7.},
       {43., 2.},{41.,-3.},{39.,-8.0},{37.,-13.},
       {35.,-18.},{33.,-23.},
       {64.,19.},{64.,19.},{61.,13.0},{58., 7.},
       {55., 1.},{52.,-5.},{49.,-11.0},{46.,-17.},
       {43.,-23.},{40.,-29.},
       {79.,21.},{79.,21.},{75.,14.0},{71., 7.},
       {67., 0.},{63.,-7.},{59.,-14.0},{55.,-21.},
       {51.,-28.},{47.,-35.},
       {94., 28.},{94., 28.},{89., 20.0},{84., 12.},
       {79., 4.},{74.,-4.},{69.,-12.0},{64.,-20.},
       {59.,-28.},{54.,-36.},
       {109.,30.},{109.,30.},{103.,21.0},{97.,12.},
       {91., 3.},{85.,-6.},{79.,-15.0},{73.,-24.},
       {67.,-33.},{61.,-42.},
       {124., 32.},{124., 32.},{117.,22.0},{110.,12.},
       {103., 2.},{96.,-8.},{89.,-18.0},{82.,-28.},
       {75.,-38.},{68.,-48.},
       {139.,34.},{139.,34.},{131.,23.0},{123.,12.},
       {115., 1.},{107.,-10.},{99.,-21.0},{91.,-32.},
       {83.,-43.},{75.,-54.}};
 
      dcmplx zresa5c[100] = {
       {25.,3.},{25.,3.},{23.,2.0},{21., 1.},
       {19., 0.},{17.,-1.},{5.,-2.0},{13.,-3.},
       {11.,-4.},{9.,-5.},
       {25.,3.},{25.,3.},{23.,2.0},{21., 1.},
       {19., 0.},{17.,-1.},{15.,-2.0},{13.,-3.},
       {11.,-4.},{9.,-5.},
       {40.,5.},{40.,5.},{37.,3.0},{34., 1.},
       {31.,-1.},{28.,-3.},{25.,-5.0},{22.,-7.},
       {19.,-9.},{16.,-11.},
       {55.,7.},{55.,7.},{51.,4.0},{47., 1.},
       {43.,-2.},{39.,-5.},{35.,-8.0},{31.,-11.},
       {27.,-14.},{23.,-17.},
       {70.,9.},{70.,9.},{65.,5.0},{60., 1.},
       {55.,-3.},{50.,-7.},{45.,-11.0},{40.,-15.},
       {35.,-19.},{30.,-23.},
       {85.,11.},{85.,11.},{79.,6.0},{73., 1.},
       {67.,-4.},{61.,-9.},{55.,-14.0},{49.,-19.},
       {43.,-24.},{37.,-29.},
       {100., 18.},{100., 18.},{93., 12.0},{86., 6.},
       {79., 0.},{72.,-6.},{65.,-12.0},{58.,-18.},
       {51.,-24.},{44.,-30.},
       {115.,20.},{115.,20.},{107.,13.0},{99.,6.},
       {91.,-1.},{83.,-8.},{75.,-15.0},{67.,-22.},
       {59.,-29.},{51.,-36.},
       {130., 22.},{130., 22.},{121.,14.0},{112.,6.},
       {103.,-2.},{94.,-10.},{85.,-18.0},{76.,-26.},
       {67.,-34.},{58.,-42.},
       {145.,24.},{145.,24.},{135.,15.0},{125.,6.},
       {115.,-3.},{105.,-12.},{95.,-21.0},{85.,-30.},
       {75.,-39.},{65.,-48.}};
 
/* PDSYRK, PZSYRK, PZHERK    START : 8, PDTRMM, PZTRMM : 9     */
 
   double dta8[40], dtc8[64];
 
   double dtas8[40] = {  0.0,  1.0,  2.0,  3.0,  4.0,  5.0,  6.0,  7.0,
                         8.0,  9.0, 10.0, 11.0, 12.0, 13.0, 14.0, 15.0,
                        16.0, 17.0, 18.0, 19.0, 20.0, 21.0, 22.0, 23.0,
                        24.0, 25.0, 26.0, 27.0, 28.0, 29.0, 30.0, 31.0,
                        32.0, 33.0, 34.0, 35.0, 36.0, 37.0, 38.0, 39.0 };
 
   double dtcs8[64] = {  0.0,  1.0,  2.0,  3.0,  4.0,  5.0,  6.0,  7.0,
                         0.0,  8.0,  9.0, 10.0, 11.0, 12.0, 13.0, 14.0,
                         0.0,  0.0, 15.0, 16.0, 17.0, 18.0, 19.0, 20.0,
                         0.,   0.,   0.,  21.0, 22.0, 23.0, 24.0, 25.0,
                         0.,   0.,   0.,   0.,  26.0, 27.0, 28.0, 29.0,
                         0.,   0.,   0.,   0.,   0.,  30.0, 31.0, 32.0,
                         0.,   0.,   0.,   0.,   0.,   0.,  33.0, 34.0,
                         0.,   0.,   0.,   0.,   0.,   0.,   0.,  35.0 };
 
   double dresc8[64] = {1920., 2001., 2082., 2163., 2244., 2325., 2406., 2487.,
                           0., 2093., 2179., 2265., 2351., 2437., 2523., 2609.,
                           0.,    0., 2275., 2366., 2457., 2548., 2639., 2730.,
                           0.,    0.,    0., 2466., 2562., 2658., 2754., 2850.,
                           0.,    0.,    0.,    0., 2666., 2767., 2868., 2969.,
                           0.,    0.,    0.,    0.,    0., 2875., 2981., 3087.,
                           0.,    0.,    0.,    0.,    0.,    0., 3093., 3204.,
                           0.,    0.,    0.,    0.,    0.,    0.,    0., 3320.};
 
/*  PZSYRK, PZHERK : 8     */
      dcmplx    zta8[15],ztc8[9];
 
      dcmplx ztas8[15] = {
                   {2.0, 0.0}, {3.0, 3.0}, {1.0, 3.0},
                   {3.0, 2.0}, {8.0, 0.0}, {2.0, 1.0},
                   {4.0, 1.0}, {2.0, 5.0}, {6.0, 0.0},
                   {1.0, 7.0}, {2.0, 4.0}, {3.0, 2.0},
                   {0.0, 0.0}, {1.0, 2.0}, {2.0, 2.0}};
 
      dcmplx ztcs8[9] = {
                   {2.0, 1.0}, {0.0, 0.0}, {0.0, 0.0},
                   {1.0, 9.0}, {3.0, 1.0}, {0.0, 0.0},
                   {4.0, 5.0}, {6.0, 7.0}, {8.0, 1.0}};
 
      dcmplx zresc8[9] ={
                    {-57.0, 13.0}, {  0.0,  0.0}, {0.0,  0.0},
                    {-63.0, 79.0}, {-28.0, 90.0}, {0.0,  0.0},
                    {-24.0, 70.0}, {-55.0,103.0}, {13.0,75.0}};
 
      dcmplx     zta8h[15], ztc8h[9];
 
      dcmplx ztas8h[15] = {
       {2.0,0.0},{3.0,3.0},{1.0,3.0},{3.0,3.0},{1.0,9.0},
       {3.0,2.0},{8.0,0.0},{2.0,1.0},{8.0,0.0},{3.0,0.0},
       {4.0,1.0},{2.0,5.0},{6.0,0.0},{2.0,5.0},{6.0,7.0}};
 
      dcmplx ztcs8h[9] = {
                   {6.0, 0.0}, {3.0, 4.0}, {9.0, 1.0},
                   {0.0, 0.0}, {10.0,0.0}, {12.0,2.0},
                   {0.0, 0.0}, {0.0, 0.0}, {3.0, 0.0}};
 
      dcmplx zresc8h[9] = {
                    {138.0,  0.0}, { 65.0, 80.0}, {134.0, 46.0},
                    {  0.0,  0.0}, {165.0,  0.0}, { 88.0,-88.0},
                    {  0.0,  0.0}, {  0.0,  0.0}, {199.0, 0.0}};
 
  double dta9[25],dtb9[15];
  double dtas9[25] = {  3.,   0.,   0.,   0.,   0.,
                       -1.,  -2.,   0.,   0.,   0.,
                        2.,   4.,  -3.,   0.,   0.,
                        2.,  -1.,   0.,   4.,   0.,
                        1.,   3.,   2.,  -2.,   1. };
 
  double dtbs9[15] = {  2.,   5.,   0.,   3.,  -1.,
                        3.,   5.,   1.,   1.,   2.,
                        1.,   4.,   2.,  -3.,   1. };
 
  double dresb9[15] = {  6., -16.,  -2.,  14.,  -1.,
                        10.,  -1.,   1.,   0.,   2.,
                        -2.,   6.,  -4., -14.,   1. };
 
 
/*  PZTRMM : 9     */
  dcmplx zta9[25],ztb9[5];
  dcmplx ztas9[25] = {
       {-4.0, 1.0}, { 0.0, 0.0}, { 0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0},
       { 4.0,-3.0}, {-2.0, 0.0}, { 0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0},
       {-1.0, 3.0}, {-3.0,-1.0}, {-5.0, 3.0}, {0.0, 0.0}, {0.0, 0.0},
       { 0.0, 0.0}, {-2.0,-1.0}, {-3.0,-3.0}, {4.0,-4.0}, {0.0, 0.0},
       {-1.0, 0.0}, { 4.0, 3.0}, {-5.0,-5.0}, {2.0, 0.0}, {2.0,-1.0}};
 
  dcmplx ztbs9[ 5] = {
       { 3.0, 4.0}, {-4.0, 2.0}, {-5.0, 0.0}, {1.0, 3.0}, {3.0, 1.0}};
 
  dcmplx zresb9[ 5] = {
       {-8.0,-19.0},{ 8.0, 21.0},{44.0,-8.0},{13.0,-7.0},{19.0, 2.0}};
 
/* PDSYMV, PZHEMV    START :10   */
 
   double dta10[64], dtx10[8];
   double dtas10[64] = {  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,
                         -1.0,  1.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,
                         -1.0,  0.0, -1.0,  0.0,  0.0,  0.0,  0.0,  0.0,
                          0.0,  1.0, -1.0, -1.0,  0.0,  0.0,  0.0,  0.0,
                          0.0,  0.0,  0.0,  1.0, -1.0,  0.0,  0.0,  0.0,
                          0.0,  1.0,  0.0,  1.0,  0.0,  1.0,  0.0,  0.0,
                          0.0,  0.0,  1.0,  0.0,  0.0,  0.0,  0.0,  0.0,
                          0.0,  1.0,  0.0,  1.0,  0.0,  0.0,  0.0,  0.0 };
 
   double dtxs10[8] = {  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0 };
   double dty10[8] = {  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0 };
   double dresy10[8] = { -2.0,  3.0, -2.0,  2.0,  0.0,  3.0,  1.0,  2.0 };
 
      dcmplx  zta10[36], ztx10[6];
      dcmplx ztas10[36] = {
       {0., 0.},{0., 0.},{0., 0.},{0., 0.},{0., 0.},
       {0., 0.},
       {-1.,1.},{2.,-1.},{0., 0.},{0., 0.},{0., 0.},
       {0., 0.},
       {-1.,-2.},{5., 4.},{0.,0.},{0., 0.},{0., 0.},
       {0., 0.},
       { 0., 3.},{2., 0.},{-1.,0.},{4.,2.},{0., 0.},
       {0., 0.},
       {2., 1.},{-1.,-1.},{0.,0.},{2.,-1.},{-1.,2.},
       {0., 0.},
       {1., 0.},{ 0., 2.},{1.,1.},{-1.,-1.},{0.,2.},
       {1., 0.}};
 
      dcmplx ztxs10[6] = {   {-1.,1.},{2.,1.},{1.,2.},{-2.,-3.},
                     {0., 1.},{1.,0.}};
      dcmplx zty10[6] = {   { 0.,0.},{0.,0.},{0.,0.},{ 0., 0.},
                     {0., 0.},{0.,0.}};
      dcmplx  zresy10 [6] = { {9.,-7.},{0.,11.},{16.,-2.},
                     {-2.,-8.},{-5.,-3.},{12.,-1.}};
 
/* PDGEMM    START :11*/
 
   double dta11[30], dtb11[20], dtc11[24];
   double dtas11[30] = {  1.0,  2.0,  1.0, -3.0,  4.0, -1.0,
                          2.0,  0.0, -1.0,  2.0,  0.0, -1.0,
                         -1.0,  1.0, -1.0,  2.0, -2.0,  1.0,
                         -1.0,  1.0,  1.0,  2.0,  1.0, -3.0,
                          4.0, -1.0,  2.0,  0.0, -1.0,  2.0 };
 
   double dtbs11[20] = {  1.0,  2.0,  1.0, -3.0,  4.0,
                         -1.0,  2.0,  0.0, -1.0,  2.0,
                          0.0, -1.0, -1.0,  1.0, -1.0,
                          2.0, -2.0,  1.0, -1.0,  1.0 };
 
   double dtcs11[24] = {  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                          0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                          0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                          0.5,  0.5,  0.5,  0.5,  0.5,  0.5 };
 
   double dresc11[24] = { 24.0, -3.0,  4.0, -2.0, -4.0, 16.0,
                          13.0, -4.0,  1.0,  6.0, -6.0,  7.0,
                          -5.0,  2.0,  2.0, -1.0,  5.0, -4.0,
                           3.0,  4.0,  5.0, -9.0,  5.0,  7.0 };
 
/* PZGEMM    */
 
  dcmplx  zta11[18], ztb11[6], ztc11[12];
  dcmplx  ztas11[18] = {
       {1.0, 5.0}, {2.0, 4.0}, {3.0, 3.0},
       {4.0, 2.0}, {5.0, 1.0}, {6.0, 6.0},
       {9.0, 2.0}, {8.0, 3.0}, {7.0, 5.0},
       {4.0, 7.0}, {5.0, 1.0}, {3.0, 6.0},
       {1.0, 9.0}, {1.0, 8.0}, {1.0, 7.0},
       {1.0, 5.0}, {1.0, 6.0}, {1.0, 4.0}};
 
  dcmplx  ztbs11[6] = {
       {1.0, 8.0}, {4.0, 4.0}, {6.0, 2.0},
       {2.0, 7.0}, {6.0, 8.0}, {4.0, 5.0}};
 
  dcmplx  ztcs11[12] = {
       {0.5, 0.0}, {0.5, 0.0}, {0.5, 0.0},
       {0.5, 0.0}, {0.5, 0.0}, {0.5, 0.0},
       {0.5, 0.0}, {0.5, 0.0}, {0.5, 0.0},
       {0.5, 0.0}, {0.5, 0.0}, {0.5, 0.0}};
 
 
  dcmplx  zresc11[12] = {
      {-22.0,113.0}, {-19.0,114.0}, {-20.0,119.0},
      {-27.0,110.0}, {  8.0,103.0}, {-55.0,116.0},
      {-35.0,142.0}, {-35.0,141.0}, {-43.0,146.0},
      {-58.0,131.0}, {  0.0,112.0}, {-75.0,135.0}};
 
/* PDGEMV, PZGEMV    START :12    */
 
   double dta12[30], dtx12[20], dty12[24];
   double dtas12[30] = {  1.0,  2.0,  1.0, -3.0,  4.0, -1.0,
                          2.0,  0.0, -1.0,  2.0,  0.0, -1.0,
                         -1.0,  1.0, -1.0,  2.0, -2.0,  1.0,
                         -1.0,  1.0,  1.0,  2.0,  1.0, -3.0,
                          4.0, -1.0,  2.0,  0.0, -1.0,  2.0 };
 
   double dtxs12[20] = {  1.0,  2.0,  1.0, -3.0,  4.0,
                         -1.0,  2.0,  0.0, -1.0,  2.0,
                          0.0, -1.0, -1.0,  1.0, -1.0,
                          2.0, -2.0,  1.0, -1.0,  1.0 };
 
   double dtys12[24] = {  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                          0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                          0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                          0.5,  0.5,  0.5,  0.5,  0.5,  0.5 };
 
   double dresy12[24] = {  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                           0.5,  0.5,  1.0,  6.0, -6.0,  7.0,
                           0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
                           0.5,  0.5,  0.5,  0.5,  0.5,  0.5 };
 
      dcmplx     zta12[18], ztx12[6], zty12[12];
      dcmplx ztas12[18] = {
       {1.,5.},{2.,4.},{3.,3.},{4.,2.},{5.,1.},{6.,6.},
       {9.,2.},{8.,3.},{7.,5.},{4.,7.},{5.,1.},{3.,6.},
       {1.,9.},{1.,8.},{1.,7.},{1.,5.},{1.,6.},{1.,4.}};
 
      dcmplx ztxs12[6] = {
                      {1., 8.},{4., 4.},{6., 2.},
                      {2., 7.},{6., 8.},{4., 5.}};
 
      dcmplx ztys12[12] = {
       {0.5, 0.},{0.5, 0.},{0.5, 0.},{0.5, 0.},
       {0.5, 0.},{0.5, 0.},
       {0.5, 0.},{0.5, 0.},{0.5, 0.},{0.5, 0.},
       {0.5, 0.},{0.5, 0.}};
 
      dcmplx zresy12[12] = {
       {0.5, 0.},{0.5, 0.},{0.5, 0.},{0.5, 0.},
       {0.5, 0.},{0.5, 0.},
       {-35.,142.},{-35.,141.},{-43.,146.},{-58.,131.},
       {0., 112.},{-75., 135.}};
 
/* PDSYR, PZHER     START :13   */
 
   double dta13[81], dtx13[9];
   double dtxs13[9]  = { 1., 1., 1., 1., 1., 1., 1., 1., 1. };
   double dtas13[81] = { 1.,   2.,   3.,   4.,   5.,   6.,   7.,   8.,   9.,
                        11.,  12.,  13.,  14.,  15.,  16.,  17.,  18.,  19.,
                        21.,  22.,  23.,  24.,  25.,  26.,  27.,  28.,  29.,
                        31.,  32.,  33.,  34.,  35.,  36.,  37.,  38.,  39.,
                        41.,  42.,  43.,  44.,  45.,  46.,  47.,  48.,  49.,
                        51.,  52.,  53.,  54.,  55.,  56.,  57.,  58.,  59.,
                        61.,  62.,  63.,  64.,  65.,  66.,  67.,  68.,  69.,
                        71.,  72.,  73.,  74.,  75.,  76.,  77.,  78.,  79.,
                        81.,  82.,  83.,  84.,  85.,  86.,  87.,  88.,  89. };
 
   double dresa13[81] = {2.,   3.,   4.,   5.,   6.,   7.,   8.,   9.,  10.,
                        11.,  13.,  14.,  15.,  16.,  17.,  18.,  19.,  20.,
                        21.,  22.,  24.,  25.,  26.,  27.,  28.,  29.,  30.,
                        31.,  32.,  33.,  35.,  36.,  37.,  38.,  39.,  40.,
                        41.,  42.,  43.,  44.,  46.,  47.,  48.,  49.,  50.,
                        51.,  52.,  53.,  54.,  55.,  57.,  58.,  59.,  60.,
                        61.,  62.,  63.,  64.,  65.,  66.,  68.,  69.,  70.,
                        71.,  72.,  73.,  74.,  75.,  76.,  77.,  79.,  80.,
                        81.,  82.,  83.,  84.,  85.,  86.,  87.,  88.,  90. };
 
      dcmplx     zta13[9], ztx13[3];
      dcmplx ztxs13[3] = { {1., 2.},{4., 0.},{3., 4.}};
      dcmplx ztas13[9] = {
       {1., 0.}, {3., -5.}, {2., 3.},
       {0., 0.}, {7.,  0.}, {4., 8.},
       {0., 0.}, {0.,  0.}, {6., 0.}};
 
      dcmplx zresa13[9] = {
       {6., 0.}, {7., -13.}, {13., 1.},
       {0., 0.}, {23.,  0.}, {16., 24.},
       {0., 0.}, {0.,  0.},  {31., 0.}};
 
/* PDSYR2, PZHER2    START :14     */
 
   double dta14[81], dtx14[9], dty14[9];
   double dtxs14[9]  = { 1., 1., 1., 1., 1., 1., 1., 1., 1. };
   double dtys14[9]  = { 2., 2., 2., 2., 2., 2., 2., 2., 2. };
   double dtas14[81] = { 1.,   2.,   3.,   4.,   5.,   6.,   7.,   8.,   9.,
                        11.,  12.,  13.,  14.,  15.,  16.,  17.,  18.,  19.,
                        21.,  22.,  23.,  24.,  25.,  26.,  27.,  28.,  29.,
                        31.,  32.,  33.,  34.,  35.,  36.,  37.,  38.,  39.,
                        41.,  42.,  43.,  44.,  45.,  46.,  47.,  48.,  49.,
                        51.,  52.,  53.,  54.,  55.,  56.,  57.,  58.,  59.,
                        61.,  62.,  63.,  64.,  65.,  66.,  67.,  68.,  69.,
                        71.,  72.,  73.,  74.,  75.,  76.,  77.,  78.,  79.,
                        81.,  82.,  83.,  84.,  85.,  86.,  87.,  88.,  89. };
 
   double dresa14[81] = { 5.,   6.,   7.,   8.,   9.,  10., 11.,  12.,  13.,
                         11.,  16.,  17.,  18.,  19.,  20., 21.,  22.,  23.,
                         21.,  22.,  27.,  28.,  29.,  30., 31.,  32.,  33.,
                         31.,  32.,  33.,  38.,  39.,  40., 41.,  42.,  43.,
                         41.,  42.,  43.,  44.,  49.,  50., 51.,  52.,  53.,
                         51.,  52.,  53.,  54.,  55.,  60., 61.,  62.,  63.,
                         61.,  62.,  63.,  64.,  65.,  66., 71.,  72.,  73.,
                         71.,  72.,  73.,  74.,  75.,  76., 77.,  82.,  83.,
                         81.,  82.,  83.,  84.,  85.,  86., 87.,  88.,  93. };
 
   dcmplx zta14[9], ztx14[3], zty14[3];
   dcmplx ztxs14[3] = { {1., 2.},{4., 0.},{3., 4.}};
   dcmplx ztys14[3] = { {1., 0.},{2.,-1.},{2., 1.}};
   dcmplx ztas14 [9] = {
       {1., 0.}, {3., -5.}, {2., 3.},
       {0., 0.}, {7.,  0.}, {4., 8.},
       {0., 0.}, {0.,  0.}, {6., 0.}};
 
   dcmplx zresa14 [9] = {
       {3., 0.}, {7., -10.}, {9., 4.},
       {0., 0.}, {23.,  0.}, {14., 23.},
       {0., 0.}, {0.,  0.},  {26., 0.}};
 
/* PDTRSM, PZTRSM    START :15  */
 
   double dta15[25], dtb15[15];
   double dtas15[25] = {  3.,   0.,   0.,   0.,   0.,
                         -1.,  -2.,   0.,   0.,   0.,
                          2.,   4.,  -3.,   0.,   0.,
                          2.,  -1.,   0.,   4.,   0.,
                          1.,   3.,   2.,  -2.,   1. };
 
   double dtbs15[15] = {  6., -16.,  -2.,  14.,  -1.,
                         10.,  -1.,   1.,   0.,   2.,
                         -2.,   6.,  -4., -14.,   1. };
 
   double dresb15[15] = {  2.,   5.,   0.,   3.,  -1.,
                           3.,   5.,   1.,   1.,   2.,
                           1.,   4.,   2.,  -3.,   1. };
 
      dcmplx   zta15[25], ztb15[10];
      dcmplx   ztas15[25] = {
                      {-4.,1.},{0., 0.},{0.,0.},{0.,0.},{0.,0.},
        { 4.,-3.},{-2., 0.},{0.,0.},{0.,0.},{0.,0.},
       {-1., 3.},{-3.,-1.},{-5.,3.},{0.,0.},{0.,0.},
      {0., 0.},{-2.,-1.},{-3.,-3.},{4.,-4.},{0.,0.},
      {-1.,0.},{ 4., 3.},{-5.,-5.},{2.,0.},{2.,-1.}};
 
      dcmplx ztbs15[10] = {
        {5.5,-13.50},{-6.5,14.5},{26.0,18.0},{10.0,3.0},{8.5,10.5},
        {-3.0,-5.0},{-3.0, 5.0},{ 4.0,-3.0},{ 6.0,-6.0},{13.0,-12.0}};
 
      dcmplx zresb15[10] = {
         { 3.,4.},{-4.,2.},{-5.,0.},{1.,3.},{3.,1.},
         { 2.,0.},{3.,-1.},{-1.,2.},{0.,-2.},{1.,3.}};
 
  char modn[9], *modname=modn;
  modname = "PDTRMV  ";
 
  blacs_pinfo(&iam,&nnodes);
  nprow = 1;
  npcol = nnodes;
  blacs_get(0,0,&ictx);
  blacs_gridinit(&ictx,"R",nprow,npcol);
  blacs_gridinfo(ictx,&nprow,&npcol,&myrow,&mycol);
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 12;
  desc_a[N_] = 12;
  desc_a[MB_] = 3;
  desc_a[NB_] = 3;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(12,3,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 12;
  desc_x[N_] = 1;
  desc_x[MB_] = 3;
  desc_x[NB_] = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(12,3,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  pdmtxcpy(12, 12, 1, 1, dta1, desc_a, dtas1);
  pdvtrcpy(12, 1, 1, dtx1, desc_x, dtxs1, 1);
 
  pdtrmv("U","N","N", 12, dta1, 1, 1, desc_a, dtx1, 1, 1, desc_x, 1);
  pdvtchk(modname,12, 1, 1, dtx1, desc_x, dresx1, 1, tol, ierr,&ilerr,
         &error, ifail);
 
  modname = "PDTRSV  ";
  pdvtrcpy(12, 1, 1, dtx1, desc_x, dresx1, 1);
  pdtrsv("U","N","N", 12, dta1, 1, 1, desc_a, dtx1, 1, 1, desc_x, 1);
  pdvtchk(modname,12, 1, 1, dtx1, desc_x, dtxs1, 1, tol, ierr,&ilerr,
         &error, ifail);
 
  modname = "PZTRMV  ";
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 4;
  desc_a[N_] = 4;
  desc_a[MB_] = 3;
  desc_a[NB_] = 3;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(4,3,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 4;
  desc_x[N_] = 1;
  desc_x[MB_] = 3;
  desc_x[NB_] = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(4,3,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  pzmtxcpy(4, 4, 1, 1, zta1, desc_a, ztas1);
  pzvtrcpy(4, 1, 1, ztx1, desc_x, ztxs1, 1);
 
  pztrmv("L","N","U", 4, zta1, 1, 1, desc_a, ztx1, 1, 1, desc_x, 1);
  pzvtchk(modname,4, 1, 1, ztx1, desc_x, zresx1, 1, tol, ierr,&ilerr,
         &error, ifail);
 
  modname = "PZTRSV  ";
  pzvtrcpy(4, 1, 1, ztx1, desc_x, zresx1, 1);
  pztrsv("L","N","U", 4,  zta1, 1, 1, desc_a, ztx1, 1, 1, desc_x, 1);
  pzvtchk(modname,4, 1, 1, ztx1, desc_x, ztxs1, 1, tol, ierr,&ilerr,
         &error, ifail);
 
/* PDTRMV, PZTRMV and PDTRSV,PZTRSV END  */
 
/* PDTRAN, PZTRANU, PZTRANC      START           */
 
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 8;
  desc_a[N_] = 9;
  desc_a[MB_] = 2;
  desc_a[NB_] = 4;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_c[DT_] = BLOCK_CYCLIC_2D;
  desc_c[M_] = 9;
  desc_c[N_] = 8;
  desc_c[MB_] = 4;
  desc_c[NB_] = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(9,4,myrow,0,nprow);
  desc_c[LLD_] = MAX(1,il_lda);
 
  modname = "PDTRAN  ";
  pdmtxcpy(8, 9, 1, 1, dta2, desc_a, dtas2);
  pdmtxcpy(9, 8, 1, 1, dtc2, desc_c, dtcs2);
 
  pdtran(9, 8, 1.0, dta2, 1, 1, desc_a, 1.0, dtc2, 1, 1, desc_c);
  pdmtchk(modname,9, 8, 1, 1, dtc2, desc_c, dresc2, tol, ierr,&ilerr,
         &error, ifail);
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 8;
  desc_a[N_]    = 7;
  desc_a[MB_]   = 2;
  desc_a[NB_]   = 3;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_c[DT_]   = BLOCK_CYCLIC_2D;
  desc_c[M_]    = 7;
  desc_c[N_]    = 8;
  desc_c[MB_]   = 3;
  desc_c[NB_]   = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(7,3,myrow,0,nprow);
  desc_c[LLD_]  = MAX(1,il_lda);
 
  modname = "PZTRANU ";
  CSET(alpha ,1.0,0.0);
  CSET(beta ,1.0,0.0);
  pzmtxcpy(8, 7, 1, 1, zta2, desc_a, ztas2);
  pzmtxcpy(7, 8, 1, 1, ztc2, desc_c, ztcs2);
 
  pztranu(7, 8,alpha, zta2, 1, 1, desc_a, beta, ztc2, 1, 1, desc_c);
  pzmtchk(modname,7, 8, 1, 1, ztc2, desc_c, zresc2u, tol, ierr,&ilerr,
         &error, ifail);
 
  modname = "PZTRANC ";
  pzmtxcpy(8, 7, 1, 1, zta2, desc_a, ztas2);
  pzmtxcpy(7, 8, 1, 1, ztc2, desc_c, ztcs2);
 
  pztranc(7, 8,alpha, zta2, 1, 1, desc_a, beta, ztc2, 1, 1, desc_c);
  pzmtchk(modname,7, 8, 1, 1, ztc2, desc_c, zresc2c, tol, ierr,&ilerr,
         &error, ifail);
 
/* PDTRAN END, PZTRANU, PZTRANC     */
 
/* PDSYR2K, PZSYR2K, PZHER2K START  */
 
 
   for (i=0; i<81; i++)
     dtc3[i] = 0.0;
 
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 8;
  desc_a[N_] = 9;
  desc_a[MB_] = 2;
  desc_a[NB_] = 4;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_b[M_] = 8;
  desc_b[N_] = 9;
  desc_b[MB_] = 2;
  desc_b[NB_] = 4;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_b[LLD_] = MAX(1,il_lda);
 
  desc_c[DT_] = BLOCK_CYCLIC_2D;
  desc_c[M_] = 9;
  desc_c[N_] = 9;
  desc_c[MB_] = 4;
  desc_c[NB_] = 4;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(9,4,myrow,0,nprow);
  desc_c[LLD_] = MAX(1,il_lda);
 
  modname = "PDSYR2K ";
  CSET(alpha,1.0,0.0);
  CSET(beta,0.0,0.0);
 
  pdmtxcpy(8, 9, 1, 1, dta3, desc_a, dtas3);
  pdmtxcpy(8, 9, 1, 1, dtb3, desc_b, dtbs3);
 
  pdsyr2k("U","T", 9, 8, 1.0, dta3, 1, 1, desc_a, dtb3, 1, 1, desc_b, 0.0,
          dtc3, 1, 1, desc_c);
  pdmtchk(modname,9, 9, 1, 1, dtc3, desc_c, dresc3, tol, ierr,&ilerr,
         &error, ifail);
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 8;
  desc_a[N_] = 7;
  desc_a[MB_] = 2;
  desc_a[NB_] = 3;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_b[M_] = 8;
  desc_b[N_] = 7;
  desc_b[MB_] = 2;
  desc_b[NB_] = 3;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_b[LLD_] = MAX(1,il_lda);
 
  desc_c[DT_] = BLOCK_CYCLIC_2D;
  desc_c[M_] = 7;
  desc_c[N_] = 7;
  desc_c[MB_] = 3;
  desc_c[NB_] = 3;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(7,3,myrow,0,nprow);
  desc_c[LLD_] = MAX(1,il_lda);
 
   for (i=0; i<49; i++)
   CSET(ztc3[i],0.0,0.0);
  modname = "PZSYR2K ";
  CSET(alpha,1.0,0.0);
  CSET(beta,0.0,0.0);
  pzmtxcpy(8, 7, 1, 1, zta3, desc_a, ztas3);
  pzmtxcpy(8, 7, 1, 1, ztb3, desc_b, ztbs3);
 
  pzsyr2k("U","T", 7, 8,alpha, zta3,1,1,desc_a,ztb3, 1,1, desc_b,beta,
          ztc3, 1, 1, desc_c);
  pzmtchk(modname,7, 7, 1, 1, ztc3, desc_c, zresc3, tol, ierr,&ilerr,
         &error, ifail);
 
   for (i=0; i<49; i++)
   CSET(ztc3[i],0.0,0.0);
  modname = "PZHER2K ";
  pzmtxcpy(8, 7, 1, 1, zta3, desc_a, ztas3);
  pzmtxcpy(8, 7, 1, 1, ztb3, desc_b, ztbs3);
 
  pzher2k("U","C", 7, 8,alpha, zta3,1,1,desc_a,ztb3, 1,1, desc_b, 0.0,
          ztc3, 1, 1, desc_c);
  pzmtchk(modname,7, 7, 1, 1, ztc3, desc_c, zresc3h, tol, ierr,&ilerr,
         &error, ifail);
 
/* PDSYR2K, PZSYR2K, PZHER2K END   */
 
/* PDSYMM, PZSYMM, PZHEMM  START  */
 
 
   for (i=0; i<128; i++)
     dtc4[i] = 0.0;
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 8;
  desc_a[N_] = 8;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_b[M_] = 16;
  desc_b[N_] = 8;
  desc_b[MB_] = 4;
  desc_b[NB_] = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(16,4,myrow,0,nprow);
  desc_b[LLD_] = MAX(1,il_lda);
 
  desc_c[DT_] = BLOCK_CYCLIC_2D;
  desc_c[M_] = 16;
  desc_c[N_] = 8;
  desc_c[MB_] = 4;
  desc_c[NB_] = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(16,4,myrow,0,nprow);
  desc_c[LLD_] = MAX(1,il_lda);
 
  modname = "PDSYMM  ";
  pdmtxcpy(8, 8, 1, 1, dta4, desc_a, dtas4);
  pdmtxcpy(16, 8, 1, 1, dtb4, desc_b, dtbs4);
 
  pdsymm("R","U", 16, 8, 1.0, dta4, 1, 1, desc_a, dtb4, 1, 1, desc_b, 0.0,
          dtc4, 1, 1, desc_c);
  pdmtchk(modname,16, 8, 1, 1, dtc4, desc_c, dresc4, tol, ierr,&ilerr,
         &error, ifail);
 
/* ztc4 was zero out because PZSYMM and PZHEMM update upper or lover
   triangular part but PZMTCHK cheks all data in the ztc4          */
 
   for (i=0; i<128; i++)
     CSET(ztc4[i], 0.0,0.0);
     CSET(alpha, 1.0,2.0);
     CSET(beta, 0.0,0.0);
 
  modname = "PZSYMM  ";
  pzmtxcpy(8, 8, 1, 1, zta4, desc_a, ztas4);
  pzmtxcpy(16, 8, 1, 1, ztb4, desc_b, ztbs4);
 
  pzsymm("R","U", 16, 8,alpha, zta4, 1, 1, desc_a, ztb4, 1, 1, desc_b,
          beta, ztc4, 1, 1, desc_c);
  pzmtchk(modname,16, 8, 1, 1, ztc4, desc_c, zresc4, tol, ierr,&ilerr,
         &error, ifail);
 
   for (i=0; i<128; i++)
     CSET(ztc4h[i], 0.0,0.0);
     CSET(alpha, 1.0,0.0);
     CSET(beta, 0.0,0.0);
 
  modname = "PZHEMM  ";
  pzmtxcpy(8, 8, 1, 1, zta4h, desc_a, ztas4h);
  pzmtxcpy(16, 8, 1, 1, ztb4h, desc_b, ztbs4h);
 
  pzhemm("R","U", 16, 8,alpha, zta4h, 1, 1, desc_a,ztb4h, 1, 1, desc_b,
          beta, ztc4h, 1, 1, desc_c);
  pzmtchk(modname,16, 8, 1, 1, ztc4h, desc_c, zresc4h, tol, ierr,&ilerr,
         &error, ifail);
 
/* PDSYMM, PZSYMM, PZHEMM END      */
 
/* PDGER, PZGERU, PZGERC START : 5 */
 
   for (i=1; i<101; i++)
     dtas5[i-1] = (double)i;
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 10;
  desc_a[N_] = 10;
  desc_a[MB_] = 4;
  desc_a[NB_] = 4;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(10,4,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 11;
  desc_x[N_] = 1;
  desc_x[MB_] = 4;
  desc_x[NB_] = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(11,4,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 1;
  desc_y[N_] = 11;
  desc_y[MB_] = 1;
  desc_y[NB_] = 4;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(1,1,myrow,0,nprow);
  desc_y[LLD_] = MAX(1,il_lda);
 
  modname = "PDGER   ";
  pdmtxcpy(10, 10, 1, 1, dta5, desc_a, dtas5);
  pdvtrcpy(11, 1, 1, dtx5, desc_x, dtxs5, 1);
  pdmtxcpy(1, 11, 1, 1, dty5, desc_y, dtys5);
 
  pdger(9, 9, 1., dtx5, 2, 1, desc_x, 1, dty5, 1, 2, desc_y, 1, dta5, 2, 2,
        desc_a);
  pdmtchk(modname, 9,  9, 2, 2, dta5, desc_a, dresa5, tol, ierr,&ilerr,
         &error, ifail);
 
  modname = "PZGERU  ";
  CSET(alpha,1.,-1.);
  pzmtxcpy(10, 10, 1, 1, zta5, desc_a, ztas5);
  pzvtrcpy(11, 1, 1, ztx5, desc_x, ztxs5, 1);
  pzmtxcpy(1, 11, 1, 1, zty5, desc_y, ztys5);
 
  pzgeru(9,9,alpha, ztx5, 2, 1, desc_x, 1, zty5, 1,2,desc_y,1,zta5, 2,2,
        desc_a);
  pzmtchk(modname, 9,  9, 2, 2, zta5, desc_a, zresa5u, tol, ierr,&ilerr,
         &error, ifail);
 
  modname = "PZGERC  ";
  pzmtxcpy(10, 10, 1, 1, zta5, desc_a, ztas5);
  pzvtrcpy(11, 1, 1, ztx5, desc_x, ztxs5, 1);
  pzmtxcpy(1, 11, 1, 1, zty5, desc_y, ztys5);
 
  pzgerc(9,9,alpha, ztx5, 2, 1, desc_x, 1, zty5, 1,2,desc_y,1,zta5, 2,2,
        desc_a);
  pzmtchk(modname, 9,  9, 2, 2, zta5, desc_a, zresa5c, tol, ierr,&ilerr,
         &error, ifail);
 
/* PDGER, PZGERU, PZGERC END                           */
 
/* PDSYRK, PZSYRK, PZHERK    START : 8, PDTRMM, PZTRMM : 9     */
 
   for (i=0; i<64; i++)
     dtc8[i] = 0.0;
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 8;
  desc_a[N_] = 5;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_c[DT_] = BLOCK_CYCLIC_2D;
  desc_c[M_] = 8;
  desc_c[N_] = 8;
  desc_c[MB_] = 2;
  desc_c[NB_] = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_c[LLD_] = MAX(1,il_lda);
 
 
  modname = "PDSYRK  ";
  pdmtxcpy(8, 5, 1, 1, dta8, desc_a, dtas8);
  pdmtxcpy(8, 8, 1, 1, dtc8, desc_c, dtcs8);
  pdsyrk("L","N",8, 5, 1.0, dta8, 1, 1, desc_a, 1.0, dtc8, 1, 1, desc_c);
 
  pdmtchk(modname,8, 8, 1, 1, dtc8, desc_c, dresc8, tol, ierr,&ilerr,
         &error, ifail);
 
/* PDSYRK    END     */
 
   for (i=0; i<9; i++)
    CSET(ztc8[i], 0., 0.);
    CSET(alpha, 1., 1.);
    CSET(beta, 1., 1.);
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 3;
  desc_a[N_] = 5;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_c[DT_] = BLOCK_CYCLIC_2D;
  desc_c[M_] = 3;
  desc_c[N_] = 3;
  desc_c[MB_] = 2;
  desc_c[NB_] = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_c[LLD_] = MAX(1,il_lda);
 
 
  modname = "PZSYRK  ";
  pzmtxcpy(3, 5, 1, 1, zta8, desc_a, ztas8);
  pzmtxcpy(3, 3, 1, 1, ztc8, desc_c, ztcs8);
  pzsyrk("U","N",3,5,alpha,zta8,1,1, desc_a, beta, ztc8, 1, 1, desc_c);
 
  pzmtchk(modname,3, 3, 1, 1, ztc8, desc_c, zresc8, tol, ierr,&ilerr,
         &error, ifail);
 
/* PZSYRK    END     */
 
 
   for (i=0; i<9; i++)
    CSET(ztc8h[i], 0., 0.);
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 5;
  desc_a[N_] = 3;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_c[DT_] = BLOCK_CYCLIC_2D;
  desc_c[M_] = 3;
  desc_c[N_] = 3;
  desc_c[MB_] = 2;
  desc_c[NB_] = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_c[LLD_] = MAX(1,il_lda);
 
 
  modname = "PZHERK  ";
  pzmtxcpy(5, 3, 1, 1, zta8h, desc_a, ztas8h);
  pzmtxcpy(3, 3, 1, 1, ztc8h, desc_c, ztcs8h);
  pzherk("L","C",3,5, 1.0 ,zta8h,1,1, desc_a, 1.0, ztc8h, 1, 1,desc_c);
 
  pzmtchk(modname,3, 3, 1, 1, ztc8h, desc_c, zresc8h, tol, ierr,&ilerr,
         &error, ifail);
 
/* PZHERK    END     */
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 5;
  desc_a[N_] = 5;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_b[M_] = 5;
  desc_b[N_] = 3;
  desc_b[MB_] = 2;
  desc_b[NB_] = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_b[LLD_] = MAX(1,il_lda);
 
  modname = "PDTRMM  ";
  pdmtxcpy(5, 5, 1, 1, dta9, desc_a, dtas9);
  pdmtxcpy(5, 3, 1, 1, dtb9, desc_b, dtbs9);
  pdtrmm("L","U","N", "N", 5, 3, 1.0, dta9, 1, 1, desc_a, dtb9, 1, 1, desc_b);
 
  pdmtchk(modname,5, 3, 1, 1, dtb9, desc_b, dresb9, tol, ierr,&ilerr,
         &error, ifail);
 
/* PDTRMM    END         */
 
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 5;
  desc_a[N_] = 5;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_b[M_] = 5;
  desc_b[N_] = 1;
  desc_b[MB_] = 2;
  desc_b[NB_] = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_b[LLD_] = MAX(1,il_lda);
 
  modname = "PZTRMM  ";
  CSET(alpha, 1., 0.);
  pzmtxcpy(5, 5, 1, 1, zta9, desc_a, ztas9);
  pzmtxcpy(5, 1, 1, 1, ztb9, desc_b, ztbs9);
  pztrmm("L","U","C", "N", 5, 1,alpha,zta9, 1, 1, desc_a, ztb9, 1, 1, desc_b);
 
  pzmtchk(modname,5, 1, 1, 1, ztb9, desc_b, zresb9, tol, ierr,&ilerr,
         &error, ifail);
 
/* PZTRMM    END         */
 
/* PDSYMV, PZHEMV    START :10   */
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 8;
  desc_a[N_] = 8;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 8;
  desc_x[N_] = 1;
  desc_x[MB_] = 2;
  desc_x[NB_] = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 8;
  desc_y[N_] = 1;
  desc_y[MB_] = 2;
  desc_y[NB_] = 1;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(8,2,myrow,0,nprow);
  desc_y[LLD_] = MAX(1,il_lda);
 
  modname = "PDSYMV  ";
  pdmtxcpy(8, 8, 1, 1, dta10, desc_a, dtas10);
  pdvtrcpy(8, 1, 1, dtx10, desc_x, dtxs10, 1);
 
  pdsymv("U", 8, 1.0, dta10, 1, 1, desc_a, dtx10, 1, 1, desc_x, 1, 0.0,
          dty10, 1, 1, desc_y, 1);
  pdvtchk(modname,8, 1, 1, dty10, desc_y, dresy10, 1, tol, ierr,&ilerr,
         &error, ifail);
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 6;
  desc_a[N_] = 6;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(6,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 6;
  desc_x[N_] = 1;
  desc_x[MB_] = 2;
  desc_x[NB_] = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(6,2,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 6;
  desc_y[N_] = 1;
  desc_y[MB_] = 2;
  desc_y[NB_] = 1;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(6,2,myrow,0,nprow);
  desc_y[LLD_] = MAX(1,il_lda);
 
  modname = "PZHEMV  ";
  CSET(alpha,1.0,0.0);
  CSET(beta,0.0,0.0);
  pzmtxcpy(6, 6, 1, 1, zta10, desc_a, ztas10);
  pzvtrcpy(6, 1, 1, ztx10, desc_x, ztxs10, 1);
 
  pzhemv("U", 6,alpha,zta10, 1, 1, desc_a, ztx10, 1, 1, desc_x, 1,beta,
          zty10, 1, 1, desc_y, 1);
  pzvtchk(modname,6, 1, 1, zty10, desc_y, zresy10, 1, tol, ierr,&ilerr,
         &error, ifail);
 
/* PDSYMV PZHEMV END      */
 
/* PDGEMM    START :11*/
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 6;
  desc_a[N_] = 5;
  desc_a[MB_] = 3;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(6,3,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_b[M_] = 5;
  desc_b[N_] = 4;
  desc_b[MB_] = 2;
  desc_b[NB_] = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_b[LLD_] = MAX(1,il_lda);
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_c[M_] = 6;
  desc_c[N_] = 4;
  desc_c[MB_] = 3;
  desc_c[NB_] = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(6,3,myrow,0,nprow);
  desc_c[LLD_] = MAX(1,il_lda);
 
  modname = "PDGEMM  ";
  pdmtxcpy(6, 5, 1, 1, dta11, desc_a, dtas11);
  pdmtxcpy(5, 4, 1, 1, dtb11, desc_b, dtbs11);
  pdmtxcpy(6, 4, 1, 1, dtc11, desc_c, dtcs11);
 
  pdgemm("N", "N", 6, 4, 5, 1.0, dta11, 1, 1, desc_a, dtb11, 1, 1, desc_b,
          2.0, dtc11, 1, 1, desc_c);
  pdmtchk(modname,6, 4, 1, 1, dtc11, desc_c, dresc11, tol, ierr,&ilerr,
         &error, ifail);
 
/* PDGEMM    END          */
 
/* PZGEMM    START        */
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 6;
  desc_a[N_]    = 3;
  desc_a[MB_]   = 2;
  desc_a[NB_]   = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(6,2,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_b[DT_]   = BLOCK_CYCLIC_2D;
  desc_b[M_]    = 3;
  desc_b[N_]    = 2;
  desc_b[MB_]   = 2;
  desc_b[NB_]   = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_b[LLD_]  = MAX(1,il_lda);
 
  desc_c[DT_]   = BLOCK_CYCLIC_2D;
  desc_c[M_]    = 6;
  desc_c[N_]    = 2;
  desc_c[MB_]   = 2;
  desc_c[NB_]   = 2;
  desc_c[RSRC_] = 0;
  desc_c[CSRC_] = 0;
  desc_c[CTXT_] = ictx;
  il_lda = numroc(6,2,myrow,0,nprow);
  desc_c[LLD_]  = MAX(1,il_lda);
 
  modname = "PZGEMM  ";
  pzmtxcpy(6, 3, 1, 1, zta11, desc_a, ztas11);
  pzmtxcpy(3, 2, 1, 1, ztb11, desc_b, ztbs11);
  pzmtxcpy(6, 2, 1, 1, ztc11, desc_c, ztcs11);
 
  CSET(alpha, 1., 0.);
  CSET(beta, 2., 0.);
 
  pzgemm("N", "N", 6, 2, 3, alpha, zta11, 1, 1, desc_a, ztb11,
          1, 1, desc_b, beta, ztc11, 1, 1, desc_c);
  pzmtchk(modname,6, 2, 1, 1, ztc11, desc_c, zresc11, tol,
          ierr,&ilerr, &error, ifail);
 
/* PZGEMM    END           */
 
/* PDGEMV, PZGEMV    START :12    */
 
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 6;
  desc_a[N_] = 5;
  desc_a[MB_] = 3;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(6,3,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 5;
  desc_x[N_] = 4;
  desc_x[MB_] = 2;
  desc_x[NB_] = 2;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 6;
  desc_y[N_] = 4;
  desc_y[MB_] = 3;
  desc_y[NB_] = 2;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(6,3,myrow,0,nprow);
  desc_y[LLD_] = MAX(1,il_lda);
 
  modname = "PDGEMV  ";
  pdmtxcpy(4, 5, 3, 1, dta12, desc_a, dtas12);
  pdmtxcpy(5, 3, 1, 2, dtx12, desc_x, dtxs12);
  pdmtxcpy(4, 3, 3, 2, dty12, desc_y, dtys12);
 
  pdgemv("N", 4, 5, 1.0, dta12, 3, 1, desc_a, dtx12, 1, 2, desc_x, 1,
          2.0, dty12, 3, 2, desc_y, 1);
  pdmtchk(modname,4, 3, 3, 2, dty12, desc_y, dresy12, tol, ierr,&ilerr,
         &error, ifail);
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 6;
  desc_a[N_] = 3;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(6,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 3;
  desc_x[N_] = 2;
  desc_x[MB_] = 2;
  desc_x[NB_] = 2;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 6;
  desc_y[N_] = 2;
  desc_y[MB_] = 2;
  desc_y[NB_] = 2;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(6,2,myrow,0,nprow);
  desc_y[LLD_] = MAX(1,il_lda);
 
  modname = "PZGEMV  ";
  CSET(alpha,1.0,0.0);
  CSET(beta,2.0,0.0);
 
  pzmtxcpy(6, 3, 1, 1, zta12, desc_a, ztas12);
  pzmtxcpy(3, 1, 1, 2, ztx12, desc_x, ztxs12);
  pzmtxcpy(6, 1, 1, 2, zty12, desc_y, ztys12);
 
  pzgemv("N", 6, 3,alpha, zta12, 1, 1, desc_a, ztx12, 1, 2, desc_x, 1,
         beta, zty12, 1, 2, desc_y, 1);
  pzmtchk(modname,6, 1, 1, 2, zty12, desc_y, zresy12, tol, ierr,&ilerr,
         &error, ifail);
 
/* PDGEMV, PZGEMV    END         */
 
/* PDSYR, PZHER     START :13   */
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 9;
  desc_a[N_] = 9;
  desc_a[MB_] = 4;
  desc_a[NB_] = 4;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(9,4,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 9;
  desc_x[N_] = 1;
  desc_x[MB_] = 4;
  desc_x[NB_] = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(9,4,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  modname = "PDSYR   ";
  pdmtxcpy(9, 9, 1, 1, dta13, desc_a, dtas13);
  pdvtrcpy(9, 1, 1, dtx13, desc_x, dtxs13, 1);
 
  pdsyr("L",9, 1., dtx13, 1, 1, desc_x, 1, dta13, 1, 1, desc_a);
  pdmtchk(modname,9, 9, 1, 1, dta13, desc_a, dresa13, tol, ierr,&ilerr,
         &error, ifail);
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 3;
  desc_a[N_] = 3;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 3;
  desc_x[N_] = 1;
  desc_x[MB_] = 2;
  desc_x[NB_] = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  modname = "PZHER   ";
  pzmtxcpy(3, 3, 1, 1, zta13, desc_a, ztas13);
  pzvtrcpy(3, 1, 1, ztx13, desc_x, ztxs13, 1);
 
  pzher("L",3, 1., ztx13, 1, 1, desc_x, 1, zta13, 1, 1, desc_a);
  pzmtchk(modname,3, 3, 1, 1, zta13, desc_a, zresa13, tol, ierr,&ilerr,
         &error, ifail);
 
/* PDSYR, PZHER END               */
 
/* PDSYR2, PZHER2   START :14     */
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 9;
  desc_a[N_] = 9;
  desc_a[MB_] = 4;
  desc_a[NB_] = 4;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(9,4,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 9;
  desc_x[N_] = 1;
  desc_x[MB_] = 4;
  desc_x[NB_] = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(9,4,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 9;
  desc_y[N_] = 1;
  desc_y[MB_] = 4;
  desc_y[NB_] = 1;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(9,4,myrow,0,nprow);
  desc_y[LLD_] = MAX(1,il_lda);
 
  modname = "PDSYR2  ";
  pdmtxcpy(9, 9, 1, 1, dta14, desc_a, dtas14);
  pdvtrcpy(9, 1, 1, dtx14, desc_x, dtxs14, 1);
  pdvtrcpy(9, 1, 1, dty14, desc_x, dtys14, 1);
 
  pdsyr2("L",9, 1., dtx14, 1, 1, desc_x, 1, dty14, 1, 1, desc_y, 1,
         dta14, 1, 1, desc_a);
  pdmtchk(modname,9, 9, 1, 1, dta14, desc_a, dresa14, tol, ierr,&ilerr,
         &error, ifail);
 
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 3;
  desc_a[N_] = 3;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 3;
  desc_x[N_] = 1;
  desc_x[MB_] = 2;
  desc_x[NB_] = 1;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 3;
  desc_y[N_] = 1;
  desc_y[MB_] = 2;
  desc_y[NB_] = 1;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_y[LLD_] = MAX(1,il_lda);
 
  modname = "PZHER2  ";
  CSET(alpha,1., 0.);
  pzmtxcpy(3, 3, 1, 1, zta14, desc_a, ztas14);
  pzvtrcpy(3, 1, 1, ztx14, desc_x, ztxs14, 1);
  pzvtrcpy(3, 1, 1, zty14, desc_x, ztys14, 1);
 
  pzher2("L",3, alpha, ztx14, 1, 1, desc_x, 1, zty14, 1, 1, desc_y, 1,
         zta14, 1, 1, desc_a);
  pzmtchk(modname,3, 3, 1, 1, zta14, desc_a, zresa14, tol, ierr,&ilerr,
         &error, ifail);
 
/* PDSYR2, PZHER2  END          */
 
/* PDTRSM, PZTRSM    START :15  */
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 5;
  desc_a[N_] = 5;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_b[M_] = 5;
  desc_b[N_] = 3;
  desc_b[MB_] = 2;
  desc_b[NB_] = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_b[LLD_] = MAX(1,il_lda);
 
  modname = "PDTRSM  ";
  pdmtxcpy(5, 5, 1, 1, dta15, desc_a, dtas15);
  pdmtxcpy(5, 3, 1, 1, dtb15, desc_b, dtbs15);
 
  pdtrsm("L", "U", "N", "N", 5, 3, 1.0, dta15, 1, 1, desc_a,
          dtb15, 1, 1, desc_b);
  pdmtchk(modname,5, 3, 1, 1, dtb15, desc_b, dresb15, tol, ierr,&ilerr,
         &error, ifail);
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_b[M_] = 5;
  desc_b[N_] = 2;
  desc_b[MB_] = 2;
  desc_b[NB_] = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(5,2,myrow,0,nprow);
  desc_b[LLD_] = MAX(1,il_lda);
 
  modname = "PZTRSM  ";
  CSET(alpha,1.0,-1.0);
  tol = 0.1;
  pzmtxcpy(5, 5, 1, 1, zta15, desc_a, ztas15);
  pzmtxcpy(5, 2, 1, 1, ztb15, desc_b, ztbs15);
 
  pztrsm("L", "U", "C", "N", 5, 2,alpha, zta15, 1, 1, desc_a,
          ztb15, 1, 1, desc_b);
  pzmtchk(modname,5, 2, 1, 1, ztb15, desc_b, zresb15, tol, ierr,&ilerr,
         &error, ifail);
 
 
/* PDTRSM, PZTRSM  END    */
 
 
  blacs_gridexit(ictx);
}
 
 
/*********************************************************************
*          PARALLEL EIGEN SYSTEM ANALYSIS SUBROUTINES                *
*********************************************************************/
 
void peig(int *ifail)
{
/* PDSYTRD  PZHETRD  START : 16     */
 
   int iam, nnodes,myrow,mycol,istor[1],ra[1],ca[1],ix;
   int desc_a[DLEN_], desc_x[DLEN_], il_lda, desc_b[DLEN_],il_ldb;
   int ictx, nprow, npcol,ierr=5, ilerr, i, INCX, info=0, lwork=0,iret;
   double error=0.0, tol=0.01; /*will not pass if tol is lower 0.01*/
   double dta16[16], dtd16[4], dte16[4], dtt16[4], work[5000];
   dcmplx zta16[16], ztt16[4], workz[1],rworkz[1];
   double dtas16[16] = { 5.,  0.,  0.,  0.,
                         4.,  5.,  0.,  0.,
                         1.,  1.,  4.,  0.,
                         1.,  1.,  2.,  4. };
 
   double dresa16[16] = { 1.,   0.,    0.,   0.,
                          0.,   6.,    0.,   0.,
                          0.41, 2.83,  7.0,  0.,
                          0.22, 0.22, -2.45, 4.0 };
 
   double dresd16[4] = { 1.,  6.,  7.,    4. };
   double drese16[4] = { 0.,  0.,  2.83, -2.45 };
   double drest16[4] = { 0.,  0.,  1.71,  1.82 };
 
      dcmplx ztas16[16] = {{5.,0.},{4.,1.},{1.,2.},{2.,3.},
                         {0.,0.},{5.,0.},{1.,0.},{3.,2.},
                         {0.,0.},{0.,0.},{4.,0.},{5.,1.},
                         {0.,0.},{0.,0.},{0.,0.},{4.,0.}};
 
      dcmplx zresa16[16] = {
          {5.,0.},{-5.92,0.00},{0.12,0.19},{0.23,0.28},
          {0.,0.},{10.09,0.00},{2.36,0.00},{0.14,0.19},
          {0.,0.},{ 0.00,0.00},{4.16,0.00},{1.62,0.00},
          {0.,0.},{ 0.00,0.00},{0.00,0.00},{-1.25,0.00}};
 
      double dresd161[4] = { 5., 10.09,  4.16, -1.25 };
      double drese161[4] = {-5.92,  2.36,  1.62,  0.00 };
      dcmplx zrest16[4] = {{1.68,.17},{1.87,.21},{1.96,-.27},{0.,0.}};
 
/* PDGEHRD   START : 17     */
 
   double dta17[9], dtt17[2];
   double dtas17[9] = {  33.00, -24.00,  -8.00,
                         16.00, -10.00,  -4.00,
                         72.00, -57.00, -17.00 };
 
   double dresa17[9] = {  33.00,  25.30,  0.16,
                         -37.95, -29.00,  0.00,
                          63.25,  53.00,  2.00 };
 
   double drest17[2] = { 1.95, 0.0 };
 
/* PDGEBRD PZGEBRD PDGESVD PZGESVD   START : 18  */
 
   int desc_xr[DLEN_], desc_xc[DLEN_];
   int desc_u[DLEN_],  desc_vt[DLEN_];
   int   lrwork=0;
   double rwork[1];
   double dts18[3],dtvt18[12],dtu18[12];
   double ztd18[3],zte18[2],zts18[3];
   dcmplx zta18[12],ztvt18[12],ztu18[9];
   dcmplx zttp18[3],zttq18[12];
   double dta18[12], dtd18[3], dte18[2], dttq18[3], dttp18[3];
 
   dcmplx ztvts18[12] = {
          {   .02,  -.08}, {  -.42,  -.48}, {   .20,   .16},
          {   .03,   .03}, {  -.41,   .11}, {   .17,   .78},
          {   .39,   .51}, {  -.32,   .39}, {  -.39,  -.11},
          {  -.48,  -.59}, {  -.21,   .33}, {  -.35,  -.03}
   };
 
   dcmplx zresa18[12] = {
          {-3.46,  .00}, {-2.08,  .00}, {  .11, -.40},
          {  .17, -.26}, { 4.51,  .00}, {-1.73,  .00},
          {  .43, -.10}, {  .10, -.59}, { 2.51,  .00},
          { -.10, -.43}, {  .33,  .52}, {  .01,  .11}
   };
 
   dcmplx ztas18[12] = {
          { 1.0, 1.0}, {-1.0, 1.0}, { 0.0,-1.0},
          { 1.0,-1.0}, {-1.0, 0.0}, {-1.0,-1.0},
          { 2.0, 0.0}, { 0.0,-2.0}, {-2.0, 2.0},
          { 0.0,-2.0}, { 2.0, 2.0}, { 2.0,-2.0}
   };
 
   double zresd18[3] = { -3.46, 4.51, 2.51 };
   double zrese18[2] = { -2.08, -1.73 };
 
   dcmplx zrestp18[3] = { {1.29,-0.29},{1.03,0.36},{1.88,-0.43} };
 
   dcmplx zrestq18[3] = { {1.69,-0.14},{1.91,0.41},{0.0,0.0} };
   double dtvts18[12] = {
          -0.25, 0.26, 0.93, 0.00,
          -0.54, -0.84, 0.09, 0.00,
          -0.81, 0.48, -0.35, 0.00
   };
 
   double dtus18[12] = {
          -0.37, -0.51, -0.64, -0.44,
          0.28, -0.82,  0.50, 0.00,
          0.86, -0.02, -0.51, 0.03
   };
 
   dcmplx ztus18[12] = {
          {   .34,   .00}, {  -.60,  -.11}, {   .08,   .71},
          {  -.82,   .00}, {  -.06,  -.37}, {   .34,   .25},
          {  -.45,   .00}, {  -.35,   .60}, {  -.55,   .08},
          {   .00,   .00}, {   .00,   .00}, {   .00,   .00}
   };
 
   double dtas18[12] = { 10.00,   2.00,   3.00,  4.0,
                          5.00,  16.00,   7.00,  8.0,
                          9.00,  10.00,  21.00, 12.0  };
 
   double dresa18[12] = { -11.36,   0.09,  0.14, 0.19,
                           22.80,  23.32,  0.46, 0.22,
                            0.56,   1.67, -9.68, 0.08 };
   double dtss18[3] = {33.64, 9.89, 7.70};
 
   double dresd18[3] = {-11.36, 23.32, -9.68  };
   double drese18[2] = { 22.80,  1.67 };
   double drestq18[3] = { 1.88,  1.59,  1.99  };
   double drestp18[3] = { 1.52,  0.0 ,  0.0   };
 
   double ztss18[3] = {5.51, 3.31, 2.15};
/* PDSYEVX PZHEEVX  START : 19  */
 
   int M,NZ,ifaile[4],desc_z[DLEN_],iwork[1],liwork=0;
   double dta19[16], dtz19[16], dtw19[4];
   dcmplx zta19[16], ztz19[16];
   int ifailres[4] = { 0, 0, 0, 0 };
   double *dtg19=0;
   double *dtg19res=0;
   int    *iclustr=0;
   int    *iclustres=0;
 
   double dtas19[16] = { 5.,  0.,  0.,  0.,
                         4.,  5.,  0.,  0.,
                         1.,  1.,  4.,  0.,
                         1.,  1.,  2.,  4. };
 
   double dresz19[16] = { 0.7071, -0.7071,  0.0,     0.0,
                          0.0,     0.0,    -0.7071,  0.7071,
                         -0.3162, -0.3162,  0.6325,  0.6325,
                         -0.6325, -0.6325, -0.3162, -0.3162  };
 
   double dresw19[4] = { 1.,  2.,  5.,   10. };
 
      dcmplx ztas19[16] =  { {5.,0.},{4.,1.},{1.,2.},{2.,3.},
                             {0.,0.},{5.,0.},{1.,0.},{3.,2.},
                             {0.,0.},{0.,0.},{4.,0.},{5.,1.},
                             {0.,0.},{0.,0.},{0.,0.},{4.,0.}};
 
      dcmplx zresz19[16] = { { .0926, .0000},{ .0840,-.2873},
                             { .5013,-.3461},{-.6703, .2854},
                             { .7591, .0000},{-.5874, .0177},
                             { .0526,-.0674},{-.0155,-.2662},
                             { .3758, .0000},{ .5370,-.2067},
                             {-.6287,-.2149},{-.2294,-.1834},
                             {-.5234, .0000},{-.4515,-.1735},
                             {-.2864,-.3134},{-.3059,-.4672}};
 
      double zresw19[4] = { -1.765, 0.727, 4.664, 14.374 };
 
/* PDSYGST   PDSYGVX START : 20  */
/* PZHEGST   PZHEGVX START : 20  */
 
      int     ibtype;
      double dta20[16], dtb20[16], scale, scaler=1.0,dtw20[4],dtz20[16];
      dcmplx zta20[16], ztb20[16], ztz20[16];
      double  dtc20[16], tstnrm, ztw20[4];
      double  dtas20[16] =  { -1.,  1., -1.,  1.,
                               0.,  1., -1.,  1.,
                               0.,  0.,  1., -1.,
                               0.,  0.,  0.,  1.};
 
      double dtbs20[16] =  {   2.,  1.,  0.,  0.,
                               0.,  2.,  1.,  0.,
                               0.,  0.,  2.,  1.,
                               0.,  0.,  0.,  2.};
 
      double  dtac20[16] =  { -1.,  1., -1.,  1.,
                               1.,  1., -1.,  1.,
                              -1., -1.,  1., -1.,
                               1.,  1., -1.,  1.};
 
      double dtbc20[16] =  {   2.,  1.,  0.,  0.,
                               1.,  2.,  1.,  0.,
                               0.,  1.,  2.,  1.,
                               0.,  0.,  1.,  2.};
 
      double dresa20[16]= { -0.50000, 0.86603,-1.22474, 1.58114,
                             0.00000,-0.16667,-0.23570, 0.54772,
                             0.00000, 0.00000, 1.16667,-1.93649,
                            0.00000, 0.00000, 0.00000, 3.10000};
 
      double dresz20[16]= { 0.88420,-0.56190, 0.30720,-0.11980,
                            0.00000, 0.36340, 0.42660, 0.06310,
                            0.00000,-0.40980, 0.13430, 0.54410,
                            0.13490,-0.76430, 0.95160,-0.69690};
 
      double dresw20[4]= { -1.55261, 0.00000, 0.00000, 5.15261};
 
      dcmplx ztas20[16] = {{1.,0.},{5.,-2.},{7.,4.},{9.,-6.},
                         {0.,0.},{10.,0.},{15.,6.},{20.,-4.},
                         {0.,0.},{0.,0.},{20.,0.},{25.,-9.},
                         {0.,0.},{0.,0.},{0.,0.},{30.,0.}};
 
      dcmplx ztbs20[16] = {{9.,0.},{3.,-3.},{3.,3.},{3.,-3.},
                         {0.,0.},{18.,0.},{8.,6.},{8.,-6.},
                         {0.,0.},{0.,0.},{27.,0.},{12.,-9.},
                         {0.,0.},{0.,0.},{0.,0.},{61.,0.}};
 
      dcmplx zresa20[16] = {
          {.1111,0.},{.3889,-.1389},{0.2919,0.2485},{0.2422,-.2175},
          {0.,0.},{0.3472,0.00},{.5714,-.0652},{0.2001,-.0009},
          {0.,0.},{ 0.00,0.00},{.0757,0.0},{0.4003,.2645},
          {0.,0.},{ 0.00,0.00},{0.00,0.00},{-.0488,.0}};
 
      dcmplx zresb20[16] = {
          {3.0,0.},{1.0,-1.0},{1.0,1.0},{1.0,-1.0},
          {0.,0.},{4.0,0.00},{2.0,1.0},{1.5,-1.5},
          {0.,0.},{ 0.00,0.00},{4.4721,0.0},{2.3479,-.5590},
          {0.,0.},{ 0.00,0.00},{0.00,0.00},{6.9767,.0}};
 
      dcmplx zresz20[16] = {
          {.2784,-.0218},{.0601, .1506},{0.0,-.1675},{-.0406, .0627},
          {-.1300,-.0562},{.1844,.1016},{.0004,-.0082},{-.0726,-.0362},
          {.1842,-.0064},{-.0064,-.0516},{-.0591,.1230},{-.0414,-.0635},
          {-.0692,.0118},{-.0948,.0008},{-.0946,.0178},{-.0513,.0023}};
 
      double zresw20[4]= { -0.7969,-0.1152, 0.1669, 1.2304};
 
/* PDSYNTRD START : 21     */
 
   double dta21[16], dtd21[4], dte21[4], dtt21[4];
   double dtas21[16] = { 5.,  0.,  0.,  0.,
                         4.,  5.,  0.,  0.,
                         1.,  1.,  4.,  0.,
                         1.,  1.,  2.,  4. };
 
   double dresa21[16] = { 1.,   0.,    0.,   0.,
                          0.,   6.,    0.,   0.,
                          0.41, 2.83,  7.0,  0.,
                          0.22, 0.22, -2.45, 4.0 };
 
   double dresd21[4] = { 1.,  6.,  7.,    4. };
   double drese21[4] = { 0.,  0.,  2.83, -2.45 };
   double drest21[4] = { 0.,  0.,  1.71,  1.82 };
 
   double dresd211[4] = { 5., 10.09,  4.16, -1.25 };
   double drese211[4] = {-5.92,  2.36,  1.62,  0.00 };
 
/* PDSYNGST   START : 22  */
 
      double dta22[16], dtb22[16];
      double  dtas22[16] =  { -1.,  1., -1.,  1.,
                               0.,  1., -1.,  1.,
                               0.,  0.,  1., -1.,
                               0.,  0.,  0.,  1.};
 
      double dtbs22[16] =  {   2.,  1.,  0.,  0.,
                               0.,  2.,  1.,  0.,
                               0.,  0.,  2.,  1.,
                               0.,  0.,  0.,  2.};
 
      double dresa22[16]= { -0.50000, 0.86603,-1.22474, 1.58114,
                             0.00000,-0.16667,-0.23570, 0.54772,
                             0.00000, 0.00000, 1.16667,-1.93649,
                            0.00000, 0.00000, 0.00000, 3.10000};
/* PDSYEVD        START : 23  */
 
   double dta23[16], dtz23[16], dtw23[4];
 
   double dtas23[16] = { 5.,  0.,  0.,  0.,
                         4.,  5.,  0.,  0.,
                         1.,  1.,  4.,  0.,
                         1.,  1.,  2.,  4. };
 
   double dresz23[16] = { 0.7071, -0.7071,  0.0,     0.0,
                          0.0,     0.0,     0.7071, -0.7071,
                          0.3162,  0.3162, -0.6325, -0.6325,
                          0.6325,  0.6325,  0.3162,  0.3162  };
 
   double dresw23[4] = { 1.,  2.,  5.,   10. };
 
 
  char modn[9], *modname=modn;
 
   for (i=0; i< 16; i++)
    dta16[i] = 0.0;
   for (i=0; i< 16; i++)
    CSET(zta16[i], 0.0,0.0);
  blacs_pinfo(&iam,&nnodes);
  nprow = 1;
  npcol = nnodes;
  blacs_get(0,0,&ictx);
  blacs_gridinit(&ictx,"R",nprow,npcol);
  blacs_gridinfo(ictx,&nprow,&npcol,&myrow,&mycol);
 
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 4;
  desc_a[N_] = 4;
  desc_a[MB_] = 1;
  desc_a[NB_] = 1;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(4,1,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
/* following desc_x is created to check vectors d, e, and t
// IX = 1
// JX = JA     */
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 1;
  INCX = desc_x[M_];
  desc_x[N_] = desc_a[N_];
  desc_x[MB_] = 1;
  desc_x[NB_] = desc_a[NB_];
  desc_x[CSRC_] = desc_a[CSRC_];
  desc_x[CTXT_] = ictx;
  desc_x[LLD_] = INCX;
 
  modname = "PDSYTRD ";
  pdmtxcpy(4, 4, 1, 1, dta16, desc_a, dtas16);
  pdsytrd("U", 4, dta16, 1, 1, desc_a, dtd16, dte16, dtt16, work, lwork,&info);
  if (info != 0)
  {
   if (iam == 0)
    printf( " INFO from PDSYTRD = %4d\n", info);
    *ifail = *ifail + 1;
  }
  else
  {
  pdmtchk(modname, 4, 4, 1, 1, dta16, desc_a, dresa16, tol, ierr,&ilerr,
         &error, ifail);
 
  for (i=0; i< nprow; i++)
   {
    desc_x[RSRC_] = i;
    pdvtchk(modname,4, 1, 1, dtd16, desc_x, dresd16, 1, tol, ierr,&ilerr,
         &error, ifail);
    pdvtchk(modname,4, 1, 1, dte16, desc_x, drese16, 1, tol, ierr,&ilerr,
         &error, ifail);
    pdvtchk(modname,4, 1, 1, dtt16, desc_x, drest16, 1, tol, ierr,&ilerr,
         &error, ifail);
   }
  }
 
  modname = "PZHETRD ";
  pzmtxcpy(4, 4, 1, 1, zta16, desc_a, ztas16);
  pzhetrd("L", 4, zta16, 1, 1, desc_a, dtd16, dte16, ztt16,workz, lwork,&info);
  if (info != 0)
  {
   if (iam == 0)
    printf( " INFO from PZHETRD = %4d\n", info);
    *ifail = *ifail + 1;
  }
  else
  {
  pzmtchk(modname, 4, 4, 1, 1, zta16, desc_a, zresa16, tol, ierr,&ilerr,
         &error, ifail);
 
  for (i=0; i< nprow; i++)
   {
    desc_x[RSRC_] = i;
    pdvtchk(modname,4, 1, 1, dtd16, desc_x,dresd161, 1, tol, ierr,&ilerr,
         &error, ifail);
    pdvtchk(modname,4, 1, 1, dte16, desc_x,drese161, 1, tol, ierr,&ilerr,
         &error, ifail);
    pzvtchk(modname,4, 1, 1, ztt16, desc_x, zrest16, 1, tol, ierr,&ilerr,
         &error, ifail);
   }
  }
 
/* PDSYTRD  PZHETRD  END          */
 
/* PDGEHRD   START : 17   */
 
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 3;
  desc_a[N_] = 3;
  desc_a[MB_] = 1;
  desc_a[NB_] = 1;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(3,1,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
/* following desc_x is created to check vectors d, e, and t
// IX = 1
// JX = JA       */
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 1;
  INCX = desc_x[M_];
  desc_x[N_] = desc_a[N_]-1;
  desc_x[MB_] = 1;
  desc_x[NB_] = desc_a[NB_];
  desc_x[CSRC_] = desc_a[CSRC_];
  desc_x[CTXT_] = ictx;
  desc_x[LLD_] = INCX;
 
  modname = "PDGEHRD ";
  pdmtxcpy(3, 3, 1, 1, dta17, desc_a, dtas17);
  pdgehrd(3, 1, 3, dta17, 1, 1, desc_a, dtt17, work, lwork,&info);
  if (info != 0)
  {
   if (iam == 0)
    printf( " INFO from PDGEHRD = %4d\n",info);
    *ifail = *ifail + 1;
  }
  else
  {
  pdmtchk(modname, 3, 3, 1, 1, dta17, desc_a, dresa17, tol, ierr,&ilerr,
         &error, ifail);
 
  for (i=0; i< nprow; i++)
   {
    desc_x[RSRC_] = i;
    pdvtchk(modname,2, 1, 1, dtt17, desc_x, drest17, 1, tol, ierr,&ilerr,
         &error, ifail);
   }
  }
 
/* PDGEHRD   END      */
 
/* PDGEBRD   START : 18  */
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 4;
  desc_a[N_] = 3;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(4,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
/* following desc_xr, and desc_xc are created to check vectors d,e,tp,tq*/
 
  desc_xr[DT_] = BLOCK_CYCLIC_2D;
  desc_xr[M_] = 1;
  INCX = desc_xr[M_];
  desc_xr[N_] = MAX(desc_a[M_],desc_a[N_]);
  desc_xr[MB_] = desc_a[MB_];
  desc_xr[NB_] = desc_a[NB_];
  desc_xr[RSRC_] = desc_a[RSRC_];
  desc_xr[CSRC_] = desc_a[CSRC_];
  desc_xr[CTXT_] = desc_a[CTXT_];
  desc_xr[LLD_] = INCX;
 
  desc_xc[DT_] = BLOCK_CYCLIC_2D;
  desc_xc[M_] = MAX(desc_a[M_],desc_a[N_]);
  desc_xc[N_] = 1;
  desc_xc[MB_] = desc_a[MB_];
  desc_xc[NB_] = desc_a[NB_];
  desc_xc[RSRC_] = desc_a[RSRC_];
  desc_xc[CSRC_] = desc_a[CSRC_];
  desc_xc[CTXT_] = desc_a[CTXT_];
  il_lda = numroc(desc_a[N_],desc_a[NB_],mycol,desc_a[CSRC_],npcol);
  desc_xc[LLD_] = MAX(desc_a[LLD_],il_lda);
 
  modname = "PDGEBRD ";
  pdmtxcpy(4, 3, 1, 1, dta18, desc_a, dtas18);
  pdgebrd(4, 3, dta18, 1, 1, desc_a, dtd18, dte18, dttq18, dttp18, work,
          lwork,&info);
  if (info != 0)
  {
   if (iam == 0)
    printf(" INFO from PDGEBRD = %4d\n",info);
    *ifail = *ifail + 1;
  }
  else
  {
  pdmtchk(modname, 4, 3, 1, 1, dta18, desc_a, dresa18, tol, ierr,&ilerr,
         &error, ifail);
 
  for (i=0; i< nprow; i++)
   {
     /* as M >= N, only vector D and Tq are neede to check as row vector */
    desc_xr[N_] = 3;
    desc_xr[RSRC_] = i;
    pdvtchk(modname,3, 1, 1, dtd18, desc_xr, dresd18, 1, tol, ierr,&ilerr,
         &error, ifail);
    pdvtchk(modname,3, 1, 1, dttq18, desc_xr, drestq18, 1, tol, ierr,&ilerr,
         &error, ifail);
   }
 
  for (i=0; i< npcol; i++)
   {
     /* as M >= N, only vector E and Tp are neede to check as column vecto*/
    desc_xc[M_] = 2;
    desc_xc[CSRC_] = i;
    pdvtchk(modname,2, 1, 1, dte18, desc_xc, drese18, 1, tol, ierr,&ilerr,
         &error, ifail);
    desc_xc[M_] = 3;
    pdvtchk(modname,3, 1, 1, dttp18, desc_xc, drestp18, 1, tol, ierr,&ilerr,
         &error, ifail);
   }
  }
 
/* PDGEBRD   END         */
/* PDGESVD   START : 18  */
 
  desc_a[DT_]    = BLOCK_CYCLIC_2D;
  desc_a[M_]     = 4;
  desc_a[N_]     = 3;
  desc_a[MB_]    = 2;
  desc_a[NB_]    = 2;
  desc_a[RSRC_]  = 0;
  desc_a[CSRC_]  = 0;
  desc_a[CTXT_]  = ictx;
  il_lda         = numroc(4,2,myrow,0,nprow);
  desc_a[LLD_]   = MAX(1,il_lda);
 
/* following desc_u and desc_vt are created to check vectors u, vt */
 
  desc_vt[DT_]   = BLOCK_CYCLIC_2D;
  desc_vt[M_]    = 4;
  desc_vt[N_]    = 3;
  desc_vt[MB_]   = 2;
  desc_vt[NB_]   = 2;
  desc_vt[RSRC_] = 0;
  desc_vt[CSRC_] = 0;
  desc_vt[CTXT_] = desc_a[CTXT_];
  il_lda         = numroc(4,2,myrow,0,nprow);
  desc_vt[LLD_]  = MAX(1,il_lda);
 
  desc_u[DT_]    = BLOCK_CYCLIC_2D;
  desc_u[M_]     = 4;
  desc_u[N_]     = 3;
  desc_u[MB_]    = 2;
  desc_u[NB_]    = 2;
  desc_u[RSRC_]  = 0;
  desc_u[CSRC_]  = 0;
  desc_u[CTXT_]  = desc_a[CTXT_];
  il_lda         = numroc(4,2,myrow,0,nprow);
  desc_u[LLD_]   = MAX(1,il_lda);
 
  desc_xr[DT_]   = BLOCK_CYCLIC_2D;
  desc_xr[M_]    = 1;
  INCX           = desc_xr[M_];
  desc_xr[N_]    = 3;
  desc_xr[MB_]   = desc_a[MB_];
  desc_xr[NB_]   = 3;
  desc_xr[RSRC_] = desc_a[RSRC_];
  desc_xr[CSRC_] = desc_a[CSRC_];
  desc_xr[CTXT_] = desc_a[CTXT_];
  desc_xr[LLD_]  = INCX;
 
  for(ix=0;ix<12;ix++) dtvt18[ix] = 0.00; /* zero out array so that extra
                                           * space has no garbage during
                                           * the error check */
  modname = "PDGESVD ";
  pdmtxcpy(4, 3, 1, 1, dta18, desc_a, dtas18);
 
  pdgesvd ("V", "V", 4, 3, dta18, 1, 1, desc_a, dts18,
                dtu18, 1, 1, desc_u, dtvt18, 1, 1, desc_vt,
                work, lwork, &info);
 
  if (info != 0) {
   if (iam == 0) printf( " INFO from PDGESVD = %d\n", info);
   *ifail = *ifail + 1;
  } else {
   pdmtchk(modname, 4, 3, 1, 1, dtu18, desc_u, dtus18, tol,
           ierr,&ilerr, &error, ifail);
   pdmtchk(modname, 4, 3, 1, 1, dtvt18, desc_vt, dtvts18, tol,
           ierr,&ilerr, &error, ifail);
   for (i=0; i< 3; i++)
       if ( fabs(dts18[i] - dtss18[i]) > tol)
        {
          *ifail = *ifail + 1;
         printf(" PDGESVD dts18[%4d ]= %4g\n",i,dts18[i]);
         printf("  dtss18[%4d] = %4g\n",i,dtss18[i]);
        }
  }
 
/* PDGESVD   END         */
/* PZGEBRD   START : 18  */
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 3;
  desc_a[N_] = 4;
  desc_a[MB_] = 2;
  desc_a[NB_] = 2;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(3,2,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
/* following desc_xr, and desc_xc are created to check vectors d,e,tp,tq*/
 
  desc_xr[DT_] = BLOCK_CYCLIC_2D;
  desc_xr[M_] = 1;
  INCX = desc_xr[M_];
  desc_xr[N_] = MAX(desc_a[M_],desc_a[N_]);
  desc_xr[MB_] = desc_a[MB_];
  desc_xr[NB_] = desc_a[NB_];
  desc_xr[RSRC_] = desc_a[RSRC_];
  desc_xr[CSRC_] = desc_a[CSRC_];
  desc_xr[CTXT_] = desc_a[CTXT_];
  desc_xr[LLD_] = INCX;
 
  desc_xc[DT_] = BLOCK_CYCLIC_2D;
  desc_xc[M_] = MAX(desc_a[M_],desc_a[N_]);
  desc_xc[N_] = 1;
  desc_xc[MB_] = desc_a[MB_];
  desc_xc[NB_] = desc_a[NB_];
  desc_xc[RSRC_] = desc_a[RSRC_];
  desc_xc[CSRC_] = desc_a[CSRC_];
  desc_xc[CTXT_] = desc_a[CTXT_];
  il_lda = numroc(desc_a[N_],desc_a[NB_],mycol,desc_a[CSRC_],npcol);
  desc_xc[LLD_] = MAX(desc_a[LLD_],il_lda);
 
  modname = "PZGEBRD ";
  pzmtxcpy(3, 4, 1, 1, zta18, desc_a, ztas18);
  pzgebrd(3, 4, zta18, 1, 1, desc_a, ztd18, zte18, zttq18, zttp18, workz,
          lwork,&info);
  if (info != 0) {
   if (iam == 0) printf(" INFO from PZGEBRD = %4d\n",info);
   *ifail = *ifail + 1;
  } else {
   pzmtchk(modname, 3, 4, 1, 1, zta18, desc_a, zresa18, tol, ierr,&ilerr,
           &error, ifail);
 
   for (i=0; i< nprow; i++) {
    /* as M >= N, only vector D and Tq are neede to check as row vector */
    desc_xr[N_] = 2;
    desc_xr[RSRC_] = i;
    pdvtchk(modname,2, 1, 1, zte18, desc_xr, zrese18, 1, tol, ierr,&ilerr,
            &error, ifail);
    desc_xr[N_] = 3;
    pzvtchk(modname,3, 1, 1, zttq18, desc_xr, zrestq18, 1, tol, ierr,&ilerr,
            &error, ifail);
   }
 
   for (i=0; i< npcol; i++) {
    /* as M >= N, only vector E and Tp are neede to check as column vector*/
    desc_xc[M_] = 3;
    desc_xc[CSRC_] = i;
    pdvtchk(modname,3, 1, 1, ztd18, desc_xc, zresd18, 1, tol, ierr,&ilerr,
            &error, ifail);
    pzvtchk(modname,3, 1, 1, zttp18, desc_xc, zrestp18, 1, tol, ierr,&ilerr,
            &error, ifail);
   }
  }
 
/* PZGEBRD   END         */
/* PZGESVD   START : 18  */
 
  desc_a[DT_]    = BLOCK_CYCLIC_2D;
  desc_a[M_]     = 3;
  desc_a[N_]     = 4;
  desc_a[MB_]    = 2;
  desc_a[NB_]    = 2;
  desc_a[RSRC_]  = 0;
  desc_a[CSRC_]  = 0;
  desc_a[CTXT_]  = ictx;
  il_lda         = numroc(3,2,myrow,0,nprow);
  desc_a[LLD_]   = MAX(1,il_lda);
 
/* following desc_u and desc_vt are created to check vectors u, vt */
 
  desc_xr[DT_]   = BLOCK_CYCLIC_2D;
  desc_xr[M_]    = 1;
  INCX           = desc_xr[M_];
  desc_xr[N_]    = 3;
  desc_xr[MB_]   = desc_a[MB_];
  desc_xr[NB_]   = 3;
  desc_xr[RSRC_] = desc_a[RSRC_];
  desc_xr[CSRC_] = desc_a[CSRC_];
  desc_xr[CTXT_] = desc_a[CTXT_];
  desc_xr[LLD_]  = INCX;
 
  desc_vt[DT_]   = BLOCK_CYCLIC_2D;
  desc_vt[M_]    = 3;
  desc_vt[N_]    = 4;
  desc_vt[MB_]   = 2;
  desc_vt[NB_]   = 2;
  desc_vt[RSRC_] = 0;
  desc_vt[CSRC_] = 0;
  desc_vt[CTXT_] = desc_a[CTXT_];
  il_lda         = numroc(3,2,myrow,0,nprow);
  desc_vt[LLD_]  = MAX(1,il_lda);
 
  desc_u[DT_]   = BLOCK_CYCLIC_2D;
  desc_u[M_]    = 3;
  desc_u[N_]    = 3;
  desc_u[MB_]   = 2;
  desc_u[NB_]   = 2;
  desc_u[RSRC_] = 0;
  desc_u[CSRC_] = 0;
  desc_u[CTXT_] = desc_a[CTXT_];
  il_lda        = numroc(3,2,myrow,0,nprow);
  desc_u[LLD_]  = MAX(1,il_lda);
 
  for(ix=0;ix<12;ix++) {
   RE(ztu18[ix]) = 0.00; /* zero out array so that extra space has */
   IM(ztu18[ix]) = 0.00; /* no garbage during the error check */
  }
 
  modname = "PZGESVD ";
  pzmtxcpy(3, 4, 1, 1, zta18, desc_a, ztas18);
 
  pzgesvd ("V", "V", 3, 4, zta18, 1, 1, desc_a, zts18,
           ztu18, 1, 1, desc_u, ztvt18, 1, 1, desc_vt,
           workz, lwork, work, lrwork, &info);
 
  if (info != 0) {
   if (iam == 0) printf( " INFO from PZGESVD = %d\n", info);
   *ifail = *ifail + 1;
  } else {
   pzmtchk(modname, 3, 4, 1, 1, ztu18, desc_u, ztus18, tol,
           ierr,&ilerr, &error, ifail);
   pzmtchk(modname, 3, 4, 1, 1, ztvt18, desc_vt, ztvts18, tol,
           ierr,&ilerr, &error, ifail);
   for (i=0; i< 3; i++)
       if ( fabs(zts18[i] - ztss18[i]) > tol)
        {
          *ifail = *ifail + 1;
         printf(" PZGESVD zts18[%4d ]= %4g\n",i,zts18[i]);
         printf("  ztss18[%4d] = %4g\n",i,ztss18[i]);
        }
  }
 
/* PZGESVD   END    */
/* PDSYEVX   PZHEEVX START : 19  */
 
   dtg19 = (double *) malloc(nprow*npcol*sizeof(double));
   dtg19res = (double *) malloc(nprow*npcol*sizeof(double));
   iclustr  = (int *) malloc(2*nprow*npcol*sizeof(int));
   iclustres= (int *) malloc(2*nprow*npcol*sizeof(int));
     istor[0]=0;
   if (!dtg19 || !dtg19res || !iclustr || !iclustres)
     istor[0]=1;
     igamx2d(ictx,"All", "1-Tree", 1, 1,
                    istor, 1, ra, ca, 1, 0, 0 );
     if(istor[0]==1)
     {
      if (iam==0)
      printf(" Memory allocation error for PDSYEVX\n");
      free(dtg19);         /* free memory  */
      free(dtg19res);
      free(iclustr);
      free(iclustres);
      *ifail=*ifail+1;
      blacs_gridexit(ictx);
      return;
     }
   for (i=0; i< nprow*npcol; i++)
     dtg19res[i] = -1.0;
   for (i=0; i< 2*nprow*npcol; i++)
     iclustres[i] = 0;
 
 
  M = 0;
  NZ = 0;
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 4;
  desc_a[N_] = 4;
  desc_a[MB_] = 1;
  desc_a[NB_] = 1;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(4,1,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_z[DT_] = BLOCK_CYCLIC_2D;
  desc_z[M_] = 4;
  desc_z[N_] = 4;
  desc_z[MB_] = 1;
  desc_z[NB_] = 1;
  desc_z[RSRC_] = 0;
  desc_z[CSRC_] = 0;
  desc_z[CTXT_] = ictx;
  il_lda = numroc(4,1,myrow,0,nprow);
  desc_z[LLD_] = MAX(1,il_lda);
 
  modname = "PDSYEVX ";
  pdmtxcpy(4, 4, 1, 1, dta19, desc_a, dtas19);
  pdsyevx("V","A","U", 4, dta19, 1, 1, desc_a, 0.0, 0.0, 0, 0, -1.0,&M,&NZ,
          dtw19, -1.0, dtz19, 1, 1, desc_z, work, lwork, iwork, liwork, ifaile,
          iclustr, dtg19,&info);
  if (info != 0) {
   if (iam == 0)
    printf("INFO from PDSYEVX = %4d\n", info);
  }
  if (M != 4) {
   if (iam == 0)
    printf(" M from PDSYEVX = %4d\n",M);
  }
  if (NZ != 4) {
   if (iam == 0)
    printf(" NZ from PDSYEVX = %4d\n",NZ);
  }
  if ( info != 0 || M != 4 || NZ != 4) *ifail = *ifail + 1;
  if (info == 0) {
  pdmtchk(modname, 4, 4, 1, 1, dtz19, desc_z, dresz19, tol, ierr,&ilerr,
         &error, ifail);
   for (i=0; i< nprow*npcol; i++)
       if ((fabs(dtg19[i]) - fabs(dtg19res[i])) > tol)
        {
          *ifail = *ifail + 1;
         printf(" PDSYEVX dtg19[%4d ] = %4g\n",i,dtg19[i]);
         printf("  dtg19res[%4d] = %4g\n",i,dtg19res[i] );
        }
   for (i=0; i< 2*nprow*npcol; i++)
       if (iclustr[i] != iclustres[i])
        {
          *ifail = *ifail + 1;
         printf("PDSYEVX iclustr[%4d] = %4d\n",i,iclustr[i]);
         printf(" iclustres[%4d] = %4d\n",i,iclustres[i]);
        }
   for (i=0; i< 4; i++)
       if (ifaile[i] != ifailres[i])
        {
          *ifail = *ifail + 1;
         printf(" PDSYEVX ifaile[%4d ] = %4d\n",i,ifaile[i]);
         printf("ifailres[%4d] = %4d\n",i,ifailres[i]);
        }
   for (i=0; i< 4; i++)
       if ((fabs(dtw19[i]) - fabs(dresw19[i])) > tol)
        {
          *ifail = *ifail + 1;
         printf(" PDSYEVX dtw19[%4d ]= %4g\n",i,dtw19[i]);
         printf("  dresw19[%4d] = %4g\n",i,dresw19[i]);
        }
   }
 
  modname = "PZHEEVX ";
  pzmtxcpy(4, 4, 1, 1, zta19, desc_a, ztas19);
  pzheevx("V","A","L", 4, zta19, 1, 1, desc_a, 0.0, 0.0, 0, 0, -1.0,&M,&NZ,
          dtw19,-1.0,ztz19,1,1,desc_z,workz,0, work,lwork,iwork,liwork, ifaile,
          iclustr, dtg19,&info);
  if (info != 0) {
   if (iam == 0)
    printf("INFO from PZHEEVX = %4d\n", info);
  }
  if (M != 4) {
   if (iam == 0)
    printf(" M from PZHEEVX = %4d\n",M);
  }
  if (NZ != 4) {
   if (iam == 0)
    printf(" NZ from PZHEEVX = %4d\n",NZ);
  }
  if ( info != 0 || M != 4 || NZ != 4) *ifail = *ifail + 1;
  if (info == 0) {
  pzmtchk(modname, 4, 4, 1, 1, ztz19, desc_z, zresz19, tol, ierr,&ilerr,
         &error, ifail);
   for (i=0; i< nprow*npcol; i++)
       if ((fabs(dtg19[i]) - fabs(dtg19res[i])) > tol)
        {
          *ifail = *ifail + 1;
         printf(" PZHEEVX dtg19[%4d ] = %4g\n",i,dtg19[i]);
         printf("  dtg19res[%4d] = %4g\n",i,dtg19res[i] );
        }
   for (i=0; i< 2*nprow*npcol; i++)
       if (iclustr[i] != iclustres[i])
        {
          *ifail = *ifail + 1;
         printf("PZHEEVX iclustr[%4d] = %4d\n",i,iclustr[i]);
         printf(" iclustres[%4d] = %4d\n",i,iclustres[i]);
        }
   for (i=0; i< 4; i++)
       if (ifaile[i] != ifailres[i])
        {
          *ifail = *ifail + 1;
         printf(" PZHEEVX ifaile[%4d ] = %4d\n",i,ifaile[i]);
         printf("ifailres[%4d] = %4d\n",i,ifailres[i]);
        }
   for (i=0; i< 4; i++)
       if ((fabs(dtw19[i]) - fabs(zresw19[i])) > tol)
        {
          *ifail = *ifail + 1;
         printf(" PZHEEVX dtw19[%4d ]= %4g\n",i,dtw19[i]);
         printf("  zresw19[%4d] = %4g\n",i,zresw19[i]);
        }
   }
 
/* PDSYEVX  PZHEEVX  END      */
 
/* PDSYGST   PZHEGST START : 20  */
 
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 4;
      desc_a[N_]    = 4;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
      desc_b[DT_]   = BLOCK_CYCLIC_2D;
      desc_b[M_]    = 4;
      desc_b[N_]    = 4;
      desc_b[MB_]   = 1;
      desc_b[NB_]   = 1;
      desc_b[RSRC_] = 0;
      desc_b[CSRC_] = 0;
      desc_b[CTXT_] = ictx;
      il_ldb = numroc(4,1,myrow,0,nprow);
      desc_b[LLD_]  = MAX(1,il_ldb);
 
      modname = "PDSYGST ";
      pdmtxcpy(4, 4, 1, 1, dta20, desc_a, dtas20);
      pdmtxcpy(4, 4, 1, 1, dtb20, desc_b, dtbs20);
      pdpotrf("L", 4, dtb20, 1, 1, desc_b,&iret);
      pdsygst(1,"L", 4, dta20, 1, 1, desc_a, dtb20, 1, 1, desc_b,
                   &scale, &info);
      if (info != 0)
      {
       if (iam == 0)
       printf("INFO from PDSYGST = %4d\n", info);
       *ifail = *ifail + 1;
      }
      else
      {
        pdmtchk(modname, 4, 4, 1, 1, dta20, desc_a, dresa20, tol,
                    ierr,&ilerr, &error, ifail);
 
       if ((fabs(scale) - fabs(scaler)) > tol)
        {
          *ifail = *ifail + 1;
         printf(" PDSYGST scale = %4g\n",scale);
         printf("  scaler = %4g\n",scaler);
        }
      }
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 4;
      desc_a[N_]    = 4;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
      desc_b[DT_]   = BLOCK_CYCLIC_2D;
      desc_b[M_]    = 4;
      desc_b[N_]    = 4;
      desc_b[MB_]   = 1;
      desc_b[NB_]   = 1;
      desc_b[RSRC_] = 0;
      desc_b[CSRC_] = 0;
      desc_b[CTXT_] = ictx;
      il_ldb = numroc(4,1,myrow,0,nprow);
      desc_b[LLD_]  = MAX(1,il_ldb);
 
      modname = "PZHEGST ";
      pzmtxcpy(4, 4, 1, 1, zta20, desc_a, ztas20);
      pzmtxcpy(4, 4, 1, 1, ztb20, desc_b, ztbs20);
      pzpotrf("L", 4, ztb20, 1, 1, desc_b,&iret);
      pzhegst(1,"L", 4, zta20, 1, 1, desc_a, ztb20, 1, 1, desc_b,
                   &scale, &info);
      if (info != 0)
      {
       if (iam == 0)
       printf("INFO from PZHEGST = %4d\n", info);
       *ifail = *ifail + 1;
      }
      else
      {
        pzmtchk(modname, 4, 4, 1, 1, zta20, desc_a, zresa20, tol,
                    ierr,&ilerr, &error, ifail);
 
       if ((fabs(scale) - fabs(scaler)) > tol)
        {
          *ifail = *ifail + 1;
         printf(" PZHEGST scale = %4g\n",scale);
         printf("  scaler = %4g\n",scaler);
        }
      }
/* PDSYGVX PZHEGVX         START : 20   */
 
   for (i=0; i< nprow*npcol; i++)
     dtg19res[i] = -1.0;
   for (i=0; i< 2*nprow*npcol; i++)
     iclustres[i] = 0;
 
      M = 0;
      NZ = 0;
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 4;
      desc_a[N_]    = 4;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
      desc_b[DT_]   = BLOCK_CYCLIC_2D;
      desc_b[M_]    = 4;
      desc_b[N_]    = 4;
      desc_b[MB_]   = 1;
      desc_b[NB_]   = 1;
      desc_b[RSRC_] = 0;
      desc_b[CSRC_] = 0;
      desc_b[CTXT_] = ictx;
      il_ldb = numroc(4,1,myrow,0,nprow);
      desc_b[LLD_]  = MAX(1,il_ldb);
 
      desc_z[DT_]   = BLOCK_CYCLIC_2D;
      desc_z[M_]    = 4;
      desc_z[N_]    = 4;
      desc_z[MB_]   = 1;
      desc_z[NB_]   = 1;
      desc_z[RSRC_] = 0;
      desc_z[CSRC_] = 0;
      desc_z[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_z[LLD_]  = MAX(1,il_lda);
 
      modname = "PDSYGVX ";
      lwork = 5000;
      pdmtxcpy(4, 4, 1, 1, dta20, desc_a, dtas20);
      pdmtxcpy(4, 4, 1, 1, dtb20, desc_b, dtbs20);
      pdsygvx(1,"V","A","L", 4, dta20, 1, 1, desc_a,
        dtb20, 1, 1, desc_b,  0.0, 0.0,
        0, 0, -1.0,&M,&NZ, dtw20, -1.0, dtz20, 1, 1, desc_z,work,
        lwork, iwork, liwork, ifaile, iclustr, dtg19,&info);
  if (info != 0) {
   if (iam == 0)
    printf("INFO from PDSYGVX = %4d\n", info);
  }
  if (M != 4) {
   if (iam == 0)
    printf(" M from PDSYGVX = %4d\n",M);
  }
  if (NZ != 4) {
   if (iam == 0)
    printf(" NZ from PDSYGVX = %4d\n",NZ);
  }
  if ( info != 0 || M != 4 || NZ != 4) *ifail = *ifail + 1;
  if (info == 0) {
  pdmtxcpy(4, 4, 1, 1, dta20, desc_a, dtac20);
  pdmtxcpy(4, 4, 1, 1, dtb20, desc_b, dtbc20);
  pdgspchk(1, 4, 4, dta20, 1, 1, desc_a, dtb20, 1, 1, desc_b,
           500.0, dtz20, 1, 1, desc_z, dtc20, 1, 1, desc_a,
           dtw20, work, lwork, &tstnrm, &iret);
       if (iret != 0 || (tstnrm-tstnrm) != 0.0)
        {
         *ifail = *ifail + 1;
         printf(" PDSYGVX dtz20 error = %4g %4d\n",tstnrm,iret);
        }
   for (i=0; i< nprow*npcol; i++)
       if ((fabs(dtg19[i]) - fabs(dtg19res[i])) > tol)
        {
          *ifail = *ifail + 1;
         printf(" PDSYGVX dtg19[%4d ] = %4g\n",i,dtg19[i]);
         printf("  dtg19res[%4d] = %4g\n",i,dtg19res[i] );
        }
   for (i=0; i< 2*nprow*npcol; i++)
       if (iclustr[i] != iclustres[i])
        {
          *ifail = *ifail + 1;
         printf("PDSYGVX iclustr[%4d] = %4d\n",i,iclustr[i]);
         printf(" iclustres[%4d] = %4d\n",i,iclustres[i]);
        }
   for (i=0; i< 4; i++)
       if (ifaile[i] != ifailres[i])
        {
          *ifail = *ifail + 1;
         printf(" PDSYGVX ifaile[%4d ] = %4d\n",i,ifaile[i]);
         printf("ifailres[%4d] = %4d\n",i,ifailres[i]);
        }
   for (i=0; i< 4; i++)
       if ((fabs(dtw20[i]) - fabs(dresw20[i])) > tol)
        {
          *ifail = *ifail + 1;
         printf(" PDSYGVX dtw20[%4d ]= %4g\n",i,dtw19[i]);
         printf("  dresw20[%4d] = %4g\n",i,dresw19[i]);
        }
   }
 
   for (i=0; i< nprow*npcol; i++)
     dtg19res[i] = -1.0;
   for (i=0; i< 2*nprow*npcol; i++)
     iclustres[i] = 0;
 
      M = 0;
      NZ = 0;
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 4;
      desc_a[N_]    = 4;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
      desc_b[DT_]   = BLOCK_CYCLIC_2D;
      desc_b[M_]    = 4;
      desc_b[N_]    = 4;
      desc_b[MB_]   = 1;
      desc_b[NB_]   = 1;
      desc_b[RSRC_] = 0;
      desc_b[CSRC_] = 0;
      desc_b[CTXT_] = ictx;
      il_ldb = numroc(4,1,myrow,0,nprow);
      desc_b[LLD_]  = MAX(1,il_ldb);
 
      desc_z[DT_]   = BLOCK_CYCLIC_2D;
      desc_z[M_]    = 4;
      desc_z[N_]    = 4;
      desc_z[MB_]   = 1;
      desc_z[NB_]   = 1;
      desc_z[RSRC_] = 0;
      desc_z[CSRC_] = 0;
      desc_z[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_z[LLD_]  = MAX(1,il_lda);
 
      modname = "PZHEGVX ";
      pzmtxcpy(4, 4, 1, 1, zta20, desc_a, ztas20);
      pzmtxcpy(4, 4, 1, 1, ztb20, desc_b, ztbs20);
      pzhegvx(1,"V","A","L", 4, zta20, 1, 1, desc_a,
        ztb20, 1, 1, desc_b,  0.0, 0.0,
        0, 0, -1.0,&M,&NZ, ztw20, -1.0, ztz20, 1, 1, desc_z,workz,0,
        work,0, iwork, 0, ifaile, iclustr, dtg19,&info);
  if (info != 0) {
   if (iam == 0)
    printf("INFO from PZHEGVX = %4d\n", info);
  }
  if (M != 4) {
   if (iam == 0)
    printf(" M from PZHEGVX = %4d\n",M);
  }
  if (NZ != 4) {
   if (iam == 0)
    printf(" NZ from PZHEGVX = %4d\n",NZ);
  }
  if ( info != 0 || M != 4 || NZ != 4) *ifail = *ifail + 1;
  if (info == 0) {
        pzmtchk(modname, 4, 4, 1, 1, ztz20, desc_z, zresz20, tol,
                    ierr,&ilerr, &error, ifail);
   for (i=0; i< nprow*npcol; i++)
       if ((fabs(dtg19[i]) - fabs(dtg19res[i])) > tol)
        {
          *ifail = *ifail + 1;
         printf(" PZHEGVX dtg19[%4d ] = %4g\n",i,dtg19[i]);
         printf("  dtg19res[%4d] = %4g\n",i,dtg19res[i] );
        }
   for (i=0; i< 2*nprow*npcol; i++)
       if (iclustr[i] != iclustres[i])
        {
          *ifail = *ifail + 1;
         printf("PZHEGVX iclustr[%4d] = %4d\n",i,iclustr[i]);
         printf(" iclustres[%4d] = %4d\n",i,iclustres[i]);
        }
   for (i=0; i< 4; i++)
       if (ifaile[i] != ifailres[i])
        {
          *ifail = *ifail + 1;
         printf(" PZHEGVX ifaile[%4d ] = %4d\n",i,ifaile[i]);
         printf("ifailres[%4d] = %4d\n",i,ifailres[i]);
        }
   for (i=0; i< 4; i++)
       if ((fabs(ztw20[i]) - fabs(zresw20[i])) > tol)
        {
          *ifail = *ifail + 1;
         printf(" PZHEGVX ztw20[%4d ]= %4g\n",i,ztw20[i]);
         printf("  zresw20[%4d] = %4g\n",i,zresw20[i]);
        }
   }
 
/* PDSYGST PDSYGVX  END : 20  */
/* PZHEGST PZHEGVX  END : 20  */
/* PDSYNTRD START : 21  */
  for (i=0; i< 16; i++) dta21[i] = 0.0;
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 4;
  desc_a[N_] = 4;
  desc_a[MB_] = 1;
  desc_a[NB_] = 1;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(4,1,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  /* following desc_x is created to check vectors d, e, and t
  // IX = 1
  // JX = JA     */
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 1;
  INCX = desc_x[M_];
  desc_x[N_] = desc_a[N_];
  desc_x[MB_] = 1;
  desc_x[NB_] = desc_a[NB_];
  desc_x[CSRC_] = desc_a[CSRC_];
  desc_x[CTXT_] = ictx;
  desc_x[LLD_] = INCX;
 
  modname = "PDSYNTRD";
 
  pdmtxcpy(4, 4, 1, 1, dta21, desc_a, dtas21);
 
  pdsyntrd("U", 4, dta21, 1, 1, desc_a, dtd21, dte21, dtt21, work, 0, &info);
 
  if (info != 0) {
      if (iam == 0) printf( " INFO from PDSYNTRD = %4d\n", info);
      *ifail = *ifail + 1;
  } else {
      pdmtchk(modname, 4, 4, 1, 1, dta21, desc_a, dresa21, tol, ierr, &ilerr,
              &error, ifail);
 
      for (i=0; i< nprow; i++) {
          desc_x[RSRC_] = i;
          pdvtchk(modname,4, 1, 1, dtd21, desc_x, dresd21, 1, tol, ierr,&ilerr,
                  &error, ifail);
          pdvtchk(modname,4, 1, 1, dte21, desc_x, drese21, 1, tol, ierr,&ilerr,
                  &error, ifail);
          pdvtchk(modname,4, 1, 1, dtt21, desc_x, drest21, 1, tol, ierr,&ilerr,
                  &error, ifail);
      }
  }
 
/* PDSYNTRD END :   21  */
/* PDSYNGST START : 22  */
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 4;
  desc_a[N_]    = 4;
  desc_a[MB_]   = 1;
  desc_a[NB_]   = 1;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(4,1,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_b[DT_]   = BLOCK_CYCLIC_2D;
  desc_b[M_]    = 4;
  desc_b[N_]    = 4;
  desc_b[MB_]   = 1;
  desc_b[NB_]   = 1;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_ldb = numroc(4,1,myrow,0,nprow);
  desc_b[LLD_]  = MAX(1,il_ldb);
 
  modname = "PDSYNGST";
 
  pdmtxcpy(4, 4, 1, 1, dta22, desc_a, dtas22);
  pdmtxcpy(4, 4, 1, 1, dtb22, desc_b, dtbs22);
 
  pdpotrf("L", 4, dtb22, 1, 1, desc_b, &iret);
 
  pdsyngst(1,"L", 4, dta22, 1, 1, desc_a, dtb22, 1, 1, desc_b,
           scale, work, 0, &info);
 
  if (info != 0) {
      if (iam == 0) printf("INFO from PDSYNGST = %4d\n", info);
      *ifail = *ifail + 1;
  } else {
      pdmtchk(modname, 4, 4, 1, 1, dta22, desc_a, dresa22, tol,
              ierr,&ilerr, &error, ifail);
 
      if ((fabs(scale) - fabs(scaler)) > tol) {
          *ifail = *ifail + 1;
          printf(" PDSYNGST scale = %4g\n", scale);
          printf("  scaler = %4g\n", scaler);
      }
  }
 
/* PDSYNGST END :   22  */
/* PDSYEVD  START : 23  */
 
  desc_a[DT_]   = BLOCK_CYCLIC_2D;
  desc_a[M_]    = 4;
  desc_a[N_]    = 4;
  desc_a[MB_]   = 1;
  desc_a[NB_]   = 1;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda        = numroc(4,1,myrow,0,nprow);
  desc_a[LLD_]  = MAX(1,il_lda);
 
  desc_z[DT_]   = BLOCK_CYCLIC_2D;
  desc_z[M_]    = 4;
  desc_z[N_]    = 4;
  desc_z[MB_]   = 1;
  desc_z[NB_]   = 1;
  desc_z[RSRC_] = 0;
  desc_z[CSRC_] = 0;
  desc_z[CTXT_] = ictx;
  il_lda        = numroc(4,1,myrow,0,nprow);
  desc_z[LLD_]  = MAX(1,il_lda);
 
  modname = "PDSYEVD ";
 
  pdmtxcpy(4, 4, 1, 1, dta23, desc_a, dtas23);
 
  pdsyevd("V","U", 4, dta23, 1, 1, desc_a, dtw23, dtz23,
          1, 1, desc_z, work, 0, iwork, 0, &info);
 
  if (info != 0 && iam == 0) printf("INFO from PDSYEVD = %4d\n", info);
  if (info != 0) *ifail = *ifail + 1;
 
  if (info == 0) {
      pdmtchk(modname, 4, 4, 1, 1, dtz23, desc_z, dresz23, tol, ierr,
              &ilerr, &error, ifail);
 
      for (i=0; i< 4; i++)
          if ((fabs(dtw23[i]) - fabs(dresw23[i])) > tol) {
              *ifail = *ifail + 1;
              printf(" PDSYEVD dtw23[%4d ]= %4g\n", i, dtw23[i]);
              printf("  dresw23[%4d] = %4g\n", i, dresw23[i]);
          }
  }
 
/* PDSYEVD   END : 23  */
 
      if(dtg19)     free(dtg19);         /* free memory  */
      if(dtg19res)  free(dtg19res);
      if(iclustr)   free(iclustr);
      if(iclustres) free(iclustres);
 
      blacs_gridexit(ictx);
 
}
 
/*********************************************************************
*          PARALLEL LINEAR ALGEBRAIC EQUATIONS SUBROUTINES           *
*********************************************************************/
 
void pleq(int *ifail)
{
/* PDPOTRF/S PDPOSV START : 6     */
 
   int iam, nnodes,myrow,mycol,myrows,mycols;
   int desc_a[DLEN_], desc_b[DLEN_], desc_x[DLEN_], il_lda;
   int desc_r[DLEN_], desc_atr[DLEN_1D_], desc_btr[DLEN_1D_];
   int ictx, ictxs, nprow, npcol,ierr=5, ilerr,nprows,npcols,INCX,i;
   double error=0.0, tol=0.0, anorm, rcond, rwork[100];
   double ntol= 0.01;
   double ztol= 0.1;
   double rconds2=0.02773;
   double rcondd=0.03053, rcondz=0.006173;
   int iwork[10], liwork=0, lrwork=0;
   int ipvt[12], iret, laf, lwork,info,istor[1],ra[1],ca[1];
   dcmplx workz[100];
   double dta6[81], dtb6[54];
   double dtas6[81] = { 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,
                        0.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,
                        0.,  0.,  3.,  3.,  3.,  3.,  3.,  3.,  3.,
                        0.,  0.,  0.,  4.,  4.,  4.,  4.,  4.,  4.,
                        0.,  0.,  0.,  0.,  5.,  5.,  5.,  5.,  5.,
                        0.,  0.,  0.,  0.,  0.,  6.,  6.,  6.,  6.,
                        0.,  0.,  0.,  0.,  0.,  0.,  7.,  7.,  7.,
                        0.,  0.,  0.,  0.,  0.,  0.,  0.,  8.,  8.,
                        0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  9. };
 
   double dresa6[81] = { 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,
                         0.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,
                         0.,  0.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,
                         0.,  0.,  0.,  1.,  1.,  1.,  1.,  1.,  1.,
                         0.,  0.,  0.,  0.,  1.,  1.,  1.,  1.,  1.,
                         0.,  0.,  0.,  0.,  0.,  1.,  1.,  1.,  1.,
                         0.,  0.,  0.,  0.,  0.,  0.,  1.,  1.,  1.,
                         0.,  0.,  0.,  0.,  0.,  0.,  0.,  1.,  1.,
                         0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  1. };
   double dtas6b[81] = {
  1., 0., 0., 0., 0., 0., 0., 0., 0.,
  1., 2., 0., 0., 0., 0., 0., 0., 0.,
  1., 2., 3., 0., 0., 0., 0., 0., 0.,
  1., 2., 3., 4., 0., 0., 0., 0., 0.,
  1., 2., 3., 4., 5., 0., 0., 0., 0.,
  1., 2., 3., 4., 5., 6., 0., 0., 0.,
  1., 2., 3., 4., 5., 6., 7., 0., 0.,
  1., 2., 3., 4., 5., 6., 7., 8., 0.,
  1., 2., 3., 4., 5., 6., 7., 8., 9.
 };
 
   double dresa61[81] = {
        2., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,
        0.,  2., -1.,  0.,  0.,  0.,  0.,  0.,  0.,
        0.,  0.,  2., -1.,  0.,  0.,  0.,  0.,  0.,
        0.,  0.,  0.,  2., -1.,  0.,  0.,  0.,  0.,
        0.,  0.,  0.,  0.,  2., -1.,  0.,  0.,  0.,
        0.,  0.,  0.,  0.,  0.,  2., -1.,  0.,  0.,
        0.,  0.,  0.,  0.,  0.,  0.,  2., -1.,  0.,
        0.,  0.,  0.,  0.,  0.,  0.,  0.,  2., -1.,
        0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  1. };
 
 
 
   double dtbs6[54] = { 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,
                       18., 34., 48., 60., 70., 78., 84., 88., 90.,
                       27., 51., 72., 90.,105.,117.,126.,132.,135.,
                       36., 68., 96.,120.,140.,156.,168.,176.,180.,
                       45., 85.,120.,150.,175.,195.,210.,220.,225.,
                        9., 17., 24., 30., 35., 39., 42., 44., 45. };
 
   double dresb6[54] = { 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,
                         2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,
                         3.,  3.,  3.,  3.,  3.,  3.,  3.,  3.,  3.,
                         4.,  4.,  4.,  4.,  4.,  4.,  4.,  4.,  4.,
                         5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,
                         1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1. };
 
/* PDGETRF/S PDGESV PDGETRI PDLANGE PDGECON START : 7     */
  double dta7[81], dtb7[45];
  double dtas7[81] =  { 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6,
                        1.2, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4,
                        1.4, 1.2, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2,
                        1.6, 1.4, 1.2, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0,
                        1.8, 1.6, 1.4, 1.2, 1.0, 1.2, 1.4, 1.6, 1.8,
                        2.0, 1.8, 1.6, 1.4, 1.2, 1.0, 1.2, 1.4, 1.6,
                        2.2, 2.0, 1.8, 1.6, 1.4, 1.2, 1.0, 1.2, 1.4,
                        2.4, 2.2, 2.0, 1.8, 1.6, 1.4, 1.2, 1.0, 1.2,
                        2.6, 2.4, 2.2, 2.0, 1.8, 1.6, 1.4, 1.2, 1.0 };
 
  double dresa7[81] = {
   2.6000,   .3846,   .4615,   .5385,  .6154,  .6923,  .7692, .8462, .9231,
   2.4000,   .2769,  -.3889,  -.3333, -.2778, -.2222, -.1667,-.1111,-.0556,
   2.2000,   .5538,   .4000,   .0000,  .0000,  .0000,  .0000, .0000, .0000,
   2.0000,   .8308,   .8000,   .4000,  .0000,  .0000,  .0000, .0000, .0000,
   1.8000,  1.1077,  1.2000,   .8000,  .4000,  .0000,  .0000, .0000, .0000,
   1.6000,  1.3846,  1.6000,  1.2000,  .8000,  .4000,  .0000, .0000, .0000,
   1.4000,  1.6615,  2.0000,  1.6000, 1.2000,  .8000,  .4000, .0000, .0000,
   1.2000,  1.9385,  2.4000,  2.0000, 1.6000, 1.2000,  .8000, .4000, .0000,
   1.0000,  2.2154,  2.8000,  2.4000, 2.0000, 1.6000, 1.2000, .8000, .4000 };
 
    double   dresa71[81] = {
     +      -2.3611,  2.5000,   .0000,   .0000,  .0000,
     +        .0000,  .0000, .0000, .1389,
     +       2.5000, -5.0000,  2.5000,   .0000,  .0000,
     +        .0000,  .0000, .0000, .0000,
     +       0.0000,  2.5000, -5.0000,  2.5000,  .0000,
     +        .0000,  .0000, .0000, .0000,
     +       0.0000,   .0000,  2.5000, -5.0000, 2.5000,
     +        .0000,  .0000, .0000, .0000,
     +       0.0000,  0.0000,  0.0000,  2.5000,-5.0000,
     +       2.5000,  .0000, .0000, .0000,
     +       0.0000,  0.0000,  0.0000,  0.0000, 2.5000,
     +      -5.0000, 2.5000, .0000, .0000,
     +       0.0000,  0.0000,  0.0000,  0.0000, 0.0000,
     +       2.5000,-5.0000,2.5000, .0000,
     +       0.0000,  0.0000,  0.0000,  0.0000, 0.0000,
     +       0.0000, 2.5000,-5.0000,2.5000,
     +       0.1389,  0.0000,  0.0000,  0.0000, 0.0000,
     +       0.0000, 0.0000,2.5000,-2.3611 };
 
  double  dtbs7[45]= {
            93.,     84.4000, 76.6000, 70.0000, 65.0000,
            62.0000, 61.4000, 63.60,   69.0000,
           186.,    168.8000,153.2000, 140.0000,130.0000,
           124.000, 122.8000,127.2000, 138.0000,
           279.000, 253.2000,229.8000, 210.0000,195.0000,
           186.0000,184.2000,190.8000, 207.0000,
           372.000, 337.6000,306.4000, 280.0000,260.0000,
           248.0000,245.6000,254.4000, 276.0000,
           465.000, 422.0000,383.0000, 350.0000,325.0000,
           310.0000,307.0000,318.0000, 345.0000};
 
  double  dresb7[45]= {
          1.0000,  2.0000,  3.0000,  4.0000,  5.0000,
          6.0000, 7.0000, 8.0000,  9.0000,
          2.0000,  4.0000,  6.0000,  8.0000, 10.0000,
         12.0000,14.0000,16.0000, 18.0000,
          3.0000,  6.0000,  9.0000, 12.0000, 15.0000,
         18.0000,21.0000,24.0000, 27.0000,
          4.0000,  8.0000, 12.0000, 16.0000, 20.0000,
         24.0000,28.0000,32.0000, 36.0000,
          5.0000, 10.0000, 15.0000, 20.0000, 25.0000,
         30.0000,35.0000,40.0000, 45.0000};
 
/* PZPOTRF, PZPOSV    */
  dcmplx zta6[81], ztb6[54];
  dcmplx ztas6[81] = {
       {18.0,0.0},{1.0, 1.0},{1.0, 1.0},{1.0, 1.0},{1.0,1.0},{1.0,1.0},
       {1.0, 1.0}, {1.0, 1.0}, {1.0, 1.0},   /* 1st           col end */
       {0.0, 0.0},{18.0,0.0},{3.0, 1.0},{3.0,1.0},{3.0,1.0},{3.0,1.0},
       {3.0, 1.0}, {3.0, 1.0}, {3.0, 1.0},   /* 2nd           col end */
       {0.0, 0.0},{0.0, 0.0},{18.0,0.0},{5.0,1.0},{5.0,1.0},{5.0,1.0},
       {5.0, 1.0}, {5.0, 1.0}, {5.0, 1.0},   /* 3rd           col end */
       {0.0, 0.0},{0.0, 0.0},{0.0, 0.0},{18.0,0.0},{7.0,1.0},{7.0,1.0},
       {7.0, 1.0}, {7.0, 1.0}, {7.0, 1.0},   /* 4th           col end */
       {0.0, 0.0},{0.0, 0.0},{0.0, 0.0},{0.0,0.0},{18.0,0.0},{9.0,1.0},
       {9.0, 1.0}, {9.0, 1.0}, {9.0, 1.0},   /* 5th           col end */
       {0.0, 0.0},{0.0, 0.0},{0.0, 0.0},{0.0,0.0},{0.0,0.0},{18.0,0.0},
       {11.0,1.0}, {11.0,1.0}, {11.0,1.0},   /* 6th           col end */
       {0.0, 0.0},{0.0, 0.0},{0.0, 0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0},
       {18.0,0.0}, {13.0,1.0}, {13.0,1.0},   /* 7th           col end */
       {0.0, 0.0},{0.0, 0.0},{0.0, 0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0},
       {0.0, 0.0}, {18.0,0.0}, {15.0,1.0},   /* 8th           col end */
       {0.0, 0.0},{0.0, 0.0},{0.0, 0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0},
       {0.0, 0.0}, {0.0, 0.0}, {18.0,0.0}};  /* 9th           col end */
 
 
  dcmplx ztas6b[81] = {
       {18.0,1.0}, {1.0, 1.0}, {1.0, 1.0}, {1.0, 1.0}, {1.0, 1.0}, {1.0, 1.0},
       {1.0, 1.0}, {1.0, 1.0}, {1.0, 1.0}, /* end col 1 */
       {0.0, 0.0}, {18.0,2.0}, {3.0, 1.0}, {3.0, 1.0}, {3.0, 1.0}, {3.0, 1.0},
       {3.0, 1.0}, {3.0, 1.0}, {3.0, 1.0}, /* end col 2 */
       {0.0, 0.0}, {0.0, 0.0}, {18.0,3.0}, {5.0, 1.0}, {5.0, 1.0}, {5.0, 1.0},
       {5.0, 1.0}, {5.0, 1.0}, {5.0, 1.0}, /* end col 3 */
       {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {18.0,4.0}, {7.0, 1.0}, {7.0, 1.0},
       {7.0, 1.0}, {7.0, 1.0}, {7.0, 1.0}, /* end col 4 */
       {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {18.0,5.0}, {9.0, 1.0},
       {9.0, 1.0}, {9.0, 1.0}, {9.0, 1.0}, /* end col 5 */
       {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {18.0,6.0},
       {11.0,1.0}, {11.0,1.0}, {11.0,1.0}, /* end col 6 */
       {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0},
       {18.0,7.0}, {13.0,1.0}, {13.0,1.0}, /* end col 7 */
       {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0},
       {0.0, 0.0}, {18.0,8.0}, {15.0,1.0}, /* end col 8 */
       {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0},
       {0.0, 0.0}, {0.0, 0.0}, {18.0,9.0}  /* end col 9 */
  };
 
 
  dcmplx zresa61[81] = {
       { 0.06, 0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0},
       { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0},
       { 0.0,  0.0}, { 0.06, 0.0}, {-0.01, 0.0}, { 0.0,  0.0}, { 0.0,  0.0},
       { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0},
       { 0.0,  0.0}, { 0.0,  0.0}, { 0.07, 0.0}, {-0.01, 0.0}, {-0.01, 0.0},
       {-0.01, 0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0},
       { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.07, 0.0}, {-0.01,-0.01},
       {-0.01, 0.0}, {-0.01, 0.0}, {-0.01, 0.0}, { 0.0,  0.0},
       { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.09, 0.0},
       {-0.02,-0.01}, {-0.01, 0.0}, {-0.01, 0.0}, {-0.01, 0.0},
       { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0},
       { 0.11, 0.0}, {-0.02,-0.01}, {-0.02,-0.01}, {-0.02, 0.01},
       { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0},
       { 0.0,  0.0}, { 0.15, 0.0}, {-0.04,-0.02}, {-0.05, 0.01},
       { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0},
       { 0.0,  0.0}, { 0.0,  0.0}, { 0.22, 0.0}, {-0.13,-0.03},
       { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0},
       { 0.0,  0.0}, { 0.0,  0.0}, { 0.0,  0.0}, { 0.22, 0.0}
  };
 
 
  dcmplx ztbs6[54] = {
       {0.0, 0.0},{0.0, 0.0},{0.0, 0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0},
       {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0},
       {60.0, 10.0},{86.0, 28.0},{108.0, 44.0},{126.0,58.0},
       {140.0, 70.0},{150.0, 80.0},{156.0,88.0},{158.0,94.0},
       {156.0, 98.0},{86.0,  2.0},{126.0, 22.0},{160.0, 40.0},
       {188.0, 56.0},{210.0, 70.0},{226.0, 82.0},{236.0, 92.0},
       {240.0,100.0},{238.0,106.0},{112.0,-6.0}, {166.0,16.0},
       {212.0, 36.0},{250.0, 54.0},{280.0, 70.0},{302.0, 84.0},
       {316.0, 96.0},{322.0,106.0},{320.0,114.0},{138.0,-14.0},
       {206.0, 10.0},{264.0, 32.0},{312.0, 52.0},{350.0, 70.0},
       {378.0, 86.0},{396.0,100.0},{404.0,112.0},{402.0,122.0},
       {34.0, 18.0},{46.0, 34.0},{56.0, 48.0},{64.0, 60.0},
       {70.0, 70.0},{ 74.0, 78.0}, {76.0, 84.0},
       {76.0, 88.0},{ 74.0, 90.0}};
 
  dcmplx zresa6[81] = {
       {4.20,  0.0},{0.24, 0.24},{0.24, 0.24},{0.24, 0.24},
       {0.24, 0.24},{0.24, 0.24},{0.24, 0.24},{0.24,0.24},
       {0.24, 0.24},{0.00,  0.0},{4.20, 0.00},{0.68, 0.24},
       {0.68, 0.24},{0.68, 0.24},{0.68, 0.24},{0.68, 0.24},
       {0.68, 0.24},{0.68, 0.24},{0.0,0.0},{0.0,0.0},{4.20,0.0},
       {1.10, 0.24},{1.10, 0.24},{1.10, 0.24},{1.1,0.24},{1.1,0.24},
       {1.10, 0.24}, {0.00,  0.0},{0.00, 0.00},{0.00, 0.00},
       {4.00, 0.00},{1.30, 0.25},{1.30, 0.25},{1.3,0.25},{1.3,0.25},
       {1.3, 0.25},{0.00,  0.0},{0.00, 0.00},{0.00, 0.00},{0.0,0.0},
       {3.80, 0.00},{1.40, 0.26},{1.4,0.26},{1.4,0.26},
       {1.4,0.26},{0.0, 0.0},{0.0, 0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0},
       {3.5,0.0},{1.50, 0.28},{1.50, 0.28},{1.50, 0.28},{0.0,0.0},
       {0.00, 0.00},{0.00, 0.00},{0.0,0.0},{0.0,0.0},{0.0,0.0},
       {3.20,0.00},{1.60,0.32},{1.60,0.32},{0.0,0.},{0.0,0.0},{0.0,0.0},
       {0.00, 0.00},{0.00, 0.00},{0.00, 0.00},{0.0,0.0},{2.7,0.0},
       {1.6,0.37},{0.00,  0.0},{0.00, 0.00},{0.00, 0.00},{0.0,0.},
       {0.0, 0.0},{0.0, 0.0},{0.0,0.},{0.,0.},{2.2,0.0}};
 
 
  dcmplx zresb6[54] = {
       {0.0, 0.0},{0.0, 0.0},{0.0, 0.0},{0.0,0.},{0.0,0.0},{0.0,0.0},
       {0.0, 0.0},{0.0, 0.0},{0.0, 0.0},  /* 1st              col end */
       {2.0, 1.0}, {2.0, 1.0}, {2.0, 1.0},{2.0,1.0},{2.0,1.0},
       {2.0,1.0},{2.0, 1.0}, {2.0, 1.0}, {2.0, 1.0},  /* 2nd  col end */
       {3.0, 1.0}, {3.0, 1.0}, {3.0, 1.0}, {3.0,1.0},{3.0,1.0},
       {3.0,1.0},{3.0, 1.0}, {3.0, 1.0}, {3.0, 1.0},  /* 3rd  col end */
       {4.0, 1.0}, {4.0, 1.0},{4.0, 1.0},{4.0,1.0}, {4.0,1.0},
       {4.0,1.0},{4.0, 1.0}, {4.0, 1.0}, {4.0, 1.0},  /* 4th  col end */
       {5.0, 1.0}, {5.0, 1.0},{5.0, 1.0},{5.0,1.0},{5.0,1.0},
       {5.0, 1.0},{5.0, 1.0}, {5.0, 1.0}, {5.0, 1.0},  /* 5th col end */
       {1.0, 1.0}, {1.0, 1.0},{1.0, 1.0},{1.0,1.0},{1.0,1.0},
       {1.0,1.0},{1.0, 1.0}, {1.0, 1.0}, {1.0, 1.0}}; /* 6th  col end */
 
/* PZGETRF/S PZGESV PZGETRI PZLANGE PZGECON START : 7     */
  dcmplx zta7[81], ztb7[54];
  dcmplx ztas7[81] = {
       {2.0,1.0},{2.4, 1.0},{2.8, 1.0},{3.2,1.0},{3.6,1.0},
       {4.0,1.0},{4.4, 1.0},{4.8, 1.0}, {5.2, 1.0},  /* 1st   col end */
       {2.4,-1.0},{2.0, 1.0}, {2.4, 1.0},{2.8,1.0},{3.2,1.0},
       {3.6,1.0},{4.0, 1.0},{4.4, 1.0}, {4.8, 1.0},  /* 2nd   col end */
       {2.8,-1.0},{2.4,-1.0},{2.0, 1.0},{2.4,1.0},{2.8,1.0},
       {3.2,1.0},{3.6, 1.0}, {4.0, 1.0}, {4.4, 1.0},  /* 3rd  col end */
       {3.2,-1.0},{2.8,-1.0},{2.4,-1.0},{2.0,1.0},{2.4,1.0},
       {2.8, 1.0},{3.2, 1.0}, {3.6, 1.0}, {4.0, 1.0},  /* 4th col end */
       {3.6,-1.0}, {3.2,-1.0}, {2.8,-1.0},{2.4,-1.0},{2.0,1.0},
       {2.4, 1.0},{2.8, 1.0}, {3.2, 1.0}, {3.6, 1.0},  /* 5th col end */
       {4.0,-1.0}, {3.6,-1.0}, {3.2,-1.0},{2.8,-1.0},{2.4,-1.0},
       {2.0, 1.0},{2.4, 1.0}, {2.8, 1.0}, {3.2, 1.0},  /* 6th col end */
       {4.4,-1.0}, {4.0,-1.0}, {3.6,-1.0},{3.2,-1.0},{2.8,-1.0},
       {2.4, -1.0},{2.0, 1.0}, {2.4, 1.0}, {2.8, 1.0},  /* 7th col end*/
       {4.8,-1.0}, {4.4,-1.0}, {4.0,-1.0},{3.6,-1.0},{3.2,-1.0},
       {2.8, -1.0},{2.4,-1.0}, {2.0, 1.0}, {2.4, 1.0},  /* 8th col end*/
       {5.2,-1.0}, {4.8,-1.0}, {4.4,-1.0},{4.0,-1.0},{3.6,-1.0},
       {3.2, -1.0},{2.8,-1.0}, {2.4,-1.0}, {2.0, 1.0}}; /* 9th col end*/
 
  dcmplx zresa7[81] = {
       {5.20,1.0},{0.4, 0.1}, {0.5, 0.1},{0.6,0.1},{0.6,0.1},
       {0.7, 0.1},{0.8, 0.0}, {0.9, 0.0}, {0.9, 0.0},  /* 1st col end */
       {4.80,1.0}, {0.6,-2.0},{0.0,-0.1},{0.0,-0.1},{0.0,-0.1},
       {0.0,-0.1}, {0.0,0.0},{0.0, 0.0}, {0.0, 0.0},  /* 2nd  col end */
       {4.40,1.0}, {1.1,-1.9}, {0.6,-1.9},{-0.1,-0.1},{-0.1,-0.1},
       {0.0,0.0},{0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0},  /* 3rd  col end */
       {4.00,1.0}, {1.7,-1.9}, {1.2,-1.8},{0.7,-1.9},{-0.1,0.0},
       {-0.1,0.0},{0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0},  /* 4th col end */
       {3.60,1.0}, {2.3,-1.8}, {1.8,-1.7},{1.3,-1.7},{0.7,-1.9},
       {-0.1,0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0},  /* 5th col end*/
       {3.20,1.0}, {2.8,-1.8}, {2.5,-1.6},{2.0,-1.6},{1.5,-1.7},
       {0.8,-1.9},{0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0},  /* 6th  col end*/
       {2.80,1.0}, {3.4,-1.7}, {3.1,-1.5},{2.7,-1.5},{2.2,-1.6},
       {1.6,-1.8},{0.8,-1.9}, {0.0, 0.0}, {0.0, 0.0},  /* 7th  col end*/
       {2.40,1.0}, {3.9,-1.7}, {3.7,-1.4},{3.4,-1.4},{2.9,-1.5},
       {2.4,-1.6},{1.7,-1.8}, {0.8,-2.0}, {0.0, 0.0},  /* 8th  col end*/
       {2.00,1.0}, {4.5,-1.6}, {4.3,-1.3},{4.0,-1.2},{3.7,-1.3},
       {3.2,-1.5},{2.5,-1.8}, {1.7,-1.9}, {0.8,-2.0}}; /* 9th  col end*/
 
  dcmplx zresa71[81] = {
       {-.17,-.42}, {0.18, 0.43}, {0.01, 0.00},
       {0.01, 0.00}, {0.00, 0.01}, {0.00, 0.01},
       {0.00, 0.01}, {0.00, 0.01}, {-.01, 0.01},
       {-0.12,0.13}, {-0.04,-0.55},{0.18, 0.43},
       {0.01, 0.00}, {0.01, 0.00}, {0.00, 0.01},
       {0.00, 0.01}, {0.00, 0.01}, {0.00, 0.01},
       {-0.06,0.15}, {-0.06,-0.03}, {-0.04,-0.55},
       { 0.18, 0.43},{ 0.01, 0.00}, {0.01, 0.00},
       {0.00, 0.01}, {0.00, 0.01}, {0.00, 0.01},
       {0.00, 0.15}, {-0.06,-0.01}, {-0.06,-0.03},
       {-0.04,-0.55},{ 0.18, 0.43},{ 0.01, 0.00},
       {0.01, 0.00}, { 0.00, 0.01}, {0.00, 0.01},
       {0.05,0.13}, {-0.05, 0.01}, {-0.06,-0.01},
       {-0.06,-0.03},{-0.04,-0.55},{ 0.18, 0.43},
       {0.01, 0.00}, {0.01, 0.00}, {0.00, 0.01},
       {0.09,0.09}, {-0.04, 0.03}, {-0.05, 0.01},
       {-0.06,-0.01},{-0.06,-0.03},{-0.04,-0.55},
       {0.18, 0.43}, {0.01, 0.00}, {0.01, 0.00},
       {0.11,0.05}, {-0.03, 0.04}, {-0.04, 0.03},
       {-0.05, 0.01},{-0.06,-0.01},{-0.06,-0.03},
       {-0.04,-0.55}, {0.18, 0.43}, {0.01, 0.00},
       {0.11,0.00}, {-0.01, 0.04}, {-0.03, 0.04},
       {-0.04, 0.03},{-0.05, 0.01},{-0.06,-0.01},
       {-0.06,-0.03}, {-0.04,-0.55}, {0.18, 0.43},
       {0.04,-0.28}, {0.11, 0.00}, {0.11, 0.05},
       { 0.09, 0.09},{ 0.05, 0.13},{ 0.00, 0.15},
       {-.06, 0.15}, {-0.12, 0.13}, {-0.17,-0.42}};
 
  dcmplx ztbs7[54] = {
       {0.0, 0.0}, {0.0, 0.0},{0.0, 0.0},{0.0,0.0},{0.0,0.0},
       {0.0,0.0},{0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0},
       {193.0,-10.60}, {173.8,-9.40}, {156.2, -5.40},{141.0,1.4},
       {129.0,11.00},{121.0, 23.40},{117.8,38.6},{120.2,56.6},
       {129.0,77.4},{200.0,21.80},{178.8,20.20}, {159.2, 22.20 },
       {142.0, 27.80}, {128.0,37.00}, {118.0,49.80}, {112.8,66.2},
       {113.2,86.20},{ 120.0,109.80},{207.0,54.2},{183.8,49.8},
       {162.2,49.8},{143.0, 54.20}, {127.0,63.00}, {115.0, 76.20},
       {107.8, 93.80}, {106.2,115.8}, {111.0,142.20},{214.0,86.6},
       {188.8,79.40}, {165.2, 77.40},{144.0,80.6},{126.0,89.0},
       {112.0,102.60},{102.8,121.40}, { 99.2,145.4}, {102.0,174.60},
       {221.0,119.00}, {193.8,109.0}, {168.2,105.00},{145.0,107.0},
       {125.0,115.0}, {109.0,129.00},{97.8,149.0},{92.2,175.0},
       {93.0,207.00}};
 
 
  dcmplx zresb7[54] = {
       {0.0, 0.0}, {0.0, 0.0},{0.0, 0.0},{0.0,0.0},{0.0,0.0},
       {0.0,0.0},{0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0},  /* 1st  col end */
       {1.0, 1.0}, {2.0, 1.0},{3.0, 1.0},{4.0,1.0},{5.0,1.0},
       {6.0, 1.0},{7.0, 1.0}, {8.0, 1.0}, {9.0, 1.0},  /* 2nd col end */
       {1.0, 2.0},{2.0, 2.0}, {3.0, 2.0},{4.0,2.0},{5.0,2.0},
       {6.0,2.0},{7.0,2.0}, {8.0, 2.0}, {9.0, 2.0},  /* 3rd   col end */
       {1.0, 3.0}, {2.0, 3.0},{3.0, 3.0},{4.0,3.0},{5.0,3.0},
       {6.0, 3.0},{7.0, 3.0}, {8.0, 3.0}, {9.0, 3.0},  /* 4th col end */
       {1.0, 4.0}, {2.0, 4.0}, {3.0, 4.0},{4.0,4.0},{5.0,4.0},
       {6.0,4.0},{7.0, 4.0}, {8.0, 4.0}, {9.0, 4.0},  /* 5th  col end */
       {1.0, 5.0}, {2.0, 5.0},{3.0, 5.0},{4.0,5.0},{5.0,5.0},
       {6.0,5.0},{7.0, 5.0}, {8.0, 5.0}, {9.0, 5.0}}; /* 6th  col end */
 
/* PDGTTRF/S  */
 
      double *af,work[100];
      double dtl8[12], dtd8[12], dtu8[12], dtb8[36], dt2u8[12];
      double dtls8[12] = {0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
                          1.0, 1.0, 1.0, 1.0};
 
      double dtds8[12] = {2.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0,
                          3.0, 3.0, 3.0, 3.0};
 
      double dtus8[12] = {2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0,
                          2.0, 2.0, 2.0, 0.0};
 
      double dtbs8[36] = {46.0, 65.0, 59.0, 53.0,
                          47.0, 41.0, 35.0, 29.0,
                          23.0, 17.0, 11.0,  5.0,
                           6.0, 13.0, 19.0, 25.0,
                          31.0, 37.0, 43.0, 49.0,
                          55.0, 61.0, 67.0, 47.0,
                           4.0,  6.0,  6.0,  6.0,
                           6.0,  6.0,  6.0,  6.0,
                           6.0,  6.0,  6.0,  4.0};
 
      double dresb8[36] =  {12.0, 11.0, 10.0,  9.0,
                             8.0,  7.0,  6.0,  5.0,
                             4.0,  3.0,  2.0,  1.0,
                             1.0,  2.0,  3.0,  4.0,
                             5.0,  6.0,  7.0,  8.0,
                             9.0, 10.0, 11.0, 12.0,
                             1.0,  1.0,  1.0,  1.0,
                             1.0,  1.0,  1.0,  1.0,
                             1.0,  1.0,  1.0,  1.0};
 
/* PDPBTRF/S START */
      int k;
      double dta9[72];
      double dtb9[27];
      double dtas9[72] = {
         1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
         2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 1.0,
         3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 2.0, 0.0,
         4.0, 4.0, 4.0, 4.0, 4.0, 3.0, 0.0, 0.0,
         5.0, 5.0, 5.0, 5.0, 4.0, 0.0, 0.0, 0.0,
         6.0, 6.0, 6.0, 5.0, 0.0, 0.0, 0.0, 0.0,
         7.0, 7.0, 6.0, 0.0, 0.0, 0.0, 0.0, 0.0,
         8.0, 7.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
         8.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
 
      double dtbs9[27] = {
         8.0, 16.0, 23.0, 29.0, 34.0, 38.0,  41.0, 43.0, 36.0,
        36.0, 80.0,122.0,161.0,196.0,226.0, 250.0,267.0,240.0,
        44.0, 80.0,108.0,129.0,144.0,154.0, 160.0,163.0,120.0 };
 
      double dresb9[27] = {
           1.00,  1.00,  1.00,  1.00,  1.00,  1.00, 1.00,  1.00,  1.00,
           1.00,  2.00,  3.00,  4.00,  5.00,  6.00, 7.00,  8.00,  9.00,
           9.00,  8.00,  7.00,  6.00,  5.00,  4.00, 3.00,  2.00,  1.00 };
 
/* PDPTTRF/S, PDPTSV       */
      double dtd10[12], dtu10[12], dtb10[36];
      double dtds10[12] = {4.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0,
                           5.0, 5.0, 5.0, 5.0};
      double dtus10[12] = {2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0,
                           2.0, 2.0, 2.0, 0.0};
      double dtbs10[36] = { 70.0, 99.0, 90.0, 81.0,
                            72.0, 63.0, 54.0, 45.0,
                            36.0, 27.0, 18.0,  9.0,
                             8.0, 18.0, 27.0, 36.0,
                            45.0, 54.0, 63.0, 72.0,
                            81.0, 90.0, 99.0, 82.0,
                             6.0,  9.0,  9.0,  9.0,
                             9.0,  9.0,  9.0,  9.0,
                             9.0,  9.0,  9.0,  7.0};
 
/* PDGEQRF PZGEQRF         */
      double dga11[12], dgt11[3];
      dcmplx zga11[12], zgt11[3];
      double dgas11[12] = {  1.0,  2.0,  2.0,  4.0,
                           -2.0,  0.0, -4.0,  0.0,
                           -1.0,  1.0,  2.0,  0.0};
 
      double dresa11[12] = { -5.0,  0.33,  0.33,  0.67,
                            2.0, -4.00, -0.50,  0.50,
                           -1.00, 1.00, -2.00,  0.00};
 
      double drest11[3] = {  1.2,  1.33,  2.00};
 
  dcmplx zgas11[12] = {
       {1.0, 0.0}, {1.0,-1.0},{1.0,-2.0},
       {-2.0,1.0},{2.0, 2.0}, {-2.0, 3.0},
       {-3.0,-1.0}, {-3.0, 0.0},{-3.0, 1.0},
       {4.0,-3.0},{-4.0,-2.0}, {4.0,-1.0}};
 
  dcmplx zresa11[12] = {
       {-2.83, 0.0}, {0.26,-0.26},{0.26,-0.52},
       { 3.54,-1.41},{-3.39, 0.0}, {-0.29,-0.09},
       { 3.89, 3.18}, { 0.37,-0.37},{-1.87, 0.0},
       {-2.83, 0.71},{ 6.78, 0.74}, {1.87,-1.87}};
 
  dcmplx zrest11[3] = {
       { 1.35, 0.0}, {1.83,-0.02},{1.47,-0.88}};
/* PDGELS  PZGELS  START : 12  */
      double dga12[12], dgb12[20];
      dcmplx zga12[12], zgb12[12];
      double dgas12[12]=  {  1.0,  2.0,  2.0,  4.0,
                            -2.0,  0.0, -4.0,  0.0,
                            -1.0,  1.0,  2.0,  0.0};
 
      double dgbs12[20]=  { -1.0,  1.0,  1.0, -2.0,
                            -2.0,  3.0,  0.0,  4.0,
                            -7.0,  4.0,  4.0,  4.0,
                             0.0,  3.0,  2.0,  0.0,
                            -5.0,  5.0,  2.0,  4.0};
 
      double dresb12[20]=  { -0.4,  0.00,  1.00, -1.00,
                              1.0,  1.00,  1.00,  0.00,
                              0.80, 1.50,  4.00,  2.00,
                              0.2,  0.00,  1.00, -2.00,
                              1.00, 1.50,  3.00,  0.00};
 
  dcmplx zgas12[12] = {
       {1.0, 0.0}, {1.0,-1.0},{1.0,-2.0},
       {-2.0,1.0},{2.0, 2.0}, {-2.0, 3.0},
       {-3.0,-1.0}, {-3.0, 0.0},{-3.0, 1.0},
       {4.0,-3.0},{-4.0,-2.0}, {4.0,-1.0}};
 
      dcmplx zgbs12[12] = {
       { 1.0,0.0}, {-1.0,1.0}, {2.0, 1.0}, {0.0, 0.0},
       { 0.0,1.0}, { 1.0,-1.0},{1.0, 2.0}, {0.0, 0.0},
       { 1.0,1.0}, { 0.0, 0.0},{-1.0,-1.0},{0.0, 0.0}};
 
      dcmplx zresb12[12] = {
       {-.16,.15},{ .11,.02}, {-.13,-.32}, {.37,-.05},
       {-.08,.18}, { .21,-.50},{.16, .12}, {.04, .06},
       { .16,-.31}, {-.38,.65},{-.27,-.28},{-.19,.32}};
 
/* PDGELS  PZGELS  END   : 12  */
/* PDPOTRF/S  PDPOTRI PDLANSY         PDPOCON PDPOSV START : 6    */
/* PZPOTRF/S  PZPOTRI PZLANSY PZLANHE PZPOCON PZPOSV START : 6    */
 
  char modn[9], *modname=modn;
  blacs_pinfo(&iam,&nnodes);
  nprow = 1;
  npcol = nnodes;
  blacs_get(0,0,&ictx);
  blacs_gridinit(&ictx,"R",nprow,npcol);
  blacs_gridinfo(ictx,&nprow,&npcol,&myrow,&mycol);
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 9;
  desc_a[N_] = 9;
  desc_a[MB_] = 3;
  desc_a[NB_] = 3;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(9,3,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_b[M_] = 9;
  desc_b[N_] = 6;
  desc_b[MB_] = 3;
  desc_b[NB_] = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(9,3,myrow,0,nprow);
  desc_b[LLD_] = MAX(1,il_lda);
 
  modname = "PDPOTRF ";
  pdmtxcpy(9, 9, 1, 1, dta6, desc_a, dtas6);
  pdpotrf("L", 9, dta6, 1, 1, desc_a,&iret);
  if (iret != 0)
  {
   if(iam == 0)
    printf( " IRET from PDPOTRF =%4d\n", iret);
   *ifail = *ifail + 1;
  }
  else
  {
  pdmtchk(modname,9, 9, 1, 1, dta6, desc_a, dresa6, tol, ierr,&ilerr,
         &error, ifail);
 
  modname = "PDPOTRS ";
  pdmtxcpy(9, 5, 1, 2, dtb6, desc_b, dtbs6);
  pdpotrs("L", 9, 5, dta6, 1, 1, desc_a, dtb6, 1, 2, desc_b,&iret);
  if (iret != 0)
  {
   if(iam == 0) printf( " IRET from PDPOTRS = %4d\n", iret);
   *ifail = *ifail + 1;
  }
  pdmtchk(modname,9, 5, 1, 2, dtb6, desc_b, dresb6, tol, ierr,&ilerr,
         &error, ifail);
  }
 
  modname = "PDPOTRI ";
  tol = 0.001;
  pdmtxcpy(9, 9, 1, 1, dta6, desc_a, dtas6);
  pdpotrf("L", 9, dta6, 1, 1, desc_a,&iret);
  pdpotri("L", 9, dta6, 1, 1, desc_a,&iret);
  if (iret != 0) {
   if (iam == 0) printf( " IRET from PDPOTRI = %d\n",iret);
   *ifail = *ifail + 1;
  } else {
   pdmtchk(modname,9, 9, 1, 1, dta6, desc_a, dresa61, tol,
           ierr,&ilerr, &error, ifail);
  }
 
  modname = "PDPOCON ";
  pdmtxcpy(9, 9, 1, 1, dta6, desc_a, dtas6);
  anorm=pdlansy("I","L",9,dta6,1,1,desc_a, work);
  if ( fabs(anorm-45.00) > ntol) {
   if (iam == 0) printf( " ANORM from PDLANSY  = %f\n", anorm);
   *ifail = *ifail + 1;
  }
  pdpotrf("L", 9, dta6, 1, 1, desc_a,&iret);
  pdpocon("L",9,dta6,1,1,desc_a,anorm,&rcond,work,0,iwork,0,&iret);
  if (iret != 0) {
   if (iam == 0) printf( " IRET from PDPOCON = %d\n",iret);
   *ifail = *ifail + 1;
  } else {
   if(fabs(rcond-rcondz) > tol) {
    if (iam == 0) printf( " RCOND from PDPOCON = %f %f",rcond,rcondz);
    *ifail = *ifail + 1;
   }
  }
 
  modname = "PDPOSV  ";
  pdmtxcpy(9, 9, 1, 1, dta6, desc_a, dtas6);
  pdmtxcpy(9, 5, 1, 2, dtb6, desc_b, dtbs6);
  pdposv("L", 9, 5, dta6, 1, 1, desc_a, dtb6, 1, 2, desc_b,&iret);
  if (iret != 0)
  {
   if(iam == 0)
    printf( " IRET from PDPOSV  = %4d\n", iret);
   *ifail = *ifail + 1;
      printf("FAIL in pdposv: %d\n", *ifail);
  }
  pdmtchk(modname,9, 9, 1, 1, dta6, desc_a, dresa6, tol, ierr,&ilerr,
         &error, ifail);
  pdmtchk(modname,9, 5, 1, 2, dtb6, desc_b, dresb6, tol, ierr,&ilerr,
         &error, ifail);
 
  tol = 0.1;
  modname = "PZPOTRF ";
  pzmtxcpy(9, 9, 1, 1, zta6, desc_a, ztas6);
  pzpotrf("L", 9, zta6, 1, 1, desc_a,&iret);
  if (iret != 0)
  {
   if(iam == 0)
   printf( " IRET from PZPOTRF =%4d\n",iret);
   *ifail = *ifail + 1;
      printf("FAIL in pzpotrf: %d\n", *ifail);
  }
  else
  {
  pzmtchk(modname,9, 9, 1, 1, zta6, desc_a, zresa6, tol, ierr,&ilerr,
         &error, ifail);
 
  modname = "PZPOTRS ";
  pzmtxcpy(9, 5, 1, 2, ztb6, desc_b, ztbs6);
  pzpotrs("L", 9, 5, zta6, 1, 1, desc_a, ztb6, 1, 2, desc_b,&iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PZPOTRS = %4d\n", iret);
    *ifail = *ifail + 1;
      printf("FAIL in pzpotrs: %d\n", *ifail);
   }
  pzmtchk(modname,9, 5, 1, 2, ztb6, desc_b, zresb6, tol, ierr,&ilerr,
         &error, ifail);
  }
 
  modname = "PZPOTRI ";
  pzmtxcpy(9, 9, 1, 1, zta6, desc_a, ztas6);
  pzpotrf("L", 9, zta6, 1, 1, desc_a,&iret);
 
  pzpotri("L", 9, zta6, 1, 1, desc_a,&iret);
 
  if (iret != 0) {
   if (iam == 0) printf( " IRET from PZPOTRI = %d\n",iret);
   *ifail = *ifail + 1;
  } else {
   pzmtchk(modname,9, 9, 1, 1, zta6, desc_a, zresa61, ztol,
           ierr,&ilerr, &error, ifail);
  }
 
  modname = "PZPOCON ";
  ntol = 0.1;
  pzmtxcpy(9, 9, 1, 1, zta6, desc_a, ztas6);
  anorm=pzlanhe("1","L",9,zta6,1,1,desc_a, work);
  if ( fabs(anorm-82.92) > ntol) {
    if (iam == 0) printf( " ANORM from PZLANHE  = %f\n", anorm);
    *ifail = *ifail + 1;
  }
  pzpotrf("L", 9, zta6, 1, 1, desc_a,&iret);
  pzpocon("L",9,zta6,1,1,desc_a,anorm,&rcond,workz,0,work,0,&iret);
  if (iret != 0) {
   if (iam == 0) printf( " IRET from PZPOCON = %d\n",iret);
   *ifail = *ifail + 1;
  } else {
   if(fabs(rcond-rconds2) > tol) {
    if (iam == 0) printf( " RCOND from PZPOCON = %f %f\n",rcond,rconds2);
    *ifail = *ifail + 1;
   }
  }
 
  modname = "PZPOSV  ";
  pzmtxcpy(9, 9, 1, 1, zta6, desc_a, ztas6);
  pzmtxcpy(9, 5, 1, 2, ztb6, desc_b, ztbs6);
 
  pzposv("L", 9, 5, zta6, 1, 1, desc_a, ztb6, 1, 2, desc_b,&iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PZPOSV  = %4d\n", iret);
    *ifail = *ifail + 1;
   }
  pzmtchk(modname,9, 9, 1, 1, zta6, desc_a, zresa6, tol, ierr,&ilerr,
         &error, ifail);
  pzmtchk(modname,9, 5, 1, 2, ztb6, desc_b, zresb6, tol, ierr,&ilerr,
         &error, ifail);
 
  modname = "PZLANSY ";
  ntol = 0.01;
  pzmtxcpy(9, 9, 1, 1, zta6, desc_a, ztas6b);
  anorm=pzlansy("F","L",9,zta6,1,1,desc_a, work);
  if ( fabs(anorm-81.88) > ntol) {
    if (iam == 0) printf( " ANORM from PZLANSY  = %f\n", anorm);
    *ifail = *ifail + 1;
  }
 
/* PDPOTRF/S  PDPOTRI PDLANSY         PDPOCON PDPOSV END : 6    */
/* PZPOTRF/S  PZPOTRI PZLANSY PZLANHE PZPOCON PZPOSV END : 6    */
 
/* PDGETRF/S & PZGETRF/S & PDGESV & PZGESV START : 7      */
/* PDGETRI   & PZGETRI                     START : 7      */
/* PDGECON   & PZGECON, PDLANGE  & PZLANGE START : 7      */
 
 
  desc_a[DT_] = BLOCK_CYCLIC_2D;
  desc_a[M_] = 9;
  desc_a[N_] = 9;
  desc_a[MB_] = 3;
  desc_a[NB_] = 3;
  desc_a[RSRC_] = 0;
  desc_a[CSRC_] = 0;
  desc_a[CTXT_] = ictx;
  il_lda = numroc(9,3,myrow,0,nprow);
  desc_a[LLD_] = MAX(1,il_lda);
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_b[M_] = 9;
  desc_b[N_] = 5;
  desc_b[MB_] = 3;
  desc_b[NB_] = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(9,3,myrow,0,nprow);
  desc_b[LLD_] = MAX(1,il_lda);
 
  modname = "PDGETRF ";
  tol = 0.001;
  pdmtxcpy(9, 9, 1, 1, dta7, desc_a, dtas7);
  pdgetrf(9, 9, dta7, 1, 1, desc_a, ipvt,&iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PDGETRF = %4d\n", iret);
   *ifail = *ifail + 1;
  }
  else
  {
 
  pdmtchk(modname,9, 9, 1, 1, dta7, desc_a, dresa7, tol, ierr,& ilerr,
         &error, ifail);
 
  modname = "PDGETRS ";
  pdmtxcpy(9, 5, 1, 1, dtb7, desc_b, dtbs7);
  pdgetrs("N", 9, 5, dta7, 1, 1, desc_a, ipvt, dtb7, 1, 1, desc_b,&iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PDGETRS = %4d\n", iret);
   *ifail = *ifail + 1;
  }
  pdmtchk(modname,9, 5, 1, 1, dtb7, desc_b, dresb7, tol, ierr,&ilerr,
         &error, ifail);
  }
 
  modname = "PDGESV  ";
  pdmtxcpy(9, 9, 1, 1, dta7, desc_a, dtas7);
  pdmtxcpy(9, 5, 1, 1, dtb7, desc_b, dtbs7);
  pdgesv(9, 5, dta7, 1, 1, desc_a, ipvt, dtb7, 1, 1, desc_b,&iret)      ;
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PDGESV  = %4d\n", iret);
   *ifail = *ifail + 1;
  }
  pdmtchk(modname,9, 9, 1, 1, dta7, desc_a, dresa7, tol, ierr,& ilerr,
         &error, ifail);
  pdmtchk(modname,9, 5, 1, 1, dtb7, desc_b, dresb7, tol, ierr,&ilerr,
         &error, ifail);
 
  modname = "PDGETRI ";
  lwork = 0;
  tol = 0.001;
  pdmtxcpy(9, 9, 1, 1, dta7, desc_a, dtas7);
  pdgetrf(9, 9, dta7, 1, 1, desc_a, ipvt,&iret);
  pdgetri(9, dta7, 1, 1, desc_a, ipvt,work,lwork,iwork,liwork,&iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PDGETRI = %4d\n", iret);
   *ifail = *ifail + 1;
  }
 
  pdmtchk(modname,9, 9, 1, 1, dta7, desc_a, dresa71,tol, ierr,& ilerr,
         &error, ifail);
 
  modname = "PDGECON ";
  liwork = 0;
  tol = 0.001;
  pdmtxcpy(9, 9, 1, 1, dta7, desc_a, dtas7);
  anorm = pdlange("O", 9, 9, dta7, 1, 1, desc_a, work);
  pdgetrf(9, 9, dta7, 1, 1, desc_a, ipvt,&iret);
  pdgecon("O", 9, dta7, 1, 1, desc_a, anorm, &rcond, work, lwork,
          iwork, liwork, &iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PDGECON = %4d\n", iret);
   *ifail = *ifail + 1;
  }
 
       if ((fabs(rcond) - fabs(rcondd)) > tol)
        {
          *ifail = *ifail + 1;
         printf(" PDGECON rcond= %4g\n",rcond);
         printf("  rcondd = %4g\n",rcondd);
        }
 
  modname = "PZGETRF ";
  tol = 0.1;
  pzmtxcpy(9, 9, 1, 1, zta7, desc_a, ztas7);
  pzgetrf(9, 9, zta7, 1, 1, desc_a, ipvt,&iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PZGETRF = %4d\n", iret);
   *ifail = *ifail + 1;
  }
  else
  {
    pzmtchk(modname,9, 9, 1, 1, zta7, desc_a, zresa7, tol, ierr,&ilerr,
           &error, ifail);
 
  desc_b[DT_] = BLOCK_CYCLIC_2D;
  desc_b[M_] = 9;
  desc_b[N_] = 6;
  desc_b[MB_] = 3;
  desc_b[NB_] = 2;
  desc_b[RSRC_] = 0;
  desc_b[CSRC_] = 0;
  desc_b[CTXT_] = ictx;
  il_lda = numroc(9,3,myrow,0,nprow);
  desc_b[LLD_] = MAX(1,il_lda);
 
  modname = "PZGETRS ";
  pzmtxcpy(9, 6, 1, 2, ztb7, desc_b, ztbs7);
  pzgetrs("N", 9, 5, zta7, 1, 1, desc_a, ipvt, ztb7, 1, 2, desc_b,&iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PZGETRS = %4d\n", iret);
   *ifail = *ifail + 1;
  }
  pzmtchk(modname,9, 6, 1, 2, ztb7, desc_b, zresb7, tol, ierr,&ilerr,
         &error, ifail);
  }
 
  modname = "PZGESV  ";
  pzmtxcpy(9, 9, 1, 1, zta7, desc_a, ztas7);
  pzmtxcpy(9, 6, 1, 2, ztb7, desc_b, ztbs7);
  pzgesv(9, 5, zta7, 1, 1, desc_a, ipvt, ztb7, 1, 2, desc_b,&iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PZGESV  = %4d\n", iret);
   *ifail = *ifail + 1;
  }
    pzmtchk(modname,9, 9, 1, 1, zta7, desc_a, zresa7, tol, ierr,&ilerr,
           &error, ifail);
  pzmtchk(modname,9, 6, 1, 2, ztb7, desc_b, zresb7, tol, ierr,&ilerr,
         &error, ifail);
 
  modname = "PZGETRI ";
  tol = 0.1;
  pzmtxcpy(9, 9, 1, 1, zta7, desc_a, ztas7);
  pzgetrf(9, 9, zta7, 1, 1, desc_a, ipvt,&iret);
  pzgetri(9, zta7, 1, 1, desc_a, ipvt,workz,lwork,iwork,liwork,&iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PZGETRI = %4d\n", iret);
   *ifail = *ifail + 1;
  }
    pzmtchk(modname,9, 9, 1, 1, zta7, desc_a,zresa71, tol, ierr,&ilerr,
           &error, ifail);
 
  modname = "PZGECON ";
  tol = 0.1;
  pzmtxcpy(9, 9, 1, 1, zta7, desc_a, ztas7);
  anorm = pzlange("O", 9, 9, zta7, 1, 1, desc_a, rwork);
  pzgetrf(9, 9, zta7, 1, 1, desc_a, ipvt,&iret);
  pzgecon("O", 9, zta7, 1, 1, desc_a, anorm, &rcond, workz, lwork,
           rwork, lrwork, &iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PZGECON = %4d\n", iret);
   *ifail = *ifail + 1;
  }
       if ((fabs(rcond) - fabs(rcondz)) > tol)
        {
          *ifail = *ifail + 1;
         printf(" PZGECON rcond= %4g\n",rcond);
         printf("  rcondz = %4g\n",rcondz);
        }
 
/* PDGETRF/S & PZGETRF/S PDGESV, PZGESV END : 7  */
/* PDGECON & PZGECON, PDLANGE & PZLANGE END : 7  */
/* PDGETRI & PZGETRI                    END : 7  */
 
/* PDGEQRF  PZGEQRF  START        : 11     */
 
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 4;
      desc_a[N_]    = 3;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
/* following desc_x is created to check vector tau    */
 
      desc_x[DT_]   = BLOCK_CYCLIC_2D;
      desc_x[M_]    = 1;
      INCX          = desc_x[M_];
      desc_x[N_]    = desc_a[N_];
      desc_x[MB_]   = 1;
      desc_x[NB_]   = desc_a[NB_];
      desc_x[CSRC_] = desc_a[CSRC_];
      desc_x[CTXT_] = ictx;
      desc_x[LLD_]  = INCX;
 
      modname = "PDGEQRF ";
      tol = 0.1;
      pdmtxcpy(4, 3, 1, 1, dga11, desc_a, dgas11);
      pdgeqrf(4, 3, dga11, 1, 1, desc_a, dgt11, work, 0, &iret);
      if (iret != 0)
  {
    if(iam == 0)
    printf( " IRET from PDGEQRF = %4d\n", iret);
   *ifail = *ifail + 1;
  }
      else
  {
    pdmtchk(modname, 4, 3, 1, 1, dga11, desc_a, dresa11, tol,
                    ierr,&ilerr, &error, ifail);
    for (i=0; i< nprow; i++)
      {
         desc_x[RSRC_] = i;
         pdvtchk(modname,3, 1, 1, dgt11, desc_x, drest11, 1, tol,
                      ierr,&ilerr, &error, ifail);
       }
   }
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 3;
      desc_a[N_]    = 4;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(3,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
/* following desc_x is created to check vector tau    */
 
      desc_x[DT_]   = BLOCK_CYCLIC_2D;
      desc_x[M_]    = 1;
      INCX          = desc_x[M_];
      desc_x[N_]    = desc_a[M_];
      desc_x[MB_]   = 1;
      desc_x[NB_]   = desc_a[NB_];
      desc_x[CSRC_] = desc_a[CSRC_];
      desc_x[CTXT_] = ictx;
      desc_x[LLD_]  = INCX;
 
      modname = "PZGEQRF ";
      tol = 0.1;
      pzmtxcpy(3, 4, 1, 1, zga11, desc_a, zgas11);
      pzgeqrf(3, 4, zga11, 1, 1, desc_a, zgt11, workz, 0, &iret);
      if (iret != 0)
  {
    if(iam == 0)
    printf( " IRET from PZGEQRF = %4d\n", iret);
   *ifail = *ifail + 1;
  }
      else
  {
    pzmtchk(modname, 3, 4, 1, 1, zga11, desc_a, zresa11, tol,
                    ierr,&ilerr, &error, ifail);
    for (i=0; i< nprow; i++)
      {
         desc_x[RSRC_] = i;
         pzvtchk(modname,3, 1, 1, zgt11, desc_x, zrest11, 1, tol,
                      ierr,&ilerr, &error, ifail);
       }
   }
/* PDGEQRF  PZGEQRF END : 11  */
/* PDGELS   PZGELS  START : 12  */
 
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 4;
      desc_a[N_]    = 3;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
/* following desc_b   */
 
      desc_b[DT_]   = BLOCK_CYCLIC_2D;
      desc_b[M_]    = 4;
      desc_b[N_]    = 5;
      desc_b[MB_]   = 1;
      desc_b[NB_]   = 2;
      desc_b[RSRC_] = 0;
      desc_b[CSRC_] = 0;
      desc_b[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_b[LLD_]  = MAX(1,il_lda);
 
      modname = "PDGELS  ";
      tol = 0.1;
      pdmtxcpy(4, 3, 1, 1, dga12, desc_a, dgas12);
      pdmtxcpy(4, 5, 1, 1, dgb12, desc_b, dgbs12);
      pdgels("N",4,3,5,dga12,1,1,desc_a,dgb12,1,1,desc_b, work,0,
                 &iret);
      if (iret != 0)
  {
    if(iam == 0)
    printf( " IRET from PDGELS  = %4d\n", iret);
   *ifail = *ifail + 1;
  }
      else
  {
    pdmtchk(modname, 4, 5, 1, 1, dgb12, desc_b, dresb12, tol,
                    ierr,&ilerr, &error, ifail);
   }
 
      desc_a[DT_]   = BLOCK_CYCLIC_2D;
      desc_a[M_]    = 3;
      desc_a[N_]    = 4;
      desc_a[MB_]   = 1;
      desc_a[NB_]   = 1;
      desc_a[RSRC_] = 0;
      desc_a[CSRC_] = 0;
      desc_a[CTXT_] = ictx;
      il_lda = numroc(3,1,myrow,0,nprow);
      desc_a[LLD_]  = MAX(1,il_lda);
 
/* following desc_b   */
 
      desc_b[DT_]   = BLOCK_CYCLIC_2D;
      desc_b[M_]    = 4;
      desc_b[N_]    = 3;
      desc_b[MB_]   = 1;
      desc_b[NB_]   = 1;
      desc_b[RSRC_] = 0;
      desc_b[CSRC_] = 0;
      desc_b[CTXT_] = ictx;
      il_lda = numroc(4,1,myrow,0,nprow);
      desc_b[LLD_]  = MAX(1,il_lda);
 
      modname = "PZGELS  ";
      tol = 0.1;
      pzmtxcpy(3, 4, 1, 1, zga12, desc_a, zgas12);
      pzmtxcpy(4, 3, 1, 1, zgb12, desc_b, zgbs12);
      pzgels("N",3,4,3,zga12,1,1,desc_a,zgb12,1,1,desc_b,workz,0,&iret);
      if (iret != 0)
  {
    if(iam == 0)
    printf( " IRET from PZGELS  = %4d\n", iret);
   *ifail = *ifail + 1;
  }
      else
  {
    pzmtchk(modname, 4, 3, 1, 1, zgb12, desc_b, zresb12, tol,
                    ierr,&ilerr, &error, ifail);
   }
 
/* PDGELS  PZGELS  END : 12  */
  lwork = 0;
  tol = 0.00000001;
 
  nprows = nnodes;
  npcols = 1;
  blacs_get(0,0,&ictxs);
  blacs_gridinit(&ictxs,"R",nprows,npcols);
  blacs_gridinfo(ictxs,&nprows,&npcols,&myrows,&mycols);
 
/* following desc_x is used to distribute vector in 1xnpcol grid */
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 1;
  desc_x[N_] = 12;
  desc_x[MB_] = 1;
  desc_x[NB_] = (12+(nprow*npcol)-1)/(nprow*npcol);
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  desc_x[LLD_] = 1;
 
/* following desc_r is used to distribute RHS matrix in nprowsx1 grid*/
  desc_r[DT_] = BLOCK_CYCLIC_2D;
  desc_r[M_] = 12;
  desc_r[N_] = 3;
  desc_r[MB_] = (12+(nprow*npcol)-1)/(nprow*npcol);
  desc_r[NB_] = 3;
  desc_r[RSRC_] = 0;
  desc_r[CSRC_] = 0;
  desc_r[CTXT_] = ictxs;
  il_lda = numroc(12,desc_r[MB_],myrows,0,nprows);
   desc_r[LLD_] = MAX(1,il_lda);
 
/* following desc_a used by DGTTRF/S,DDTTRF/S,DGTSV/DDTSV routines */
  desc_a[DT_1D_] = BLOCK_COLUMN_1D;
  desc_a[N_1D_] = 12;
  desc_a[NB_1D_] = (12+(nprow*npcol)-1)/(nprow*npcol);
  desc_a[CSRC_1D_] = 0;
  desc_a[CTXT_1D_] = ictx;
  desc_a[LLD_1D_] = 0;
  desc_a[RSVD_1D_] = 0;
 
/* following desc_b is used by DGTTRS, DDTTRS,DGTSV/DDTSV routines */
  desc_b[DT_1D_] = BLOCK_ROW_1D;
  desc_b[M_1D_] = 12;
  desc_b[MB_1D_] = (12+(nprow*npcol)-1)/(nprow*npcol);
  desc_b[RSRC_1D_] = 0;
  desc_b[CTXT_1D_] = ictx;
  desc_b[LLD_1D_] = desc_r[LLD_];   /* for a check */
  desc_b[RSVD_1D_] = 0;
  laf = 12*nprows*npcols + 3*desc_a[NB_1D_];
   istor[0]=0;
   af = (double *) malloc(laf*sizeof(double));
   if (!af)
     istor[0]=1;
     igamx2d(ictx,"All", "1-Tree", 1, 1,
                    istor, 1, ra, ca, 1, 0, 0 );
     if(istor[0]==1)
     {
      if (iam==0)
      printf(" Memory allocation error for PDGTTRF\n");
      free(af);         /* free memory  */
      *ifail=*ifail+1;
      blacs_gridexit(ictx);
      blacs_gridexit(ictxs);
      return;
     }
 
  modname = "PDGTTRF ";
  pdvtrcpy(12, 1, 1, dtl8, desc_x, dtls8, 1);
  pdvtrcpy(12, 1, 1, dtd8, desc_x, dtds8, 1);
  pdvtrcpy(12, 1, 1, dtu8, desc_x, dtus8, 1);
  pdgttrf(12,dtl8,dtd8,dtu8,dt2u8,1,desc_a,ipvt,af,laf,
           work, lwork, &iret);
 
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PDGTTRF = %4d\n", iret);
   *ifail = *ifail + 1;
  }
  else
  {
 
  modname = "PDGTTRS ";
  pdmtxcpy(12, 3, 1, 1, dtb8, desc_r, dtbs8);
  pdgttrs("N",12,3,dtl8,dtd8,dtu8,dt2u8,1,desc_a, ipvt,
        dtb8,1,desc_b, af, laf, work,lwork,&iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PDGTTRS = %4d\n", iret);
   *ifail = *ifail + 1;
  }
 
  pdmtchk(modname, 12, 3, 1, 1, dtb8, desc_r, dresb8, tol,
                   ierr,&ilerr, &error, ifail);
  }
 
  modname = "PDGTSV  ";
  pdvtrcpy(12, 1, 1, dtl8, desc_x, dtls8, 1);
  pdvtrcpy(12, 1, 1, dtd8, desc_x, dtds8, 1);
  pdvtrcpy(12, 1, 1, dtu8, desc_x, dtus8, 1);
  pdmtxcpy(12, 3, 1, 1, dtb8, desc_r, dtbs8);
  pdgtsv(12,3,dtl8,dtd8,dtu8,1,desc_a,
       dtb8,1,desc_b, work,lwork,&iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PDGTSV = %4d\n", iret);
   *ifail = *ifail + 1;
  }
 
  pdmtchk(modname, 12, 3, 1, 1, dtb8, desc_r, dresb8, tol,
                   ierr,&ilerr, &error, ifail);
 
  modname = "PDDTTRF ";
  pdvtrcpy(12, 1, 1, dtl8, desc_x, dtls8, 1);
  pdvtrcpy(12, 1, 1, dtd8, desc_x, dtds8, 1);
  pdvtrcpy(12, 1, 1, dtu8, desc_x, dtus8, 1);
  pddttrf(12,dtl8,dtd8,dtu8,1,desc_a, af, laf,
         work,lwork, &iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PDDTTRF = %4d\n", iret);
   *ifail = *ifail + 1;
  }
  else
  {
 
  modname = "PDDTTRS ";
  pdmtxcpy(12, 3, 1, 1, dtb8, desc_r, dtbs8);
  pddttrs("N",12,3,dtl8,dtd8,dtu8,1,desc_a,
        dtb8,1,desc_b, af, laf, work, lwork, &iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PDDTTRS = %4d\n", iret);
   *ifail = *ifail + 1;
  }
 
  pdmtchk(modname, 12, 3, 1, 1, dtb8, desc_r, dresb8, tol,
                   ierr,&ilerr, &error, ifail);
  }
 
  modname = "PDDTSV  ";
  pdvtrcpy(12, 1, 1, dtl8, desc_x, dtls8, 1);
  pdvtrcpy(12, 1, 1, dtd8, desc_x, dtds8, 1);
  pdvtrcpy(12, 1, 1, dtu8, desc_x, dtus8, 1);
  pdmtxcpy(12, 3, 1, 1, dtb8, desc_r, dtbs8);
  pddtsv(12,3,dtl8,dtd8,dtu8,1,desc_a,
              dtb8,1,desc_b, work,lwork,&iret);
  if (iret != 0)
   {
   if(iam == 0)
    printf( " IRET from PDDTSV = %4d\n", iret);
   *ifail = *ifail + 1;
  }
 
  pdmtchk(modname, 12, 3, 1, 1, dtb8, desc_r, dresb8, tol,
                   ierr,&ilerr, &error, ifail);
 
/* PDPTTRF/S PDPTSV      */
 
  modname = "PDPTTRF ";
  pdvtrcpy(12, 1, 1, dtd10, desc_x, dtds10, 1);
  pdvtrcpy(12, 1, 1, dtu10, desc_x, dtus10, 1);
  pdpttrf(12,dtd10,dtu10,1,desc_a, af, laf,
         work,lwork, &info);
  if (info != 0)
   {
   if(iam == 0)
    printf(   " info from PDPTTRF = %4d\n", info);
   *ifail = *ifail + 1;
  }
  else
  {
 
  modname = "PDPTTRS ";
  pdmtxcpy(12, 3, 1, 1, dtb10, desc_r, dtbs10);
  pdpttrs(12,3,dtd10,dtu10,1,desc_a,
        dtb10,1,desc_b, af, laf, work, lwork, &info);
  if (info != 0)
   {
   if(iam == 0)
    printf(  " info from PDPTTRS = %4d\n",  info);
   *ifail = *ifail + 1;
  }
 
  pdmtchk(modname, 12, 3, 1, 1, dtb10, desc_r, dresb8, tol,
                   ierr,&ilerr, &error, ifail);
  }
 
  modname = "PDPTSV  ";
  pdvtrcpy(12, 1, 1, dtd10, desc_x, dtds10, 1);
  pdvtrcpy(12, 1, 1, dtu10, desc_x, dtus10, 1);
  pdmtxcpy(12, 3, 1, 1, dtb10, desc_r, dtbs10);
  pdptsv(12,3,dtd10,dtu10,1,desc_a,
              dtb10,1,desc_b, work,lwork,&info);
  if (info != 0)
   {
   if(iam == 0)
    printf(  " info from PDPTSV = %4d\n", info);
   *ifail = *ifail + 1;
  }
 
  pdmtchk(modname, 12, 3, 1, 1, dtb10, desc_r, dresb8, tol,
                   ierr,&ilerr, &error, ifail);
 
 
/* PDPBTRF/S  PDPBSV START : */
/* following desc_x is used to distribute matrix in 1xnpcol grid  */
      desc_x[DT_] = BLOCK_CYCLIC_2D;
      desc_x[M_] = 8;
      desc_x[N_] = 9;
      desc_x[MB_] = 8;
      desc_x[NB_] = (9+(nprow*npcol)-1)/(nprow*npcol);
      desc_x[RSRC_] = 0;
      desc_x[CSRC_] = 0;
      desc_x[CTXT_] = ictx;
      desc_x[LLD_] = 8;
 
/* following desc_r  used to distribute RHS matrix in nprows x 1 grid*/
      desc_r[DT_] = BLOCK_CYCLIC_2D;
      desc_r[M_] = 9;
      desc_r[N_] = 3;
      desc_r[MB_] = (9+(nprow*npcol)-1)/(nprow*npcol);
      desc_r[NB_] = 3;
      desc_r[RSRC_] = 0;
      desc_r[CSRC_] = 0;
      desc_r[CTXT_] = ictxs;
      il_lda = numroc(9,desc_r[MB_],myrows,0,nprows);
      desc_r[LLD_] = MAX(1,il_lda);
 
/* following desc_a is used by DPBTRF/S routines   */
      desc_a[DT_1D_] = BLOCK_COLUMN_1D;
      desc_a[N_1D_] = 9;
      desc_a[NB_1D_] = (9+(nprow*npcol)-1)/(nprow*npcol);
      desc_a[CSRC_1D_] = 0;
      desc_a[CTXT_1D_] = ictx;
      desc_a[LLD_1D_] = 8;
      desc_a[RSVD_1D_] = 0;
 
/* following desc_b is used by DGTTRS routine     */
      desc_b[DT_1D_] = BLOCK_ROW_1D;
      desc_b[M_1D_] = 9;
      desc_b[MB_1D_] = (9+(nprow*npcol)-1)/(nprow*npcol);
      desc_b[RSRC_1D_] = 0;
      desc_b[CTXT_1D_] = ictx;
      desc_b[LLD_1D_] = desc_r[LLD_];
      desc_b[RSVD_1D_] = 0;
 
      k = 7;
 
      modname = "PDPBTRF ";
      pdmtxcpy(8, 9, 1, 1, dta9, desc_x, dtas9);
      pdpbtrf("L",9,7,dta9,1,desc_a, af, laf, work,lwork,&iret);
 
      if (iret != 0)
   {
      if(iam == 0)
       printf( " IRET from PDPBTRF = %4d\n", iret);
      *ifail = *ifail + 1;
   }
  else
  {
 
      modname = "PDPBTRS ";
      pdmtxcpy(9, 3, 1, 1, dtb9, desc_r, dtbs9);
      pdpbtrs("L", 9, 7, 3, dta9, 1, desc_a, dtb9,1,desc_b, af,
               laf, work,lwork,&iret);
 
      if (iret != 0)
   {
      if(iam == 0)
       printf( " IRET from PDPBTRS = %4d\n", iret);
      *ifail = *ifail + 1;
   }
 
      pdmtchk(modname, 9, 3, 1, 1, dtb9, desc_r, dresb9, tol,
              ierr,&ilerr, &error, ifail);
   }
 
      modname = "PDPBSV  ";
      pdmtxcpy(8, 9, 1, 1, dta9, desc_x, dtas9);
      pdmtxcpy(9, 3, 1, 1, dtb9, desc_r, dtbs9);
      pdpbsv("L",9,7,3,dta9,1,desc_a, dtb9,1,desc_b, work, lwork,&iret);
 
      if (iret != 0)
   {
      if(iam == 0)
       printf( " IRET from PDPBSV  = %4d\n", iret);
      *ifail = *ifail + 1;
   }
 
      pdmtchk(modname, 9, 3, 1, 1, dtb9, desc_r, dresb9, tol,
              ierr,&ilerr, &error, ifail);
 
 
/* PDPBTRF/S END   */
 
  blacs_gridexit(ictxs);
  blacs_gridexit(ictx);
}
 
 
/*********************************************************************
*          PARALLEL SPARSE LINEAR ALGEBRAIC EQUATIONS SUBROUTINES    *
*********************************************************************/
 
void psprs(int *ifail)
{
/**********************************************************************
!
! This program illustrates how to use the Parallel ESSL F77 Sparse Iterative
! Solver and its supporting utility subroutines.  A very simple problem
! (DSRIS Example 1 from the ESSL Guide and Reference) using an
! HPF BLOCK data distribution is solved.
!
***********************************************************************/
    int  izero=0, ione=1, dupflag=2, n=9, nnz=22;
    int  ierr=5, ilerr, istor[1], ra[1], ca[1];
    double error=0.0, tol=0.0001;
    int desc_r[DLEN_],il_lda;
    char modn[9], *modname=modn;
    char stor[]="DEF  ";
    char mtype[]="GEN  ";
    char order[]="R";
    int *desc_a;
 
/**********************************************************************
! Sparse Matrices and related information
!  What size should these be? I dont find that in the prolog
***********************************************************************/
    double            as[22];
    int               ia1[50],ia2[50];
    int               infoa[30];
 
    double            bs[22];
    int               ib1[10],ib2[22];
    int               infob[30];
 
/**********************************************************************
! Preconditioner Data Structure
***********************************************************************/
    double                  prcs[110];
    int                     lprcs;
 
/**********************************************************************
! Dense Vectors
***********************************************************************/
    double                  b[9],x[9];
 
/**********************************************************************
! BLACS parameters
***********************************************************************/
    int            ictxt,nprow,npcol,myrow,mycol,nnodes,iam,ercode=-1;
 
/**********************************************************************
! Solver parameters
***********************************************************************/
      int                   iter,itmax,info,itrace;
      int                   iprec,methd,istopc;
      double                err,eps;
 
/**********************************************************************
! We will not have duplicates so PV used by the PARTS subroutine
! PART_BLOCK only needs to be of length 1.
***********************************************************************/
 
      int                   pv[1];
 
/* Other variables  */
      int                   nb, ldb, ldbg;
      int                   nx, ldx, ldxg;
      int                   nrhs ;
      int                   i,j;
      int                   global_indx, nv_count;
      int                   global_indx_owner, nv;
      int                   local_indx;
 
/**********************************************************************
!
!     Global Problem
!      DSRIS Example 1 from the ESSL Guide and Reference
!
***********************************************************************/
      double a_global[22] = {2.0,2.0,-1.0,1.0,2.0,1.0,2.0,-1.0,
                         1.0,2.0,-1.0,1.0,2.0,-1.0,1.0,2.0,
                        -1.0,1.0,2.0,-1.0,1.0,2.0};
/**********************************************************************
!
!     ia and ja arrays used by Fortran 77 subrouitines
!      that why their elements are 1 origin
!
***********************************************************************/
      int  ja[22] = {1,2,3,2,3,1,4,5,4,5,6,5,6,7,6,7,8,
                         7,8,9,8,9};
      int  ia[10] = {1,2,4,6,9,12,15,18,21,23};
 
      int  iparm[20] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
 
      double  rparm[20] = {0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,
                           0.,0.,0.,0.,0.,0.,0.,0.,0.};
 
      double b_global[9] = {2.0,1.0,3.0,2.0,2.0,2.0,2.0,2.0,
                           3.0};
      double xinit_global[9] = {0.0,0.0,0.0,0.0,0.0,0.0,0.0,
                               0.0, 0.0};
      double xres_global[9] = {1.0,1.0,1.0,1.0,1.0,1.0,1.0,
                               1.0, 1.0};
 
/**********************************************************************
! Initialize BLACS
! Define a P x 1 Process Grid
***********************************************************************/
 
    blacs_pinfo(&iam,&nnodes);
    nprow = nnodes;
    npcol = 1;
    blacs_get(0,0,&ictxt);
    blacs_gridinit(&ictxt,"R",nprow,npcol);
    blacs_gridinfo(ictxt,&nprow,&npcol,&myrow,&mycol);
 
/**********************************************************************
!
! Initialize some elements of the sparse matrix A
! and its descriptor vector, DESC_A
!
***********************************************************************/
      istor[0] = 0;
      desc_a = (int *) malloc((33+3*nnodes+4*n)*sizeof(int));
 
   if (!desc_a)
     istor[0]=1;
     igamx2d(ictxt,"All", "1-Tree", 1, 1,
                    istor, 1, ra, ca, 1, 0, 0 );
     if(istor[0]==1)
     {
      if (iam == 0)
      printf(" Memory allocation error for SPARSE\n");
      *ifail=*ifail+1;
      free(desc_a);
      blacs_gridexit(ictxt);
      return;
     }
      desc_a[10] = 33+3*nnodes+4*n;
      padinit(n,part_block,desc_a,ictxt);
 
      infoa[0] = sizeof(as)/sizeof(double);
      infoa[1] = sizeof(ia1)/sizeof(int);
      infoa[2] = sizeof(ia2)/sizeof(int);
      pdspinit(as,ia1,ia2,infoa,desc_a);
 
/**********************************************************************
!
! In this simple example, all processes have a copy of
! the global sparse matrix, A
!
! Only the owning process will insert the row into the
! local part of the sparse matrix
!
! global_indx used by c and Fortran 77 subroutines
!  that why global_indx is 1 origin
!
***********************************************************************/
      nb = 1;
      ldb = sizeof(b)/sizeof(double);
      ldbg = sizeof(b_global)/sizeof(double);
      nx = 1;
      ldx = sizeof(x)/sizeof(double);
      ldxg = sizeof(xinit_global)/sizeof(double);
 
      for(global_indx=1;global_indx<=n;++global_indx)
       {
        part_block(&global_indx,&n,&nnodes,pv,&nv);
/**********************************************************************
!
! In this simple example, NV will always be 1
! since there will not be duplicate coefficents
!
***********************************************************************/
        for(nv_count=0;nv_count<nv;++nv_count)
         {
          global_indx_owner = pv[nv_count];
          if (global_indx_owner == myrow) {
           ib2[0] = 1;
           ib2[1] = 1;
           for(j=ia[global_indx-1]-1;j<ia[global_indx]-1;j++)
            {
             bs[ib2[1]-1] = a_global[j];
             ib1[ib2[1]-1]= ja[j];
             ib2[1] = ib2[1] + 1;
            }
           infob[0] = ib2[1] - 1;
           infob[1] = ib2[1] - 1;
           infob[2] = 2;
           infob[3] = 1;
           infob[4] = 1;
           infob[5] = 1;
           infob[6] = n;
           pdspins(as,ia1,ia2,infoa,desc_a,global_indx, 1,
                      bs,ib1,ib2,infob);
           pdgeins(nb,b,ldb,global_indx,1,1,1,
                   &b_global[global_indx-1],ldbg,desc_a);
           pdgeins(nx,x,ldx,global_indx,1,1,1,
                   &xinit_global[global_indx-1],ldxg,desc_a);
          }
         }
       }
 
/*! Assemble A and DESC_A       */
      pdspasb(as,ia1,ia2,infoa,desc_a,
              mtype,stor,dupflag,&info);
 
      if (info != 0) {
       if (iam == 0) {
        printf ("Error in assembly: %4d\n",info);
        free(desc_a);
        *ifail = *ifail + 1;
        blacs_gridexit(ictxt);
        return;
       }
      }
 
/**********************************************************************
!
! In this simple example, all processes have a copy of
! the global rhs vector, B, and the global initial guess vector, X
!
! Only the owning process will insert the element into the
! local part of the vector(s)
!
***********************************************************************/
 
 
/*! Assemble B and X     */
 
      pdgeasb(nb,b,ldb,desc_a);
      pdgeasb(nx,x,ldx,desc_a);
 
/**********************************************************************
!
!  Preconditioning
!
!  We are using ILU for the preconditioner
!
***********************************************************************/
      iprec = 2;
      lprcs = sizeof(prcs)/sizeof(double);
 
      pdspgpr(iprec,as,ia1,ia2,infoa,prcs,
              lprcs,desc_a,&info);
 
      if (info != 0) {
       if (iam == 0) {
        printf ("Error in preconditioner: %4d\n",info);
        free(desc_a);
        *ifail = *ifail + 1;
        blacs_gridexit(ictxt);
        return;
       }
      }
 
/**********************************************************************
!
!  Iterative Solver - use the BICGSTAB method
!
***********************************************************************/
      nrhs  = 1;
      itmax = 1000;
      eps   = 1.e-8;
      methd = 1;
      istopc = 1;
      itrace = 0;
      iparm[0] = methd;
      iparm[1] = istopc;
      iparm[2] = itmax;
      iparm[3] = itrace;
      rparm[0] = eps;
 
 
      pdspgis(as,ia1,ia2,infoa,nrhs,b,ldb,x,ldx,prcs,desc_a,
                   iparm,rparm,&info);
 
 
      if (info != 0) {
       if (iam == 0) {
        printf ("Error in solver: %4d\n",info);
        free(desc_a);
        *ifail = *ifail + 1;
        blacs_gridexit(ictxt);
        return;
       }
      }
 
/*    err = rparm[1];
      iter = iparm[4];
      if (iam == 0) {
       printf ("Number of iteration: %d\n",iter);
       printf ("Error on exit: %g\n",err);
      }                                 */
 
/* following desc_r  RHS vector in nprowsx1 grid*/
  desc_r[DT_] = BLOCK_CYCLIC_2D;
  desc_r[M_] = 9;
  desc_r[N_] = 1;
  desc_r[MB_] = (9+(nprow*npcol)-1)/(nprow*npcol);
  desc_r[NB_] = 1;
  desc_r[RSRC_] = 0;
  desc_r[CSRC_] = 0;
  desc_r[CTXT_] = ictxt;
  il_lda = numroc(9,desc_r[MB_],myrow,0,nprow);
   desc_r[LLD_] = MAX(1,il_lda);
 
  modname = "PDSPGIS ";
  pdmtchk(modname, 9, 1, 1, 1, x, desc_r, xres_global, tol,
                   ierr,&ilerr, &error, ifail);
 
/**********************************************************************
!
!  Terminate the process grid and the BLACS
!
***********************************************************************/
       free(desc_a);
       blacs_gridexit(ictxt);
}
 
void part_block(global_indx,n,nnodes,pv,nv)
int  *global_indx,*n,*nnodes,*pv,*nv;
{
  int dim_block;
     dim_block = (*n + *nnodes -1)/(*nnodes);
     *nv = 1;
     pv[*nv-1] = (*global_indx - 1)/dim_block;
}
 
 
/*********************************************************************
*          PARALLEL TWO DIMENSIONAL SHORT PRECISION FFT SUBROUTINES  *
*********************************************************************/
 
void psfft2(int *ifail)
{
/* PSCFT2    START : 1   */
 
   int i, iam, nnodes,myrow,mycol,ip[40];
   int desc_x[DLEN_], desc_y[DLEN_], il_lda;
   int ictx, nprow, npcol,ierr=5,ilerr;
   char modn[9], *modname=modn;
   double error=0.0, tol=0.0;
   cmplx cx1[48], cy1[48],cxs1[48],cresy1[48];
   cmplx cx3[16],cxs3[16];
   float   sy3[48], sys3[48], sresy3[48];
 
   ip[0] = 1;
   ip[1] = 1;
   ip[19] = 0;
   ip[20] = 0;
 
   CSET(cxs1[0],48.0,0.0);
   for (i=1; i<48; i++)
   CSET(cxs1[i],0.0,0.0);
 
   for (i=0; i<48; i++)
   CSET(cresy1[i],1.0,0.0);
 
   for (i=1; i<16; i++)
   CSET(cxs3[i],1.0,0.0);
   CSET(cxs3[0],1.0,1.0);
   CSET(cxs3[2],1.0,1.0);
 
    sresy3[0] = 1.0;
   for (i=1; i<48; i++)
    sresy3[i] = 0.0;
 
  blacs_pinfo(&iam,&nnodes);
  if (nnodes > 1)
   {
     nprow = 1;
     npcol = 2;
   }
  else
   {
     nprow = 1;
     npcol = 1;
   }
  blacs_get(0,0,&ictx);
  blacs_gridinit(&ictx,"R",nprow,npcol);
  blacs_gridinfo(ictx,&nprow,&npcol,&myrow,&mycol);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 8;
  desc_x[N_] = 6;
  desc_x[MB_] = 8;
  desc_x[NB_] = 6/npcol;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 8;
  desc_y[N_] = 6;
  desc_y[MB_] = 8;
  desc_y[NB_] = 6/npcol;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
 
  modname = "PSCFT2  ";
  if (iam < 2 && nnodes > 1 )
  {
    il_lda = numroc(8,8,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(8,8,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    pcmtxcpy(8, 6, 1, 1, cx1,  desc_x, cxs1);
    pscft2(cx1, cy1, 8, 6, -1, 1.0/48.0, ictx, ip);
    pcmtchk(modname,8, 6, 1, 1, cy1, desc_y, cresy1, tol, ierr, &ilerr,
           &error, ifail);
  }
  else if (iam == 0)
  {
    il_lda = numroc(8,8,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(8,8,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    pcmtxcpy(8, 6, 1, 1, cx1,  desc_x, cxs1);
    pscft2(cx1, cy1, 8, 6, -1, 1.0/48.0, ictx, ip);
    pcmtchk(modname,8, 6, 1, 1, cy1, desc_y, cresy1, tol, ierr, &ilerr,
           &error, ifail);
  }
 
   ip[0] = 1;
   ip[19] = 4;
   ip[20] = 12;
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 4;
  desc_x[N_] = 4;
  desc_x[MB_] = 4;
  desc_x[NB_] = 8/(2*npcol);
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  desc_x[LLD_] = 4;
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 8;
  desc_y[N_] = 4;
  desc_y[MB_] = 8;
  desc_y[NB_] = 4/npcol;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  desc_y[LLD_] = 12;
  if (iam < 2 && nnodes > 1 )
  {
    modname = "PSCRFT2 ";
    pcmtxcpy(4, 4, 1, 1, cx3,  desc_x, cxs3);
    pscrft2(cx3, sy3, 8, 4, -1, 1.0/32.0, ictx, ip);
    psmtchk(modname,8, 4, 1, 1, sy3, desc_y, sresy3, tol, ierr, &ilerr,
           &error, ifail);
    ip[19] = 12;
    ip[20] = 4;
    modname = "PSRCFT2 ";
    psrcft2(sy3, cx3, 8, 4, 1, 1.0, ictx, ip);
    pcmtchk(modname,4, 4, 1, 1, cx3, desc_x, cxs3, tol, ierr, &ilerr,
           &error, ifail);
    blacs_gridexit(ictx);
  }
  else if (iam == 0)
  {
    modname = "PSCRFT2 ";
    pcmtxcpy(4, 4, 1, 1, cx3,  desc_x, cxs3);
    pscrft2(cx3, sy3, 8, 4, -1, 1.0/32.0, ictx, ip);
    psmtchk(modname,12, 4, 1, 1, sy3, desc_y, sresy3, tol, ierr, &ilerr,
           &error, ifail);
    ip[19] = 12;
    ip[20] = 4;
    modname = "PSRCFT2 ";
    psrcft2(sy3, cx3, 8, 4, 1, 1.0, ictx, ip);
    pcmtchk(modname,4, 4, 1, 1, cx3, desc_x, cxs3, tol, ierr, &ilerr,
           &error, ifail);
    blacs_gridexit(ictx);
  }
 
}
 
/*********************************************************************
*          PARALLEL TWO DIMENSIONAL LONG PRECISION FFT SUBROUTINES   *
*********************************************************************/
 
 
void pfft2(int *ifail)
{
/* PDCFT2    START : 1,      */
 
   int i, iam, nnodes,myrow,mycol,ip[40];
   int desc_x[DLEN_], desc_y[DLEN_], il_lda;
   int ictx, nprow, npcol,ierr=5,ilerr;
   double error=0.0, tol=0.0;
   dcmplx zx1[48], zy1[48],zxs1[48],zresy1[48];
   dcmplx zx3[16];
   double  dy3[48], dys3[48], dresy3[48];
   dcmplx zxs3[16] = {
     {1.0,1.0}, {1.0,0.0}, {1.0,1.0}, {1.0,0.0},
     {1.0,0.0}, {1.0,0.0}, {1.0,0.0}, {1.0,0.0},
     {1.0,0.0}, {1.0,0.0}, {1.0,0.0}, {1.0,0.0},
     {1.0,0.0}, {1.0,0.0}, {1.0,0.0}, {1.0,0.0} };
   char modn[9], *modname=modn;
 
   ip[0] = 1;
   ip[1] = 1;
   ip[19] = 0;
   ip[20] = 0;
 
   CSET(zxs1[0],48.0,0.0);
   for (i=1; i<48; i++)
   CSET(zxs1[i],0.0,0.0);
 
   for (i=0; i<48; i++)
   CSET(zresy1[i],1.0,0.0);
 
    dresy3[0] = 1.0;
   for (i=1; i<48; i++)
    dresy3[i] = 0.0;
 
  blacs_pinfo(&iam,&nnodes);
  if (nnodes > 1)
   {
     nprow = 1;
     npcol = 2;
   }
  else
   {
     nprow = 1;
     npcol = 1;
   }
  blacs_get(0,0,&ictx);
  blacs_gridinit(&ictx,"R",nprow,npcol);
  blacs_gridinfo(ictx,&nprow,&npcol,&myrow,&mycol);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 8;
  desc_x[N_] = 6;
  desc_x[MB_] = 8;
  desc_x[NB_] = 6/npcol;
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(8,8,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 8;
  desc_y[N_] = 6;
  desc_y[MB_] = 8;
  desc_y[NB_] = 6/npcol;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(8,8,myrow,0,nprow);
  desc_y[LLD_] = MAX(1,il_lda);
 
  modname = "PDCFT2  ";
  if (iam < 2 && nnodes > 1 )
  {
    pzmtxcpy(8, 6, 1, 1, zx1,  desc_x, zxs1);
    pdcft2(zx1, zy1, 8, 6, -1, 1.0/48.0, ictx, ip);
    pzmtchk(modname,8, 6, 1, 1, zy1, desc_y, zresy1, tol, ierr,&ilerr,
           &error, ifail);
  }
  else if (iam == 0)
  {
    pzmtxcpy(8, 6, 1, 1, zx1,  desc_x, zxs1);
    pdcft2(zx1, zy1, 8, 6, -1, 1.0/48.0, ictx, ip);
    pzmtchk(modname,8, 6, 1, 1, zy1, desc_y, zresy1, tol, ierr,&ilerr,
           &error, ifail);
  }
 
   ip[0] = 1;
   ip[19] = 4;
   ip[20] = 12;
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 4;
  desc_x[N_] = 4;
  desc_x[MB_] = 4;
  desc_x[NB_] = 8/(2*npcol);
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  desc_x[LLD_] = 4;
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 8;
  desc_y[N_] = 4;
  desc_y[MB_] = 8;
  desc_y[NB_] = 4/npcol;
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  desc_y[LLD_] = 12;
  if (iam < 2 && nnodes > 1 )
  {
    modname = "PDCRFT2 ";
    pzmtxcpy(4, 4, 1, 1, zx3,  desc_x, zxs3);
    pdcrft2(zx3, dy3, 8, 4, -1, 1.0/32.0, ictx, ip);
    pdmtchk(modname,8, 4, 1, 1, dy3, desc_y, dresy3, tol, ierr,&ilerr,
           &error, ifail);
    ip[19] = 12;
    ip[20] = 4;
    modname = "PDRCFT2 ";
    pdrcft2(dy3, zx3, 8, 4, 1, 1.0, ictx, ip);
    pzmtchk(modname,4, 4, 1, 1, zx3, desc_x, zxs3, tol, ierr,&ilerr,
           &error, ifail);
    blacs_gridexit(ictx);
  }
  else if (iam == 0)
  {
    modname = "PDCRFT2 ";
    pzmtxcpy(4, 4, 1, 1, zx3,  desc_x, zxs3);
    pdcrft2(zx3, dy3, 8, 4, -1, 1.0/32.0, ictx, ip);
    pdmtchk(modname,12, 4, 1, 1, dy3, desc_y, dresy3, tol, ierr,&ilerr,
           &error, ifail);
    ip[19] = 12;
    ip[20] = 4;
    modname = "PDRCFT2 ";
    pdrcft2(dy3, zx3, 8, 4, 1, 1.0, ictx, ip);
    pzmtchk(modname,4, 4, 1, 1, zx3, desc_x, zxs3, tol, ierr,&ilerr,
           &error, ifail);
    blacs_gridexit(ictx);
  }
 
}
 
/*********************************************************************
*          PARALLEL THREE DIMENSIONAL SHORT PRECISION FFT SUBROUTINES*
*********************************************************************/
 
void psfft3(int *ifail)
{
/* PSCFT3    START : 2,    */
 
   int i, iam, nnodes,myrow,mycol,ip[40];
   int desc_x[DLEN_], desc_y[DLEN_], il_lda;
   int ictx, nprow, npcol,ierr=5,ilerr;
   double error=0.0, tol=0.0;
   char modn[9], *modname=modn;
   cmplx cx2[32], cy2[32],cxs2[32],cresy2[32];
   cmplx cx4[64], cxs4[64];
   float  sy4[128], sresy4[128];
 
   ip[0] = 1;
   ip[1] = 1;
   ip[19] = 0;
   ip[20] = 0;
   ip[21] = 0;
   ip[22] = 0;
 
   CSET(cxs2[0],1.0,0.0);
   for (i=1; i<32; i++)
    CSET(cxs2[i],0.0,0.0);
 
   for (i=0; i<32; i++)
    CSET(cresy2[i],1.0,0.0);
 
    sresy4[0] = 1.0;
   for (i=1; i<128; i++)
    sresy4[i] = 0.0;
 
   for (i=0; i<64; i++)
    CSET(cxs4[i],1.0,0.0);
 
    CSET(cxs4[0],1.0,1.0);
    CSET(cxs4[2],1.0,1.0);
    CSET(cxs4[8],1.0,1.0);
    CSET(cxs4[10],1.0,1.0);
 
  blacs_pinfo(&iam,&nnodes);
  if (nnodes > 1)
   {
     nprow = 1;
     npcol = 2;
   }
  else
   {
     nprow = 1;
     npcol = 1;
   }
  blacs_get(0,0,&ictx);
  blacs_gridinit(&ictx,"R",nprow,npcol);
  blacs_gridinfo(ictx,&nprow,&npcol,&myrow,&mycol);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 4*4;     /* n1*n2    */
  desc_x[N_] = 2;       /* n3       */
  desc_x[MB_] = 4*4;     /* n1*n2    */
  desc_x[NB_] = 2/npcol; /* n3/npcol */
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 4*4;     /* n1*n2    */
  desc_y[N_] = 2;       /* n3       */
  desc_y[MB_] = 4*4;     /* n1*n2    */
  desc_y[NB_] = 2/npcol; /* n3/npcol */
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
 
  modname = "PSCFT3  ";
  if (iam < 2 && nnodes > 1 )
  {
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    pcmtxcpy(16, 2, 1, 1, cx2,  desc_x, cxs2);
    pscft3(cx2, cy2, 4, 4, 2, 1, 1.0, ictx, ip);
    pcmtchk(modname,16, 2, 1, 1, cy2, desc_y, cresy2, tol, ierr, &ilerr,
           &error, ifail);
  }
  else if (iam == 0)
  {
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    pcmtxcpy(16, 2, 1, 1, cx2,  desc_x, cxs2);
    pscft3(cx2, cy2, 4, 4, 2, 1, 1.0, ictx, ip);
    pcmtchk(modname,16, 2, 1, 1, cy2, desc_y, cresy2, tol, ierr, &ilerr,
           &error, ifail);
  }
 
   ip[0] = 1;
   ip[19] = 4;
   ip[20] = 4;
   ip[21] = 8;
   ip[22] = 4;
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 4*4;     /* n3*n2        */
  desc_x[N_] = 4;       /* n1/2         */
  desc_x[MB_] = 4*4;     /* n3*n2        */
  desc_x[NB_] = 4/npcol; /* n1/(2*npcol) */
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 8*4;     /* n1*n2        */
  desc_y[N_] = 4;       /* n3           */
  desc_y[MB_] = 8*4;     /* n1*n2        */
  desc_y[NB_] = 4/npcol; /* n3/npcol     */
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
 
  if (iam < 2 && nnodes > 1 )
  {
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(32,32,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    modname = "PSCRFT3 ";
    pcmtxcpy(16, 4, 1, 1, cx4, desc_x, cxs4);
    pscrft3(cx4, sy4, 8, 4, 4, -1, 1.0/(8*4*4), ictx, ip);
    psmtchk(modname,32, 4, 1, 1, sy4, desc_y, sresy4, tol, ierr,&ilerr,
           &error, ifail);
    ip[19] = 8;
    ip[21] = 4;
    modname = "PSRCFT3 ";
    psrcft3(sy4, cx4, 8, 4, 4, 1, 1.0, ictx, ip);
    pcmtchk(modname,16, 4, 1, 1, cx4, desc_x, cxs4, tol, ierr,&ilerr,
           &error, ifail);
    blacs_gridexit(ictx);
  }
  else if (iam == 0)
  {
    il_lda = numroc(16,16,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
    il_lda = numroc(32,32,myrow,0,nprow);
    desc_y[LLD_] = MAX(1,il_lda);
    modname = "PSCRFT3 ";
    pcmtxcpy(16, 4, 1, 1, cx4, desc_x, cxs4);
    pscrft3(cx4, sy4, 8, 4, 4, -1, 1.0/(8*4*4), ictx, ip);
    psmtchk(modname,32, 4, 1, 1, sy4, desc_y, sresy4, tol, ierr,&ilerr,
           &error, ifail);
    ip[19] = 8;
    ip[21] = 4;
    modname = "PSRCFT3 ";
    psrcft3(sy4, cx4, 8, 4, 4, 1, 1.0, ictx, ip);
    pcmtchk(modname,16, 4, 1, 1, cx4, desc_x, cxs4, tol, ierr,&ilerr,
           &error, ifail);
    blacs_gridexit(ictx);
  }
}
 
/*********************************************************************
*          PARALLEL THREE DIMENSIONAL LONG PRECISION FFT SUBROUTINES *
*********************************************************************/
 
void pfft3(int *ifail)
{
/* PDCFT3    START : 2,    */
 
   int i, iam, nnodes,myrow,mycol,ip[40];
   int desc_x[DLEN_], desc_y[DLEN_], il_lda;
   int ictx, nprow, npcol,ierr=5,ilerr;
   double error=0.0, tol=0.0;
   dcmplx zx2[32], zy2[32],zxs2[32],zresy2[32];
   dcmplx zx4[64], zxs4[64];
   double dy4[128], dresy4[128];
   char modn[9], *modname=modn;
 
   ip[0] = 1;
   ip[1] = 1;
   ip[19] = 0;
   ip[20] = 0;
   ip[21] = 0;
   ip[22] = 0;
 
   CSET(zxs2[0],1.0,0.0);
   for (i=1; i<32; i++)
   CSET(zxs2[i],0.0,0.0);
 
   for (i=0; i<32; i++)
   CSET(zresy2[i],1.0,0.0);
 
    dresy4[0] = 1.0;
   for (i=1; i<128; i++)
    dresy4[i] = 0.0;
 
   for (i=0; i<64; i++)
   CSET(zxs4[i],1.0,0.0);
 
   CSET(zxs4[0],1.0,1.0);
   CSET(zxs4[2],1.0,1.0);
   CSET(zxs4[8],1.0,1.0);
   CSET(zxs4[10],1.0,1.0);
 
  blacs_pinfo(&iam,&nnodes);
  if (nnodes > 1)
   {
     nprow = 1;
     npcol = 2;
   }
  else
   {
     nprow = 1;
     npcol = 1;
   }
  blacs_get(0,0,&ictx);
  blacs_gridinit(&ictx,"R",nprow,npcol);
  blacs_gridinfo(ictx,&nprow,&npcol,&myrow,&mycol);
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 4*4;     /* n1*n2  */
  desc_x[N_] = 2;       /* n3     */
  desc_x[MB_] = 4*4;     /* n1*n2  */
  desc_x[NB_] = 2/npcol; /* n3/npcol */
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(16,16,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 4*4;     /* n1*n2  */
  desc_y[N_] = 2;       /* n3     */
  desc_y[MB_] = 4*4;     /* n1*n2  */
  desc_y[NB_] = 2/npcol; /* n3/npcol */
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(16,16,myrow,0,nprow);
  desc_y[LLD_] = MAX(1,il_lda);
 
  modname = "PDCFT3  ";
  if (iam < 2 && nnodes > 1 )
  {
    pzmtxcpy(16, 2, 1, 1, zx2,  desc_x, zxs2);
    pdcft3(zx2, zy2, 4, 4, 2, 1, 1.0, ictx, ip);
    pzmtchk(modname,16, 2, 1, 1, zy2, desc_y, zresy2, tol, ierr,&ilerr,
           &error, ifail);
  }
  else if (iam == 0)
  {
    pzmtxcpy(16, 2, 1, 1, zx2,  desc_x, zxs2);
    pdcft3(zx2, zy2, 4, 4, 2, 1, 1.0, ictx, ip);
    pzmtchk(modname,16, 2, 1, 1, zy2, desc_y, zresy2, tol, ierr,&ilerr,
           &error, ifail);
  }
 
   ip[0] = 1;
   ip[19] = 4;
   ip[20] = 4;
   ip[21] = 8;
   ip[22] = 4;
 
  desc_x[DT_] = BLOCK_CYCLIC_2D;
  desc_x[M_] = 4*4;     /* n3*n2    */
  desc_x[N_] = 4;       /* n1/2     */
  desc_x[MB_] = 4*4;     /* n3*n2    */
  desc_x[NB_] = 4/npcol; /* n1/(2*npcol)   */
  desc_x[RSRC_] = 0;
  desc_x[CSRC_] = 0;
  desc_x[CTXT_] = ictx;
  il_lda = numroc(16,16,myrow,0,nprow);
  desc_x[LLD_] = MAX(1,il_lda);
 
  desc_y[DT_] = BLOCK_CYCLIC_2D;
  desc_y[M_] = 8*4;     /* n1*n2      */
  desc_y[N_] = 4;       /* n3         */
  desc_y[MB_] = 8*4;     /* n1*n2      */
  desc_y[NB_] = 4/npcol; /* n3/npcol   */
  desc_y[RSRC_] = 0;
  desc_y[CSRC_] = 0;
  desc_y[CTXT_] = ictx;
  il_lda = numroc(32,32,myrow,0,nprow);
  desc_y[LLD_] = MAX(1,il_lda);
 
  if (iam < 2 && nnodes > 1 )
  {
    modname = "PDCRFT3 ";
    pzmtxcpy(16, 4, 1, 1, zx4, desc_x, zxs4);
    pdcrft3(zx4, dy4, 8, 4, 4, -1, 1.0/(8*4*4), ictx, ip);
    pdmtchk(modname,32, 4, 1, 1, dy4, desc_y, dresy4, tol, ierr,&ilerr,
           &error, ifail);
    ip[19] = 8;
    ip[21] = 4;
    modname = "PDRCFT3 ";
    pdrcft3(dy4, zx4, 8, 4, 4, 1, 1.0, ictx, ip);
    pzmtchk(modname,16, 4, 1, 1, zx4, desc_x, zxs4, tol, ierr,&ilerr,
           &error, ifail);
    blacs_gridexit(ictx);
  }
  else if (iam == 0)
  {
    modname = "PDCRFT3 ";
    pzmtxcpy(16, 4, 1, 1, zx4, desc_x, zxs4);
    pdcrft3(zx4, dy4, 8, 4, 4, -1, 1.0/(8*4*4), ictx, ip);
    pdmtchk(modname,32, 4, 1, 1, dy4, desc_y, dresy4, tol, ierr,&ilerr,
           &error, ifail);
    ip[19] = 8;
    ip[21] = 4;
    modname = "PDRCFT3 ";
    pdrcft3(dy4, zx4, 8, 4, 4, 1, 1.0, ictx, ip);
    pzmtchk(modname,16, 4, 1, 1, zx4, desc_x, zxs4, tol, ierr,&ilerr,
           &error, ifail);
    blacs_gridexit(ictx);
  }
}
 
/*********************************************************************
*          PARALLEL RANDOM NUMBER GENERATION SUBROUTINE              *
*********************************************************************/
 
void prnd(int *ifail)
{
/* PDURNG    START :     */
 
   int iam, nnodes,myrow,mycol;
   int desc_x[DLEN_], il_lda;
   int ictx, nprow, npcol,ierr=5, ilerr;
   double dx1r[30], seed, error=0.0, tol=0.0;
   double dresx1r[30]= {
     0.6838215161, 0.0588744078, 0.3918552508, 0.7559946530, 0.5577643014,
     0.0013338017, 0.0568559327, 0.3310630362, 0.3473397944, 0.6494290203,
     0.3861448762, 0.4572248550, 0.8925181341, 0.0745487482, 0.9123793668,
     0.1128094991, 0.8575476050, 0.7564809018, 0.0469933644, 0.8894576840,
     0.1677757661, 0.5049527226, 0.9997259245, 0.6962694873, 0.6718965980,
     0.2714721560, 0.5664184066, 0.4646848657, 0.9824425397, 0.0224404825 };
 
   char modn[9], *modname=modn;
  blacs_pinfo(&iam,&nnodes);
  if (nnodes > 4)
   {
     nprow = 1;
     npcol = 5;
   }
  else
   {
     nprow = 1;
     npcol = 1;
   }
  blacs_get(0,0,&ictx);
  blacs_gridinit(&ictx,"R",nprow,npcol);
  blacs_gridinfo(ictx,&nprow,&npcol,&myrow,&mycol);
 
  tol = 0.00000001;
  seed = 31415926535897.0;
  modname = "PDURNG  ";
  if ( iam < 5 && nnodes > 4 )
  {
    desc_x[DT_] = BLOCK_CYCLIC_2D;
    desc_x[M_] = 1; /*consider that X is distributed row wise to Check*/
    desc_x[N_] = 30;
    desc_x[MB_] = 1;
    desc_x[NB_] = 3;
    desc_x[RSRC_] = 0;
    desc_x[CSRC_] = 0;
    desc_x[CTXT_] = ictx;
    il_lda = numroc(1,1,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
 
    pdurng(&seed, 30, 3, dx1r, 0, ictx);
    pdvtchk(modname,30, 1, 1, dx1r, desc_x, dresx1r, 1, tol, ierr,&ilerr,
           &error, ifail);
    blacs_gridexit(ictx);
  }
  else if (iam == 0)
  {
    desc_x[DT_] = BLOCK_CYCLIC_2D;
    desc_x[M_] = 30;/*consider that X is ditributed colmn wise to chec*/
    desc_x[N_] = 1;
    desc_x[MB_] = 3;
    desc_x[NB_] = 1;
    desc_x[RSRC_] = 0;
    desc_x[CSRC_] = 0;
    desc_x[CTXT_] = ictx;
    il_lda = numroc(30,3,myrow,0,nprow);
    desc_x[LLD_] = MAX(1,il_lda);
 
    pdurng(&seed, 30, 3, dx1r, 0, ictx);
    pdvtchk(modname,30, 1, 1, dx1r, desc_x, dresx1r, 1, tol, ierr,&ilerr,
           &error, ifail);
    blacs_gridexit(ictx);
  }
}
 
/*********************************************************************
*          PARALLEL UTILITY SUBROUTINE                               *
*********************************************************************/
 
void putil(int *ifail)
{
 
  int iam,nnodes,myrow,mycol;
  int ictx,nprow,npcol,ipsl;
  int m,n,mb,nb,irsrc,crsrc,lld,info,ifails;
  int ix,ceil,inum,idenom,i1,i2,lcm;
  int iic,jjc,ic,jc,m_c,mb_c,nb_c,rsrc_c,csrc_c,icrow,iccol;
  int testvalue,rc,li;
 
  int desc[DLEN_];
  int desc_a[DLEN_]={ 1,0,9,9,4,4,0,0,0 };
 
/* IPESSL    START :     */
 
  blacs_pinfo(&iam,&nnodes);
 
  nprow = 1;
  npcol = nnodes;
 
  blacs_get(0,0,&ictx);
  blacs_gridinit(&ictx,"R",1,nnodes);
  blacs_gridinfo(ictx,&nprow,&npcol,&myrow,&mycol);
 
  ipsl = ipessl();
  if (ipsl < 3010000) {
    printf( " IPESSL FAIL\n");
    *ifail = *ifail + 1;
  }
 
/* IPESSL      END         */
/* DESCINIT    START :     */
 
  m     = 9;
  n     = 9;
  mb    = 4;
  nb    = 4;
  irsrc = 0;
  crsrc = 0;
  lld   = numroc(m,mb,myrow,irsrc,nprow);
  lld   = MAX(1, lld);
 
  desc_a[1] = ictx;
  desc_a[8] = lld;
 
  ifails = 0;
 
  descinit(desc,m,n,mb,nb,irsrc,crsrc,ictx,lld,&info);
 
  if( info != 0 ) {
    if (!iam) printf(" INFO from DESCINIT = %d\n",info);
    *ifail = *ifail + 1;
  } else {
    for(ix=0;ix<9;ix++) {
      if(desc[ix] != desc_a[ix] ) {
        printf(" ERROR IN DESCINIT:  ELEMENT %d = %d instead of %d\n",
               ix,desc[ix],desc_a[ix] );
        ifails = 1;
      }
    }
  }
 
  if(ifails == 1) *ifail = *ifail + 1;
 
/* DESCINIT    END         */
/* DESCSET     START :     */
 
  ifails = 0;
 
  descset(desc,m,n,mb,nb,irsrc,crsrc,ictx,lld);
 
  for(ix=0;ix<9;ix++) {
    if(desc[ix] != desc_a[ix] ) {
      printf(" ERROR IN DESCSET:  ELEMENT %d = %d instead of %d\n",
             ix,desc[ix],desc_a[ix] );
      ifails = 1;
    }
  }
 
  if(ifails == 1) *ifail = *ifail + 1;
 
/* DESCSET     END         */
/* ICEIL       START :     */
 
  inum   = 5;
  idenom = 3;
 
  ceil   = iceil(inum, idenom);
 
  if( ceil != 2 ) {
    if(!iam) printf(" ERROR IN ICEIL:  RESULT: %d EXPECTED: 2\n",ceil);
    *ifail = *ifail + 1;
  }
 
/* ICEIL       END         */
/* ILCM        START :     */
 
  i1  = 5;
  i2  = 3;
 
  lcm = ilcm(i1, i2);
 
  if( lcm != 15 ) {
    if(!iam) printf(" ERROR IN ILCM:  RESULT: %d EXPECTED: 15\n",lcm);
    *ifail = *ifail + 1;
  }
 
/* ILCM        END         */
/* INDXG2L     START :     */
 
  ic     = 6;
  jc     = 9;
  mb_c   = 4;
  nb_c   = 4;
  rsrc_c = 0;
  csrc_c = 0;
 
  ifails = 0;
 
  iic = indxg2l(ic,mb_c,0,rsrc_c,nprow);
  jjc = indxg2l(jc,nb_c,0,csrc_c,npcol);
 
  testvalue = mb_c*((ic-1)/(mb_c*nprow))+( (ic-1) % mb_c)+1;
 
  if(iic != testvalue) {
    if(!iam) printf(" ERROR IN INDXG2L:  RESULT: %d EXPECTED: %d\n",iic,
                    testvalue);
    ifails = 1;
  }
 
  testvalue = nb_c*((jc-1)/(nb_c*npcol))+( (jc-1) % nb_c)+1;
 
  if(jjc != testvalue) {
    if(!iam) printf(" ERROR IN INDXG2L:  RESULT: %d EXPECTED: %d\n",jjc,
                    testvalue);
    ifails = 1;
  }
 
  if(ifails == 1) *ifail = *ifail + 1;
 
/* INDXG2L     END         */
/* INDXG2P     START :     */
 
  ic     = 6;
  jc     = 9;
  mb_c   = 4;
  nb_c   = 4;
  rsrc_c = 0;
  csrc_c = 0;
 
  ifails = 0;
 
  iic = indxg2p(ic,mb_c,0,rsrc_c,nprow);
  jjc = indxg2p(jc,nb_c,0,csrc_c,npcol);
 
  testvalue = (rsrc_c + (ic - 1) / mb_c) % nprow;
 
  if(iic != testvalue) {
    if(!iam) printf(" ERROR IN INDXG2P:  RESULT: %d EXPECTED: %d\n",iic,
                    testvalue);
    ifails = 1;
  }
 
  testvalue = (csrc_c + (jc - 1) / nb_c) % npcol;
 
  if(jjc != testvalue) {
    if(!iam) printf(" ERROR IN INDXG2P:  RESULT: %d EXPECTED: %d\n",jjc,
                    testvalue);
    ifails = 1;
  }
 
  if(ifails == 1) *ifail = *ifail + 1;
 
/* INDXG2P     END         */
/* INDXL2G     START :     */
 
  ic        = 1;
  jc        = 1;
 
  mb_c      = 4;
  nb_c      = 4;
  rsrc_c    = 0;
  csrc_c    = 0;
 
  ifails    = 0;
 
  iic = indxl2g(ic,mb_c,myrow,rsrc_c,nprow);
  jjc = indxl2g(jc,nb_c,mycol,csrc_c,npcol);
 
  testvalue = nprow*mb_c*((ic-1)/mb_c) + (ic-1) % mb_c + ((nprow+myrow-rsrc_c) %
              nprow)*mb_c + 1;
 
  if(iic != testvalue ) {
    if(!iam) printf(" ERROR IN INDXL2G:  RESULT: %d EXPECTED: %d\n",iic,
                    testvalue);
    ifails = 1;
  }
 
  testvalue = npcol*nb_c*((jc-1)/nb_c) + (jc-1) % nb_c + ((npcol+mycol-csrc_c) %
              npcol)*nb_c + 1;
 
  if(jjc != testvalue) {
    if(!iam) printf(" ERROR IN INDXL2G:  RESULT: %d EXPECTED: %d\n",jjc,
                    testvalue);
    ifails = 1;
  }
 
  if(ifails == 1) *ifail = *ifail + 1;
 
/* INDXL2G     END         */
/* INFOG1L     START :     */
 
  ic          = 5;
  jc          = 7;
  m_c         = 9;
 
  desc[DT_]   = 1;
  desc[CTXT_] = ictx;
  desc[M_]    = 9;
  desc[N_]    = 9;
  desc[MB_]   = 4;
  desc[NB_]   = 4;
  desc[RSRC_] = 0;
  desc[CSRC_] = 0;
  desc[LLD_]  = MAX(1,numroc(m_c,mb_c,myrow,rsrc_c,nprow));
 
  ifails      = 0;
 
  infog1l(ic,desc[MB_],nprow,myrow,desc[RSRC_],&iic,&icrow);
  infog1l(jc,desc[NB_],npcol,mycol,desc[CSRC_],&jjc,&iccol);
 
  testvalue = ( ((ic-1) / desc[MB_]) + desc[RSRC_] ) % nprow;
 
  if(icrow != testvalue) {
    printf(" ERROR IN INFOG1L:  RESULT: %d EXPECTED: %d\n",icrow,testvalue);
    ifails = 1;
  }
 
  testvalue = ( ((jc-1) / desc[NB_]) + desc[CSRC_] ) % npcol;
 
  if(iccol != testvalue) {
    printf(" ERROR IN INFOG1L:  RESULT: %d EXPECTED: %d\n",iccol,testvalue);
    ifails = 1;
  }
 
  /*  Test iic  */
 
  rc = ( ((ic-1)/desc[MB_]) + desc[RSRC_]) % nprow;
  li = ( ((ic-1)/desc[MB_]) / nprow + 1) * desc[MB_] + 1;
 
  if( ((myrow+nprow-desc[RSRC_]) % nprow) >= ( ((ic-1)/desc[MB_]) % nprow) ) {
     if(myrow == rc) li = li + (ic-1) % desc[MB_];
     li = li - desc[MB_];
  }
 
  testvalue = li;
 
  if(iic != testvalue ) {
    printf(" ERROR IN INFOG1L:  RESULT: %d EXPECTED: %d\n",iic, testvalue);
    ifails = 1;
  }
 
  /*  Test jjc  */
 
  rc = ( ((jc-1)/desc[NB_]) + desc[CSRC_]) % npcol;
  li = ( ((jc-1)/desc[NB_]) / npcol +1) * desc[NB_] + 1;
 
  if( ((mycol+npcol-desc[CSRC_]) % npcol) >= ( ((jc-1)/desc[NB_]) % npcol) ) {
    if(mycol == rc) li = li + (jc-1) % desc[NB_];
    li = li - desc[NB_];
  }
 
  testvalue = li;
 
  if(jjc != testvalue) {
    printf(" ERROR IN INFOG1L:  RESULT: %d EXPECTED: %d\n",jjc, testvalue);
    ifails = 1;
  }
 
  if(ifails == 1) *ifail = *ifail + 1;
 
/* INFOG1L     END         */
/* INFOG2L     START :     */
 
  ic          = 5;
  jc          = 7;
 
  desc[DT_]   = 1;
  desc[CTXT_] = ictx;
  desc[M_]    = 9;
  desc[N_]    = 9;
  desc[MB_]   = 4;
  desc[NB_]   = 4;
  desc[RSRC_] = 0;
  desc[CSRC_] = 0;
  desc[LLD_]  = MAX(1,numroc(m_c,mb_c,myrow,rsrc_c,nprow));
 
  ifails      = 0;
 
  infog2l(ic,jc,desc,nprow,npcol,myrow,mycol,&iic,&jjc,&icrow,&iccol);
 
  testvalue =  ( ( (ic-1) / desc[MB_] ) + desc[RSRC_] ) % nprow;
 
  if(icrow != testvalue) {
    printf(" ERROR IN INFOG2L:  RESULT: %d EXPECTED: %d\n",icrow,testvalue);
    ifails = 1;
  }
 
  testvalue = ( ( (jc-1) / desc[NB_] ) + desc[CSRC_] ) % npcol;
 
  if(iccol != testvalue) {
    printf(" ERROR IN INFOG2L:  RESULT: %d EXPECTED: %d\n",iccol,testvalue);
    ifails = 1;
  }
 
  rc = ( ( (ic-1) / desc[MB_] ) + desc[RSRC_]) % nprow;
  testvalue = ( ( (ic-1) / desc[MB_] ) / nprow + 1 ) * desc[MB_] + 1;
 
  if( (myrow+nprow-desc[RSRC_]) % nprow >= ( ( (ic-1) / desc[MB_] ) % nprow) ) {
    if( myrow == rc ) testvalue = testvalue + (ic-1) % desc[MB_];
    testvalue = testvalue - desc[MB_];
  }
 
  if(iic != testvalue ) {
    printf(" ERROR IN INFOG2L:  RESULT: %d EXPECTED: %d\n",iic,testvalue);
    ifails = 1;
  }
 
  rc = ( ( (jc-1) / desc[NB_] ) + desc[CSRC_]) % npcol;
  testvalue = ( ( (jc-1) / desc[NB_] ) / npcol + 1 ) * desc[NB_] + 1;
 
  if( (mycol+npcol-desc[CSRC_]) % npcol >= ( ( (jc-1) / desc[NB_] ) % npcol) ) {
    if( mycol == rc ) testvalue = testvalue + (jc-1) % desc[NB_];
    testvalue = testvalue - desc[NB_];
  }
  if(jjc != testvalue ) {
    printf(" ERROR IN INFOG2L:  RESULT: %d EXPECTED: %d\n",jjc,testvalue);
    ifails = 1;
  }
 
  if(ifails == 1) *ifail = *ifail + 1;
 
/* INFOG2L     END         */
 
  blacs_gridexit(ictx);
}
